/*
 *
 * File: InstructionTrace.cpp
 * Author: Jing Que, Babak Yadegari
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#include "pin.H"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <set>
#include <sstream>
#include <map>

typedef struct _IMAGE_DOS_HEADER {      // DOS .EXE header
    unsigned short   e_magic;                     // Magic number
    unsigned short   e_cblp;                      // unsigned chars on last page of file
    unsigned short   e_cp;                        // Pages in file
    unsigned short   e_crlc;                      // Relocations
    unsigned short   e_cparhdr;                   // Size of header in paragraphs
    unsigned short   e_minalloc;                  // Minimum extra paragraphs needed
    unsigned short   e_maxalloc;                  // Maximum extra paragraphs needed
    unsigned short   e_ss;                        // Initial (relative) SS value
    unsigned short   e_sp;                        // Initial SP value
    unsigned short   e_csum;                      // Checksum
    unsigned short   e_ip;                        // Initial IP value
    unsigned short   e_cs;                        // Initial (relative) CS value
    unsigned short   e_lfarlc;                    // File address of relocation table
    unsigned short   e_ovno;                      // Overlay number
    unsigned short   e_res[4];                    // Reserved unsigned shorts
    unsigned short   e_oemid;                     // OEM identifier (for e_oeminfo)
    unsigned short   e_oeminfo;                   // OEM information; e_oemid specific
    unsigned short   e_res2[10];                  // Reserved unsigned shorts
    long   e_lfanew;                    // File address of new exe header
} IMAGE_DOS_HEADER, *PIMAGE_DOS_HEADER;

typedef struct _IMAGE_FILE_HEADER {
    unsigned short    Machine;
    unsigned short    NumberOfSections;
    unsigned long   TimeDateStamp;
    unsigned long   PointerToSymbolTable;
    unsigned long   NumberOfSymbols;
    unsigned short    SizeOfOptionalHeader;
    unsigned short    characteristics;
} IMAGE_FILE_HEADER, *PIMAGE_FILE_HEADER;

typedef struct _IMAGE_DATA_DIRECTORY {
    unsigned long   VirtualAddress;
    unsigned long   Size;
} IMAGE_DATA_DIRECTORY, *PIMAGE_DATA_DIRECTORY;

#define IMAGE_NUMBEROF_DIRECTORY_ENTRIES 16

typedef struct _IMAGE_OPTIONAL_HEADER {
    //
    // Standard fields.
    //

    unsigned short    Magic;
    unsigned char    MajorLinkerVersion;
    unsigned char    MinorLinkerVersion;
    unsigned long   SizeOfCode;
    unsigned long   SizeOfInitializedData;
    unsigned long   SizeOfUninitializedData;
    unsigned long   AddressOfEntryPoint;
    unsigned long   BaseOfCode;
    unsigned long   BaseOfData;

    //
    // NT additional fields.
    //

    unsigned long   ImageBase;
    unsigned long   SectionAlignment;
    unsigned long   FileAlignment;
    unsigned short    MajorOperatingSystemVersion;
    unsigned short    MinorOperatingSystemVersion;
    unsigned short    MajorImageVersion;
    unsigned short    MinorImageVersion;
    unsigned short    MajorSubsystemVersion;
    unsigned short    MinorSubsystemVersion;
    unsigned long   Win32VersionValue;
    unsigned long   SizeOfImage;
    unsigned long   SizeOfHeaders;
    unsigned long   CheckSum;
    unsigned short    Subsystem;
    unsigned short    Dllcharacteristics;
    unsigned long   SizeOfStackReserve;
    unsigned long   SizeOfStackCommit;
    unsigned long   SizeOfHeapReserve;
    unsigned long   SizeOfHeapCommit;
    unsigned long   LoaderFlags;
    unsigned long   NumberOfRvaAndSizes;
    IMAGE_DATA_DIRECTORY DataDirectory[IMAGE_NUMBEROF_DIRECTORY_ENTRIES];
} IMAGE_OPTIONAL_HEADER32, *PIMAGE_OPTIONAL_HEADER32;

typedef struct _IMAGE_NT_HEADERS {
    unsigned long Signature;
    IMAGE_FILE_HEADER FileHeader;
    IMAGE_OPTIONAL_HEADER32 OptionalHeader;
} IMAGE_NT_HEADERS, *PIMAGE_NT_HEADERS;

// Trick to avoid playing with new strings every time we see an API call
map<string, ADDRINT> sys_APIs;
map<ADDRINT, string> sys_APIs2;


/**
* Output file the shellcode information is dumped to.
**/
std::ofstream traceFile;

/**
* Command line option to specify the name of the output file.
* Default is shellcode.out.
**/
KNOB<string> outputFile(KNOB_MODE_WRITEONCE, "pintool", "o", "trace.out", "specify trace file name");
//KNOB<string> outputFile2(KNOB_MODE_WRITEONCE, "pintool", "o", "api.out", "specify api file name");

bool bStopTrace = false;

ADDRINT OEP = 0;
bool OEP_has_reached = false;

std::string target_path;

PIN_MUTEX lock;

/**
* Prints usage information.
**/
INT32 usage()
{
    cerr << "This tool produces an instruction trace." << endl << endl;
    cerr << KNOB_BASE::StringKnobSummary() << endl;
    return -1;
}

VOID dump_reg(std::string *instructionString, ADDRINT ip, ADDRINT eax, ADDRINT ecx, ADDRINT edx, ADDRINT ebx, ADDRINT esp, ADDRINT ebp, ADDRINT esi, ADDRINT edi)
{
    if (bStopTrace) {
        return;
    }
	
	if (sys_APIs2.find(ip) != sys_APIs2.end())
	{
		traceFile << PIN_GetTid() << "\t"<< "API CALL " << sys_APIs2[ip].c_str() << endl;
	}
    traceFile << PIN_GetTid() << "\t" << *instructionString << "\t" << hex << eax << "\t" << ecx << "\t" << edx << "\t" << ebx << "\t" << esp << "\t" << ebp << "\t" << esi << "\t" << edi << endl;
}

std::string extractFilename(const std::string &filename)
{
    unsigned int lastBackslash = filename.rfind("\\");

    if (lastBackslash == std::string::npos) {
        return filename;
    } else {
        return filename.substr(lastBackslash + 1);
    }
}

std::string extractPath(const std::string &filename)
{
    unsigned int lastBackslash = filename.rfind("\\");

    if (lastBackslash == -1) {
        return NULL;
    } else {
        return filename.substr(0, lastBackslash + 1);
    }
}

VOID Instruction(INS ins, VOID *v)
{
    std::stringstream ss;

    if (bStopTrace) {
        return;
    }

    ADDRINT address = INS_Address(ins);

    if (!OEP_has_reached) {
        if (address == OEP) {
            OEP_has_reached = true;
        } else {
            return;
        }
    }

    ss << hex << address << "\t";

    IMG img = IMG_FindByAddress(address);

    if (IMG_Valid(img)) {
        if(IMG_Name(img).find(target_path) != std::string::npos) {
            ss << "1";
        } else {
            ss << "0";
        }
    } else {
        ss << "2";
    }

    // Generate address and module information


    // Generate diassembled string
    ss << "\t" << INS_Disassemble(ins) << "\t" << hex << INS_Size(ins);

    // Generate instruction byte encoding
    /*  	for (int i=0;i<INS_Size(ins);i++)
    	{
    		ss << setfill('0') << setw(2) << (((unsigned int) *(unsigned char*)(address + i)) & 0xFF) << " ";
    	} */

    INS_InsertCall(
        ins, IPOINT_BEFORE, (AFUNPTR)dump_reg, IARG_PTR, new string(ss.str()),
		IARG_INST_PTR,
        IARG_REG_VALUE, REG_EAX,
        IARG_REG_VALUE, REG_ECX,
        IARG_REG_VALUE, REG_EDX,
        IARG_REG_VALUE, REG_EBX,
        IARG_REG_VALUE, REG_ESP,
        IARG_REG_VALUE, REG_EBP,
        IARG_REG_VALUE, REG_ESI,
        IARG_REG_VALUE, REG_EDI,
        IARG_END);
}

int count = 0;
int last_time = 393440186;
VOID timeGetTimeHack(CONTEXT *ctxt, ADDRINT return_ip)
{
    PIN_SetContextReg(ctxt, REG_EAX, last_time);
    count++;

    if (count % 2 == 0) {
        last_time += rand() % 100;
    }

    ADDRINT esp = PIN_GetContextReg(ctxt, REG_ESP);
    PIN_SetContextReg(ctxt, REG_ESP, esp + 4);
    PIN_SetContextReg(ctxt, REG_INST_PTR, return_ip);
    PIN_ExecuteAt(ctxt);
}

VOID BlockInputHack(CONTEXT *ctxt, ADDRINT return_ip)
{
    PIN_SetContextReg(ctxt, REG_EAX, 0);

    ADDRINT esp = PIN_GetContextReg(ctxt, REG_ESP);
    PIN_SetContextReg(ctxt, REG_ESP, esp + 8);
    PIN_SetContextReg(ctxt, REG_INST_PTR, return_ip);
    PIN_ExecuteAt(ctxt);
}

VOID StopTrace()
{
	if (!OEP_has_reached) {
        return;
    }

	bStopTrace = true;
}

VOID imgInstrumentation(IMG img, VOID *val)
{
    if (bStopTrace) {
        return;
    }

    if (!IMG_IsMainExecutable(img)) {
        for (SEC sec = IMG_SecHead(img); SEC_Valid(sec); sec = SEC_Next(sec)) {
            // Forward pass over all routines in a section
            for (RTN rtn = SEC_RtnHead(sec); RTN_Valid(rtn); rtn = RTN_Next(rtn)) {
                string nameAPI = RTN_Name(rtn);
                int api_addr = RTN_Address(rtn);

                if ((nameAPI != ".text") && (nameAPI != "unnamedImageEntryPoint")) {
                    RTN_Open(rtn);

                   if (nameAPI == "BlockInput") {
                        RTN_InsertCall(rtn, IPOINT_BEFORE, (AFUNPTR)BlockInputHack,
                                       IARG_CONTEXT,
                                       IARG_RETURN_IP,
                                       IARG_END);
                    }
                    else if (nameAPI == "GetMessageW" || nameAPI == "ShowWindow") { // || nameAPI == "CreateWindowW" || nameAPI == "CreateWindowExW" || nameAPI=="SendMessage")
                        RTN_InsertCall(rtn, IPOINT_BEFORE, (AFUNPTR)StopTrace, IARG_END);
                    } else {
                        if (sys_APIs.find(nameAPI) == sys_APIs.end())
						{
							sys_APIs[nameAPI] = api_addr;
							sys_APIs2[api_addr] = nameAPI;
						}
                    }

                    RTN_Close(rtn);
                }
            }
        }
    }
}


/**
* Finalizer function that is called at the end of the trace process.
* In this script, the finalizer function is responsible for closing
* the shellcode output file.
**/
VOID fini(INT32, VOID *)
{
    traceFile.close();
}

void get_OEP(char *filename)
{
    ifstream inf(filename, ifstream::binary);

    if (!inf.is_open()) {
        printf("Cannot open target file\n");
        return ;
    }

    char *buf = new char[1024];
    inf.read(buf, 1024);
    IMAGE_DOS_HEADER *dos_header = (IMAGE_DOS_HEADER *)buf;
    IMAGE_NT_HEADERS *pINH = (IMAGE_NT_HEADERS *)((unsigned long)dos_header + dos_header->e_lfanew);
    OEP = pINH->OptionalHeader.AddressOfEntryPoint + pINH->OptionalHeader.ImageBase;
    printf("EntryPoint:0x%.8x\n", OEP);

    delete [] buf;
}


int main(int argc, char *argv[])
{
    char *target_filename;

    PIN_InitSymbols();

    if ( PIN_Init(argc, argv)) {
        return usage();
    }

    for (int i = 1; i < argc; i++) {
        if (strcmp(argv[i], "--") == 0) {
            target_filename = argv[i + 1];
            break;
        }
    }

    get_OEP(target_filename);
    target_path = extractPath(target_filename);

    traceFile.open(outputFile.Value().c_str());
    PIN_MutexInit(&lock);
    INS_AddInstrumentFunction(Instruction, 0);
    IMG_AddInstrumentFunction(imgInstrumentation, 0);
    PIN_AddFiniFunction(fini, 0);

    // Never returns
    PIN_StartProgram();

    return 0;
}

