/*
 *
 * File: flowgraph.h
 * Author: Babak Yadegari, Saumya Debray
 */

/*
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#ifndef _FLOWGRAPH_H_
#define _FLOWGRAPH_H_

#include "lynx_types.h"
#include "instr.h"

/*
 * Flags definitions used to mark each block
 */

#define BBL_START               (1 << 0)
#define BBL_END                 (1 << 1)
#define BBL_ENTRY               (1 << 2)
#define BBL_FORWARD_TAINTED     (1 << 3)
#define BBL_BACKWARD_TAINTED    (1 << 4)
#define BBL_IS_CONDITIONAL      (1 << 5)
#define BBL_EXIT                (1 << 6)
#define BBL_IN_BASE_MODULE		 (1 << 7)
#define	 BBL_DOES_UNPACKING		 (1 << 8)

#define NODE_HAS_FLAG(node, flag)     ((node) ? (((node)->flags) &  (flag)) : 0)
#define NODE_SET_FLAG(node, flag)     ((node) ? (((node)->flags) |= (flag)) : 0)
#define NODE_REMOVE_FLAG(node, flag)  ((node) ? (((node)->flags) &= ~(flag)) : 0)


#define FORWARD_TAINT_COLOR				"turquoise"
#define BACKWARD_TAINT_COLOR			"red"
#define FORWARD_TAINTED_CONDITIONAL		"violet"
#define SYSTEM_CALL_COLOR				"grey"

#define HASHTABSZ  32768

/*
 * InsNode holds information about an instruction.
 */
typedef struct InsNode {
    Instruction *ins;
    int flags;
    struct iListNode *succs, *preds;
    int numSuccs, numPreds;
} InsNode;

/*
 * iListNode defines a doubly-linked list of pointers to instruction nodes.
 */
typedef struct iListNode {
    InsNode *iptr;
    struct iListNode *prev, *next; // links in the doubly-linked list of instructions
    struct iListNode *hnext; // next node in the hash table
} iListNode;

/*
 * Bbl defines the data structure used to represent basic blocks
 */
typedef struct Bbl {
    int id;
    int flags;
    iListNode *ins_hd, *ins_tl; // first and last instruction in the block
    struct bListNode *succs, *pDoms, *preds, *final_succs, *ipDoms;
    struct Bbl *next, *prev;
    struct iListNode *instr_succs_hd, *instr_succs_tl;		// keep a list of successors of type instrs for a block
    struct iListNode *instr_preds_hd, *instr_preds_tl;      // keep a list of predecessors of a block
    int numSuccs;           //number of successors
    int numPreds;
    int numInstrs;
    struct Bbl *cDepend;
    ArrayList *occurrences;          // keeps the positions in the trace where the bbl is executed
    int hasVariation;                   // is set if the last instr in a basic
                                        //block (control transfer) has the same
                                        //address but different target: we have call edx in original trace,
                                        //after simplification, there are different targets depending on different edx values!
    int variation;
    
    int start_addr;
    int instrs_updated;
    int trace_instr_start;

} Bbl;

typedef struct BblEdge {
    int id;
    Bbl *from;
    Bbl *to;
    struct BblEdge *next, *prev;
} BblEdge;

typedef struct bListNode {
    Bbl *bptr;
    struct bListNode *next;
    struct bListNode *reserved;					// as a helper variable. it might be used in some functions
} bListNode;


typedef struct SuperBbl{
    int id;
    int flags;
    iListNode *ins_hd, *ins_tl;
    struct sbListNode *succs;
    struct SuperBbl *next, *prev;
    int numSuccs;
    bListNode *bbl_list_hd, *bbl_list_tl;
} SuperBbl;

typedef struct sbListNode {
    SuperBbl *bptr;
    struct sbListNode *next;
} sbListNode;

/*
 * Internal data structures
 */

typedef struct MemDefWithIndex{
    Range *memDef;
    uint64_t index;
} MemDefWithIndex;

typedef struct ConstantsinScope{
    REGISTER_BITVECTOR  regs;
    WriteSet            *mem;
    uint64_t            startIndex;
} ConstantsinScope;


typedef struct bblList{
	int id;
	Bbl *bbl_hd, *bbl_tl;
	struct bblList *next, *prev;
} BblList;

typedef struct lastOp{
    iListNode   *backToInstr;           // last instruction [i.e trace line]
    Bbl         *from, *to, *prevBbl, *lastFound;
    bool        removeEdge;
    bListNode   *otherSuccessors;
    ArrayList   *bbls;
} lastOp;
/*
 * External API
 */
void ConstructOroginalCFG(InstrList *iList);
void ConstructCFG(InstrList *iList, int);
void ConstructInitialCFG(InstrList *iList, int DependencyAnlaysis);
void PrintCFG(FILE *f, int printInstrs, uint64_t start, uint64_t end);
void PrintFinalCFG(FILE *f, int);
void PrintBblInstrs(FILE *f);
void PrintDominatorTree(FILE *f);
void FindBblSuccsAndPreds();
void DependencyAnalysis(InstrList *iList);
void PrintCFGSvet(FILE *f, int printInstrs, uint64_t, uint64_t);
void PrintCFG(FILE *f, int printInstrs, uint64_t start, uint64_t end);
void CleanUpCFGData();
void MarkUnsimplifiableInstrs(InstrList *iList);
void MergeBblsNew();
bool FixSimplifiedCFGAddresses(InstrList *iList);
void PrintPlainCFG(FILE *f, uint64_t start, uint64_t end);
void ModifyBblsInstrs(InstrList *iList);

void PatternSuccessors(InstrList *iList);
void BuildPatterns();
#endif  /* _FLOWGRAPH_H_ */
