/*
 *
 * File: instr.h
 * functions related to instructions, their manipulation, or
 *      their characteristics
 * Author: Babak Yadegari, Kevin Coogan, Saumya Debray
 */

/*
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */


#ifndef INSTR_H_
#define INSTR_H_


#define NO_INSTR -1

#define OFFSET_TO_REG	0
#define REG_TO_OFFSET  1

#include "lynx_types.h"

#ifdef __cplusplus
extern "C" {
#endif

#define CACHE_SIZE (2*PAGE_SIZE)

/* Instruction Characteristics */
void WriteInstrFileHeader(InstrList *iList, FILE *file, bool mmap);
Instruction *AddInstruction(InstrList *i_list, Instruction *, int after);
int FindChangedRegs(InstrList *iList, Instruction *iins1);
void SetCurrentInstr(InstrList *iList, uint64_t order);
Instruction *FetchPrevInstr(InstrList *iList);
Instruction *FetchNextInstr(InstrList *iList);
Instruction *GetNextLiveInstr(InstrList *iList);
Instruction *GetPrevLiveInstr(InstrList *iList);
ud_t *LoadUDisInstrObj(Instruction *instr);
void CopyUdins2InsStr(ins_structure *uIns, ud_t *udIns);
WriteSet *GetWriteTargets(InstrList *iList, Instruction *instr);
REGISTER_BITVECTOR RegDef(InstrList *iList, Instruction *instr);
REGISTER_BITVECTOR RegUsed(InstrList *iList, Instruction *instr);
REGISTER_BITVECTOR MemRegUsed(Instruction *instr);
REGISTER_BITVECTOR NonMemRegUsed(InstrList *iList, Instruction *instr);
//REGISTER_BITVECTOR MemRegDef(Instruction *instr);
//REGISTER_BITVECTOR NonMemRegDef(Instruction *instr);
PSW_BITVECTOR FlagsDef(Instruction *instr);
PSW_BITVECTOR FlagsUsed(Instruction *instr);
WriteSet *MemDef(InstrList *iList, Instruction *instr);
WriteSet *MemUsed(InstrList *iList, Instruction *instr);
void InstrRegsFalgsUseDef(InstrList *iList, Instruction *iins);
void CalcKnownDefsAndUses(InstrList *iList);
void *CalcKnownDefsAndUsesPar(void *args);       // used in parallel version!
enum ud_type UDRegSizeMask(enum ud_type src, int size);

// TODO: obsolete ?
REGISTER_BITVECTOR RegUsedInclSP(InstrList *iList, Instruction *iins);
REGISTER_BITVECTOR RegDefInclSP(InstrList *iList, Instruction *iins);
REGISTER_BITVECTOR GetImplicitRegDef(Instruction *instr);

/* List operations */
InstrList *BuildCloneIList(InstrList *orig, char *filename, int flag);
InstrList *BuildCloneSimplifiedIList(InstrList *orig, char *filename);
void AddInstrToFile(FILE *f, Instruction *instr);
void ClearFlagFromIList(InstrList *iList, uint32_t flag);


/* helpers */
int CountInstructionsBetweenTwoInstrs2(InstrList *iList, Instruction *iins, Instruction *jins);
int CountInstructionsBetweenTwoInstrs(InstrList *iList, Instruction *iins, Instruction *jins);
void MakeInstrAlive(InstrList *iList, Instruction *instr);
bool InstrXorAsMov(Instruction *instr);
bool InstrCompare(Instruction *iins1, Instruction *iins2);
int InstrIsaCallToInputFunction(char *str);
bool InstrIsaCallToOutputFunction(char *str);
bool InstrIsIndirectJump(Instruction *instr);
bool InstrIsSetOnCondition(enum ud_mnemonic_code opcode);
bool IsMxcsrInstr(Instruction *instr);
bool IsStringCompare(Instruction *instr);
void RInstrDestroy(InstrList *iList, Instruction *instr);
void UpdateInstrCache(InstrList *iList, int index);
REGISTER_BITVECTOR GetImplicitMemRegDef(Instruction *instr);
bool IsStringOperation(Instruction *instr);
Instruction *GetInstruction(InstrList *iList, uint64_t index);
Instruction *GetNextInstrAtAddr(InstrList *iList, int index, ADDRESS next_addr);
int SaveInstrChange(InstrList *iList, Instruction *instr);
InstrList *InitInstructionCache(InstrList *iList);
int OperandSize(ud_operand_t op);
uint GetRegisterValue(Instruction *instr, enum ud_type reg);
uint SetRegisterValue(Instruction *instr, int reg, int Val);
bool GetMemoryValueFWD(InstrList *, Instruction *, WriteSet *, uint32_t *);
bool InstrMasksRegister(Instruction *instr);
bool IsControlFlowInstr(Instruction *instr);
bool IsCondJump(Instruction *instr);
bool IsMov(Instruction *instr);
bool IsPush(Instruction *instr);
bool IsPushAll(Instruction *instr);
bool IsPop(Instruction *instr);
bool IsPopAll(Instruction *instr);
bool InstrIsDirectJmp(Instruction *iins);
int UDRegSize(REGISTER_BITVECTOR reg);
REGISTER_BITVECTOR UDRegisterToBitVector(enum ud_type reg);
enum ud_type BitVectorToUDRegisterForTest(REGISTER_BITVECTOR reg);
enum ud_type BitVectorToUDRegister(REGISTER_BITVECTOR reg);
Instruction *GetInstructionFromFile(InstrList *iList, int index);
size_t WriteInstrToFile(FILE *f, Instruction *instr);
void StripTraceForAlignmentCode(FILE *f, InstrList *iList);
void ExcludeNonRealMain(InstrList *iList);
void GenerateAllMnemonicSequence(InstrList *iList, FILE *mnemonicFile);
int WriteInstrBytesToFile(FILE *f, InstrList *iList, uint64_t start, uint64_t end);
bool InstrSetsFlags(Instruction *instr);
bool CondJumpIsTaken(Instruction *iins);
REGISTER_BITVECTOR UpToAllRegs(REGISTER_BITVECTOR reg);
REGISTER_BITVECTOR UpMaskDefinedReg(REGISTER_BITVECTOR reg);
REGISTER_BITVECTOR MaskDefinedReg(REGISTER_BITVECTOR reg);
REGISTER_BITVECTOR MaskDefinedRegOld(REGISTER_BITVECTOR reg);
unsigned int ArithOp(enum ud_mnemonic_code op, unsigned int val1, int val2, int size,
        REGISTER_BITVECTOR *flags, Instruction *iins);
uint BitvectorRegisterSize(REGISTER_BITVECTOR reg);
enum ud_type UDOpIsIndirectThruReg(ud_operand_t *op);
int bSwap(int val);
int rotate(unsigned int invalue, int places, int direction, int size);
bool OpcodeIsArithWithNoEffect(enum ud_mnemonic_code opcode);
bool OpcodeIsArith(enum ud_mnemonic_code opcode);
void CarefullCopyUdins2InsStr(InstrList *iList, Instruction *iins, ins_structure *uIns, ud_t *udIns, int doNotChangeFlags);
WriteSet *MemoryMask(int offset, REGISTER_BITVECTOR reg);
int RegMask(REGISTER_BITVECTOR reg);
enum ud_type UDOpIsRegister(ud_operand_t *op);
int floor_log2(REGISTER_BITVECTOR v);
int SignExtend(int value);
Instruction *TransformXchgNotXchgToNot(InstrList *iList, Instruction *iins);
Instruction *ThreeXorsToXchg(InstrList *iList, Instruction *iins);
void ConstantMemoryMap(InstrList* iList);


bool InstrWritesToMemory(Instruction *iins);
bool InstrHasDest(Instruction *instr);
Range *GetImplicitTarget(InstrList *iList, Instruction *instr);

void GetLocsAccessedByOp(Instruction *instr,
			 ud_operand_t op,
			 REGISTER_BITVECTOR *regs,
			 WriteSet **mem);
void ProcessXCHG(Instruction *instr, REGISTER_BITVECTOR *regs, WriteSet **mems);
Range *GetTarget(Instruction *instr, ud_operand_t op);

/* creating udis86 instruction representations */
void InitUdis(ud_t *ud_obj);
ud_t *UDCreateJz(int val);
ud_t *UDCreatePushReg(enum ud_type reg);
ud_t *UDCreatePopReg(enum ud_type reg);
ud_t *UDCreateMovImmToReg(enum ud_type reg, unsigned int val);
ud_t *UDCreateMovRegToReg(enum ud_type destination, enum ud_type source);
ud_t *UDCreateAddRegToReg(enum ud_type src, enum ud_type dst);
ud_t *UDCreateAddImmToOffset(int offset, int value, int size);
ud_t *UDCreateAddRegToOffset(int offset, enum ud_type src, int size);
ud_t* UDCreateAddImmToReg(enum ud_type reg, int val);
ud_t *UDCreateSubRegFromReg(enum ud_type src, enum ud_type dst);
ud_t *UDCreateSubImmFromReg(enum ud_type reg, int val);
ud_t *UDCreateXorImmWithReg(enum ud_type reg, int val);
ud_t *UDCreateAndImmWithReg(enum ud_type reg, int val);
ud_t *UDCreateCmpImmWithReg(enum ud_type reg, int val);
ud_t *UDCreateCmpRegWithImm(enum ud_type src, int value);
ud_t *UDCreateCmpOffsetWithReg(int offset, enum ud_type dest, int type);
ud_t *UDCreateCmpOffsetWithImm(int offset, int value, int size);
ud_t *UDCreateOrImmWithReg(enum ud_type reg, int val);
ud_t *UDCreateTestImmWithReg(enum ud_type reg, int val);
ud_t *UDCreateShlImmToReg(enum ud_type reg, int val);
ud_t *UDCreateShrImmToReg(enum ud_type reg, int val);
ud_t *UDCreateRorImmToReg(enum ud_type reg, int val);
ud_t *UDCreateRolImmToReg(enum ud_type reg, int val);
ud_t *UDCreateNotOnReg(enum ud_type reg);
ud_t *UDCreateNotMem(int offset, int size);
ud_t *UDCreatePushImm(enum ud_type reg, int val);
ud_t *UDCreateJump(int val);
ud_t *UDCreateJumpToAddr(int val);
ud_t *UDCreateRet(int val);
ud_t *UDCreateCallToAddr(int val);
ud_t *UDCreateMovImmToMem(unsigned char *bytes, enum ud_type dest, int val);
ud_t *UDCreateMovImmToOffset(int offset, int value, int size);
ud_t *UDCreateAddImmToMem(unsigned char *bytes, enum ud_type dest, int val);
ud_t *UDCreateSubImmFromMem(unsigned char *bytes, enum ud_type dest, int val);
ud_t *UDCreateSubImmFromOffset(int offset, int value, int size);

ud_t *UDCreateCmpMemWithImm(unsigned char *bytes, enum ud_type dest, int val);
ud_t *UDCreateXorMemWithImm(unsigned char *bytes, enum ud_type dest, int val);
ud_t *UDCreateOrMemWithImm(unsigned char *bytes, enum ud_type dest, int val);
ud_t *UDCreateAndMemWithImm(unsigned char *bytes, enum ud_type dest, int val);
ud_t *ChangeXchgMemRegToMov(unsigned char *bytes, int exchangeOps);
ud_t *UDCreatePopOffset(unsigned char *bytes, int offset);
ud_t *UDCreatePushOffset(unsigned char *bytes, int offset);
ud_t *UDCreateMovOffsetToReg(enum ud_type dest, int offset, int direction /*1 : reg to memory offset, 0 : offset to reg*/);
ud_t *UDCreateIndirectStoreConst(enum ud_type base,
		enum ud_type index,
		int disp,
		int val);
ud_t *UDCreateNop();
ud_t *UDCreateImulImmToReg(enum ud_type reg, int val);

#ifdef __cplusplus
}
#endif

#endif /* INSTR_H_ */
