/*
 *
 * File: lynx_types.h
 * Author: Babak Yadegari, Kevin Coogan
 */

/*
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */


#ifndef _LYNX_TYPES_H_
#define _LYNX_TYPES_H_

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdbool.h>
#include "udis86.h"
#include <assert.h>
#include <ctype.h>
#include <pthread.h>
#include "array_list.h"
#include "flags.h"
#include "x86-table-defs.h"
#include "x86-classes.h"
//#include "mem_maps.h"

#ifdef __APPLE__
#include <sys/types.h>
#endif


#ifdef __cplusplus
extern "C" {
#endif

// Enable large file sizes
#define _FILE_OFFSET_BITS 64

//#define ENABLE_DLL_POINTER_CALCULATION 1

// page size for instr file
#define SYS_PAGE_SIZE sysconf(_SC_PAGE_SIZE)

#define PAGE_SIZE ((uint64_t)256*SYS_PAGE_SIZE)

//set to true to store instruction objects in file database instead of memory
#define STORE_INSTRS_IN_FILE true   //This is never false anymore -- kpc 6/8/12

//set to true if mmap'ing file to memory
#define USE_MMAP true

// marker for main function of program, Used for Cristian's bianries!

#define PIN_TRACE	0

#define OLDITERSTIONAL 0		// for performance and functional tests only, not important for actual use!

#define DEBUG_MESSAGES      0

#define TMP_DIR_PREFIX		"./files/"

#define MAIN_MARKER 1

#define MEMBITVARS_MAPPING  1

#define MARK_THREAD_BOUNDARIES      1



// set to true to ignore processing non-base_module instructions
#define IGNORE_API_CALLS            0

#define MAX_CHARS_PER_LINE 500
#define MAX_INSTR_NAME_LEN 51       //50 + terminating null char
#define MAX_MODULE_LEN 21           //20 + terminating null char
#define MAX_FUNC_NAME_LEN 41        //40 + terminating null char
#define MAX_INSTRLEN 15             //max number of bytes for x86 instr
#define InstrFileHeaderSize sizeof(InstrFileHeader)

//true ignores use/def of sp reg in u-d calc for push/pop's
#define IGNORE_SP_REG true
#define ALLOW_TASKSWITCH_CALLS false

typedef uint32_t REGISTER_BITVECTOR;
typedef uint32_t PSW_BITVECTOR;

typedef uint32_t ADDRESS;
/*
 *  BITVECTOR REGISTER ASSIGNMENTS
 */
#define LYNX_EAX  (REGISTER_BITVECTOR)(1 << 0)
#define LYNX_ECX  (REGISTER_BITVECTOR)(1 << 1)
#define LYNX_EDX  (REGISTER_BITVECTOR)(1 << 2)
#define LYNX_EBX  (REGISTER_BITVECTOR)(1 << 3)
#define LYNX_ESP  (REGISTER_BITVECTOR)(1 << 4)
#define LYNX_EBP  (REGISTER_BITVECTOR)(1 << 5)
#define LYNX_ESI  (REGISTER_BITVECTOR)(1 << 6)
#define LYNX_EDI  (REGISTER_BITVECTOR)(1 << 7)
#define LYNX_ES   (REGISTER_BITVECTOR)(1 << 8)
#define LYNX_CS   (REGISTER_BITVECTOR)(1 << 9)
#define LYNX_SS   (REGISTER_BITVECTOR)(1 << 10)
#define LYNX_DS   (REGISTER_BITVECTOR)(1 << 11)
#define LYNX_FS   (REGISTER_BITVECTOR)(1 << 12)
#define LYNX_GS   (REGISTER_BITVECTOR)(1 << 13)
#define LYNX_ALL  (REGISTER_BITVECTOR)(LYNX_EAX | LYNX_ECX | LYNX_EBX | LYNX_EDX | LYNX_ESP | \
		   LYNX_EBP | LYNX_ESI | LYNX_EDI)
#define	LYNX_SEGMENT_REGS	(LYNX_ES | LYNX_CS | LYNX_SS | LYNX_DS | LYNX_FS | LYNX_GS)

/*
 * 16 bit registers
 */
#define LYNX_AX   (REGISTER_BITVECTOR)(1 << 14)
#define LYNX_CX   (REGISTER_BITVECTOR)(1 << 15)
#define LYNX_DX   (REGISTER_BITVECTOR)(1 << 16)
#define LYNX_BX   (REGISTER_BITVECTOR)(1 << 17)
#define LYNX_SP   (REGISTER_BITVECTOR)(1 << 18)
#define LYNX_BP   (REGISTER_BITVECTOR)(1 << 19)
#define LYNX_SI   (REGISTER_BITVECTOR)(1 << 20)
#define LYNX_DI   (REGISTER_BITVECTOR)(1 << 21)

/*
 * 8 bit registers
 */
#define LYNX_AL   (REGISTER_BITVECTOR)(1 << 22)
#define LYNX_AH   (REGISTER_BITVECTOR)(1 << 23)
#define LYNX_CL   (REGISTER_BITVECTOR)(1 << 24)
#define LYNX_CH   (REGISTER_BITVECTOR)(1 << 25)
#define LYNX_DL   (REGISTER_BITVECTOR)(1 << 26)
#define LYNX_DH   (REGISTER_BITVECTOR)(1 << 27)
#define LYNX_BL   (REGISTER_BITVECTOR)(1 << 28)
#define LYNX_BH   (REGISTER_BITVECTOR)(1 << 29)

#define LYNX_NONE 0

/*
 *  Floating Point Registers, need LLU suffix on 1 for 32-bit machine
 *  to handle 64 bit numbers
 */

#define LYNX_ST0 (REGISTER_BITVECTOR)(1LLU << 32)
#define LYNX_ST1 (REGISTER_BITVECTOR)(1LLU << 33)
#define LYNX_ST2 (REGISTER_BITVECTOR)(1LLU << 34)
#define LYNX_ST3 (REGISTER_BITVECTOR)(1LLU << 35)
#define LYNX_ST4 (REGISTER_BITVECTOR)(1LLU << 36)
#define LYNX_ST5 (REGISTER_BITVECTOR)(1LLU << 37)
#define LYNX_ST6 (REGISTER_BITVECTOR)(1LLU << 38)
#define LYNX_ST7 (REGISTER_BITVECTOR)(1LLU << 39)
#define LYNX_ST8 (REGISTER_BITVECTOR)(1LLU << 40)
#define LYNX_ALL_FPU (REGISTER_BITVECTOR) (LYNX_ST0 | LYNX_ST1 | LYNX_ST2 | LYNX_ST3 | LYNX_ST4 | \
    LYNX_ST5 | LYNX_ST6 | LYNX_ST7 | LYNX_ST8)


/*
 * Data types
 */
#define MAX_MODULE_NAMES	5000			// maximum module table size
#define MAX_FUNCTION_NAMES	10000			// maximum function table size

char *module_names[MAX_MODULE_NAMES];
char *function_names[MAX_FUNCTION_NAMES];


#define module_names_file	    "module_names"
#define function_names_file	    "function_nemes"

void InitializeModuleTable(char **module_names);
char *GetModuleNameById(int id);
int ModuleNameExists(char *m_name);
int AddModuleNametoTable(char *m_name);

void InitializeFunctionNameTable(char **function_names);
char *GetFunctionNameById(int id);
int FunctionNameExists(char *f_name);
int AddFunctionNametoTable(char *f_name);



#define     FlagsUsedKnown      (1 << 0)
#define     FlagsDefKnown       (1 << 1)
#define     MemRegUsedKnown     (1 << 2)          //mem regs used already stored
#define     NonMemRegUsedKnown  (1 << 3)          //non mem regs used already stored
#define     RegDefKnown         (1 << 4)          //regs def already stored
#define     MemUsedKnown        (1 << 5)          //mem used stored in abbreviated form
#define     MemDefKnown         (1 << 6)          //mem def stored in abbreviated form


/*
 * ins_structure -- a struct that contains information about the structure
 * of the instruction, namely, its mnemonic, operands, etc.  This is an
 * excerpted version of the structure ud_t from udis86.
 */
typedef struct ins_structure {
    enum ud_mnemonic_code mnemonic;
    struct ud_operand operand[3];
    unsigned char bytes[MAX_INSTRLEN];
    uint8_t length;
    //char name[MAX_INSTR_NAME_LEN];
    enum ud_type pfx_seg; // segment register prefix, if present
} ins_structure;


/* Instruction struct
 *      IMPORTANT: Pointers are not allowed for fields in this struct
 *
 *      notes: name, module, and funcName lengths are hardcoded to
 *      standardize size of Instruction when stored in file
 */
typedef struct Instruction {
    ADDRESS                 addr;
    ins_structure           uInstr;
    int                     module;
    //char                  funcName[MAX_FUNC_NAME_LEN];
    int                     funcName;
    //int                   count;
    uint64_t                flags;
    unsigned int            eax;
    unsigned int            ecx;
    unsigned int            edx;
    unsigned int            ebx;
    unsigned int            esp;
    unsigned int            ebp;
    unsigned int            esi;
    unsigned int            edi;
    int                     order;
    int                     tid; //represents the threadID of an insruction in the trace
    uint8_t                    regFlagsKnown;
    /*bool                    flagsUsedKnown; //flags used already stored in flagsUsed
    bool                    flagsDefKnown; //flags def already stored
    bool                    memRegUsedKnown; //mem regs used already stored
    bool                    nonMemRegUsedKnown; //non mem regs used already stored
    bool                    regDefKnown; //regs def already stored
    bool                    memUsedKnown; //mem used stored in abbreviated form
    bool                    memDefKnown; //mem def stored in abbreviated form*/
    REGISTER_BITVECTOR      memRegUsed;
    REGISTER_BITVECTOR      nonMemRegUsed;
    REGISTER_BITVECTOR      regDef;
    PSW_BITVECTOR           flagsUsed;
    PSW_BITVECTOR           flagsDef;
    uint32_t                memUsedMin;
    uint8_t                 memUsedSize;
    int32_t					memUsedVal;
    uint32_t                memDefMin;
    uint8_t                 memDefSize;
    int32_t					memDefVal;
    int                     next; // pointing to next instruction(used for expanded instructions)!
    int                     prev; // pointer to the previous instruction as well
    int                     ControlDependentOn;     // keeps the instruction order that this instruction is control dependent on!
    int     				origNext;
    int     				origPrev;
} Instruction;

typedef struct PageTable {
    int                     startAddr; // holding start address of each memory range
    char                    *pageAddr; // holding its associated page address allocated in run-time
    int                     numPages;
} PageTable;

typedef struct AddrRange{
    uint64_t        start;              // end addr is start + PAGE_SIZE!
    uint64_t        end;
    struct AddrRange *next;
} AddrRange;

typedef struct InstrList {
    uint32_t            numInstrs;
    Instruction         *newInstrCache;
    uint32_t            newFirstInstr;
    uint32_t            newLastInstr;
    Instruction         *instrCache;
    int            firstInstr;
    int            lastInstr; // in a page, used for chace
    int            currentInstr;
    int            lastFetchedInstr;
    int            last; // last instr's order in the trace
    int            lastBaseModuelInstr;      // keeps the order of last instr of base program
    int	    		startMarker;
    int 			endMarker;					// these two variables keep the start and end instruction of real main function
    int            numBaseModuleInstrs;
    int					stackBottom;				// contains the highest esp value, setting by the first instruction in the trace
    int                 loop;
    int                 tid; // holds threadID of an instruction
    int                 threadID; // holds the index of current thread in the list
    int                 rangeID; // the range id of a specific thread in the list
    ADDRESS				nextAvailabelAddr;			// keeps the biggest addr in base module instrs
    PageTable           *pageDir;
    struct BytetoByteMap		*constantPageDir;
    //struct BytetoByteMap		*memoryState;
    int					constantNumPageEntries;				// hold total number of pages a program access


    /*union {
        int               numAddrRanges;
        struct AddrRange *addrRangeHead;
    };*/
    union {
        Instruction     *iMMap;         //pointer to mmap of instructions
        FILE            *iListFile;
        ArrayList       *iListMem;
    };
} InstrList;

typedef struct LiveInfo {
    PSW_BITVECTOR           FlagsLive;
    REGISTER_BITVECTOR      RegsLive;
    PageTable               *MemLive;
} LiveInfo;

/*********************************************************************
 *                           BASIC BLOCKS                            *
 *********************************************************************/

typedef enum {
    BT_NORMAL, BT_COND, BT_JUMP, BT_CALL, BT_RET, BT_EXCEPTION, BT_EXCEPTION_HANDLER
} BblType;

typedef struct Block {
    int                 id;
    BblType             btype;
    ADDRESS             addr; // address of first instruction -- used for lookups
    ArrayList           *instrs; // the instructions in the block: arraylist of Instruction
    ArrayList           *preds, *succs; // predecessors and successors: arraylist of Edge
    int                 count;
    uint32_t            flags;
    struct Block        *next; // hash chain
} Block;

/*
 * The following macros try to disguise the horrible way we abuse the preds and
 * succs fields of "special" basic blocks.
 */
#define SplBblModName(ibbl)  (BBL_HAS_FLAG(ibbl,BBL_IS_DLL) ? (char *)al_get(ibbl->preds,0) : "-error-")
#define SplBblFnName(ibbl)   (BBL_HAS_FLAG(ibbl,BBL_IS_DLL) ? (char *)al_get(ibbl->succs,0) : "-error-")

/*********************************************************************
 *                               EDGES                               *
 *********************************************************************/

typedef enum {
    ET_NORMAL, ET_COND_TAKEN, ET_COND_NOTTAKEN, ET_JUMP, ET_CALL, ET_LINK, // to connect calls and exceptions to the instr at the return address
    ET_RET,
    ET_EXCEPTION, // points to the next non-ignored instruction in the trace
    ET_EXCEPTION_EXEC, // points to the next instruction in the trace
    ET_UNEXPECTED,
    ET_UNKNOWN, // for DLL code where instr bytes are not available
    ET_UNDEF
// for edges where the type has not yet been defined
} EdgeType;

/*
 * iEdge -- edges between instructions
 */
typedef struct iEdge {
    EdgeType            etype;
    int                 id;
    Instruction         *from;
    Instruction         *to;
} iEdge;

typedef struct iEdgeNode {
    iEdge               *edge;
    struct iEdgeNode    *next;
    struct iEdgeNode    *prev;
} iEdgeNode;

typedef struct udEdge {
    EdgeType            etype;
    int                 id;
    ADDRESS             fromAddr;
    ADDRESS             toAddr;
} udEdge;

/*
 * Edge -- edges between basic blocks
 */
typedef struct Edge {
    EdgeType            etype;
    int                 id;
    Block               *from;
    Block               *to;
} Edge;

typedef struct EdgeNode {
    Edge                *edge;
    struct EdgeNode     *next;
    struct EdgeNode     *prev;
} EdgeNode;

/*********************************************************************
 *                               MISC.                               *
 *********************************************************************/

// represents a range of values between min and max (inclusive)
typedef struct Range {
    uint            max;
    uint            min;
    uint            flag;
} Range;

typedef struct WriteSet {
    ArrayList       *ranges;
} WriteSet;

typedef struct AddrKey {
    ADDRESS         ADDR;
} AddrKey;

// tracks constant values through forward propagation
typedef struct Constant {
    REGISTER_BITVECTOR      reg;
    char                    reg_fields; //ex eax = 00001111, ax = 00000011, ah = 00000010, al = 00000001
    WriteSet                *mem;
    int                     size;
    uint64_t                val;
} Constant;

typedef struct _Queue {
    void                *key;
    struct _Queue       *prev;
    struct _Queue       *next;
} Queue;

typedef struct Thread {
    ArrayList 	*ranges;
    int 		tid;
    uint64_t	numInstrs;
    int 		numRanges; // needed to recover thread strutures from .InstrFile!
    uint64_t 	numBaseModulInstrs;
} Thread;

typedef struct ThreadRange {
    uint64_t first;
    uint64_t last;
} ThreadRange;

typedef struct InstrFileHeader {
    //int       InstrFileHeaderSize;            // size of this header! may vary depending when dealing with a multithread trace!
	char 		baseModuleName[100];
    uint64_t NumInstrs; // total number of instrs stored in the .InstrFile
    uint64_t LastInstrInOriginalTrace; // last instruction order of the original trace, NOTE: we add instrs to the end of file                                  during simplification
    uint64_t	lastBaseModule;
    int Loop; // last iteration of simplification

    int TraceIsMultiThreaded; // stuff to recover threads information for multithreaded traces.
    int NumThreads;
    uint64_t RangeInfoOffset;
    size_t RangeInfoSize;
    int stackButtom;					// save the stack buttom of a trace in its instrFile file header
    int nextAvailAddr;
    uint64_t startMarker;
    uint64_t endMarker;

    //bool      Simplified;                     // indicates whether the file is simplified or not
} InstrFileHeader;

/*
 * Macro Definitions
 */

// use size of long to determine if 32 bit or 64 bit system
#define IS32BIT (sizeof(long) == 4 ? true: false)

// size of void * in bytes
#if 0
#define PTRSIZE (sizeof(long))
#else
#define PTRSIZE (sizeof(int))
#endif

extern uint64_t numInstrs; // global variable holding the total number of instructions

/*
 * Function declarations
 */

Range *RangeCreate(int min, int max);
void RangeDestroy(Range *r);
WriteSet *WriteSetCreate();
WriteSet *CloneWriteSet(WriteSet *orig);
void WriteSetDestroy(WriteSet *ws);
void AddRangeToWriteSet(WriteSet *ws, Range *r);
WriteSet *SubtractRangeFromWriteSet(WriteSet *ws, Range *r);
WriteSet *MergeWriteSets(WriteSet *ws1, WriteSet *ws2);
WriteSet *IntersectWriteSets(WriteSet *ws1, WriteSet *ws2);
WriteSet *UnionWriteSets(WriteSet *ws1, WriteSet *ws2);
bool WriteSetsOverlap(WriteSet *ws1, WriteSet *ws2);
WriteSet *SubtractWriteSets(WriteSet *ws1, WriteSet *ws2);
bool WriteSetIsEmpty(WriteSet *ws);
bool ValInWriteSet(WriteSet *ws, int val);
bool RangeInWriteSet(WriteSet *ws, Range *r);
bool WriteSetEquals(WriteSet *ws1, WriteSet *ws2);
void fPrintWriteSet(FILE *fp, WriteSet *ws);
void PrintWriteSet(WriteSet *ws);
WriteSet *CalcWriteSetFromMask(WriteSet *ws, uint32_t mask);
Range *IntersectRange(Range *r1, Range *r2);

Block *NewBbl();
void DestroyBbl(Block *bbl, bool destroyIns);
void DestroyBblList(ArrayList *bblList, bool destroyIns);
Block *CreateSpecialBbl(char *module, char *fn);
Block *GetBblById(int id);
Instruction *InstrCreate();
void InstrDestroy();
Instruction *InstrClone(Instruction *instr);
iEdge *iEdgeCreate();
void iEdgeDestroy(iEdge *edge);
iEdge *iEdgeClone(iEdge *edge);
iEdge *NewiEdge(EdgeType etype, Instruction *from, Instruction *to);
Edge *EdgeCreate();
void EdgeDestroy(Edge *edge);
Edge *NewEdge(EdgeType etype, Block *from, Block *to);
iEdgeNode *iEdgeNodeCreate();
EdgeType GetiEdgeType(Instruction *from, Instruction *to);
bool InstrIsControlTransfer(InstrList *iList, Instruction *iins);
Instruction *FindInstrInArrayList(ArrayList *list, Instruction *instr);
Instruction *FindInstrInArrayListByAddr(ArrayList *list, ADDRESS addr);
int FindInstrPosInArrayList(ArrayList *iList, Instruction *instr);
int FindInstrPosInArrayListByOrder(ArrayList *iList, Instruction *instr);
Instruction *FindInstrInList(InstrList *iList, Instruction *instr);
udEdge *udEdgeCreate(EdgeType etype, ADDRESS fromAddr, ADDRESS toAddr);
udEdge *FindudEdgeInArrayList(ArrayList *list, ADDRESS from, ADDRESS to);
bool InstrInList(InstrList *iList, Instruction *instr);
bool InstrInArraylist(ArrayList *list, Instruction *instr);
iEdgeNode *AddiEdgeToList(iEdgeNode *eList, iEdge *edge);
bool InstrIsEqual(Instruction *instr1, Instruction *instr2);
char *InstrToString(Instruction *instr);

bool WriteSetSubsetOfWriteSet(WriteSet *ws_sub, WriteSet *ws_super);
uint32_t RangeOverlapMask(Range *orig, Range *overlap);
WriteSet *MaskWriteSet(WriteSet *ws, uint32_t mask);
bool RangeIsEmpty(Range *r);
bool RangeOverlapWithWriteSet(WriteSet *ws, Range* r);

PageTable *ClonePageTables(PageTable *orig);
void DestroyPageTables(PageTable *pt);
bool RangeIsinRange(Range *s, Range *r);    // is r in s?

/*
 * Structures needed for multithreading
 */

typedef struct threadArgs {
    uint64_t first;
    uint64_t last;
    InstrList *i_list;
    LiveInfo *assumeLive;
    LiveInfo *assumeDead;
} ThreadArgs;


typedef struct ControlInstr{
    Instruction     *iins;
    bool            Occurs;
    uint64_t        Start;
    uint64_t        End;

} ControlInstr;

bool handle_rop;

#ifdef __cplusplus
}
#endif

#endif /* _LYNX_TYPES_H_ */
