/*
 *
 * File: taint_analysis.h
 * Author: Babak Yadegari
 */

/*
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */




#ifdef __cplusplus
#include <vector>
#include <string>
#include <sstream>
using std::vector;

//extern "C" {
#endif

#include "instr.h"
#include "flags.h"
#include "lynx_types.h"
#include "simplify.h"
#include "mem_maps.h"

#ifndef TAINT_ANALYSIS_H_
#define TAINT_ANALYSIS_H_

#define DATAFLAG_CONST			0x0			// reserved (not used) in PSW_FLAGS!
#define DATAFLAG_UNKNOWN		(unsigned)1 << 31


#define TAINT_GENERIC       	(uint)-1      // for marking taints from input values (use a reserved flag)
#define TAINT_NONE				0x0         // not tainted

typedef struct TaintedRes{
	int regs;
	int regsMask[32];               //eaxMask, edxMask, ecxMask, ebxMask, espMask, ebpMask, esiMask, ediMask;
	int flags;
} TaintedRes;

typedef struct TaintedNoFlagRes{
    int         regs;
    int         regsMask[32];               //eaxMask, edxMask, ecxMask, ebxMask, espMask, ebpMask, esiMask, ediMask;
} TaintedNoFlagRes;


//////////////////////////// bit-level flag data structures /////////////
typedef struct MemBitVars{
    int addr;
    int size;                   // in bytes
    int *bits;
} MemBitVars;
/////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
class Taint{
public:
    uint64_t    tag;                // the tag indicated the taint tag, generic tag used for non-flag
                                    // taints, and (1 << flag_pos) for flag bits used in registers
                                    // and memory locations
    uint32_t    src_instr;
    uint8_t     polarity;           // for flags, original flag or negated!
    Taint(uint64_t t, uint32_t src): tag(t), src_instr(src), polarity(0) {}

    bool Tainted() const{
    	return tag!=DATAFLAG_CONST;
    }
    
    bool operator==(const Taint &o) const{
    	return (tag == DATAFLAG_CONST && o.tag == DATAFLAG_CONST) || (tag==o.tag /*&& src_instr==o.src_instr*/ && polarity==o.polarity);
    }

    Taint operator!(void) const{
    	Taint o = (*this);
    	if(tag != DATAFLAG_CONST)
    		o.polarity = !o.polarity;		// flip the polarity
		return o;
	}

    Taint &Xor(const Taint &o, uint c0, uint c1){
    	if(tag != DATAFLAG_UNKNOWN && ((*this) == o || (*this == !o))){
    		tag = DATAFLAG_CONST;
    	} else if(tag == DATAFLAG_CONST){				// o is tainted then
    		if(c0)
    			(*this) = !o;
    		else
    			(*this) = o;
    	} else if (o.tag == DATAFLAG_CONST){
    		if(c1)
    			(*this) = !(*this);
    	} else {
    		tag = DATAFLAG_UNKNOWN;			// may need a new label here?
    	}
		return *this;
	}

    Taint &operator&(const Taint &o){
    	if((tag == DATAFLAG_CONST && o.tag == DATAFLAG_CONST) || (*this) == !o)
    		tag = DATAFLAG_CONST;
    	else
    		tag = DATAFLAG_UNKNOWN;
		return *this;
	}

    Taint &operator|(const Taint &o){
    	if((tag == DATAFLAG_CONST && o.tag == DATAFLAG_CONST) || (*this) == !o)
    		tag = DATAFLAG_CONST;
    	else
    		tag = DATAFLAG_UNKNOWN;
		return *this;
	}

    /*
     *
     * c0 is the concrete value of this and c1 is the concrete
     * value of o in case not tainted.
     */
    Taint &Add(const Taint &o, Taint &c, uint c0, uint c1){
		// simple truth table

    	if(this->Tainted() && o.Tainted() && c.Tainted()){
    		Taint tmp = o;
    		tmp.Add(c, c.src_instr, c1);
    		this->Add(c, c0, c1);
    	} else {
    		if(this->Tainted() && o.Tainted()){
    			Taint tmp = o;					// since c is not tainted
    			this->Add(tmp, 0, 0);
    			c = tmp;
    		} else if (this->Tainted() && c.Tainted()){
    			this->Add(c, c0, c1);
    		} else if (c.Tainted() && o.Tainted()){
    			Taint tmp = o;
    			tmp.Add(c, 0, c0);
    			(*this) = tmp;
    		} else if (o.Tainted()){
    			Taint tmp = o;
    			tmp.Add(c, 0, c0);
    			(*this) = tmp;
    		} else {			// none tainted or (c.Tainted() && !this->Tainted()) or (!c.Tainted() && this->Tainted())
    			this->Add(c, c0, c1);
    		}
    	}
    	return (*this);
    	/*if (((*this) == o && c == o) ){
            if(tag == DATAFLAG_CONST && c0){
                c.src_instr = !c.src_instr;
            }
        } else if (((*this) == o && c == !o) || ((*this) == c && (*this) == !o)){
			polarity = !polarity;
			c.polarity = !c.polarity;
    	} else if(o == c && (*this) == !o) {		//tag != o.tag
    		if(polarity == !c.polarity){
    			c.polarity = !polarity;
    		}
        } else if(tag == DATAFLAG_CONST) {
            (*this) = o;
            this->Add(c, c0, c1);
        } else if(o.tag == DATAFLAG_CONST){
            this->Add(c, c0, c1);
        } else {
            tag = DATAFLAG_UNKNOWN;
            c.tag = DATAFLAG_CONST;
            c.src_instr = 0;
            c.polarity = 0;
        }*/

    }

    /*
     * Add(Taint &c, uint c0, uint cons)
     * adds a tainted bit with previous carry and a cons
     * cons is bit, 0 or 1. If there is a carry because
     * of this operation, it is saved in c. c0 is the constant
     * value of the tainted bit (this) in case it is not
     * tainted. If carry (produced or input) is not tainted,
     * then c.src_instr is significant, showing the concrete
     * vale of the carry (0 if src_instr = 0, 1 otherwise)
     */
    Taint &Add(Taint &c, uint c0, uint cons){
        if(this->Tainted()){
            if(!c.Tainted()){
                if((cons && c.src_instr == 0) || (!cons && c.src_instr == 1)){		// t+1
                	(*this) = !(*this);
                    //c = (*this);

                    c.src_instr = 0;
                } else if(cons && c.src_instr == 1){			// t+1+1
                	// a constant 1 will be carried over
                	c.src_instr = 1;
                }
            } else {
				if(c == (*this) && c.tag != TAINT_GENERIC){
					c = (*this);
					tag = DATAFLAG_CONST;
					polarity = 0;
				} else if((*this) == !c && c.tag != TAINT_GENERIC){
					c.tag = DATAFLAG_CONST;
					c.polarity = 0;
					tag = DATAFLAG_CONST;
					polarity = 0;
				}
            }
        } else {
            if(!c.Tainted()){
                if((cons && c.src_instr == 1) || (cons && c0) || (c0 && c.src_instr == 1)){
                    c.src_instr = 1;
                } else {
                	c.src_instr = 0;
                }
            } else {
            	this->src_instr = 0;
                c.Add((*this), cons, c0);
                Taint tmp = c;
                c = (*this);
                (*this) = tmp;
            }
        }
        
        
        
    	/*if(!cons){
            if(tag != DATAFLAG_CONST){
                if(c == (*this)){
                    tag = DATAFLAG_CONST;
                    polarity = 0;
                } else if((*this) == !c){
                    c.tag = DATAFLAG_CONST;
                    c.polarity = 0;
                    tag = DATAFLAG_CONST;
                    polarity = 0;
                }
			} else {
				if(c.tag == DATAFLAG_CONST){
					//
				} else {
					if(tag != DATAFLAG_CONST){
						tag = DATAFLAG_UNKNOWN;
						polarity = 0;
					} else {
						(*this) = c;
						c.tag = DATAFLAG_CONST;
						c.polarity = 0;
					}
				}
			}
    	} else {
			if(c == (*this)){
				c = !(*this);
				tag = DATAFLAG_CONST;
				polarity = 0;
			} else if((*this) == !c){
				c.tag = DATAFLAG_CONST;
				c.polarity = 0;
				tag = DATAFLAG_CONST;
				polarity = 0;
			} else {
				if(c.tag == DATAFLAG_CONST){
					c = (*this);
					polarity = !polarity;
				} else {
					if(tag != DATAFLAG_CONST){
						tag = DATAFLAG_UNKNOWN;
						polarity = 0;
					} else {
						(*this) = !c;
						c.tag = DATAFLAG_CONST;
						c.polarity = 0;
					}
				}
			}
    	}*/
        return (*this);
    }

};

class TByte{
	vector<Taint> t;

public:
	TByte(): t(8, Taint(DATAFLAG_CONST, 0)){}
	TByte(uint64_t tag, uint32_t src): t(8, Taint(tag, src)){}
	Taint& operator[](int index){
		assert(index >= 0 && index < 8);
		return t[index];
	}
	bool Tainted(){
		for (int i=0; i<8; i++){
			if (t[i].tag != DATAFLAG_CONST) return true;
		}
		return false;
	}
};

/*
 * one or more TaintByte objects. This object is just
 * a placeholder for TBytes, makes it easier to use
 * with operator[] as well as implementing operational
 * semantics for taint propagation purposes!
 */
class TaintBlock{
private:
	vector<TByte *> bytes;
	uint size;

public:
	TaintBlock():bytes(0), size(0){}
	TaintBlock(uint s, uint64_t tag, uint src=0):size(0){
		for(uint i=0; i<s; i++){
			this->AddByte(new TByte(tag, src));
		}
	}
	Taint& operator[](uint index) const {
		assert(index >= 0 && index < size*8);
		return (*bytes[(index >> 3)])[index&7];
	}
	void AddByte(TByte *t){
		bytes.push_back(t);
		size++;
	}
	uint Size() const {return size;}
	TaintBlock &operator=(TaintBlock &o){
		if(this == &o) return *this;
		assert(size >= o.Size());
		for(uint i=0; i < o.Size()*8; i++){
			(*this)[i] = o[i];
		}
		return *this;
	}
	TByte *GetTByte(uint i) const{
		assert(i >= 0 && i < size);
		return bytes[i];
	}
	bool Tainted() const{
		for(uint i=0; i < size; i++){
			if(bytes[i]->Tainted()) return true;
		}
		return false;
	}

	// operators: actually implement the operational semantics of instructions!

	TaintBlock &operator!(void){
		for(uint i=0; i < size*8; i++){
			(*this)[i] = !(*this)[i];
		}
		return *this;
	}

	TaintBlock &operator~(void){
		/*for(uint i=0; i < size*8; i++){
			(*this)[i] = (*this)[i];
		}*/
		return (!(*this)).Add(1, 0);
	}

	TaintBlock &operator|=(TaintBlock &o){
		assert(o.Size() <= size);
		for(uint i=0; i < size*8; i++){
			(*this)[i] = (*this)[i] | o[i];
		}
		return *this;
	}
	friend TaintBlock &operator|(TaintBlock &o1, TaintBlock &o2){
		return o1 |= o2;
	}

	TaintBlock &operator|=(uint c){
		for(uint i=0; i < size*8; i++){
			uint mask = (1 << i)&c;
			if(mask)
				(*this)[i].tag = DATAFLAG_CONST;
		}
		return *this;
	}

	friend TaintBlock &operator|(TaintBlock &o1, uint c){
		return o1 |= c;
	}

	TaintBlock &operator^(uint c){
		for(uint i=0; i < size*8; i++){
			uint mask = (1 << i)&c;
			if(mask && (*this)[i].Tainted()){
				(*this)[i] = !(*this)[i];
			}
		}
		return *this;
	}

	TaintBlock &Xor(const TaintBlock &o, uint c0, uint c1){
		assert(o.Size() <= size);
		uint mask = 1;
		for(uint i=0; i < size*8; i++){
            (*this)[i].Xor(o[i], c0&mask, c1&mask);
            mask <<= 1;
		}
		return *this;
	}

	TaintBlock &operator&=(const TaintBlock &o){
		assert(o.Size() <= size);
		for(uint i=0; i < size*8; i++){
			(*this)[i] = (*this)[i] & o[i];
		}
		return *this;
	}

	friend TaintBlock &operator&(TaintBlock &o1, TaintBlock &o2){
		return o1 &= o2;
	}

	TaintBlock &operator&=(uint c){
		for(uint i=0; i < size*8; i++){
			uint mask = (1 << i)&c;
			if(mask == 0)
				(*this)[i].tag = DATAFLAG_CONST;
		}
		return *this;
	}

	friend TaintBlock &operator&(TaintBlock &o1, uint c){
		return o1 &= c;
	}

	TaintBlock &operator>>(uint shr){
		if(!shr)
			return *this;
		uint i;
		for (i = shr%(size*8); i < size*8 - shr%(size*8); i++){
			(*this)[i-shr%(size*8)] = (*this)[i];
		}
		for (; i < size*8; i++){
			(*this)[i].tag = DATAFLAG_CONST;
		}
		return *this;
	}

	TaintBlock &operator<<(uint shl){
		if(!shl){
			return *this;
		}
		int i;
        for (i = size*8-1; i >= (int)(shl%(size*8)); i--){
        	(*this)[i] = (*this)[i-shl%(size*8)];
        }
        for (; i >= 0; i--){
        	(*this)[i].tag = DATAFLAG_CONST;
        }
        return *this;
	}

	TaintBlock &Add(const TaintBlock &o, uint c0, uint c1){
		assert(o.Size() <= size);
		uint i, mask=1;
		Taint carry(DATAFLAG_CONST, 0);     // use carry.src_instr as the concrete value of carry
        
        for (i = 0; i < size*8; i++){
        	(*this)[i].Add(o[i], carry, c0&mask, c1&mask);
            mask <<= 1;
        	/*uint mask = (1 << i)&c;
        	if((*this)[i].Tainted()){
				if(o[i].Tainted()){
					if(carry.Tainted()){
	        			Taint tmp = carry;
	        			tmp = (*this)[i] & carry & o[i];
	        			(*this)[i] = carry ^ (*this)[i] ^ o[i];
	        			(*this)[i] = ~(*this)[i];			// for +1
	        			carry = tmp;
					} else {
						carry = (*this)[i] & o[i];
					}
				} else if (carry.Tainted()){
					if(mask){
	        			Taint tmp = carry;
	        			tmp = (*this)[i] & carry;
	        			(*this)[i] ^= carry;
	        			(*this)[i] = ~(*this)[i];			// for +1
	        			carry = tmp;
					} else {
						(*this)[i] = carry;
						carry.tag = DATAFLAG_CONST;
					}
				}
        	} else {
				if(o[i].Tainted()){
					if(carry.Tainted()){
	        			Taint tmp = carry;
	        			tmp = (*this)[i] & carry & o[i];
	        			(*this)[i] = carry ^ (*this)[i] ^ o[i];
	        			(*this)[i] = ~(*this)[i];			// for +1
	        			carry = tmp;
					} else {
						carry = (*this)[i] & o[i];
					}
				} else if (carry.Tainted()){
					if(mask){
	        			Taint tmp = carry;
	        			tmp = (*this)[i] & carry;
	        			(*this)[i] ^= carry;
	        			(*this)[i] = ~(*this)[i];			// for +1
	        			carry = tmp;
					} else {
						(*this)[i] = carry;
						carry.tag = DATAFLAG_CONST;
					}
				}
        	}*/
        }
        return *this;
	}

	TaintBlock &Add(uint c0, uint c1){
		uint i;
        uint mask = 1;
		Taint carry(DATAFLAG_CONST, 0);
        for (i = 0; i < size*8; i++){
            (*this)[i].Add(carry, c0&mask, c1&mask);
            mask <<= 1;
        }
        return *this;
	}

	/*friend TaintBlock &Add+(TaintBlock &o1, uint c0, unint c1){
		return o1 += c;
	}*/

	TaintBlock &Adc(TaintBlock &o, uint c0, uint c1, const TaintBlock &flags){
		uint i, mask = 1;
		Taint carry = flags[0];
        for (i = 0; i < size*8; i++){
        	carry = (*this)[i].Add(o[i], carry, c0&mask, c1&mask);
            mask <<= 1;
        }
		return *this;
	}

	TaintBlock &Adc(uint c0, uint c1, const TaintBlock &flags){
		uint i;
		Taint carry = flags[0];
        for (i = 0; i < size*8; i++){
        	uint mask = (1 << i);
        	carry = (*this)[i].Add(carry, c0&mask, c1&mask);
        }
		return *this;
	}

	// in fact, rotate right our array
	TaintBlock &Rol(uint c){
		uint i;
		vector<Taint> tmpArr(size*8, Taint(0, 0));
        for (i = 0; i < size*8; i++){
       		tmpArr[(i+c)%(size*8)] = (*this)[i];
        }
        for (i = 0; i < size*8; i++){
			(*this)[i] = tmpArr[i];
		}
		return *this;
	}

	// in fact, rotate left our array
	TaintBlock &Ror(uint c){
		uint i;
		vector<Taint> tmpArr(size*8, Taint(0,0));
        for (i = 0; i < size*8; i++){
        	tmpArr[i] = (*this)[(i+c)%(size*8)];
        }
        for (i = 0; i < size*8; i++){
			(*this)[i] = tmpArr[i];
		}
		return *this;
	}


	std::string toString(){
		std::ostringstream ss;
		for(int i=size*8-1; i >= 0; i--){
			/*if((*this)[i].Tainted())
				ss << "1 ";
			else
				ss << "0 ";*/
			if((*this)[i].polarity && (*this)[i].tag != DATAFLAG_CONST) ss << "!";
			ss << std::hex << (*this)[i].tag << std::dec << " ";
			if(i%4==0){
				ss << "  ";
			}
		}
		return ss.str();
	}


};




typedef struct BitDataDep{
    REGISTER_BITVECTOR  taintedRegs;
    TByte 		regsTaint[32];			// 8 32-bit register, an integer for every bit!
    //ArrayList *memBits;
    REGISTER_BITVECTOR	taintedFlags;
    TaintBlock*		flags;
} BitDataDep;
#endif

#if __cplusplus
extern "C"
#endif
void DebugBackwardTaintPropagationCopy(FILE *f, InstrList *iList, int StartIndex, int upperBound);

#if __cplusplus
extern "C"
#endif
void TaintAnalysis(InstrList *iList, int method, int doBackward);
//extern "C" void BackwardTaintPropagationNew(InstrList *iList, int StartIndex);
//extern "C" void ForwardTaintPropagation(InstrList *iList, int StartIndex, uint64_t n);


//TaintedNoFlagRes *InitializeTaintedNonFlagResStruct(TaintedNoFlagRes *taints);
//void ApplyRegTaintWithMaskNoFlag(TaintedNoFlagRes *taints, int taint, REGISTER_BITVECTOR reg);
//void DebugBackwardTaintPropagationCopy(FILE *f, InstrList *iList, int StartIndex, int upperBound);


/*
 * temporary definitions
 */
#define TAINT_STANDARD  1
#define TAINT_BITLEVEL  2
#define TAINT_ENHANCED  3


#ifdef __cplusplus

#endif

#endif /* TAINT_ANALYSIS_H_ */
