/*
 *
 * File: flowgraph.c
 * Author: Babak Yadegari, Saumya Debray
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#include "flowgraph.h"
#include "simplify.h"
#include "hash.h"
#include "flags.h"
#include "print.h"
#include "mem_maps.h"

iListNode *conditionalsTarget[HASHTABSZ];
iListNode *hashtbl[HASHTABSZ], *addrhashtbl[HASHTABSZ];
iListNode *Conditionals_hl = NULL, *Conditionals_tl = NULL;
iListNode *ins_list = NULL, *ins_list_tl = NULL; // instructions inserted in hash by order
iListNode *ins_list_byaddr = NULL, *ins_list_tl_byaddr = NULL;
Bbl *bbl_list_hd = NULL, *bbl_list_tl = NULL; // basic blocks
SuperBbl *superbbl_list_hd = NULL, *superbbl_list_tl = NULL;
BblEdge *bbledge_hd, *bbledge_tl;

int bbl_id = 0;

extern char *baseName;

static int hash(ADDRESS addr) {
    return addr & 0x7fff; /* low 15 bits */
}


static iListNode *hashlookupByAddress(iListNode **hashtbl, Instruction *instr) {
    int hval;
    iListNode *htmp;

    if (!instr) {
        fprintf(stderr, "hashlookup: NULL argument\n");
        exit(1);
    }

    hval = hash(instr->addr);

    for (htmp = hashtbl[hval]; htmp; htmp = htmp->hnext) {
        if (htmp->iptr->ins->addr == instr->addr) {
            break;
        }
    }

    return htmp;
}

static iListNode *hashinsertByAddress(iListNode **hashtbl, Instruction *instr) {
    int hval;
    iListNode *htmp0;
    InsNode *itmp;

    if (!instr) {
        fprintf(stderr, "hashinsert: NULL argument\n");
        exit(1);
    }

    if((htmp0 = hashlookupByAddress(hashtbl, instr)) == NULL){
        itmp = calloc(1, sizeof(InsNode));
        if (!itmp) {
            fprintf(stderr, "[hashinsert]: Out of memory!\n");
            exit(1);
        }

        itmp->ins = instr;//InstrClone(instr);;
        itmp->numSuccs = itmp->numPreds = 0;
        itmp->succs = itmp->preds = NULL;


    	htmp0 = calloc(1, sizeof(iListNode));
		hval = hash(instr->addr); // add itmp to the hash table via htmp
		htmp0->iptr = itmp;
		htmp0->next = NULL;
		htmp0->prev = NULL;
		htmp0->hnext = hashtbl[hval];
		hashtbl[hval] = htmp0;
    }



    return htmp0;
}

int NumberOfPossibleSuccessorsIns(Instruction *iins){
	if(IsCondJump(iins)){
		return 2;
	} else if(iins->uInstr.mnemonic == UD_Iret){
		return 2;										// DEBUG ONLY, IT SHOULD BE -1!
	} else if(iins->uInstr.mnemonic == UD_Ijmp){
		if(iins->uInstr.operand[0].type == UD_OP_JIMM){
			return 1;
		} else {
			return -1;
		}
	}
	return 1;
}

/*********************************************************************
 *                            PROCESSING                             *
 *********************************************************************/


static iListNode *CloneiListNode(iListNode *il) {
    iListNode *itmp = malloc(sizeof(iListNode));
    if (!itmp) {
        fprintf(stderr, "Out of memory!\n");
        exit(1);
    }

    memcpy(itmp, il, sizeof(iListNode));
    itmp->next = NULL;
    itmp->prev = NULL;
    return itmp;
}


static void CollectDistinctInstructionsByAdrress(InstrList *iList) {
    uint64_t index;
    Instruction *instr;//, *next;
    uint64_t listSz = iList->numInstrs;
    iListNode *hnode;
    bool first = true, inCall = false;
    int i = 0;

    /* collect all distinct instructions */
    //SetCurrentInstr(iList, 0);

    index = iList->startMarker;

    if(iList->startMarker != 0){
    	index = index + 2;
    }
    while (!inCall && index != -1 && (index <= iList->endMarker || index > iList->last)) {

        instr = GetInstruction(iList, index);
        if(!instr){
            return;
        } else {
        	index = instr->next;
        }

        if (instr == NULL) {
            fprintf(stderr, "NULL instruction: order# = %lld [total no. of instrs = %lld]\n",
                    (unsigned long long)index,
                    (unsigned long long)listSz);
            abort();
        }
        i++;
        if(!handle_rop && !INSTR_HAS_FLAG(instr, INSTR_IN_BASE_MODULE)){
        	inCall = true;
        }
        hnode = hashinsertByAddress(addrhashtbl, instr);
        //next = GetInstruction(iList, instr->next);

        iListNode *htmp1 = calloc(1, sizeof(iListNode));
        InsNode *iptr = malloc(sizeof(InsNode));
        iptr->numPreds = iptr->numSuccs = 0;
        iptr->succs = iptr->preds = NULL;
        iptr->flags = 0;
        iptr->ins = instr;

        if (!ins_list_byaddr)
        	ins_list_byaddr = htmp1;
        htmp1->iptr = iptr; // add itmp to the global list of distinct addresses
        htmp1->prev = ins_list_tl_byaddr;
        htmp1->next = NULL;
        if (ins_list_tl_byaddr)
        	ins_list_tl_byaddr->next = htmp1;
        ins_list_tl_byaddr = htmp1;

        if (first) { // first instruction
        	NODE_SET_FLAG(hnode->iptr, BBL_ENTRY);
        	first = false;
        }

        if(inCall){
        	SetCurrentInstr(iList, instr->next);
        	Instruction *tmp;
        	while((tmp = FetchNextInstr(iList)) != NULL){
        		if(INSTR_HAS_FLAG(tmp, INSTR_IN_BASE_MODULE)){
        			index = tmp->order;
        			inCall = false;
        			break;
        		}
        	}
        }
    }
}


 /*
  * collect successors.
  */
 void CollectInstrSuccsAndPredsOld(InstrList *iList) {
	 Instruction *instr, *jnstr;
	 iListNode *hnode, *hnext, *htmp;
	 InsNode *iins, *jins;

	 //CollectDistinctInstructions(iList);
	 /*
	  * collect successors.
	  */
	 SetCurrentInstr(iList, iList->startMarker);
	 instr = FetchNextInstr(iList);
	 while ((instr = FetchNextInstr(iList)) != NULL) {
		 if(instr->order == iList->endMarker){
			 return;
		 }

		 jnstr = GetInstruction(iList, instr->next);

		 if (jnstr == NULL){
			 return;
		 }
		 /* check whether jnstr is already on instr's succs list */
		 hnode = hashlookupByAddress(addrhashtbl, instr);
		 if(!hnode){
			 continue;
		 }
		 iins = hnode->iptr;


		 hnext = hashlookupByAddress(addrhashtbl, jnstr);
		 if(hnext == NULL){
			 continue;		/* done */
		 }
		 jins = hnext->iptr;

		 for (htmp = iins->succs; htmp; htmp = htmp->next) {
			 if (htmp->iptr == jins) {
				 break;
			 }
		 }
		 if(htmp == NULL){
			 htmp = calloc(1, sizeof(iListNode));
			 jins = hnext->iptr;
			 htmp->iptr = jins;
			 htmp->next = iins->succs;
			 iins->succs = htmp;
			 iins->numSuccs ++;
		 }

		 for(htmp = jins->preds; htmp; htmp = htmp->next){
			 if(htmp->iptr == iins){
				 break;
			 }
		 }
		 if(htmp == NULL){
			 htmp = calloc(1, sizeof(iListNode));
			 htmp->iptr = iins;
			 htmp->next = jins->preds;
			 jins->preds = htmp;
			 jins->numPreds ++;
		 }
	 }
 }


static void MarkBblBoundariesByAddr(InstrList *iList) {
    iListNode *il0, *il0InHash, *il0Prev = NULL;
    bool bblStart = true;

    //CollectDistinctInstructions(iList);
    /*
     * collect successors.
     */
    il0 = ins_list_byaddr;
    while(il0){
        if (il0->iptr->ins->order == iList->endMarker){
            NODE_SET_FLAG(il0->iptr, BBL_END);
            break; /* done */
        }

        il0InHash = hashlookupByAddress(addrhashtbl, il0->iptr->ins);
        if(il0InHash == NULL){
        	return;
        }
        if (il0->iptr->ins->order == 0 || il0->iptr->ins->order == iList->startMarker+2) {
            NODE_SET_FLAG(il0->iptr, BBL_ENTRY);
            NODE_SET_FLAG(il0->iptr, BBL_START);
            il0 = il0->next;
            continue;
        }
        if(bblStart || (il0InHash->iptr->numPreds > 1 && (handle_rop || INSTR_HAS_FLAG(il0->iptr->ins, INSTR_IN_BASE_MODULE)))
        		|| (IsCondJump(il0->iptr->ins) && INSTR_HAS_FLAG(il0->iptr->ins, INSTR_FORWARD_TAINTED)
        				&& (handle_rop || INSTR_HAS_FLAG(il0->iptr->ins, INSTR_IN_BASE_MODULE)))){
            NODE_SET_FLAG(il0->iptr, BBL_START);
            if(il0Prev){
                NODE_SET_FLAG(il0Prev->iptr, BBL_END);
            }
            bblStart = false;
            /*if(il0Prev && ((il0InHash->iptr->numPreds > 1 || (IsCondJump(il0->iptr->ins)
            		&& INSTR_HAS_FLAG(il0->iptr->ins, INSTR_FORWARD_TAINTED))
            		&& INSTR_HAS_FLAG(il0->iptr->ins, INSTR_IN_BASE_MODULE))){
            	NODE_SET_FLAG(il0Prev->iptr, BBL_END);
            	bblStart = true;
            }*/
        }
        if ((InstrIsControlTransfer(iList, il0->iptr->ins) /*&& INSTR_HAS_FLAG(il0->iptr->ins, INSTR_FORWARD_TAINTED)*/) || (il0InHash->iptr->numSuccs > 1
        		&& (handle_rop || INSTR_HAS_FLAG(il0->iptr->ins, INSTR_IN_BASE_MODULE)))
        		|| (il0Prev && !INSTR_HAS_FLAG(il0Prev->iptr->ins, INSTR_IN_BASE_MODULE))) {
            NODE_SET_FLAG(il0->iptr, BBL_END);
            bblStart = true;

        }
        il0Prev = il0;
        il0 = il0->next;
    }
}

static void FindBblSuccsAndPredsOldMethod() {
    Bbl *ibbl, *jbbl;
    iListNode *il0;
    bListNode *bl0;

    for (ibbl = bbl_list_hd; ibbl && ibbl->ins_tl; ibbl = ibbl->next) {
        for (il0 = ibbl->ins_tl->iptr->succs; il0; il0 = il0->next) {
            /*
             * find the block whose first instruction is given by il0->iptr
             */
            for (jbbl = bbl_list_hd; jbbl; jbbl = jbbl->next) {    // we are treating base module and non-base module bbls differently
                //printf("bbl %d: successor: %d: [0x%08x] %s vs %d: [0x%08x] %s\n", jbbl->ins_hd->iptr->);
                if (jbbl->ins_hd->iptr->ins->addr == il0->iptr->ins->addr) {
                    break;
                }
            }

            if (jbbl == NULL) {
                continue;
            }
            /*
             * create a bListNode node for this and add it as a successor to ibbl
             */
            bl0 = calloc(1, sizeof(bListNode));
            if (!bl0) {
                fprintf(stderr, "Out of memory!\n");
                exit(1);
            }

            bl0->bptr = jbbl;
            bl0->next = ibbl->succs;
            bl0->reserved = NULL;
            ibbl->succs = bl0;
            ibbl->numSuccs++;
            for(bl0 = jbbl->preds; bl0; bl0 = bl0->next){   // add ibbl to jbbl's predecessors
                if(bl0->bptr->id == ibbl->id){
                    break;
                }
            }
            if(!bl0){
                bl0 = malloc(sizeof(bListNode));
                bl0->bptr = ibbl;
                bl0->reserved = NULL;
                bl0->next = jbbl->preds;
                jbbl->preds = bl0;
                jbbl->numPreds ++;
            }
        }
    }
}

int ConstructDistinctBblsOldMethod(InstrList *iList) {
    iListNode *il0, *iltmp, *ilnew, *htmp, *bblInstrPtr;
    Bbl *ibbl;
    InsNode *succtmp;
    int numInstrs;

    //MarkBblBoundaries(iList);

    il0 = ins_list_byaddr;
    while (il0) {

#if DEBUG_MESSAGES&0
    	if (!NODE_HAS_FLAG(il0->iptr, BBL_START)) {
        		fprintf(stderr, "Weird! il0 = %d: [0x%x] %s\n", il0->iptr->ins->order,
        				il0->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(il0->iptr->ins)));
    	}
#endif

        for (ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next) {
            if (ibbl->ins_hd->iptr->ins->addr == il0->iptr->ins->addr) {
#if DEBUG_MESSAGES&0
					printf("bbl %d exists with starting: %d: [0x%08x] %s\n", ibbl->id,
							il0->iptr->ins->order, il0->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(il0->iptr->ins)));
#endif
                al_add(ibbl->occurrences, &(il0->iptr->ins->order));
                bblInstrPtr = ibbl->ins_hd;
                numInstrs = ibbl->numInstrs;
                while (il0) {
                	if(bblInstrPtr && bblInstrPtr->iptr->ins->addr == il0->iptr->ins->addr){
                		if(INSTR_HAS_FLAG(bblInstrPtr->iptr->ins, INSTR_TMP0) && !INSTR_HAS_FLAG(il0->iptr->ins, INSTR_TMP0)){
                			bblInstrPtr->iptr->ins = il0->iptr->ins;
                		}
                	}
                    if ((IsCondJump(
                            il0->iptr->ins) || InstrIsIndirectJump(il0->iptr->ins)) && INSTR_HAS_FLAG(il0->iptr->ins, INSTR_FORWARD_TAINTED)
                            && (handle_rop || INSTR_HAS_FLAG(il0->iptr->ins, INSTR_IN_BASE_MODULE))) {
                        NODE_SET_FLAG(ibbl, BBL_IS_CONDITIONAL);
                    }
                    if (INSTR_HAS_FLAG(il0->iptr->ins, INSTR_FORWARD_TAINTED)) {
                        NODE_SET_FLAG(ibbl, BBL_FORWARD_TAINTED);
                    }
                    if(INSTR_HAS_FLAG(il0->iptr->ins, INSTR_BACKWARD_TAINTED)){
                        NODE_SET_FLAG(ibbl, BBL_BACKWARD_TAINTED);
                    }
                    if(INSTR_HAS_FLAG(il0->iptr->ins, INSTR_DOES_UNPACKING)){
						NODE_SET_FLAG(ibbl, BBL_DOES_UNPACKING);
					}
                    if(!handle_rop && !INSTR_HAS_FLAG(ibbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
						if(NODE_HAS_FLAG(il0->iptr, BBL_END)){
							break;
						}
                    } else {
                    	numInstrs --;
						if(numInstrs == 0){
							break;
						}
                    }
                    il0 = il0->next;
                }
                if(il0){
                	il0 = il0->next;
                }
                if (il0) {
                    succtmp = ibbl->ins_tl->iptr;
                    // now il0 points to the successor of ibbl!
                    for (htmp = succtmp->succs; htmp; htmp = htmp->next) {
                        if (htmp->iptr->ins->addr == il0->iptr->ins->addr) {
                            break;
                        }
                    }

                    if (htmp == NULL) { // jnstr not in instr's succs list
                        htmp = calloc(1, sizeof(iListNode));
                        if (!htmp) {
                            fprintf(stderr, "Out of memory!\n");
                            exit(1);
                        }

                        htmp->next = succtmp->succs;
                        htmp->iptr = il0->iptr;             // add the end of next bbl to the current bbl's successors
                        succtmp->numSuccs ++;
                        succtmp->succs = htmp;
                    }
                }
                break;
            }
        }
        if (!ibbl) {
            ibbl = calloc(1, sizeof(Bbl));
            ibbl->id = bbl_id++;
            ibbl->numInstrs = 0;
            ibbl->hasVariation = 0;     // there is only one kind of this bbl so far!

            if (bbl_list_tl) {
                bbl_list_tl->next = ibbl;
            }
            ibbl->prev = bbl_list_tl;
            bbl_list_tl = ibbl;
            if (bbl_list_hd == NULL) {
                bbl_list_hd = ibbl;
            }
            ibbl->occurrences = al_new();
            al_add(ibbl->occurrences, &(il0->iptr->ins->order));
            for (iltmp = il0; iltmp; iltmp = iltmp->next) {
                ilnew = CloneiListNode(iltmp);

                if (INSTR_HAS_FLAG(ilnew->iptr->ins, INSTR_IN_BASE_MODULE)){
					NODE_SET_FLAG(ibbl, BBL_IN_BASE_MODULE);
				}

                if (INSTR_HAS_FLAG(ilnew->iptr->ins, INSTR_FORWARD_TAINTED)){
                    NODE_SET_FLAG(ibbl, BBL_FORWARD_TAINTED);
                }

                if (INSTR_HAS_FLAG(ilnew->iptr->ins, INSTR_BACKWARD_TAINTED)){
                    NODE_SET_FLAG(ibbl, BBL_BACKWARD_TAINTED);
                }
                if(INSTR_HAS_FLAG(ilnew->iptr->ins, INSTR_DOES_UNPACKING)){
					NODE_SET_FLAG(ibbl, BBL_DOES_UNPACKING);
				}
                if ((IsCondJump(
                        ilnew->iptr->ins) /*|| InstrIsIndirectJump(ilnew->iptr->ins)*/) && INSTR_HAS_FLAG(ilnew->iptr->ins, INSTR_FORWARD_TAINTED)
                        && (handle_rop || INSTR_HAS_FLAG(ilnew->iptr->ins, INSTR_IN_BASE_MODULE)))
                    NODE_SET_FLAG(ibbl, BBL_IS_CONDITIONAL);

                if (ibbl->ins_tl) {
                    ibbl->ins_tl->next = ilnew;
                }
                ibbl->ins_tl = ilnew;

                if (ibbl->ins_hd == NULL) {
                    ibbl->ins_hd = ilnew;
                }
                ibbl->numInstrs ++;

                /*
                 * we are adding next instruction of a block to a list
                 * this list is contained within the last instruction
                 * in each block. next instructions show the successor
                 * blocks. there is a post-process which looks at this
                 * list on the last instruction of each block, finds the
                 * corresponding block and attaches the block as successor
                 * (this is because we know the actual blocks ahead of time
                 * so after all blocks are constructed, we construct the
                 * successors list.)
                 */

                if (NODE_HAS_FLAG(iltmp->iptr, BBL_END)) {          // attach bbl successors to the last instruction of bbl
                    ibbl->ins_tl->next = NULL;                      // so a successor is the last instruction of the next bbl
                    iListNode *tmp = iltmp->next;

                    if(tmp){            // add first successor
                        iListNode *succ = malloc(sizeof(iListNode));
                        succ->iptr = tmp->iptr;
                        succ->next = NULL;
                        succ->prev = NULL;
                        ilnew->iptr->succs = succ;
                        ilnew->iptr->numSuccs ++;
                    }
                    break;
                }
            }
#if 0&DEBUG_MESSAGES
            printf("bbl %d added with starting: %d: [0x%08x] %s\n", ibbl->id,
                    ibbl->ins_hd->iptr->ins->order, ibbl->ins_hd->iptr->ins->addr,
                    ibbl->ins_hd->iptr->ins->uInstr.name);
#endif

            if (iltmp) {
                il0 = iltmp->next;
            } else {
                il0 = NULL;
            }
        }

    }
    NODE_SET_FLAG(bbl_list_hd, BBL_ENTRY);
    // Mark first bbl az entry node!
    NODE_SET_FLAG(bbl_list_tl, BBL_EXIT);
    // Mark last bbl az exit node!

#if 0           // print bbls and their instrs
    printf("bblsss\n");
    for(ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next){
        printf("Bbl %d:\n", ibbl->id);
        iListNode *tmp = ibbl->ins_hd;
        for(; tmp; tmp = tmp->next){
            printf("\t%d: [0x%08x] %s\n ", tmp->iptr->ins->order, tmp->iptr->ins->addr, tmp->iptr->ins->uInstr.name);
        }
    }
#endif

    FindBblSuccsAndPredsOldMethod();
#if 0
    {
        for (ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next) {
            printf("bbl %d: %d: [0x%08x] %s\n", ibbl->id, ibbl->ins_hd->iptr->ins->order, ibbl->ins_hd->iptr->ins->addr, ibbl->ins_hd->iptr->ins->uInstr.name);
        }

    }

#endif

#if 0     // print instr succs
    il0 = ins_list;
    while(il0){
        printf("%d: [%08x]%s", il0->iptr->ins->order, il0->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(il0->iptr->ins)));
        iListNode *tmp = il0->iptr->succs;
        for(; tmp; tmp = tmp->next){
            printf("%d: [%08x]%s", tmp->iptr->ins->order, tmp->iptr->ins->addr, tmp->iptr->ins->uInstr.name);
        }
        printf("\n");
        il0 = il0->next;
    }
#endif
    return bbl_id;
}


/*
 * given a basic block, returns
 * number of possible successors
 * the bbl can have!
 */

int NumberOfPossibleSuccessors(Bbl *ibbl){
	Instruction *iins = ibbl->ins_tl->iptr->ins;
	if(IsCondJump(iins)){
		return 2;
	} else if(iins->uInstr.mnemonic == UD_Iret){
		return 2;
	} else if(iins->uInstr.mnemonic == UD_Ijmp){
		if(iins->uInstr.operand[0].type == UD_OP_JIMM){
			return 1;
		} else {
			return -1;
		}
	}
	return 1;
}


static bListNode *Intersect(bListNode *dom1, bListNode *dom2){
    bListNode *finger1, *finger2;
    bListNode *intersect, *tmp;

    if(dom1 == NULL || dom2 == NULL){
        return NULL;
    }
    intersect = NULL;
    for(finger1 = dom1; finger1; finger1 = finger1->next){
        for(finger2 = dom2; finger2; finger2 = finger2->next){
            if(finger1->bptr->id == finger2->bptr->id){
                tmp = malloc(sizeof(bListNode));
                tmp->bptr = finger1->bptr;
                tmp->next = intersect;
                intersect = tmp;
                break;
            }
        }
    }
    return intersect;
}

static bool DomsAreEqual(bListNode *dom1, bListNode *dom2){
    bool found = false;
    bListNode *finger1, *finger2;
    for(finger1 = dom1; finger1; finger1 = finger1->next){
        found = false;
        for(finger2 = dom2; finger2; finger2 = finger2->next){
            if(finger1->bptr->id == finger2->bptr->id){
                found = true;
                break;
            }
        }
        if(!found)
            return false;
    }
    return true;
}
/*
 * PostDominatorTree
 * find all post-dominators of a node
 * and put it in a link-list in node
 * algorithm:
 * do until no change {
 * for each node n, post-dom{n} = {n} union intersect post-dom{s} for all s in successors of n
 * }
 */
void PostDominatorTree() {
    Bbl *ibbl;
    bListNode *tmp, *isuccPdom, *newPdoms, *isuccs, *jsuccs;
    bool change;

    // for the first node, set dominator the node itself
    ibbl = bbl_list_tl;
    ibbl->pDoms = malloc(sizeof(bListNode));
    ibbl->pDoms->bptr = ibbl;
    ibbl->pDoms->next = NULL;

    // for every node in N - {n0}, set dominator to N (all nodes)
    for (ibbl = bbl_list_tl->prev; ibbl; ibbl = ibbl->prev) {
        Bbl *jbbl;
        ibbl->pDoms = NULL;
        for (jbbl = bbl_list_tl; jbbl; jbbl = jbbl->prev) {

            bListNode *domTmp = malloc(sizeof(bListNode));
            domTmp->bptr = jbbl;
            domTmp->next = ibbl->pDoms;
            ibbl->pDoms = domTmp;
        }
    }
    int i = 0;
    do {
        i++;
        change = false;
        for (ibbl = bbl_list_tl->prev; ibbl; ibbl = ibbl->prev) {
            newPdoms = malloc(sizeof(bListNode));
            newPdoms->bptr = ibbl;                  // add n itself to doms
            newPdoms->next = NULL;

            if(ibbl->numSuccs == 1){        // dom(ibbl) = union(ibbl ,pdom(ibbl->succs))
                isuccs = ibbl->succs;                    // add its successor's doms to node's doms!
                for(isuccPdom = isuccs->bptr->pDoms; isuccPdom; isuccPdom = isuccPdom->next){
                    tmp = malloc(sizeof(bListNode));
                    tmp->bptr = isuccPdom->bptr;
                    tmp->next = newPdoms;
                    newPdoms = tmp;
                }
            } else if (ibbl->numSuccs > 1) {        // > 1 successors
                isuccs = ibbl->succs;
                jsuccs = isuccs->next;
                bListNode *intersect = Intersect(isuccs->bptr->pDoms, jsuccs->bptr->pDoms);

                for (isuccs = jsuccs->next; isuccs; isuccs = isuccs->next) {
                    intersect = Intersect(isuccs->bptr->pDoms, intersect);
                }
                newPdoms->next = intersect;
            }

            if(!DomsAreEqual(ibbl->pDoms, newPdoms)){
                change = true;
                ibbl->pDoms = newPdoms;
            }
        }

    } while (change);
}


/*
 * returns true:
 * if node w is postdominated by
 * all post-dominators of v
 */
bool IsPostDomedByAllPostDominators(Bbl *v, Bbl *w){
    bListNode *vpdoms, *wpdoms;



    /*for(vpdoms = v->pDoms; vpdoms; vpdoms = vpdoms->next){
        vPostDominatesW = false;
        for(vpdoms2 = vpdoms->bptr->pDoms; vpdoms2; vpdoms2 = vpdoms2->next){
            if(vpdoms2->bptr->id == w->id){
                vPostDominatesW = true;
                break;
            }
        }
        if(vPostDominatesW){
            continue;
        } else {
            return false;
        }

    }*/
    bool vdomFoundinwdom;

    for(vpdoms = v->pDoms; vpdoms; vpdoms = vpdoms->next){
        if(vpdoms->bptr->id == v->id){
            continue;
        }
        vdomFoundinwdom = false;
        for(wpdoms = w->pDoms; wpdoms; wpdoms = wpdoms->next){
            if(wpdoms->bptr->id == vpdoms->bptr->id){
                vdomFoundinwdom = true;
                break;
            }
        }
        if(!vdomFoundinwdom){
            return false;
        }
    }
    return true;
}

/*
 * Finding immediate postdominators to construct
 * the (immediate) postdominator tree!
 * */
void ImmediateDominator(){
    Bbl *ibbl;
    bListNode *isuccPdom, *ipdom;

    ibbl = bbl_list_tl;
    ibbl->ipDoms = NULL;

    for(ibbl = bbl_list_tl->prev; ibbl; ibbl = ibbl->prev){
        for(isuccPdom = ibbl->pDoms; isuccPdom; isuccPdom = isuccPdom->next){    // pick one post dominator
            if(isuccPdom->bptr->id == ibbl->id){
                continue;
            }
            if(IsPostDomedByAllPostDominators(ibbl, isuccPdom->bptr)){
                ipdom = malloc(sizeof(bListNode *));
                ipdom ->bptr = isuccPdom->bptr;
                ipdom->next = ibbl->ipDoms;
                ibbl->ipDoms = ipdom;
                break;
            }
        }
    }
}

 void MergeBblsNew(){
    Bbl *ibbl, *jbbl;
    bListNode *preds, *succs;
    int numMergedBbls = 0;
    bool change = true;
    int i =0;
    while(change){
        change = false;
        i++;
        for(ibbl = bbl_list_hd; ibbl && ibbl->next; ibbl = ibbl->next){
            if(ibbl->numSuccs == 1 && INSTR_HAS_FLAG(ibbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
                if(ibbl->succs->bptr->numPreds == 1 && ibbl->succs->bptr->id != ibbl->id){
                    if(!INSTR_HAS_FLAG(ibbl->succs->bptr->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
                        continue;
                    }
                    change = true;
                    numMergedBbls ++;
                    ibbl->flags |= ibbl->succs->bptr->flags;
                    ibbl->ins_tl->next = ibbl->succs->bptr->ins_hd;
                    ibbl->ins_tl = ibbl->succs->bptr->ins_tl;
                    ibbl->numInstrs += ibbl->succs->bptr->numInstrs;
                    int removedBblID = ibbl->succs->bptr->id;

                    for(jbbl = bbl_list_hd; jbbl; jbbl = jbbl->next){
                        if(jbbl->id == ibbl->succs->bptr->id){
                            if(jbbl->next){
                                jbbl->next->prev = jbbl->prev;
                            } else {
                                jbbl->prev->next = NULL;
                                bbl_list_tl = jbbl->prev;
                            }
                            if(jbbl != bbl_list_hd){
                            	jbbl->prev->next = jbbl->next;
                            } else {
                            	bbl_list_hd = jbbl->next;
                            }
                            break;
                        }
                    }
                    ibbl->numSuccs = ibbl->succs->bptr->numSuccs;
                    ibbl->succs = ibbl->succs->bptr->succs;
                    for(succs = ibbl->succs; succs; succs = succs->next){
                        for(preds = succs->bptr->preds; preds; preds = preds->next){
                            if(preds->bptr->id == removedBblID){
                                preds->bptr = ibbl;
                                if(preds->bptr->next == NULL) ibbl->next = NULL;
                            }
                        }
                    }
                }
            }
        }
    }
}

BblEdge *NewBblEdge(){
    BblEdge *tmp;
    tmp = malloc(sizeof(BblEdge));
    if(bbledge_tl){
        tmp->id = bbledge_tl->id + 1;
        bbledge_tl->next = tmp;
        tmp->prev = bbledge_tl;
        bbledge_tl = tmp;
        tmp->next = NULL;
        return tmp;
    }
    tmp->id = 0;
    bbledge_hd = bbledge_tl = tmp;
    tmp->next = tmp->prev = NULL;
    return tmp;
}
/*
 * returns true if b immediately
 * postdominates a, false otherwise
 */
static bool PostDominates(Bbl *a, Bbl *b){
    bListNode *aipdoms;
    aipdoms = a->ipDoms;
    while(aipdoms){
        if(aipdoms->bptr->id == b->id){
            return true;
        }
        aipdoms = aipdoms->bptr->ipDoms;
    }
    return false;

}
/*
 * FindExplicitControlDependencies:
 * walks in the post-dominator tree
 * collects the conditioned basic blocks,
 * and marks the corresponding instructions
 */
void FindExplicitControlDependencies(InstrList *iList){
    Bbl *ibbl;
    bListNode *to, *from;
    bListNode *succs, *newNode;

    ibbl = bbl_list_hd;
    for(; ibbl; ibbl = ibbl->next){
        if(NODE_HAS_FLAG(ibbl, BBL_FORWARD_TAINTED) && NODE_HAS_FLAG(ibbl, BBL_IS_CONDITIONAL)){
            // find (i, j) edges which are in CFG so that j does not immediately post-dominate i (j is conditioned on j)
            for(succs = ibbl->succs; succs; succs = succs->next){
                if(!PostDominates(ibbl, succs->bptr)){
                    BblEdge *tmp = NewBblEdge();
                    tmp->from = ibbl;
                    tmp->to = succs->bptr;
                }
            }
        }
    }
    // !! from NODE IN EACH EDGE CONTAINS THE CONTROL BBL

#if DEBUG_MESSAGES       // print control dependent edges
    BblEdge *start;
    for(start = bbledge_hd; start; start = start->next){
        fprintf(stderr, "%d ==> %d\n", start->from->id, start->to->id);
    }
#endif

    //finding all dependent nodes in post-dominator tree
    BblEdge *edge;
    bListNode *path1, *path2;
    for(edge = bbledge_hd; edge; edge = edge->next){
        to = malloc(sizeof(bListNode));
        to->bptr = edge->to;
        from = malloc(sizeof(bListNode));
        from->bptr = edge->from;

        // find paths from two bbls to root, in domtree, then remove the intersection!
        path1 = malloc(sizeof(bListNode));
        path1->bptr = to->bptr;
        path1->next = NULL;
        bListNode *path1tmp = path1;
        for(to = to->bptr->ipDoms; to; to = to->bptr->ipDoms){
            newNode = malloc(sizeof(bListNode));
            newNode->bptr = to->bptr;
            newNode->next = NULL;
            path1tmp->next = newNode;

            path1tmp = newNode;
        }
        path2 = malloc(sizeof(bListNode));
        path2->bptr = from->bptr;
        path2->next = NULL;
        bListNode *path2tmp = path2;
        for(from = from->bptr->ipDoms; from; from = from->bptr->ipDoms){
            newNode = malloc(sizeof(bListNode));
            newNode->bptr = from->bptr;
            newNode->next = NULL;
            path2tmp->next = newNode;
            path2tmp = newNode;
        }
#if DEBUG_MESSAGES   // print all paths found!
        fprintf(stderr, "Path1: ");
        for(newNode = path1; newNode; newNode = newNode->next){
            fprintf(stderr, "%d -> ", newNode->bptr->id);
        }
        fprintf(stderr, "\nPath2: ");
        for(newNode = path2; newNode; newNode = newNode->next){
            fprintf(stderr, "%d -> ", newNode->bptr->id);
        }
        fprintf(stderr, "\n\n");
#endif
        // find the lowest common ancestor, which is the first common Bbl in both paths!
        newNode = NULL;
        bool found = false;
        for(newNode = path1; newNode; newNode = newNode->next){
            bListNode *tmp;
            for(tmp = path2; tmp; tmp = tmp->next){
                if(tmp->bptr && tmp->bptr->id == newNode->bptr->id){
                    found = true;
                    break;
                }
            }
            if(found){
                break;
            }
        }

        // add all control dependencies
        int i;
        int repetition2 = al_size(edge->to->occurrences);


        Instruction *next;
        for(i = 0; i < repetition2; i++){
            int order = *(int *)al_get(edge->to->occurrences, i);
            order -= 1;		// points to the last instruction in previous bbl, conditional jump

            next = GetInstruction(iList, order);
            if(!(INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED))){
            	continue;
            }
            for(to = path1; to->bptr->id != newNode->bptr->id; to = to->next){
                to->bptr->cDepend = edge->from;     // add the control
                if(!handle_rop &&  !INSTR_HAS_FLAG(to->bptr->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
                    continue;
                }
                int numInstrs = to->bptr->numInstrs;
                if(i >= al_size(to->bptr->occurrences)){
                	continue;
                }
                int start = *(int *) al_get(to->bptr->occurrences, i);
                SetCurrentInstr(iList, start);

                while(numInstrs --){
                    next = FetchNextInstr(iList);
                    INSTR_SET_FLAG(next, INSTR_COND_CF);
                    next->ControlDependentOn = order;
                    SaveInstrChange(iList, next);
                }
            }
            for(from = path2->next; path2->bptr->id != newNode->bptr->id && from->bptr->id != newNode->bptr->id; from = from->next){    // skip the first node, the control node itself!
                from->bptr->cDepend = edge->from;
                if(!handle_rop && !INSTR_HAS_FLAG(from->bptr->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
                    continue;
                }
                int numInstrs = from->bptr->numInstrs;
                int start = *(int *) al_get(to->bptr->occurrences, 0);
                SetCurrentInstr(iList, start);
                while(numInstrs --){
                    next = FetchNextInstr(iList);
                    INSTR_SET_FLAG(next, INSTR_COND_CF);
                    next->ControlDependentOn = order;
                    SaveInstrChange(iList, next);
                }
            }
        }
    }


    return;
 }

ArrayList *newControlTransferInstrs;
void SetControlSourceforRegs(REGISTER_BITVECTOR regs, int order, int *RegDependency){
    int i = 0;
    int reg = 1;
    for(i = 0; i < 32 && regs; i++){
        if(((reg << i) & regs)){
            RegDependency[i] = order;
        }
        regs &= ~((reg << i));
    }
}

bool MarkImplicitControlDependenciesNew(InstrList *iList){
    Instruction *iins;
    int i;
    ud_operand_t *op0;
    enum ud_mnemonic_code opcode;
    REGISTER_BITVECTOR regsDef, regsUsed, regsTainted = 0, flagsUsed, flagsDef, flagsTainted = 0;
    int RegDependency[32];       //an array which keeps the source control dependency of each register
    int FlagDependency[32];			// similar to the above, for flags
    bool Change = false;
    int PrevCondOrder = -1;

    ClearConstantPages(iList, 0);


    newControlTransferInstrs = al_new();

    for(i = 0; i < 32; i++){
        RegDependency[i] = -1;
        FlagDependency[i] = -1;
    }

    i = 0;
    SetCurrentInstr(iList, i);

    while ((iins = FetchNextInstr(iList)) != NULL) {

        opcode = iins->uInstr.mnemonic;
        regsDef = MaskDefinedReg(RegDef(iList, iins));
        regsUsed = MaskDefinedReg(RegUsed(iList, iins));

        flagsUsed = FlagsUsed(iins);
        flagsDef = FlagsDef(iins);
        op0 = &(iins->uInstr.operand[0]);

        if ((IsPush(iins) || IsPop(iins)) || opcode == UD_Icall || opcode == UD_Iret || opcode == UD_Ileave) {
            if (UDOpIsRegister(op0) == UD_R_ESP) {
                if (iins->uInstr.mnemonic == UD_Ipop)
                    regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
                else if (iins->uInstr.mnemonic == UD_Ipush)
                    regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
            } else {
                regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
                regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
            }
        }

        if((IsCondJump(iins) && INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) || !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
            for(i = 0; i < 32; i++){
				RegDependency[i] = -1;
			}
			ClearConstantPages(iList, 0);
			regsTainted = 0;
			PrevCondOrder = iins->order;
        }

        if (INSTR_HAS_FLAG(iins, INSTR_COND_CF)) {

            if (iins->ControlDependentOn != PrevCondOrder){
                for(i = 0; i < 32; i++){
                    RegDependency[i] = -1;
                }
                ClearConstantPages(iList, 0);
                regsTainted = 0;
                PrevCondOrder = iins->ControlDependentOn;
            }

            if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
            	DependencyMemSetConstWithSrc(iList, iins->memDefMin, iins->memDefSize, iins->ControlDependentOn);
            }
            if (regsDef != 0){
                regsTainted |= regsDef;
                SetControlSourceforRegs(regsDef, iins->ControlDependentOn, RegDependency);
            }
            if (flagsDef){
                flagsTainted |= flagsDef;
                SetControlSourceforRegs(flagsDef, iins->ControlDependentOn, FlagDependency);
            }
        } else {
            if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize)) {
                INSTR_SET_FLAG(iins, INSTR_TEMP_CONDITIONAL);
            }
            if ((regsUsed & regsTainted) != 0) {
                INSTR_SET_FLAG(iins, INSTR_TEMP_CONDITIONAL);
            }
            if ((flagsUsed & flagsTainted) != 0) {
                INSTR_SET_FLAG(iins, INSTR_TEMP_CONDITIONAL);
            }
            if (INSTR_HAS_FLAG(iins, INSTR_TEMP_CONDITIONAL) && !(opcode == UD_Ixor && op0->base
                    == iins->uInstr.operand[1].base)) {
                SaveInstrChange(iList, iins);
                if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
                    if (opcode == UD_Ipushad) {
                        int taintRegs = (regsTainted & 0xff);
                        int k;
                        for (k = 0; k < 8 && (taintRegs); k++) {
                            if ((taintRegs & 0x1)) {
                                Range *r = RangeCreate(iins->esp - (k + 1) * 4, iins->esp - (k) * 4
                                        - 1);

                                DependencyMemSetConstWithSrc(iList, r->min, r->max - r->min + 1, RegDependency[(taintRegs & 0x1)]);
                            }
                            taintRegs = (taintRegs >> 1);
                        }
                    }
                    else if (opcode != UD_Ixchg && opcode != UD_Icmpxchg) {


                        uint64_t index=0;
                        if((regsUsed & regsTainted)){        // filter out esp and memory registers
                            index = RegDependency[floor_log2((regsUsed & regsTainted))];
                        } else if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
                        	index = DependencyMemGetConstLocSrc(iList, iins->memUsedMin);
                        } else{
                            //fprintf(stderr, "Unhandled case: %d: %s\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
                        }
                        if(!(iins->uInstr.pfx_seg == UD_R_FS)){
                        	DependencyMemSetConstWithSrc(iList, iins->memDefMin, iins->memDefSize, index);
                        }
                    }
                }
                if (regsDef != 0) {
                    if (opcode == UD_Ipopad) {
                        int k;
                        for (k = 0; k < 8; k++) {
                            Range *r = RangeCreate(iins->esp + (k) * 4, iins->esp + (k + 1) * 4 - 1);
                            if (MemLocIsConstant(iList, r->min, 4)){//RangeOverlapWithWriteSet(memTainted, r)) {
                            	regsTainted |= MaskDefinedReg((0x1 << (7 - k)));
                            	int index = DependencyMemGetConstLocSrc(iList, r->min);

                            	SetControlSourceforRegs(MaskDefinedReg((0x1 << (7 - k))), index, RegDependency);
                            }
                        }
                    } else if (opcode == UD_Ixchg || opcode == UD_Icmpxchg) {
                        bool regIsTainted = false;
                        if (op0->type == UD_OP_MEM) { //op0 is memory
                            if (MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize))
                                regIsTainted = true;
                            if ((regsTainted & regsDef) == 0) {
                            	UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
                            } else if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {

                                DependencyMemSetConstWithSrc(iList, iins->memDefMin, iins->memDefSize, RegDependency[floor_log2((regsDef & regsTainted))]);
                                regsTainted &= ~(UpToAllRegs(regsDef));

                            }
                            if (regIsTainted){
                                regsTainted |= regsDef;
                                int index = DependencyMemGetConstLocSrc(iList, iins->memUsedMin);
                                SetControlSourceforRegs(regsDef, index, RegDependency);
                            }
                        } else if (iins->uInstr.operand[1].type == UD_OP_MEM) { //op0 is reg
                            if (MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize))//(WriteSetsOverlap(memTainted, memUsed)) != 0)
                                regIsTainted = true;
                            if ((regsTainted & regsDef) == 0) {
                            	UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);

                            } else {
                            	SetAddrAsConst(iList, iins->memDefMin, iins->memDefSize);
                                regsTainted &= ~(UpToAllRegs(regsDef));
                            }
                            if (regIsTainted){
                                regsTainted |= regsDef;
                                SetControlSourceforRegs(regsDef, iins->ControlDependentOn, RegDependency);
                            }
                        } else { // op0 and op1 are regs
                            if ((regsTainted & MaskDefinedReg(UDRegisterToBitVector(op0->base))) != 0
                                    && (regsTainted & MaskDefinedReg(UDRegisterToBitVector(
                                            iins->uInstr.operand[1].base))) == 0) {
                                int newRegs = (MaskDefinedReg(UDRegisterToBitVector(
                                        iins->uInstr.operand[1].base) | UpToAllRegs(UDRegisterToBitVector(
                                                iins->uInstr.operand[1].base))));
                                regsTainted |= newRegs;
                                SetControlSourceforRegs(newRegs, RegDependency[floor_log2(UDRegisterToBitVector(iins->uInstr.operand[0].base))], RegDependency);
                                regsTainted &= ~(MaskDefinedReg(UDRegisterToBitVector(op0->base)) | UpToAllRegs(UDRegisterToBitVector(op0->base)));
                            } else if ((regsTainted & MaskDefinedReg(UDRegisterToBitVector(
                                    iins->uInstr.operand[1].base))) != 0 && (regsTainted
                                            & MaskDefinedReg(UDRegisterToBitVector(op0->base))) == 0) {
                                int newRegs = (MaskDefinedReg(UDRegisterToBitVector(op0->base) | UpToAllRegs(UDRegisterToBitVector(op0->base))));
                                regsTainted |=newRegs;
                                SetControlSourceforRegs(newRegs, RegDependency[floor_log2(UDRegisterToBitVector(iins->uInstr.operand[1].base))], RegDependency);
                                regsTainted &= ~(MaskDefinedReg(UDRegisterToBitVector(
                                        iins->uInstr.operand[1].base) | UpToAllRegs(UDRegisterToBitVector(
                                                iins->uInstr.operand[1].base))));
                            }
                        }
                    } else {
                        int index=0;
                        if((regsUsed & regsTainted)){        // filter out esp and memory registers
                            index = RegDependency[floor_log2((regsUsed & regsTainted))];
                        } else if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
                        	index = DependencyMemGetConstLocSrc(iList, iins->memUsedMin);
                        } else{
                            //fprintf(stderr, "Unhandled case: %d: %s\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
                        }
                        regsTainted |= regsDef;
                        SetControlSourceforRegs(regsDef, index, RegDependency);
                    }
                }
                if (flagsDef){
                    flagsTainted |= flagsDef;
                    uint64_t index=0;
                    if((flagsUsed & flagsTainted)){
                    	index = FlagDependency[floor_log2((flagsUsed & flagsTainted))];
                    } else if((regsUsed & regsTainted)){
                    	index = RegDependency[floor_log2((regsUsed & regsTainted))];
                    } else if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize)){
                    	index = DependencyMemGetConstLocSrc(iList, iins->memUsedMin);
                    	/*MemDefWithIndex *tmp = SearchInIndexedMem(memDefIndexes, al_get(memUsed->ranges, 0));
                    	if(tmp){
                    	    index = tmp->index;
                    	}*/
                    }
                    SetControlSourceforRegs(flagsDef, index, FlagDependency);
                }
            } else {
                if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && MemLocIsConstant(iList, iins->memDefMin, iins->memDefSize)) {
                    /*memtmp = SubtractWriteSets(memTainted, memDef);
                    WriteSetDestroy(memTainted);
                    memTainted = memtmp;*/
                	UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
                }
                if ((regsDef & regsTainted) != 0) {
                    regsTainted &= ~UpToAllRegs(regsDef);
                }
                if ((flagsDef & flagsTainted) != 0) {
                    flagsTainted &= ~flagsDef;
                }
            }
        }

        if(!INSTR_HAS_FLAG(iins, INSTR_JMP_AS_CONDITIONAL) && (handle_rop || INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) && InstrIsControlTransfer(iList, iins)
        		&& ((regsUsed & regsTainted) || MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize)/*WriteSetsOverlap(memTainted, memUsed) */|| (flagsUsed & flagsTainted))){    // control instruction is implicit control transfer

        	if(INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED) && IsCondJump(iins)){			// don't add explicit conditionals here!
        		continue;
        	}

        	if((regsUsed & regsTainted)){
        		iins->ControlDependentOn = RegDependency[floor_log2((regsUsed & regsTainted))];
        		if(iins->ControlDependentOn == -1){
        			int tmp = floor_log2((regsUsed & regsTainted));
        			regsUsed &= ~(1 << tmp);
        			iins->ControlDependentOn = RegDependency[floor_log2((regsUsed & regsTainted))];
        		}
        	} else if(MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize)){//WriteSetsOverlap(memTainted, memUsed)){
        		//MemDefWithIndex *tmp = SearchInIndexedMem(memDefIndexes, al_get(memUsed->ranges, 0));
        		iins->ControlDependentOn = DependencyMemGetConstLocSrc(iList, iins->memUsedMin);
        	} else {
        		iins->ControlDependentOn = FlagDependency[floor_log2((flagsUsed & flagsTainted))];
        	}
        	//fprintf(stderr, "new %d: %s is dependent on %lu\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)), iins->ControlDependentOn);
            al_add(newControlTransferInstrs, iins);
            INSTR_SET_FLAG(iins, INSTR_JMP_AS_CONDITIONAL);
            SaveInstrChange(iList, iins);
            RInstrDestroy(iList, iins);
            Change = true;
        }
        RInstrDestroy(iList, iins);
    }

    return Change;
}

void AddNewControlDependencies(InstrList *iList){
    Instruction *iins, *jins, *ControlInstruction;
    Bbl *ibbl, *jbbl;

    int i;
    for(i=0; i < al_size(newControlTransferInstrs); i++){       // iterate on all new found implicit control transfer instrs
        iins = al_get(newControlTransferInstrs, i);
        jins = GetInstruction(iList, iins->next);       // and mark all bbls right after above instruction, until there is a bbl with more than 1 successor
        ControlInstruction = GetInstruction(iList, iins->ControlDependentOn);
        // find the corresponding basic block
        for(ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next){
            if(ibbl->ins_hd->iptr->ins->addr == jins->addr){
                break;
            }
        }
        for(jbbl = bbl_list_hd; jbbl; jbbl = jbbl->next){					// jins is the control transfer instr so it should be at the end of the bbl!
            if(ControlInstruction && jbbl->ins_tl->iptr->ins->addr == ControlInstruction->addr){
                break;
            }
        }

        if(!ibbl){
            return;
        }



        if(!jbbl || ibbl->id == jbbl->succs->bptr->id){			// loop!
        	continue;
        }

        while(ibbl->numSuccs <= 1){
            ibbl->cDepend = jbbl;
            while(jins && jins->addr != ibbl->ins_tl->iptr->ins->addr){
                INSTR_SET_FLAG(jins, INSTR_COND_CF);
                jins->ControlDependentOn = ControlInstruction->order;
                SaveInstrChange(iList, jins);
                RInstrDestroy(iList, jins);
                jins = GetInstruction(iList, jins->next);
            }
            if(!ibbl->succs){
                break;
            }
            ibbl = ibbl->succs->bptr;
        }
    }
}

bool OperandsAreConstants(InstrList *iList, ConstantsinScope *scope, Instruction *iins){
    ud_operand_t *op0, *op1;
    enum ud_mnemonic_code opcode;
    REGISTER_BITVECTOR regsDef, regsUsed, nonMemRegUsed; //, flagsUsed, flagsDef;
    WriteSet *memUsed, *memDef=NULL;

    opcode = iins->uInstr.mnemonic;
    op0 = &(iins->uInstr.operand[0]);
    op1 = &(iins->uInstr.operand[1]);

    regsDef = RegDef(iList, iins);
    regsUsed = RegUsed(iList, iins);
    nonMemRegUsed = NonMemRegUsed(iList, iins);
    if(INSTR_HAS_FLAG(iins, INSTR_USES_FTAINT)){
    	if(regsDef){
    		scope->regs |= MaskDefinedReg(regsDef);
    	}
    	if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){
    		memDef = MemDef(iList, iins);
    		if(WriteSetsOverlap(scope->mem, memDef)){
				WriteSet *tmp = SubtractWriteSets(scope->mem, memDef);
				WriteSetDestroy(scope->mem);
				scope->mem = tmp;
    		}
    	}
    	return true;
    }

    if(!INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && !INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){  // register instruction
    	if(opcode == UD_Ixchg){
    		bool op1Const, op0Const;
    		if(op1->type == UD_OP_REG && op0->type == UD_OP_REG){

    			op0Const = (scope->regs & UDRegisterToBitVector(op0->base));
    			op1Const = (scope->regs & UDRegisterToBitVector(op1->base));

    			if(op0Const){
    				scope->regs |= MaskDefinedReg(UDRegisterToBitVector(op1->base));
    			} else {
    				scope->regs &= ~MaskDefinedReg(UDRegisterToBitVector(op1->base));
    			}

    			if(op1Const){
    				scope->regs |= MaskDefinedReg(UDRegisterToBitVector(op0->base));
    			} else {
    				scope->regs &= ~MaskDefinedReg(UDRegisterToBitVector(op0->base));
    			}

    		}
    		return true;
    	}

    	else if (regsDef){           // TODO: special case: xchg
            if(!OpcodeIsArith(opcode) && (regsUsed & scope->regs) == regsUsed){     // if instr only uses known registers or constants
                scope->regs |=  UpToAllRegs(MaskDefinedReg(regsDef));
                return true;
            } else if(OpcodeIsArith(opcode) && ((regsDef & scope->regs) == regsDef || (regsUsed & scope->regs) == regsUsed)){
                scope->regs |=  UpToAllRegs(MaskDefinedReg(regsDef));
                return true;
            }
            // for arithmetic operations, if defined register is constant, we dont care about the other operands: consider the result as a constant!
        }
    } else {
        if(!INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){       // instr defines a memory location
            memDef = MemDef(iList, iins);
            if(opcode == UD_Ipushfd){
                if(!INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)){					// if flags are not affected by input, then the memory is constant!
                    scope->mem = MergeWriteSets(scope->mem, memDef);
                    return true;
                }
            } else if ((nonMemRegUsed && (nonMemRegUsed & scope->regs) == nonMemRegUsed) || !nonMemRegUsed){       // TODO: add support for popad, some regs maybe constant. question: can we easily separate regUsed and nonMemRegUsed?
                if(!WriteSetIsEmpty(memDef)){
                    scope->mem = MergeWriteSets(scope->mem, memDef);
                    return true;
                }
            } else if((nonMemRegUsed && (MaskDefinedReg(nonMemRegUsed) & scope->regs) == MaskDefinedReg(nonMemRegUsed))){		// ex.: push eax while ax is only constant!
            	WriteSet *es = MemoryMask(iins->memDefMin, nonMemRegUsed);
            	scope->mem = MergeWriteSets(scope->mem, es);
            	return true;
            }
        } else if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && !INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){
            memUsed = MemUsed(iList, iins);
            if(opcode == UD_Ipopfd){
                // for flags TODO
            } else if(!nonMemRegUsed || (nonMemRegUsed && (nonMemRegUsed & scope->regs) == nonMemRegUsed)){       // consider instrs like mul [mem] which implicitly uses eax
                if(regsDef && (WriteSetsOverlap(scope->mem, memUsed) || INSTR_HAS_FLAG(iins, INSTR_READS_GLOBAL_MEM))){
                    scope->regs |= MaskDefinedReg(regsDef);
                    return true;
                }
            }
        } else if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){
            memUsed = MemUsed(iList, iins);
            memDef = MemDef(iList, iins);
            if(opcode == UD_Ixchg || opcode == UD_Ixadd || opcode == UD_Icmpxchg){
            	// special case
            	if(opcode == UD_Ixchg){
            		bool op0Const = false, op1Const = false;
            		if(op1->type == UD_OP_REG && op0->type == UD_OP_REG){

            			op0Const = (scope->regs & UDRegisterToBitVector(op0->base));
            			op1Const = (scope->regs & UDRegisterToBitVector(op1->base));

            			if(op0Const){
            				scope->regs |= MaskDefinedReg(UDRegisterToBitVector(op1->base));
            			} else {
            				scope->regs &= ~MaskDefinedReg(UDRegisterToBitVector(op1->base));
            			}

            			if(op1Const){
            				scope->regs |= ~MaskDefinedReg(UDRegisterToBitVector(op0->base));
            			} else {
            				scope->regs &= ~MaskDefinedReg(UDRegisterToBitVector(op0->base));
            			}

            		} else if(op1->type == UD_OP_REG && op0->type == UD_OP_MEM){

            			op0Const = WriteSetsOverlap(scope->mem, memDef);
            			op1Const = (scope->regs & UDRegisterToBitVector(op1->base));

            			if(op1Const){
            				if(!WriteSetsOverlap(scope->mem, memUsed)){
            					scope->mem = MergeWriteSets(scope->mem, memDef);
            				}
            			} else {
            				if(WriteSetsOverlap(scope->mem, memUsed)){
            					WriteSet *tmp = SubtractWriteSets(scope->mem, memDef);
            					WriteSetDestroy(scope->mem);
            					scope->mem = tmp;
            				}
            			}
            			if(op0Const){
            				scope->regs |= MaskDefinedReg(UDRegisterToBitVector(op1->base));
            			} else {
            				scope->regs &= ~MaskDefinedReg(UDRegisterToBitVector(op1->base));
            			}
            		} else{		//op1->type == ud_op_mem ...
            			op1Const = WriteSetsOverlap(scope->mem, memDef);
            			op0Const = (scope->regs & UDRegisterToBitVector(op0->base));

            			if(op0Const){
            				if(!WriteSetsOverlap(scope->mem, memUsed)){
            					scope->mem = MergeWriteSets(scope->mem, memDef);
            				}
            			} else {
            				if(WriteSetsOverlap(scope->mem, memUsed)){
            					WriteSet *tmp = SubtractWriteSets(scope->mem, memDef);
            					WriteSetDestroy(scope->mem);
            					scope->mem = tmp;
            				}
            			}
            			if(op1Const){
            				scope->regs |= MaskDefinedReg(UDRegisterToBitVector(op0->base));
            			} else {
            				scope->regs &= ~MaskDefinedReg(UDRegisterToBitVector(op0->base));
            			}
            		}
            		return true;
            	}
            } else if (!nonMemRegUsed || (nonMemRegUsed && (nonMemRegUsed & scope->regs) == nonMemRegUsed)){
                if(!WriteSetIsEmpty(memDef) && !WriteSetIsEmpty(memDef) && WriteSetsOverlap(scope->mem, memUsed)){
                    scope->mem = MergeWriteSets(scope->mem, memDef);
                    return true;
                }
            }
        }
    }

    scope->regs &= ~regsDef;
    scope->regs |= (LYNX_ESP | LYNX_SP);
    if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && WriteSetsOverlap(scope->mem, memDef)){
    	WriteSet *tmp = SubtractWriteSets(scope->mem, memDef);
    	WriteSetDestroy(scope->mem);
    	scope->mem = tmp;
    }
    return false;
}

void MarkUnsimplifiableInstrs(InstrList *iList){
    Instruction *iins;
    ConstantsinScope *newScope, *currentScope = NULL;
    InitScopeHashTable();
    
    SetCurrentInstr(iList, 0);
    while((iins = FetchNextInstr(iList)) != NULL){
        if(iins->ControlDependentOn == -1 || (!handle_rop && !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE))){
            continue;
        }
        shnode *tmp = shLookup(iins->ControlDependentOn);
        if(tmp){
            currentScope = tmp->scope;
        } else {
            newScope = malloc(sizeof(ConstantsinScope));
            newScope->startIndex = iins->ControlDependentOn;
            newScope->regs = 0;
            newScope->mem = WriteSetCreate();
            shInsert(newScope);
            //al_add(ConstantScopes, newScope);
            
            currentScope = newScope;
        }
        if(!currentScope){
            fprintf(stderr, "Error, Instr is not within any known scope! %d: %s depend on %d\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)), (int) iins->ControlDependentOn);
            return;
        }
        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) && !OperandsAreConstants(iList, currentScope, iins) && OpcodeIsArith(iins->uInstr.mnemonic) && INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED) /*&& (INSTR_HAS_FLAG(iins, INSTR_READS_MEM) || INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM))*/){
            INSTR_SET_FLAG(iins, INSTR_UNSIMPLIFIABLE);
            SaveInstrChange(iList, iins);
        }
        
    }
}


void ConstructOroginalCFG(InstrList *iList){

	CleanUpCFGData();
	//fprintf(stderr, "constructing original CFG\n");
    CollectDistinctInstructionsByAdrress(iList);
    fprintf(stderr, ".");
    CollectInstrSuccsAndPredsOld(iList);
    fprintf(stderr, ".");
    MarkBblBoundariesByAddr(iList);

    fprintf(stderr, ".\n");

    fprintf(stderr, "Constructing basic blocks...");

    ConstructDistinctBblsOldMethod(iList);
    fprintf(stderr, "done\n");

    //MergeBblsNew();
}

void AddBblPreds(){
	Bbl *ibbl;
	bListNode *succs, *predNew;

	for(ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next){
		for(succs = ibbl->succs; succs; succs = succs->next){
			predNew = malloc(sizeof(bListNode));
			predNew->bptr = succs->bptr;
			predNew->next = succs->bptr->preds;
			succs->bptr->preds = predNew;
			succs->bptr->numPreds ++;
		}
	}
}

void FixBblBacktrack6(){
	int id = 0;
	int numInstrs;
	Bbl *start = NULL, *end = NULL, *ibbl = NULL, *prevBbl = NULL, *nextFound = NULL;
	iListNode *il0, *iltmp;
	bListNode *succs, *remainingSuccs = NULL;
	il0 = ins_list_byaddr;
	//lastOp *lastOperaqtion = malloc(sizeof(lastOp));

	int numPossibleSuccs;
	bool notSteppingBack = true;
	bool flag = true;

	ArrayList *steps = al_new();
	ArrayList *bbls;


	while (il0) {
		if (!NODE_HAS_FLAG(il0->iptr, BBL_START)) {
			if(DEBUG_MESSAGES){
			//fprintf(stderr, "Weird! il0 = %d: [0x%x] %s\n", il0->iptr->ins->order,
			//		il0->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(il0->iptr->ins)));
			}
		}
		bool found = false;
		if(prevBbl){
			if(remainingSuccs == NULL){
				succs = prevBbl->succs;
			} else {
				succs = remainingSuccs;
				remainingSuccs = NULL;
			}
			for(; notSteppingBack && flag && succs; succs = succs->next){
				if (succs->bptr->ins_hd->iptr->ins->addr == il0->iptr->ins->addr) {
					if(DEBUG_MESSAGES){
					printf("--bbl %d exists with starting: %d: [0x%08x] %s\n", succs->bptr->id,
							il0->iptr->ins->order, il0->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(il0->iptr->ins)));
					}
					found = true;

					lastOp *lastoped = malloc(sizeof(lastOp));
					lastoped->from = prevBbl;
					lastoped->to = succs->bptr;
					lastoped->prevBbl = prevBbl;
					lastoped->backToInstr = il0;
					lastoped->removeEdge = false;
					lastoped->otherSuccessors = NULL;
					lastoped->bbls = al_new();
					bbls = lastoped->bbls;
					nextFound = start;


					al_add(bbls, succs->bptr);

					bListNode *otherSuccs;
					for(otherSuccs = succs->next; otherSuccs; otherSuccs = otherSuccs->next){
						if(otherSuccs->bptr->ins_hd->iptr->ins->addr == il0->iptr->ins->addr){
							lastoped->otherSuccessors = otherSuccs;
							break;
						}
					}


					al_add(succs->bptr->occurrences, &il0->iptr->ins->order);
					if(DEBUG_MESSAGES){
						printf("1 entry added to %d\n", succs->bptr->id);
					}
					int numInstrs = succs->bptr->numInstrs;
					while (il0) {
						if(!INSTR_HAS_FLAG(succs->bptr->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
							if(NODE_HAS_FLAG(il0->iptr, BBL_END)){
								break;
							}
						} else {
							numInstrs --;
							if(numInstrs == 0){
								break;
							}
						}
						il0 = il0->next;
					}
					if(il0){
						il0 = il0->next;
					}



					al_add(steps, lastoped);


					prevBbl = succs->bptr;
					break;					// Have no idea why this 'break' should be here!!!
				}
			}
			if(found){
				continue;
			}
			// then we should add this coming bbl to prevBbl's successors!
			if(!found){
				// can we go from prevBbl to ibbl?
				found = false;
				if(notSteppingBack){
					for (ibbl = nextFound; ibbl; ibbl = ibbl->next){
						if(il0->iptr->ins->addr == ibbl->ins_hd->iptr->ins->addr){
							found = true;
							break;
						}
					}
				}

				if(found){
					numPossibleSuccs = NumberOfPossibleSuccessors(prevBbl);
					if(DEBUG_MESSAGES){
						printf("bbl %d exists with starting: %d: [0x%08x] %s\n", ibbl->id,
								il0->iptr->ins->order, il0->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(il0->iptr->ins)));
					}
					if(!INSTR_HAS_FLAG(prevBbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE) || (numPossibleSuccs == -1 || numPossibleSuccs > prevBbl->numSuccs)){
						succs = malloc(sizeof(Bbl));
						succs->bptr = ibbl;
						succs->next = prevBbl->succs;
						prevBbl->succs = succs;
						prevBbl->numSuccs ++;
						al_add(ibbl->occurrences, &il0->iptr->ins->order);
						if(DEBUG_MESSAGES){
							printf("1 entry added to %d\n", ibbl->id);
						}

						lastOp *lastoped = malloc(sizeof(lastOp));
						lastoped->from = prevBbl;
						lastoped->to = ibbl;
						lastoped->prevBbl = prevBbl;
						lastoped->backToInstr = il0;
						nextFound = start;
						lastoped->removeEdge = true;
						lastoped->otherSuccessors = NULL;
						lastoped->bbls = al_new();
						bbls = lastoped->bbls;
						al_add(bbls, ibbl);


						al_add(steps, lastoped);

					} else {
						if(DEBUG_MESSAGES){
							printf("al_size is: %d\n", al_size(steps));
						}
			        	int k = al_size(steps);
			        	lastOp *lastOperation;

						lastOperation = al_get(steps, k - 1);
						il0 = lastOperation->backToInstr;
						if(DEBUG_MESSAGES){
							printf("Going back to %d\n", il0->iptr->ins->order);
						}
						prevBbl = lastOperation->prevBbl;
						remainingSuccs = lastOperation->otherSuccessors;
						bListNode *succsTmp;


						bbls = lastOperation->bbls;
						int l;
						for(l = 0; l < al_size(bbls); l ++){
							Bbl *tmp = al_get(bbls, l);
							al_removeLast(tmp->occurrences);
							if(DEBUG_MESSAGES){
								printf("last occurrences popped from %d\n", tmp->id);
							}
						}


						if(lastOperation->removeEdge == true){
							flag = true;
							for(succsTmp = succs = lastOperation->from->succs; succs; succs = succs->next){
								if(succs->bptr->id == lastOperation->to->id){
									// remove from --> to edge
									if(succs == succsTmp){
										lastOperation->from->succs = lastOperation->from->succs->next;
									} else {
										succsTmp->next = succs->next;
									}
									lastOperation->from->numSuccs --;
								}
								succsTmp = succs;
							}
						} else {
							flag = false;
						}


						bool OtherBblsFound = false;
						for(ibbl = lastOperation->to; ibbl; ibbl = ibbl->next){
							if(ibbl== lastOperation->to){
								continue;
							}
							if(ibbl->ins_hd->iptr->ins->addr == il0->iptr->ins->addr){
								OtherBblsFound = true;
								break;
							}
						}
						if(OtherBblsFound){
							nextFound = ibbl;
							notSteppingBack = true;
						} else {
							nextFound = start;
							notSteppingBack = false;
						}

						if(lastOperation->removeEdge == false && remainingSuccs == NULL){
							notSteppingBack = false;
						}
						if(remainingSuccs){
							notSteppingBack = true;
							flag = true;
						}

			        	al_removeAt(steps, k -1);
			        	continue;
					}

					numInstrs = ibbl->numInstrs;
					while (il0) {
						if(!INSTR_HAS_FLAG(ibbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
							if(NODE_HAS_FLAG(il0->iptr, BBL_END)){
								break;
							}
						} else {
							numInstrs --;
							if(numInstrs == 0){
								break;
							}
						}
						il0 = il0->next;
						//c --;
					}
					if(il0){
						il0 = il0->next;
					}
					prevBbl = ibbl;
					continue;
				}
			}
		}
		if(prevBbl){
			numPossibleSuccs = NumberOfPossibleSuccessors(prevBbl);
		}
        if ((!prevBbl || ((!INSTR_HAS_FLAG(prevBbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE) || (numPossibleSuccs == -1 || numPossibleSuccs > prevBbl->numSuccs))))
        		&& (!ibbl || !notSteppingBack)) {
        	notSteppingBack = true;
            ibbl = calloc(1, sizeof(Bbl));
            ibbl->id = id ++;
            ibbl->numInstrs = 0;
            ibbl->hasVariation = 0;     // there is only one kind of this bbl so far!

            if (end) {
                end->next = ibbl;
            }
            ibbl->prev = end;
            end = ibbl;
            if (start == NULL) {
                start = ibbl;
            }
            ibbl->occurrences = al_new();
            al_add(ibbl->occurrences, &(il0->iptr->ins->order));

            for (iltmp = il0; iltmp; iltmp = iltmp->next) {
                if (ibbl->ins_hd == NULL) {
                    ibbl->ins_hd = iltmp;
                }
                ibbl->numInstrs ++;

                if (NODE_HAS_FLAG(iltmp->iptr, BBL_END)) {
                	break;
                }
            }
            // add new bbl to previous' successors
            ibbl->ins_tl = iltmp;
            if(prevBbl){
                succs = malloc(sizeof(bListNode));
                succs->bptr = ibbl;
                succs->next = prevBbl->succs;
                prevBbl->succs = succs;
                prevBbl->numSuccs ++;
            }

            prevBbl = ibbl;
            nextFound = start;
            if(DEBUG_MESSAGES){
            printf("bbl %d added with starting: %d: [0x%08x] %s\n", ibbl->id,
                    ibbl->ins_hd->iptr->ins->order, ibbl->ins_hd->iptr->ins->addr,
                    ud_insn_asm(LoadUDisInstrObj(ibbl->ins_hd->iptr->ins)));
            }
            if (iltmp) {
                il0 = iltmp->next;
            } else {
                il0 = NULL;
            }
        } else {
        	//printf("al_size is: %d\n", al_size(steps));

        	int k = al_size(steps);
        	lastOp *lastOperaqtion;

			lastOperaqtion = al_get(steps, k - 1);
			il0 = lastOperaqtion->backToInstr;
			//printf("Going back to %d\n", il0->iptr->ins->order);
			prevBbl = lastOperaqtion->prevBbl;
			remainingSuccs = lastOperaqtion->otherSuccessors;

			bbls = lastOperaqtion->bbls;
			int l;
			for(l = 0; l < al_size(bbls); l ++){
				Bbl *tmp = al_get(bbls, l);
				al_removeLast(tmp->occurrences);
				//printf("last occurrences popped from %d\n", tmp->id);
			}

			bListNode *succsTmp;
			if(lastOperaqtion->removeEdge == true){
				flag = true;
				for(succsTmp = succs = lastOperaqtion->from->succs; succs; succs = succs->next){
					if(succs->bptr->id == lastOperaqtion->to->id){
						// remove from --> to edge
						if(succs == succsTmp){
							lastOperaqtion->from->succs = lastOperaqtion->from->succs->next;
						} else {
							succsTmp->next = succs->next;
						}
						lastOperaqtion->from->numSuccs --;
					}
					succsTmp = succs;
				}
			} else {
				flag = false;
			}

			bool OtherBblsFound = false;
			for(ibbl = lastOperaqtion->to; ibbl; ibbl = ibbl->next){
				if(ibbl== lastOperaqtion->to){
					continue;
				}
				if(ibbl->ins_hd->iptr->ins->addr == il0->iptr->ins->addr){
					OtherBblsFound = true;
					break;
				}
			}
			if(OtherBblsFound){
				nextFound = ibbl;
				notSteppingBack = true;
			} else {
				nextFound = start;
				notSteppingBack = false;
			}

			if(lastOperaqtion->removeEdge == false && remainingSuccs == NULL){
				notSteppingBack = false;
			}
			if(remainingSuccs){
				notSteppingBack = true;
				flag = true;
			}
        	al_removeAt(steps, k-1);
        }
    }
	bbl_list_hd = start;
	bbl_list_tl = end;
}



/*
 * MergeSimilarBbls
 *
 * look at the basic blocks that start with the same
 * address, look at the instructions on the two bbls
 * if the only difference is dead codes, then two bbls
 * can be merged to one!
 */

bool MergeSimilarBbls(InstrList *iList){
	Bbl *ibbl, *jbbl;
	Instruction *iins, *iinsPrev=NULL, *jins, *jinsPrev=NULL;
	bool bblsEqual = true;
	bool retVal = false;
	int iBblTimes, jBblTimes;

	for(ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next){
		for(jbbl = ibbl->next; jbbl; jbbl = jbbl->next){
			if(ibbl->ins_hd->iptr->ins->addr == jbbl->ins_hd->iptr->ins->addr && INSTR_HAS_FLAG(ibbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){				// potentially ibbl and jbbl are the same
			    //fprintf(stderr, "%d --> %d\n", ibbl->id, jbbl->id);
				iBblTimes = al_size(ibbl->occurrences);
				jBblTimes = al_size(jbbl->occurrences);
				while(iBblTimes && jBblTimes){


				int iorder = *(int *)al_get(ibbl->occurrences, --iBblTimes);
				int *jjorder = (int *)al_get(jbbl->occurrences, --jBblTimes);
				int jorder = *jjorder;

				iins = GetInstruction(iList, iorder);
				jins = GetInstruction(iList, jorder);

				while(iins->addr == jins->addr){
					iinsPrev = iins;
					jinsPrev = jins;
					iins = GetInstruction(iList, iins->next);
					jins = GetInstruction(iList, jins->next);
				}
				// now
				// CountInstructionsBetweenTwoInstrs()
				int idiff = CountInstructionsBetweenTwoInstrs2(iList, iinsPrev, iins);
				int jdiff = CountInstructionsBetweenTwoInstrs2(iList, jinsPrev, jins);
				// whichever is greater, must have some dead instructions in between

				if(idiff < jdiff){
					//fprintf(stderr, "%d --> %d\n", ibbl->id, jbbl->id);
					while(jinsPrev != jins){

						jinsPrev = GetInstruction(iList, jinsPrev->origNext);
						iinsPrev = GetInstruction(iList, iinsPrev->origNext);

					/*}
					while(jinsPrev != jins){
						jinsPrev = GetInstruction(iList, jinsPrev->order + 1);
						iinsPrev = GetInstruction(iList, iinsPrev->order + 1);*/
						if(iinsPrev->addr != jinsPrev->addr){
                            Instruction *iinsPrevTmp = GetInstruction(iList, iinsPrev->origNext);
                            Instruction *jinsPrevTmp = GetInstruction(iList, jinsPrev->origNext);
                            if(jinsPrevTmp->addr != iinsPrev->addr && iinsPrevTmp->addr != jinsPrev->addr){
                                bblsEqual = false;
                                break;
                            } else {
                                if(jinsPrevTmp->addr == iinsPrev->addr){
                                    jinsPrev = jinsPrevTmp;
                                } else {
                                    iinsPrev = iinsPrevTmp;
                                }
                            }

						}

						if(INSTR_HAS_FLAG(jinsPrev, INSTR_IS_DELETED) && !INSTR_HAS_FLAG(iinsPrev, INSTR_IS_DELETED)){
							bblsEqual = true;
							MakeInstrAlive(iList, jinsPrev);
							ud_t *udins_new = UDCreateNop();
							CarefullCopyUdins2InsStr(iList, jinsPrev, &(jinsPrev->uInstr), udins_new, 0);
							//fprintf(stderr, "%d: 0x%08x marked alive! ibbl:%d ,jbbl:%d\n", jinsPrev->order, jinsPrev->addr, ibbl->id, jbbl->id);
						}
					}
				} else if(jdiff < idiff){
					//fprintf(stderr, "%d --> %d\n", ibbl->id, jbbl->id);
					while(iinsPrev != iins){
						jinsPrev = GetInstruction(iList, jinsPrev->origNext);
						iinsPrev = GetInstruction(iList, iinsPrev->origNext);

						if(!iinsPrev || !jinsPrev) break;

						if(iinsPrev->addr != jinsPrev->addr){
						    /*if(ibbl->id == 2 && jbbl->id == 7){
						        return false;
						    }*/
						    // skip 1 instruction
						    Instruction *iinsPrevTmp = GetInstruction(iList, iinsPrev->origNext);
						    Instruction *jinsPrevTmp = GetInstruction(iList, jinsPrev->origNext);
						    if(jinsPrevTmp->addr != iinsPrev->addr && iinsPrevTmp->addr != jinsPrev->addr){
						        bblsEqual = false;
							    break;
						    } else {
						        if(jinsPrevTmp->addr == iinsPrev->addr){
						            jinsPrev = jinsPrevTmp;
						        } else {
						            iinsPrev = iinsPrevTmp;
						        }
						    }
						}
						if(INSTR_HAS_FLAG(iinsPrev, INSTR_IS_DELETED) && !INSTR_HAS_FLAG(jinsPrev, INSTR_IS_DELETED)){
							bblsEqual = true;
							MakeInstrAlive(iList, iinsPrev);
                            ud_t *udins_new = UDCreateNop();
                            CarefullCopyUdins2InsStr(iList, iinsPrev, &(iinsPrev->uInstr), udins_new, 0);

							free(udins_new);
						}

					}
				} else {
					bblsEqual = false;
					continue;
				}
				if(bblsEqual){
					retVal = true;
					//fprintf(stderr, "%d and %d are equal:\n", ibbl->id, jbbl->id);
					//fprintf(stderr, "%d in %d places and %d in %d\n", ibbl->id, al_size(ibbl->occurrences), jbbl->id, al_size(jbbl->occurrences));
				}
			}
			}
		}
	}
	return retVal;
}

void ConstructCFG(InstrList *iList, int DependencyAnlaysis) {

    CollectDistinctInstructionsByAdrress(iList);
    CollectInstrSuccsAndPredsOld(iList);
    MarkBblBoundariesByAddr(iList);

    fprintf(stderr, "Constructing basic blocks...");

    FixBblBacktrack6();
    AddBblPreds();

    fprintf(stderr, "done\n");

    MergeBblsNew();

#if 1
    int iter = 0;
    while(MergeSimilarBbls(iList) && (iter < 100)){

    	CleanUpCFGData();

    	CollectDistinctInstructionsByAdrress(iList);
		CollectInstrSuccsAndPredsOld(iList);
		MarkBblBoundariesByAddr(iList);

        FixBblBacktrack6();
        AddBblPreds();


        MergeBblsNew();
    }

#endif
}

void DependencyAnalysis(InstrList *iList){

    fprintf(stderr, "Dependency analysis...\n");
    fprintf(stderr, "Creating post-dominator tree...");
    PostDominatorTree();
    fprintf(stderr, "done\n");
    fprintf(stderr, "Find immidiate post-dominators...");
    ImmediateDominator();
    fprintf(stderr, "done\n");

    fprintf(stderr, "Find explicit dependencies...");
    FindExplicitControlDependencies(iList);
    fprintf(stderr, "done\n");

    fprintf(stderr, "Mark implicit control dependencies!\n");
    while(MarkImplicitControlDependenciesNew(iList)){
        AddNewControlDependencies(iList);
    }

    MarkUnsimplifiableInstrs(iList);

}

void CleanUpCFGData(){

    ins_list = ins_list_tl = NULL;		// jsut nulling the pointers, they should be freed!
    bbl_list_hd = bbl_list_tl = NULL;
    ins_list_byaddr = ins_list_tl_byaddr = NULL;


	bbl_id = 0;
	return;

    iListNode *il, *ilcp;
	int i;
	for(i = 0; i < HASHTABSZ; i++){
        for (il=hashtbl[i]; il; il=il->next) {

            for (ilcp=il->iptr->succs; ilcp; ilcp=ilcp->next) {
                free(ilcp);
            }

            for (ilcp=il->iptr->preds; ilcp; ilcp=ilcp->next) {
                free(ilcp);
            }
            free(il->iptr);
            free(il);
        }
        hashtbl[i]=NULL;
        
        for (il=addrhashtbl[i]; il; il=il->next) {
            for (ilcp=il->iptr->succs; ilcp; ilcp=ilcp->next) {
                free(ilcp);
            }
            
            for (ilcp=il->iptr->preds; ilcp; ilcp=ilcp->next) {
                free(ilcp);
            }
            free(il->iptr);
            free(il);
        }
        addrhashtbl[i]=NULL;
	}


    ins_list = ins_list_tl = NULL;		// jsut nulling the pointers, they should be freed!
    bbl_list_hd = bbl_list_tl = NULL;
    ins_list_byaddr = ins_list_tl_byaddr = NULL;

    
	bbl_id = 0;

}


/*
 * special CFG print routine, for doing
 * CFG similarity by Patrick!
 */

void PrintCFG(FILE *f, int printInstrs, uint64_t start, uint64_t end) {
    Bbl *ibbl;
    bListNode *bl0;

    fprintf(f, "digraph G {\n");
#ifdef __APPLE__
    fprintf(f, "    graph [dpi = 1];\n");
#else
    fprintf(f, "	size=\"10,10\";\n");
#endif
    fprintf(f, "    fontname = Helvetica;\n");
    fprintf(f, "    fontsize =14;\n");
    fprintf(f, "    center = TRUE;\n");
    fprintf(f, "    ratio = auto;\n");
    //  printf("    page = \"7.5,10;\"\n");

    //fprintf(f, "    E_N_T_R_Y [shape=box];\n");
    //fprintf(f, "    E_X_I_T [shape=box];\n");

    for (ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next) {



        fprintf(f, "    B%d [shape=box, label=\"", ibbl->id);
        iListNode *itmp;
        itmp = ibbl->ins_hd;
        if (INSTR_HAS_FLAG(itmp->iptr->ins, INSTR_IN_BASE_MODULE) || handle_rop) {
            if (printInstrs) {
                for (itmp = ibbl->ins_hd; itmp ; itmp = itmp->next) {
                	if(itmp != ibbl->ins_tl && IsControlFlowInstr(itmp->iptr->ins)){
                		continue;
                	}
                	if(strcmp(GetFunctionNameById(itmp->iptr->ins->funcName), "") != 0){
                		fprintf(f, " call %s \\l", GetFunctionNameById(itmp->iptr->ins->funcName));
                	}else {
                		fprintf(f, " %s \\l", ud_insn_asm(LoadUDisInstrObj(itmp->iptr->ins)));
                	}
					if(itmp == ibbl->ins_tl){
						break;
					}
                }
                fprintf(f, "\"");

            } else {
                fprintf(f, "%d\"", ibbl->id);
            }
        } else {
            fprintf(f, " %s\", color=\"%s\", style=\"filled\",fontsize=18",
                    GetFunctionNameById(ibbl->ins_hd->iptr->ins->funcName), SYSTEM_CALL_COLOR);
        }

#if 0
        if (NODE_HAS_FLAG(ibbl, BBL_FORWARD_TAINTED) && NODE_HAS_FLAG(ibbl, BBL_IS_CONDITIONAL)) {
            fprintf(f, ", color=\"%s\", style=\"filled\"", FORWARD_TAINTED_CONDITIONAL);
        } else if (NODE_HAS_FLAG(ibbl, BBL_FORWARD_TAINTED)) {
            fprintf(f, ", color=\"%s\", style=\"filled\"", BACKWARD_TAINT_COLOR);
        }

        if(NODE_HAS_FLAG(ibbl, BBL_BACKWARD_TAINTED)){
           fprintf(f, ", color=\"%s\", style=\"filled\"", FORWARD_TAINT_COLOR);
       }
#endif

        fprintf(f, "];\n"); //close the box definition!

        for (bl0 = ibbl->succs; bl0; bl0 = bl0->next) {
#if 0
            if(INSTR_HAS_FLAG(ibbl->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE) && !INSTR_HAS_FLAG(bl0->bptr->ins_hd->iptr->ins, INSTR_IN_BASE_MODULE)){
                fprintf(f, "    B%d -> B%d[style=dashed, color=red]\n", ibbl->id, bl0->bptr->id);
            } else
#endif
            {
                fprintf(f, "    B%d -> B%d\n", ibbl->id, bl0->bptr->id);
            }
        }

    }
    fprintf(f, "}\n");
}

void PrintBblInstrs(FILE *f){
    Bbl *ibbl;
    iListNode *itmp;

    for (ibbl = bbl_list_hd; ibbl; ibbl = ibbl->next) {
        fprintf(f, "Bbl %d:\n", ibbl->id);
        itmp = ibbl->ins_hd;
        for (itmp = ibbl->ins_hd; itmp; itmp = itmp->next) {
            fprintf(f, "\t  %d: [0x%08x]%s\n", itmp->iptr->ins->order, itmp->iptr->ins->addr, ud_insn_asm(LoadUDisInstrObj(itmp->iptr->ins)));
            if(itmp == ibbl->ins_tl){
            	break;
            }
        }
        fprintf(f, "\n");
        int i;
        for(i=0; i < al_size(ibbl->occurrences); i++){
        	fprintf(f, "\t%d\n", *(int *)al_get(ibbl->occurrences, i));
        }
    }
    fprintf(f, "\n");
}


///////////////////////////////End Graph/////////////////
