/*
 *
 * File: hash.c
 * Author: Babak Yadegari, Kevin Coogan
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */


#include <stdio.h>
#include "al_helper.h"
#include "hash.h"
#include "lynx_types.h"
#include "print.h"


/*********************************************************************
 *                                                                   *
 *                    HASH TABLE FOR INSTRUCTIONS                    *
 *                                                                   *
 *********************************************************************/

ihnode *ihTbl[HASHTBLSZ];

ihInstrnode *ihInstrTbl[HASHTBLSZ];

void InitInstrHashTable() {
    int i;

    for (i = 0; i < HASHTBLSZ; i++) {
        ihTbl[i] = NULL;
    }
}

static ihnode *ihNew() {
    ihnode *iptr = zalloc(sizeof(*iptr));
    return iptr;
}

/*
 * ihLookup(istr) -- lookup the hash table for the
 * instruction istr at address addr.  If the instruction
 * ointr found at the address of istr is the same as
 * istr, return a pointer to ostr; otherwise, replace
 * ostr with istr and return a pointer to istr.
 */
ihnode *ihLookup(Instruction *iins) {
    int hval;
    ihnode *iptr;

    hval = (iins->addr) & HASHMASK;
    for (iptr = ihTbl[hval]; iptr; iptr = iptr->next) {
        if (iptr->addr == iins->addr) {
            return iptr;
        }
    }

    return NULL;
}

ihnode *ihInsert(Instruction *iins) {
    ihnode *itmp = ihNew();
    int hval = (iins->addr) & HASHMASK;
    uint8_t *ptr;
    int i, len;

    itmp->addr = iins->addr;
    itmp->len = len = iins->uInstr.length;
    ptr = iins->uInstr.bytes;
    for (i = 0; i < len; i++) {
        itmp->bytes[i] = *ptr++;
    }
    itmp->ok = true;
    //itmp->ud_ins = ud_ins;
    itmp->next = ihTbl[hval];
    ihTbl[hval] = itmp;

    return itmp;
}

/*
 * Given an instruction iins that is a candidate to be simplified, and
 * a udins representations udins of what it would be simplified to,
 * UpdateSimplificationHashTbl(iins, udins) checks the hash table to see whether
 * the instruction at this address has been encountered for simplification
 * previously, and if so whether the simplified version is the same.
 * if the instruction was previously encountered and udins is different,
 * the ok flag of the hash table entry is set to false.
 */
void UpdateSimplificationHashTbl(Instruction *iins, ud_t *udins) {
    ihnode *ih;
    int i, len;
    uint8_t *ptr;

    ih = ihLookup(iins);
    if (ih == NULL) {
        if (udins != NULL) {
            //ihInsert(iins, udins);
        }
    } else {
        if (udins != NULL) {
            len = ud_insn_len(udins);
            ptr = ud_insn_ptr(udins);
            if (len == ih->len) {
                for (i = 0; i < len; i++) {
                    if (ih->bytes[i] != *ptr++) {
                        ih->ok = false;
                        //fprintf(stderr,"[UpdateSimplificationHashTbl] false1\n");
                        break;
                    }
                }
            } else {
                //fprintf(stderr,"[UpdateSimplificationHashTbl] false2\n");
                ih->ok = false;
            }
        } else {
            //fprintf(stderr,"[UpdateSimplificationHashTbl] false3\n");
            ih->ok = false;
        }
    }
}

void InitIinstrHashTable() {
    int i;

    for (i = 0; i < HASHTBLSZ; i++) {
        ihInstrTbl[i] = NULL;
    }
}

static ihInstrnode *ihInstrNew() {
    ihInstrnode *iptr = zalloc(sizeof(*iptr));
    return iptr;
}

/*
 * ihLookup(istr) -- lookup the hash table for the
 * instruction istr at address addr.  If the instruction
 * ointr found at the address of istr is the same as
 * istr, return a pointer to ostr; otherwise, replace
 * ostr with istr and return a pointer to istr.
 */
ihInstrnode *ihInstrLookup(Instruction *iins) {
    int hval;
    ihInstrnode *iptr;

    hval = (iins->addr) & HASHMASK;
    for (iptr = ihInstrTbl[hval]; iptr; iptr = iptr->next) {
        if (iptr->iins->addr == iins->addr) {
            return iptr;
        }
    }

    return NULL;
}

ihInstrnode *ihInstrInsert(Instruction *iins) {
    ihInstrnode *itmp = ihInstrNew();
    int hval = (iins->addr) & HASHMASK;


    itmp->iins = iins;
    itmp->next = ihInstrTbl[hval];
    ihInstrTbl[hval] = itmp;

    return itmp;
}


/*********************************************************************
 *                                                                   *
 *                    HASH TABLE FOR SCOPES                          *
 *                                                                   *
 *********************************************************************/



shnode *shTbl[HASHTBLSZ];


void InitScopeHashTable() {
    int i;
    
    for (i = 0; i < HASHTBLSZ; i++) {
        shTbl[i] = NULL;
    }
}

static shnode *shnodeNew() {
    shnode *iptr = zalloc(sizeof(*iptr));
    return iptr;
}

shnode *shLookup(uint64_t index) {
    int hval;
    shnode *iptr;
    
    hval = (index) & HASHMASK;
    for (iptr = shTbl[hval]; iptr; iptr = iptr->next) {
        if (iptr->scope->startIndex == index) {
            return iptr;
        }
    }
    
    return NULL;
}

shnode *shInsert(ConstantsinScope *s) {
    shnode *itmp = shnodeNew();
    int hval = (s->startIndex) & HASHMASK;
    
    itmp->scope = s;
    itmp->next = shTbl[hval];
    shTbl[hval] = itmp;
    
    return itmp;
}

