/*
 *
 * File: lynx_types.c
 * Author: Babak Yadegari, Kevin Coogan
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#include "lynx_types.h"
#include "instr.h"
#include "al_helper.h"

//extern ArrayList *BblList;           // global list of basic blocks
extern char *baseName;



/*
 * instead of keeping an array in each instruction
 * better to keep a global table and store the index
 * in each instruction and lookup the index in these
 * table if needed.
 */

char *ltrim(char *s)
{
    while(isspace(*s)) s++;
    return s;
}

char *rtrim(char *s)
{
    char* back = s + strlen(s);
    while(isspace(*--back));
    *(back+1) = '\0';
    return s;
}

char *trim(char *s)
{
    return rtrim(ltrim(s));
}


void InitializeModuleTable(char **module_names){

	FILE *modules = fopen(module_names_file, "r");
	if(!modules){
		return;
	}
	int i = 0;
	char *s = malloc(MAX_MODULE_LEN);
	while(fgets(s, MAX_MODULE_LEN, modules) && i < MAX_MODULE_NAMES){
	    char *new = malloc(MAX_MODULE_LEN);
		strncpy(new, trim(s), MAX_MODULE_LEN);
		module_names[i] = new;
		i++;
	}
	if(i == MAX_MODULE_NAMES){
		fprintf(stderr, "Table size smaller than module names in file, consider increasing the size of MAX_MODULE_NAMES\n");
		fclose(modules);
		exit(-1);
	}
	free(s);
	fclose(modules);
}

char *GetModuleNameById(int id){
	if(id == -1){
		return baseName;			// id -1 represent base module's name!
	}
	if(id > MAX_MODULE_NAMES){
		return NULL;
	}
	return module_names[id];
}

int ModuleNameExists(char *m_name){
	int i;
	for(i = 0; i < MAX_MODULE_NAMES; i++){
		if(module_names[i] == NULL){
			return -1;
		}
		if(strcmp(m_name, module_names[i]) == 0){
			return i;
		}
	}
	return -1;
}

int AddModuleNametoTable(char *m_name){
	int i;
	for(i = 0; i < MAX_MODULE_NAMES; i++){
		if(!module_names[i]){
			break;
		}
	}
	int len = strlen(m_name) + 1;
	char *new = malloc(len);
	strncpy(new, m_name, len);
	module_names[i] = new;

	// add the module name in file as well

	FILE *modules = fopen(module_names_file, "a+");
	fprintf(modules, "%s\n", m_name);
	fclose(modules);

	return i;
}



void InitializeFunctionNameTable(char **function_names){

    FILE *functions = fopen(function_names_file, "r");
    if(!functions){
        return;
    }
    int i = 0;
    char *s = malloc(MAX_FUNC_NAME_LEN);
    while(fgets(s, MAX_FUNC_NAME_LEN, functions) && i < MAX_FUNCTION_NAMES){
        char *new = malloc(MAX_FUNC_NAME_LEN);
        strncpy(new, trim(s), MAX_FUNC_NAME_LEN);
        function_names[i] = new;
        i++;
    }
    if(i == MAX_FUNCTION_NAMES){
        fprintf(stderr, "Table size smaller than function names in file, consider increasing the size of MAX_FUNCTION_NAMES\n");
        fclose(functions);
        exit(-1);
    }
    free(s);
    fclose(functions);
}

char *GetFunctionNameById(int id){
    if(id == -1){
        return "";
    }
    if(id > MAX_FUNCTION_NAMES){
        return NULL;
    }
    return function_names[id];
}

int FunctionNameExists(char *f_name){
    int i;
    for(i = 0; i < MAX_FUNCTION_NAMES; i++){
        if(function_names[i] == NULL){
            return -1;
        }
        if(strcmp(f_name, function_names[i]) == 0){
            return i;
        }
    }
    return -1;
}

int AddFunctionNametoTable(char *f_name){
    int i;
    for(i = 0; i < MAX_FUNCTION_NAMES; i++){
        if(!function_names[i]){
            break;
        }
    }
    int len = strlen(f_name) + 1;
    char *new = malloc(len);
    strncpy(new, f_name, len);
    function_names[i] = new;

    // add the module name in file as well

    FILE *modules = fopen(function_names_file, "a+");
    fprintf(modules, "%s\n", f_name);
    fclose(modules);

    return i;
}
/********************************************************
 *                                                      *
 * end of data for tables of module and function names  *
 *                                                      *
 *******************************************************/





// GLOBAL VARIABLES AND HASHTABLE MACROS
ADDRESS firstAddr= 0;
int iedgeIdctr = 0;
int edgeIdCtr = 0;
int blockIdCtr = 0;
int udedgeIdCtr = 0;

Range *RangeCreate(int min, int max) {
    Range *new = malloc(sizeof(Range));
    new->min = min;
    new->max = max;
    new->flag = 0;

    return (new);
}


void RangeDestroy(Range *r) {
    if (r) {
        free(r);
        r = NULL;
    }
}


Range *IntersectRange(Range *r1, Range *r2) {
    Range *new;
    int newMin = r1->min;
    int newMax = r1->max;

    if(r2->min > newMin) {
        newMin = r2->min;
    }

    if(r2->max < newMax) {
        newMax = r2->max;
    }

    if(newMax < newMin) {
        new = RangeCreate(1,0);
    } else {
        new = RangeCreate(newMin, newMax);
    }
    return (new);
}

bool RangeIsinRange(Range *s, Range *r){
    return ((s->min <= r->min && s->max >= r->max));
}

WriteSet *WriteSetCreate() {
    WriteSet *new = malloc(sizeof(WriteSet));
    new->ranges = al_new();

    return (new);
}



WriteSet *CloneWriteSet(WriteSet *orig) {
    if(orig == NULL) {
        return NULL;
    }

    WriteSet *clone = NULL;

    int i = 0;
    int size = al_size(orig->ranges);

    if(size > 0) {
        clone = WriteSetCreate();

        for (i = 0; i < size; i++) {
            Range *next = al_get(orig->ranges, i);
            Range *new = RangeCreate(next->min, next->max);
            new->flag = next->flag;
            AddRangeToWriteSet(clone, new);
        }
    }

    return (clone);
}


void WriteSetDestroy(WriteSet *ws) {
    if(ws != NULL) {
        al_freeWithElements(ws->ranges);
        free(ws);
    }
    ws = NULL;
}

/*
 * MaskWriteSet
 * returns a new WriteSet based on the mask
 * Assumes only one range in ws
 * Assumes mask refers to complete bytes of range (e.g. 000000ff)
 */
WriteSet *MaskWriteSet(WriteSet *ws, uint32_t mask) {
    if(ws == NULL) {
        return (NULL);
    }

    int size = al_size(ws->ranges);
    if((size == 0) || mask == 0x0) {
        return (NULL);
    }

    assert(size == 1);

    WriteSet *new = WriteSetCreate();
    Range *r = al_get(ws->ranges, 0);
    int newMin = 0, newMax = 0;

    if(mask & 0x000000ff) {
        newMin = r->min;
    } else if(mask & 0x0000ff00) {
        newMin = r->min + 1;
    } else if(mask & 0x00ff0000) {
        newMin = r->min + 2;
    } else {
        newMin = r->min + 3;
    }

    if(mask & 0xff000000) {
        newMax = r->max;
    } else if(mask & 0x00ff0000) {
        newMax = r->max - 1;
    } else if(mask & 0x0000ff00) {
        newMax = r->max - 2;
    } else {
        newMax = r->max - 3;
    }

    if(newMin <= newMax) {
        Range *newR = RangeCreate(newMin, newMax);
        AddRangeToWriteSet(new, newR);
    }

    return (new);
}

void AddRangeToWriteSet(WriteSet *ws, Range *r) {
    //check for "empty" range
	if(r == NULL) {
		return;
	}
    if(r->max < r->min) {
    	free(r);
        return;
    }

    int i, min, max;
    int len = al_size(ws->ranges);
    Range *temp = r;

    i = 0;
    while(i < len) {
        Range *next = (Range *)al_get(ws->ranges, i);

        // look for overlaps with existing ranges, we use tempmax +1 and
        // tempmin - 1 so that adjacent ranges are merged.
        // e.g. 1 to 5, and 6 to 10 --> 1 to 10
        if( ((temp->max+1) >= next->min) && ((temp->min-1) <= next->max) ) {
            // set new min and max for combined range
            min = temp->min;
            if(next->min < min) {
                min = next->min;
            }

            max = temp->max;
            if(next->max > max) {
                max = next->max;
            }

            // create new temp range
            temp->min = min;
            temp->max = max;

            // remove old range from list
            al_removeAt(ws->ranges, i);
            free(next);
		//break;
            len = al_size(ws->ranges);
        } else {
            i++;
        }
    }
    al_add(ws->ranges, temp);
}


/*
 * SubtractRangeFromWriteSet(ws, r) -- remove range r from writeset ws
 */

WriteSet *SubtractRangeFromWriteSet(WriteSet *ws, Range *r) {
  WriteSet *wtmp0, *wtmp1;

  wtmp0 = WriteSetCreate();
  AddRangeToWriteSet(wtmp0, r);

  wtmp1 = SubtractWriteSets(ws, wtmp0);

  WriteSetDestroy(wtmp0);

  return wtmp1;
}


bool ValInWriteSet(WriteSet *ws, int val) {
    bool valInWS = false;
    Iterator *itr = al_iterator(ws->ranges);

    while (!valInWS && itr->it_hasNext(itr)) {
        Range *r = (Range *)itr->it_next(itr);
        if ((val >= r->min) && (val <= r->max) ) {
            valInWS = true;
        }
    }
    free(itr);
    return (valInWS);
}





/* MergeWriteSets
 *  adds all ranges in write set 2 to write set 1, and returns write set 1
 *  NOTE: write set 1 is changed as a result of this function.
 */
WriteSet *MergeWriteSets(WriteSet *ws1, WriteSet *ws2) {
    Range *temp;
    int sz, i;

    sz = al_size(ws2->ranges);
    for (i = 0; i < sz; i++) {
        temp = (Range *)al_get(ws2->ranges, i);
        Range *new = RangeCreate(temp->min, temp->max);
        AddRangeToWriteSet(ws1, new);
    }

    return (ws1);
}

/*bool WriteSetIsEmpty(WriteSet *ws) {
	if(ws == NULL || ws->ranges == NULL) {
		return (true);
	}
    return ws->ranges->al_size;
}*/

bool WriteSetIsEmpty(WriteSet *ws) {
	if(ws == NULL) {
		return (true);
	}
    int len = al_size(ws->ranges);
    if(len == 0) {
        return (true);
    }

    bool isEmpty = false;
    Range *r = al_get(ws->ranges,0);
    if(r->max < r->min) {
        isEmpty = true;
    }

    return (isEmpty);
}



bool WriteSetEquals(WriteSet *ws1, WriteSet *ws2) {
    int i = 0, j = 0;
    bool same = true;
    int sz1 = al_size(ws1->ranges);
    int sz2 = al_size(ws2->ranges);

    for (i = 0; i < sz1; i++) {
        Range *next1 = (Range *) al_get(ws1->ranges, i);
        for(j = next1->min; j <= next1->max; j++) {
            if(!ValInWriteSet(ws2, j)) {
                same = false;
            }
        }
    }

    for (i = 0; i < sz2; i++) {
        Range *next2 = (Range *) al_get(ws2->ranges, i);
        for(j = next2->min; j <= next2->max; j++) {
            if(!ValInWriteSet(ws1, j)) {
                same = false;
            }
        }
    }

    return (same);
}



WriteSet *IntersectWriteSets(WriteSet *ws1, WriteSet *ws2) {
    WriteSet *new = WriteSetCreate();
    int sz1, sz2, i, j;

    sz1 = al_size(ws1->ranges);
    sz2 = al_size(ws2->ranges);

    for (i = 0; i < sz1; i++) {
        Range *next1 = (Range *) al_get(ws1->ranges, i);
	for (j = 0; j < sz2; j++) {
            Range *next2 = (Range *) al_get(ws2->ranges, j);
            AddRangeToWriteSet(new, IntersectRange(next1, next2));
        }
    }

    return(new);
}


WriteSet *UnionWriteSets(WriteSet *ws1, WriteSet *ws2) {
    WriteSet *new = WriteSetCreate();
    
    new = MergeWriteSets(new, ws1);
    new = MergeWriteSets(new, ws2);
    
    return new;
}

/*
 * WriteSetsOverlap(ws1, ws2) -- returns true if the intersection of
 * the ranges of ws1 and ws2 is nonempty; false otherwise.
 */
bool WriteSetsOverlap(WriteSet *ws1, WriteSet *ws2)
{
  Range *r1, *r2;
  int sz1 = al_size(ws1->ranges), sz2 = al_size(ws2->ranges);
  int i, j, max1, min1, max2, min2, newMax, newMin;

    for (i = 0; i < sz1; i++) {
      r1 = (Range *) al_get(ws1->ranges, i);
      max1 = r1->max;
      min1 = r1->min;

      for (j = 0; j < sz2; j++) {
        r2 = (Range *) al_get(ws2->ranges, j);
        max2 = r2->max;
        min2 = r2->min;

        newMax = (max1 < max2 ? max1 : max2);
        newMin = (min1 > min2 ? min1 : min2);

        if (newMin <= newMax) {
          return true;
        }
      }
    }

    return false;
}


/* RangeIsEmpty
 *  returns true is r is null, or if r represents a null range, e.g. (1,0)
 *  return false o.w.
 */
bool RangeIsEmpty(Range *r) {
    bool isEmpty = true;

    if(r && r->max >= r->min) {
        isEmpty = false;
    }

    return (isEmpty);
}


/*
 * RangeOverlapMask
 * returns a 32-bit masks that represents where two ranges overlap.
 * uses orig as base for the mask, and assumes values are byte-addressable.
 * e.g., orig = (1,4) overlap = (3,4), mask = 0xffff0000
 */
uint32_t RangeOverlapMask(Range *orig, Range *overlap) {
    uint32_t mask = 0x0;
    uint32_t byteMask = 0x000000ff;
    int i = 0;
    int origSize = orig->max - orig->min + 1;

    while (i < origSize) {
        if( ((orig->min + i) >= overlap->min) &&
                ((orig->min + i) <= overlap->max) ) {
            mask |= byteMask;
        }
        byteMask = byteMask << 8;
        i++;
    }

    return (mask);
}

/* ScrubWriteSet
 *  For given write set, removes any empty ranges to help improve speed of
 *  later calculations.  This function should *never need* to be called.
 */

void ScrubWriteSet(WriteSet *ws) {
    int size = al_size(ws->ranges);
    int cur = 0;

    while(cur < size) {
        Range *next = (Range *)al_get(ws->ranges, cur);
        if(RangeIsEmpty(next)) {
            al_removeAt(ws->ranges, cur);
            free(next);
            size = al_size(ws->ranges);
        } else {
            cur++;
        }
    }
}

/* SubtractWriteSets
 *  calculates ws1 - ws2
 *
 *  Algorithm is currently somewhat inefficient.  Subtracts ranges in ws2
 *  from each range in ws1.  In one case, this may result in two separate
 *  ranges.  If this happens, we modify existing range to account for first,
 *  create a new range for second, then add new range to the write set.
 *
 *  This requires that we reset the iterator to the beginning.  However, since
 *  write sets do not contain overlapping ranges, this is safe and will
 *  terminate.  It just causes some redundant calculation
 */
WriteSet *SubtractWriteSets(WriteSet *ws1, WriteSet *ws2) {
    WriteSet *new = WriteSetCreate();
    int sz1, i, sz2, j;
    sz1 = al_size(ws1->ranges);
    sz2 = al_size(ws2->ranges);

    for (i = 0; i < sz1; i++) {
        Range *next1 = (Range *)al_get(ws1->ranges, i);
        if(next1->min <= next1->max) {
        	AddRangeToWriteSet(new, RangeCreate(next1->min, next1->max));
        }
    }

    Iterator *itr1 = al_iterator(new->ranges);
    while(itr1->it_hasNext(itr1)) {
        Range *next1 = (Range *)itr1->it_next(itr1);

        for (j = 0; j < sz2; j++) {
            Range *next2 = (Range *) al_get(ws2->ranges, j);
            int newMin = 1, newMax = 0;
            Range *newRange = NULL;
            if((next1->max < next2->min) || (next1->min > next2->max)) {
                //then no intersection, return next1
                //AddRangeToWriteSet(new, RangeCreate(next1->min, next1->max));
            } else if((next1->min <= next2->min) && (next1->max < next2->max)) {
                //   |------|
                //       |------|
                //AddRangeToWriteSet(new, RangeCreate(next1->min, next2->min-1));
                //next1->min unchanged
                next1->max = next2->min-1;
            } else if((next1->min > next2->min) && (next1->max >= next2->max)) {
                //       |------|
                //   |------|
                //AddRangeToWriteSet(new, RangeCreate(next2->max+1, next1->max));
                next1->min = next2->max+1;
                //next1->max unchanged
            } else if((next1->min <= next2->min) && (next1->max >= next2->max)) {
                //   |---------|
                //      |---|
                // result of subtract is two ranges
                //AddRangeToWriteSet(new, RangeCreate(next1->min, next2->min-1));
                //first create new range, bc uses next1->max value
                newMin = next2->max+1;
                newMax = next1->max;
                //AddRangeToWriteSet(new, RangeCreate(next2->max+1, next1->max));
                //then adjust next1 with new max
                //next1->min unchanged
                next1->max = next2->min-1;
            } else if((next1->min > next2->min) && (next1->max < next2->max)) {
                //      |---|
                //   |---------|
                // result of subtract is empty, use empty range notation
                next1->min = 1;
                next1->max = 0;
            } else {
                printf("write set 1\n");
                PrintWriteSet(ws1);
                printf("write set 2\n");
                PrintWriteSet(ws2);
                fprintf(stderr,"UNKNOWN RANGE SUBTRACT CASE\n");
            }

            if(newMin <= newMax) {
                //add range
            	newRange = RangeCreate(newMin, newMax);
                AddRangeToWriteSet(new,newRange);

                //reset iterator to start over
                free(itr1);
                itr1 = al_iterator(new->ranges);
            }
        }
    }
    free(itr1);
    ScrubWriteSet(new);

    return(new);
}

/*
 * Assumes mask and ws are of same size!
 * Return new WriteSet that represents mask applied to ws
 *
 * WriteSet may contain up to 2 ranges!
 */
WriteSet *CalcWriteSetFromMask(WriteSet *ws, uint32_t mask) {
    assert(al_size(ws->ranges) == 1);

    WriteSet *new = CloneWriteSet(ws), *temp = NULL;
    uint32_t i = 0;
    uint32_t byteMask = 0xff;
    Range *r = al_get(ws->ranges, 0);

    for (i = r->min; i <= r->max; i++) {
        if((byteMask & mask) == 0) {
            Range *newR = RangeCreate(i, i);
            temp = SubtractRangeFromWriteSet(new, newR);
            WriteSetDestroy(new);
            new = temp;
            temp = NULL;
        }
        byteMask = byteMask << 8;
    }

    return (new);
}

/*
 * RangeInWriteSet -- returns true if range r is *completely* contained
 * within write set ws.
 */
bool RangeInWriteSet(WriteSet *ws, Range *r) {
  int i, sz;

    if((r == NULL) || (r->max < r->min)) {
        return (false);
    }

    bool inWS = false;

    sz = al_size(ws->ranges);
    for (i = 0; i < sz && !inWS; i++) {
      Range *next = (Range *) al_get(ws->ranges, i);
        if( (r->max >= next->min) && (r->max <= next->max) &&
                (r->min >= next->min) && (r->min <= next->max)) {
            inWS = true;
        }
    }

    return (inWS);
}



void fPrintWriteSet(FILE *fp, WriteSet *ws) {
    int sz, i;
    Range *r;
    
    sz = al_size(ws->ranges);
    for (i = 0; i < sz; i++) {
        r = (Range *) al_get(ws->ranges, i);
        fprintf(fp, "0x%08x--0x%08x; ", r->min, r->max);
    }
}


void PrintWriteSet(WriteSet *ws) {
    fPrintWriteSet(stdout, ws);
    fprintf(stdout, "\n");

}

Block *NewBbl() {
  Block *new = zalloc(sizeof(Block));

  new->id = blockIdCtr++;
  new->instrs = al_newGeneric(AL_LIST_SORTED, insCompareByOrder, refPrint, dealloc);
  //new->instrs = al_new();
  new->preds = al_new();
  new->succs = al_new();

  return new;
}

/*
 * DestroyBbl(bbl, destroyIns) -- deallocate the block bbl as well as its components.
 * destroyIns indicates whether the instructions in the block should also be destroyed.
 */
void DestroyBbl(Block *bbl, bool destroyIns) {
  al_free(bbl->preds);
  al_free(bbl->succs);

  if (destroyIns) {
    al_freeWithElements(bbl->instrs);
  }
  else {
    al_free(bbl->instrs);
  }

  bbl->id = -1;
  bbl->btype = 0;
  bbl->instrs = bbl->preds = bbl->succs = NULL;

  free(bbl);
}


/*
 * DestroyBblList(bblList, destroyIns) -- destroy an ArrayList of blocks.  The
 * boolean destroyIns indicates whether the instructions in each block should
 * also be destroyed.
 */
void DestroyBblList(ArrayList *bblList, bool destroyIns) {
  int i, sz;
  Block *bbl;
  
  sz = al_size(bblList);
  for (i = 0; i < sz; i++) {
    bbl = al_get(bblList, i);
    DestroyBbl(bbl, destroyIns);
  }
  
  al_free(bblList);
}

/*
 * CreateSpecialBbl -- create a "special" basic block that acts as a placeholder
 * for DLL routines.  We do this to ensure that control flow edges for DLL calls
 * don't dangle into emptiness.
 *
 * A special block abuses the fields of a "regular" basic block horribly:
 *
 *           preds[0]: name of callee module
 *           succs[0]: name of callee function
 *
 * Yeah, gross.
 */
Block *CreateSpecialBbl(char *module, char *fn) {
    Block *new = NewBbl();
    BBL_SET_FLAG(new, BBL_IS_DLL);
    al_add(new->preds, strdup(module));
    al_add(new->succs, strdup(fn));

    return new;
}



iEdge *iEdgeCreate() {
    iEdge *new = alloc(sizeof(iEdge));
    new->id = iedgeIdctr++;
    new->etype = ET_UNDEF;

    return (new);
}

void iEdgeDestroy(iEdge *edge) {
    free(edge);
    edge = NULL;
}


iEdge *iEdgeClone(iEdge *edge) {
    iEdge *new = iEdgeCreate();
    new->from = edge->from;
    new->to = edge->to;

    return (new);
}


iEdge *NewiEdge(EdgeType etype, Instruction *from, Instruction *to) {
    iEdge *new = alloc(sizeof(iEdge));

    new->id = iedgeIdctr++;
    new->etype = etype;
    new->from = from;
    new->to = to;

    return new;
}


Edge *EdgeCreate() {
    Edge *new = alloc(sizeof(Edge));
    new->etype = ET_UNDEF;

    return (new);
}

void EdgeDestroy(Edge *edge) {
    free(edge);
    edge = NULL;
}

Edge *NewEdge(EdgeType etype, Block *from, Block *to) {
    Edge *new = alloc(sizeof(Edge));

    new->id = edgeIdCtr++;
    new->etype = etype;
    new->from = from;
    new->to = to;

    return new;
}


udEdge *udEdgeCreate(EdgeType etype,
					 ADDRESS fromAddr,
					 ADDRESS toAddr) {
    udEdge *new = alloc(sizeof(udEdge));

    new->id = udedgeIdCtr++;
    new->etype = etype;
    new->fromAddr = fromAddr;
    new->toAddr = toAddr;

    return new;
}


iEdgeNode *iEdgeNodeCreate() {
  iEdgeNode *new = alloc(sizeof(iEdgeNode));
  new->prev = new->next = NULL;

  return new;
}

EdgeType GetiEdgeType(Instruction *from, Instruction *to) {
    assert(from && to);

    if (!INSTR_HAS_FLAG(from, INSTR_IN_BASE_MODULE)) {  /* undisassembled */
      return ET_UNKNOWN;
    }

    if (INSTR_HAS_FLAG(from, INSTR_USED_AS_CALL)) {
      return ET_CALL;
    }

    switch (from->uInstr.mnemonic) {
    case UD_Icall:
        return ET_CALL;

    case UD_Ihlt:
        return ET_UNEXPECTED;

    case UD_Ijo:
    case UD_Ijno:
    case UD_Ijb:
    case UD_Ijae:
    case UD_Ijz:
    case UD_Ijnz:
    case UD_Ijbe:
    case UD_Ija:
    case UD_Ijs:
    case UD_Ijns:
    case UD_Ijp:
    case UD_Ijnp:
    case UD_Ijl:
    case UD_Ijge:
    case UD_Ijle:
    case UD_Ijg:
    case UD_Ijcxz:
    case UD_Ijecxz:
    case UD_Ijrcxz:
        if (to->addr == from->addr+from->uInstr.length) {
            return ET_COND_NOTTAKEN;
        }
        else {
            return ET_COND_TAKEN;
        }

    case UD_Ijmp:
        return ET_JUMP;

    case UD_Iret:
    case UD_Iretf:
        return ET_RET;

    default:
        if (from->uInstr.length == 0) {  // for undisassembled DLL code
            return ET_UNKNOWN;
        }

        if (to->addr == from->addr+from->uInstr.length) {
            return ET_NORMAL;
        }
        else {
            return ET_EXCEPTION_EXEC;
        }
    }
    /* NOTREACHED */
}


/*
 * InstrIsControlTransfer(iins) -- returns true if iins may
 * cause a control transfer.
 */
bool InstrIsControlTransfer(InstrList *iList, Instruction *iins) {
    //Instruction *jins;
    //int lastIns, i;

    if (INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL)) {
      return true;
    }

   /* if(iins->uInstr.mnemonic == UD_Ijmp && iins->uInstr.operand[0].type == UD_OP_MEM){
        return false;
    }*/

    if (!INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL)) {   //undisassembled
        Instruction *next = GetInstruction(iList, iins->next);
        if(next && !INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE)){
            return false;
        }
    }

    switch (iins->uInstr.mnemonic) {
    case UD_Icall:
    case UD_Ijo:
    case UD_Ijno:
    case UD_Ijb:
    case UD_Ijae:
    case UD_Ijz:
    case UD_Ijnz:
    case UD_Ijbe:
    case UD_Ija:
    case UD_Ijs:
    case UD_Ijns:
    case UD_Ijp:
    case UD_Ijnp:
    case UD_Ijl:
    case UD_Ijge:
    case UD_Ijle:
    case UD_Ijg:
    case UD_Ijcxz:
    case UD_Ijecxz:
    case UD_Ijrcxz:
    case UD_Ijmp:
    case UD_Iret:
    case UD_Iretf:
      return true;

    default:
/*
      lastIns = iList.numInstrs - 1;
      i = iins->order;
      if (i < lastIns) {
        jins = GetInstruction(iList, i+1);
        if (jins->addr != iins->addr + iins->length) {  // exception
          return true;
        }
      }
*/

      return false;
    }
    /* NOTREACHED */
}



iEdgeNode *AddiEdgeToList(iEdgeNode *eList, iEdge *edge) {
    iEdgeNode *newList = alloc(sizeof(iEdgeNode));
    newList->edge = edge;
    if(eList == NULL) {
        newList->next = NULL;
        newList->prev = NULL;
    } else {
        newList->next = eList;
        newList->next->prev = newList;
        newList->prev = NULL;
    }
    return (newList);
}

/*
 * InstrIsEqual
 *  takes two instructions, returns true if they are the same, false ow
 *  "same" defined as same address and same name.
 */
bool InstrIsEqual(Instruction *instr1, Instruction *instr2) {
    bool isEqual = false;
    if(instr1->addr == instr2->addr) {
        if(strcmp(ud_insn_asm(LoadUDisInstrObj(instr1)), ud_insn_asm(LoadUDisInstrObj(instr2))) == 0) {
            isEqual = true;
        }
    }

    return (isEqual);
}


Instruction *FindInstrInList(InstrList *iList, Instruction *instr) {
    Instruction *foundInstr = NULL;
    int size = iList->numInstrs;
    int index = 0;
    bool found = false;

    while ((index < size) && !found) {
        foundInstr = GetInstruction(iList, index);
        if(InstrIsEqual(foundInstr, instr)) {
            found = true;
        } else {
            free(foundInstr);
            foundInstr = NULL;
            index++;
        }
    }

    return (foundInstr);
}



Instruction *FindInstrInArrayList(ArrayList *iList, Instruction *instr) {
    Instruction *xins = NULL;
    int sz = al_size(iList);
    int i;

    for (i = 0; i < sz; i++) {
        xins = al_get(iList, i);
        if(InstrIsEqual(instr, xins)) {
            return xins;
        }
    }

    return NULL;
}


Instruction *FindInstrInArrayListByAddr(ArrayList *list, ADDRESS addr) {
	int i = 0;
    Instruction *instr = NULL;
    int size = al_size(list);

    for (i = 0; i < size; i++) {
        instr = (Instruction *)al_get(list, i);
        if(instr->addr == addr) {
            return instr;
        }
    }

    return NULL;
}


/*
 * FindInstrPosInArrayList(list, ins) -- Given a sorted arraylist of instructions
 * list and an instruction ins, returns the position of ins within list.  If
 * ins is not found, a value of -1 is returned.
 */
int FindInstrPosInArrayList(ArrayList *iList, Instruction *instr) {
  ADDRESS iaddr = instr->addr;
  ADDRESS xaddr;
  Instruction *xinstr;
  int lo, hi, mid;  /* search bounds */

  lo = 0;
  hi = al_size(iList);
  do {
    mid = (lo+hi)/2;
    xinstr = al_get(iList, mid);
    xaddr = xinstr->addr;

    if (iaddr == xaddr) {  // found
      return mid;
    }

    if (lo >= hi) {        // not found
      break;
    }

    if (iaddr > xaddr) {  // continue in upper half
      lo = mid;
    }
    else {  // continue in lower half
      hi = mid;
    }
  } while (hi >= lo);

  return -1;
}


/*
 * FindInstrPosInArrayListByOrder(list, ins) -- Given a sorted arraylist of instructions
 * list and an instruction ins, returns the position of ins within list.  If
 * ins is not found, a value of -1 is returned.
 */
int FindInstrPosInArrayListByOrder(ArrayList *iList, Instruction *instr) {
  ADDRESS iorder = instr->order;
  ADDRESS xorder;
  bool found = false;
  Instruction *xinstr;
  int lo, hi, mid;  /* search bounds */
  int instrPos = -1;

  lo = 0;
  hi = al_size(iList) - 1;

  while ((lo <= hi) && !found) {
	  mid = (hi + lo)/2;
	  xinstr = al_get(iList, mid);
	  xorder = xinstr->order;
	  if(iorder == xorder) {
		  found = true;
		  instrPos = mid;
	  } else {
		  if(iorder > xorder) {
			  lo =  mid + 1;
		  } else {
			  hi = mid - 1;
		  }
	  }
  }

  return (instrPos);
}


/*
 * InstrInList
 *  returns true if instruction exists in the list such that
 *  addr and name are identical.
 */
bool InstrInList(InstrList *iList, Instruction *instr) {
    bool found = false;
    Instruction *foundInstr = FindInstrInList(iList, instr);
    if(foundInstr) {
        found = true;
    }
    free(foundInstr);
    return (found);
}



/*
 * InstrInArrayList
 *  returns true if instruction exists in the list such that
 *  addr and name are identical.
 */
bool InstrInArraylist(ArrayList *iList, Instruction *instr) {
    bool found = false;
    Instruction *foundInstr = FindInstrInArrayList(iList, instr);
    if(foundInstr) {
        found = true;
    }

    return (found);
}


udEdge *FindudEdgeInArrayList(ArrayList *list, ADDRESS from, ADDRESS to) {
    udEdge *edge = NULL;
    int sz = al_size(list);
    int i;

    for (i = 0; i < sz; i++) {
        edge = (udEdge *)al_get(list, i);
        if((edge->fromAddr == from) && (edge->toAddr == to)) {
            return edge;
        }
    }

    return NULL;
}


/*
 * WriteSetSubsetOfWriteSet(ws_sub, ws_super) -- returns true if
 * ws_sub is contained in ws_super, false o/w.
 */
bool WriteSetSubsetOfWriteSet(WriteSet *ws_sub, WriteSet *ws_super) {
  int i, sz;

  if (WriteSetIsEmpty(ws_sub)) {
    return true;
  }

  if (WriteSetIsEmpty(ws_super)) {
    return false;  // since ws_sub is nonempty
  }

  sz = al_size(ws_sub->ranges);
  for (i = 0; i < sz; i++) {
    Range *r = (Range *) al_get(ws_sub->ranges, i);
    if (!RangeOverlapWithWriteSet(ws_super, r)) {
      return false;
    }
  }

  return true;
}

/*
 * RangeOverlapWithWriteSet(WriteSet *ws, Range* r) -- returns true if
 * range r is contained in ws, false o/w.
 */
bool RangeOverlapWithWriteSet(WriteSet *ws, Range* r){
	int sz,i;
	Range *tmp;
	if(WriteSetIsEmpty(ws) || RangeIsEmpty(r))
		return false;
	sz = al_size(ws->ranges);
	for (i = 0; i < sz; i++){		// check to see if one of r's boundaries is within at least on range in writeset!
		tmp = al_get(ws->ranges, i);
		if((r->min >= tmp->min && r->min <= tmp->max) || (r->max >= tmp->min && r->max <= tmp->max))
			return true;
	}
	return false;
}

PageTable *ClonePageTables(PageTable *orig) {
    PageTable *walker = orig;
    int i, numPageTables = 0;
    while(walker->pageAddr != NULL) {
        numPageTables++;
        walker++;
    }

    //there are numPageTables PageTable objects, plus one which marks the end
    PageTable *newPageDir = malloc((numPageTables+1) * sizeof(PageTable));

    //now copy each entry in the pageTable array, and allocate space as nec.
    for(i = 0; i < numPageTables; i++) {
        newPageDir[i].startAddr = orig[i].startAddr;
        newPageDir[i].numPages = orig[i].numPages;
        newPageDir[i].pageAddr = malloc(
                sysconf(_SC_PAGE_SIZE)*newPageDir[i].numPages);
    }

    newPageDir[numPageTables].numPages = orig[numPageTables].numPages;
    newPageDir[numPageTables].startAddr = orig[numPageTables].startAddr;
    newPageDir[numPageTables].pageAddr = NULL; //this should be same

    return (newPageDir);
}

void DestroyPageTables(PageTable *pt) {
    PageTable *walker = pt;
    while(walker->pageAddr != NULL) {
        free(walker->pageAddr);
        walker++;
    }

    free(pt);
}

