/*
 *
 * File: mem_maps.c
 * Author: Babak Yadegari
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */


#include "mem_maps.h"


/*
 * clear all memory tags used to determine
 * a memory location is constant or not.
 * Also could be used in liveness analysis
 * (as well as in taint analysis) 0 indicating
 * dead locations and 1 for live  (tainted) memory locations.
 */
int ClearConstantPages(InstrList *iList, char val){
	int i;
	for(i = 0; i < iList->constantNumPageEntries; i++){
		memset(iList->constantPageDir[i].Constatns, val, PAGE_SIZE);
	}
	return 0;
}

int ClearMemoryPages(InstrList *iList, char val){
    int r;
    for (r = 0; r < iList->constantNumPageEntries; r++) {
        int j;
        for (j = 0; j < 8 && iList->constantPageDir[r].pages[j]; j++) {
        		memset(iList->constantPageDir[r].pages[j], val, PAGE_SIZE * sizeof(uint64_t));
        }
    }
    return 0;
}

int ClearConstantAndMemoryPages(InstrList *iList, char val){
    return (ClearConstantPages(iList, val) & ClearMemoryPages(iList, val));
}

void MemoryPageTableLookup(InstrList *iList, ADDRESS addr, int *index){
	//struct IndexAddressPair *tmpIndAddr = malloc(sizeof(IndexAddressPair));

	int i,j;
	for(i = 0; i < iList->constantNumPageEntries; i++){
		for(j = 0; j < 8; j++){
			int startAddr = iList->constantPageDir[i].startAddrs[j];
			if(addr >= startAddr && addr < startAddr + PAGE_SIZE){
				index[0] = i;
				index[1] = j;
				return;
			}
		}
	}
	index[0] = 0;
	index[1] = 0;
	fprintf(stderr, "[MemoryPageTableLookup] Error: could not find a page table for address %08x\n", addr);
}


int SetAddrAsConst(InstrList *iList, ADDRESS addr, int numBytes){

	int index[2];
	MemoryPageTableLookup(iList, addr, index);

	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int i;
	int offset = ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]));
	for(i = offset; i < offset + numBytes; i++){
		page[(index[1] * PAGE_SIZE/8) + (i >> 3)] |= (1 << (7 - (i & 0x7)));			// right? right?
	}
	return 0;
}


int UnsetConstAddr(InstrList *iList, ADDRESS addr, int numBytes){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	unsigned char *valuepage = iList->constantPageDir[index[0]].pages[index[1]];
	int i;
	int offset = ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]));
	for(i = offset; i < offset + numBytes; i++){
		page[(index[1] * PAGE_SIZE/8) + (i >> 3)] &= ~(1 << (7 - (i & 0x7)));
	}


    for(i = 0; i < numBytes; i++){
        *(valuepage + (addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]) + i) = 0;
    }

	return 0;
}

int UnsetConstAddrWithoutMemReset(InstrList *iList, ADDRESS addr, int numBytes){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    unsigned char *page = iList->constantPageDir[index[0]].Constatns;
    int i;
    int offset = ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]));
    for(i = offset; i < offset + numBytes; i++){
        page[(index[1] * PAGE_SIZE/8) + (i >> 3)] &= ~(1 << (7 - (i & 0x7)));
    }

    return 0;
}

int SetAddrAsConstWithValue(InstrList *iList, ADDRESS addr, int numBytes, int val){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	unsigned char *valuepage = iList->constantPageDir[index[0]].pages[index[1]];
	int i;

	//int offset = (index[1] * PAGE_SIZE/8) + ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]) >> 3);


	for(i = 0; i < numBytes; i++){
        //mask |= (1 << (7 - ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]] + i) & 0x7)));
        *(valuepage + (addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]) + i) = val;
        val = val >> 8;
	}

	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int offset = ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]));
	for(i = offset; i < offset + numBytes; i++){
		page[(index[1] * PAGE_SIZE/8) + (i >> 3)] |= (1 << (7 - (i & 0x7)));			// right? right?
	}

	return 0;
}

bool MemLocIsConstant(InstrList *iList, ADDRESS addr, int numBytes){

	if(addr == 0){
		return false;
	}

	bool memState = true;
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int i;
	int offset = ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]));
	//int offset2 = ((addr + numBytes - iList->constantPageDir[index[0]].startAddrs[index[1]]));
	for(i = offset; i < offset + numBytes; i++){
        //mask |= (0x80000000 >> (i & 0xf));
		memState = (page[(index[1] * PAGE_SIZE/8) + (i >> 3)] & (1 << (7 - (i & 0x7)))) ? true : false;
	}
	/*unsigned char page = iList->constantPageDir[index[0]].Constatns[(index[1] * PAGE_SIZE/8) + (offset1 >> 3)];
	if(addr == 0x12ff3c){
		fprintf(stderr, "locisconst: &page = %x consts = %x and %d\n", &page, iList->constantPageDir[index[0]].Constatns, (index[1] * PAGE_SIZE/8) + (offset1 >> 3));
	}
	uint *upage =  (uint *) &page;*/
	//memState = *upage & mask;
	return memState;
}

bool PartialMemLocIsConstant(InstrList *iList, ADDRESS addr, int numBytes){
	if(addr == 0){
		return false;
	}

	bool memState = false;
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int i;
	int offset = ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]));
	for(i = offset; i < offset + numBytes; i++){
		memState |= (page[(index[1] * PAGE_SIZE/8) + (i >> 3)] & (1 << (7 - (i & 0x7)))) ? true : false;
	}

	return memState;
}


int GetMemoryLocValue(InstrList *iList, ADDRESS addr, int numBytes){
	uint value = 0;
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	unsigned char *page = iList->constantPageDir[index[0]].pages[index[1]];

	int i;
	for(i = numBytes - 1; i >= 0; i--){
		value = value << 8;
		value = value + (unsigned char)page[addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]] + i];
	}

	return value;
}


int SetMemoryLocValue(InstrList *iList, ADDRESS addr, int val){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	unsigned char *page = iList->constantPageDir[index[0]].pages[index[1]];

	page[addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]] = val;

	return val;
}

size_t GetMemoryLocPointer(InstrList *iList, ADDRESS addr){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    size_t *valuepage = (size_t *)iList->constantPageDir[index[0]].pages[index[1]];
    size_t pointer = *(valuepage + ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]])));
    return pointer;
}

void SetMemoryLocPointer(InstrList *iList, ADDRESS addr, void *pointer){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    size_t *valuepage = (size_t *)iList->constantPageDir[index[0]].pages[index[1]];
    *(valuepage + ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]))) = (size_t)pointer;
}


/////////////////////////////////////////////
/////// Helpers for DependencyAnlysis////////
/////////////////////////////////////////////
int DependencyMemSetConstWithSrc(InstrList *iList, ADDRESS addr, int numBytes, uint64_t src){

    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	uint64_t *valuepage = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];
	int i;
	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int offset = ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]));

	for(i = offset; i < offset + numBytes; i++){
		(page[(index[1] * PAGE_SIZE/8) + (i >> 3)] |= (1 << (7 - (i & 0x7))));			// right? right?
	}

	*(valuepage + ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]))) = src;

	return 0;
}

uint64_t DependencyMemGetConstLocSrc(InstrList *iList, ADDRESS addr){

	uint64_t value = 0;
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	uint64_t *page = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];

	value = page[(addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]])];

	return value;
}



////////////////////////////////////////
/////// Helpers for TaintAnlysis////////
////////////////////////////////////////

int UnSetTaintedMemoryLocAsContainsFlags(InstrList *iList, ADDRESS addr, int numBytes){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    uint64_t *page = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];
    uint *upper32 = (uint *)&(page[(addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]])]) + 1;
    *upper32 = 0;
    return 0;
}

int SetTaintedMemoryLocAsContainsFlags(InstrList *iList, ADDRESS addr, int numBytes){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    uint64_t *valuepage = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];
    *(valuepage + ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]))) |= (1LLU << 32);

    return 0;
}

bool TaintedMemoryLocContainsFlags(InstrList *iList, ADDRESS addr, int numBytes){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    uint64_t *page = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];
    uint *upper32 = (uint *)&(page[(addr - (int)iList->constantPageDir[index[0]].startAddrs[index[1]])]) + 1;

    return (*upper32 & 1);
}



int TaintMemoryMask(int numBytes, int taint){
	int mask = 0xff;
	int i;
	for(i = 1; i < numBytes; i ++){
		mask |= (0xff << (i*8));
	}
	return (taint & mask);
}


int TaintMemoryLoc(InstrList *iList, ADDRESS addr, int numBytes, int taint){

    int index[2];
    MemoryPageTableLookup(iList, addr, index);

	uint64_t *valuepage = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];
	int i;
	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int offset = ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]));

	for(i = offset; i < offset + numBytes; i++){
		(page[(index[1] * PAGE_SIZE/8) + (i >> 3)] |= (1 << (7 - (i & 0x7))));			// right? right?
	}

	taint = TaintMemoryMask(numBytes, taint);
	//*(valuepage + ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]))) = 0;
	uint *p2ValPage = (uint *)(valuepage + ((addr - (int)iList->constantPageDir[index[0]].startAddrs[index[1]])));
	*p2ValPage = taint;

	return 0;
}


int SetMemoryLocTainted(InstrList *iList, ADDRESS addr, int numBytes){

    int index[2];
    MemoryPageTableLookup(iList, addr, index);

    int i;
    unsigned char *page = iList->constantPageDir[index[0]].Constatns;
    int offset = ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]));

    for(i = offset; i < offset + numBytes; i++){
        (page[(index[1] * PAGE_SIZE/8) + (i >> 3)] |= (1 << (7 - (i & 0x7))));          // right? right?
    }
    return 0;
}

int UntaintMemoryLoc(InstrList *iList, ADDRESS addr, int numBytes){

	// it is like tainting the memory location with 0!
    int index[2];
    MemoryPageTableLookup(iList, addr, index);
	uint64_t *valuepage = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];
	int i;
	unsigned char *page = iList->constantPageDir[index[0]].Constatns;
	int offset = ((addr - iList->constantPageDir[index[0]].startAddrs[index[1]]));

	for(i = offset; i < offset + numBytes; i++){
		(page[(index[1] * PAGE_SIZE/8) + (i >> 3)] &= ~(1 << (7 - (i & 0x7))));			// right? right?
	}

	*(valuepage + ((addr - (int) iList->constantPageDir[index[0]].startAddrs[index[1]]))) = 0;

	return 0;
}

bool MemoryLocIsTainted(InstrList *iList, ADDRESS addr, int numBytes){

	return PartialMemLocIsConstant(iList, addr, numBytes);
}

int SetMemoryLocSymbolic(InstrList *iList, ADDRESS addr, int numBytes){
    return SetMemoryLocTainted(iList, addr, numBytes);

}

int MakeMemoryLocConcrete(InstrList *iList, ADDRESS addr, int numBytes){
    return UntaintMemoryLoc(iList, addr, numBytes);
}

bool MemoryLocIsSymbolic(InstrList *iList, ADDRESS addr, int numBytes){
    
    return PartialMemLocIsConstant(iList, addr, numBytes);
}

int GetMemoryLocTaint(InstrList *iList, ADDRESS addr, int numBytes){
    int index[2];
    MemoryPageTableLookup(iList, addr, index);
	uint64_t *valuepage = (uint64_t *)iList->constantPageDir[index[0]].pages[index[1]];

	int taint = *(valuepage + ((addr - (int)iList->constantPageDir[index[0]].startAddrs[index[1]])));
	taint = TaintMemoryMask(numBytes, taint);

	return taint;
}



////////////////////////////////////////
/////// Helpers for liveness analysis///
////////////////////////////////////////
int SetMemoryMap(InstrList *iList, char val){
    return ClearConstantPages(iList, val);
}

bool MemoryLocationIsLive(InstrList *iList, ADDRESS addr, int numBytes){
    return PartialMemLocIsConstant(iList, addr, numBytes);
}

int SetMemoryLocationLive(InstrList *iList, ADDRESS addr, int numBytes){
    return SetAddrAsConst(iList, addr, numBytes);
}

int SetMemoryLocationDead(InstrList *iList, ADDRESS addr, int numBytes){
    return UnsetConstAddr(iList, addr, numBytes);
}
