/*
 *
 * File: print.c
 * Code to print out a "reasonable" assembly code program for an
 * instruction trace summary.  This code assumes that the summary
 * list has been sorted by address and that function and exception
 * handler entry points have been marked.
 *
 * Author: Babak Yadegari, Kevin Coogan
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#include <stdio.h>
#include <ctype.h>
#include "lynx_types.h"
#include "instr.h"
#include "print.h"
#include "opcode_names.h"

/*********************************************************************
 *                              EXTERNS                              *
 *********************************************************************/

extern ADDRESS entryAddr; // program entry point
extern char *inFile; // input trace file
extern char *baseName; // same as BaseModule (but still needed!)

/*********************************************************************/


/*
 * if 'full_list' is true, then all instructions in InstrList are printed
 * else only print base module.
 */

void PrintInstrListInternal(InstrList *list, int full_list) {
    int size = list->numInstrs;
    int index;
    bool printRegs = true;
    Instruction *next = NULL, *nextPlus1 = NULL;

    for (index = 0; index < size; index++) {
        next = GetInstruction(list, index);
        if (!INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE) && !full_list) {
            free(next);
            continue;
        }
        if (index < size - 1) {
            nextPlus1 = GetInstruction(list, index + 1);
        }
        printf("%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            printf("EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);

            printf("  ");
            PrintMemTargets(list, next);
            printf("\n");
        }
        RInstrDestroy(list, next);
        next = NULL;
        if (index < size - 1) {
            RInstrDestroy(list, nextPlus1);
            nextPlus1 = NULL;
        }
    }
}

void PrintInstrList(InstrList *list) {
    PrintInstrListInternal(list, 0);
}

void PrintFullInstrList(InstrList *list) {
    PrintInstrListInternal(list, 1);
}

char *OpCodeToString(enum ud_mnemonic_code code) {

    return NULL;
}

void PrintTraceString(InstrList *list) {
    int size = list->numInstrs;
    int index;
    int code;
    Instruction *next = NULL;

    for (index = 0; index < size; index++) {
        next = GetInstruction(list, index);
        if (!INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE)) {
            free(next);
            continue;
        }
        code = next->uInstr.mnemonic;
        if (code == opcode[code].code) {//sanity check
            printf("%-15.15s\n", opcode[next->uInstr.mnemonic].name);
        } else {
            //we're insane, get out now
            fprintf(stderr, "Something went really, really wrong\n");
            exit(0);
        }

        free(next);
        next = NULL;
    }
}


/*
 * StringIsNotWhitespace(str) -- returns true if the string contains
 * at least one non-whitespace character.
 */
bool StringIsNotWhitespace(char *str) {
    for (; str != NULL && *str != '\0'; str++) {
        if (!isspace(*str)) {
            return true;
        }
    }

    return false;
}

void PrintPgmInfo(FILE *fp) {
    fprintf(fp, "/*\n");
    fprintf(fp, " * +++ Program:     %s\n", baseName);
    fprintf(fp, " * +++ Input file:  %s\n", inFile);
    fprintf(fp, " * +++ Entry point: 0x%x\n", entryAddr);
    fprintf(fp, " */\n\n");
}

void PrintAsmCode(FILE *fp, ArrayList *alist, InstrList *iList, iEdgeNode *elist) {
    int i, sz;
    Instruction *iins;
#if 0
    EdgeNode *en;
    Edge *edge;
#endif

    assert(alist);

    sz = al_size(alist);

    /*
     * Mark each CALL instruction whose target function has a name.
     */
    for (i = 0; i < sz; i++) {
        iins = al_get(alist, i);
        INSTR_REMOVE_FLAG(iins, INSTR_MARKED);
    }

    //Instructions used as calls are now marked when UseDef chains calculated
    //loop below is obsolete.
    /*
     for (en = elist; en; en = en->next) {
     Instruction *to = GetInstruction(iList, en->edge->to);
     Instruction *toSum = FindInstrInArrayList(alist, to);
     if (toSum && en->edge->etype == ET_CALL && StringIsNotWhitespace(toSum->funcName)) {
     Instruction *from = GetInstruction(iList, en->edge->from);
     Instruction *fromSum = FindInstrInArrayList(alist, from);
     INSTR_SET_FLAG(fromSum, INSTR_MARKED);
     free(from);
     }
     free(to);
     }
     */
    /*
     * Traverse the instruction list and print out the instructions.
     */
    for (i = 0; i < sz; i++) {
        iins = al_get(alist, i);

        if (INSTR_HAS_FLAG(iins, INSTR_IGNORED)) {
            continue;
        }

        if (INSTR_HAS_FLAG(iins, (INSTR_FN_ENTRY|INSTR_HANDLER_ENTRY))) {
            if (i > 0) {
                fprintf(fp, "\n");
            }

            fprintf(fp, "/*\n");
            if (INSTR_HAS_FLAG(iins, INSTR_FN_ENTRY)) {
                fprintf(fp, " * -- Function: f_%x", iins->addr);
            } else if (INSTR_HAS_FLAG(iins, INSTR_HANDLER_ENTRY)) {
                fprintf(fp, " * -- Exception handler: h_%x", iins->addr);
            }

            //            if (INSTR_HAS_FLAG(iins, INSTR_PGM_ENTRY)) {
            //                fprintf(fp, " [PROGRAM ENTRY]");
            //            }
            fprintf(fp, "\n");
            /*
             * print a list of call sites
             */
            fprintf(fp, " * -- Reached from: ");
#if 0
            for (en = elist; en; en = en->next) {
                Instruction *to = GetInstruction(iList, en->edge->to);
                Instruction *toSum = FindInstrInArrayList(alist, to);
                free(to);
                if (toSum && InstrIsEqual(toSum, iins)) {
                    Instruction *from = GetInstruction(iList, en->edge->from);
                    Instruction *fromSum = FindInstrInArrayList(alist, from);
                    free(from);
                    if(fromSum) {
                        fprintf(fp, "  0x%08x", fromSum->addr);
                    }
                }
            }

            fprintf(fp, "\n */\n");
#endif
        }

        if (INSTR_HAS_FLAG(iins, INSTR_BBL_START)) {
            printf("----------\n");
        }

        if (INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
            fprintf(fp, "/* In UDChain: ");
        } else {
            fprintf(fp, "/*             ");
        }
        if (iins->module == -1) {
            fprintf(fp, "%08x */\t%s", iins->addr, ud_insn_asm(LoadUDisInstrObj(iins)));
        } else {
            fprintf(fp, "%s: %08x */\t%s", GetModuleNameById(iins->module), iins->addr, ud_insn_asm(LoadUDisInstrObj(iins)));
        }
        /*
         * For instructions that raise an exception, print the handler address.
         * In theory, the same instruction can, at different times, raise
         * exceptions that are handled by different handlers, so we search the
         * entire edge list.
         */
        //        if (INSTR_HAS_FLAG(iins, INSTR_RAISES_EXCEPTION)) {
        //          fprintf(fp, "\t/* exception --> ");
#if 0
        for (en = elist; en; en = en->next) {
            edge = en->edge;
            Instruction *from = GetInstruction(iList, edge->from);
            if (InstrIsEqual(from, iins) && edge->etype == ET_EXCEPTION) {
                Instruction *to = GetInstruction(iList, edge->to);
                fprintf(fp, "0x%08x ", to->addr);
                free(to);
            }
            free(from);
        }
        fprintf(fp, "*/");
#endif
        //  }
        /*
         * For function calls, print the target function name if available.
         * In theory, the same instruction can, at different times, call different
         * functions [indirect calls], so we search the entire edge list.
         */
        if (INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL)) {
#if 0
            for (en = elist; en; en = en->next) {
                edge = en->edge;
                Instruction *from = GetInstruction(iList, edge->from);
                if (InstrIsEqual(from, iins)/* && edge->etype == ET_CALL*/) {
                    Instruction *to = GetInstruction(iList, edge->to);
                    if(StringIsNotWhitespace(to->funcName)) {
                        fprintf(fp, "\t/* target: %s ", to->funcName);

#if 1
                        /*
                         * Lookup info about DLL routines.  This should actually
                         * be done prior to UD chain computation.
                         */
                        if (strcmp(iins->module, to->module) != 0) {
                            w32api *wptr = apiLookup(to->funcName);
                            if (wptr != NULL) {
                                printf(" [nargs: %d; retval: %d] ", wptr->nargs, wptr->retval);
                            }
                        }
                        fprintf(fp, "*/");
#endif
                    }
                    free(to);
                }
                free(from);
            }
#endif
        }

        fprintf(fp, "\n");
    }
}

void PrintMemTargets(InstrList *iList, Instruction *iins) {
    WriteSet *memUsed = MemUsed(iList, iins);
    WriteSet *memDef = MemDef(iList, iins);

    printf("R: ");
    if (!WriteSetIsEmpty(memUsed)) {
        fPrintWriteSet(stdout, memUsed);
    }

    printf("W: ");
    if (!WriteSetIsEmpty(memDef)) {
        fPrintWriteSet(stdout, memDef);
    }

    WriteSetDestroy(memUsed);
    WriteSetDestroy(memDef);
}

void fPrintMemTargets(FILE *f, InstrList *iList, Instruction *iins) {
    WriteSet *memUsed = MemUsed(iList, iins);
    WriteSet *memDef = MemDef(iList, iins);

    fprintf(f, "R: ");
    if (!WriteSetIsEmpty(memUsed)) {
        fPrintWriteSet(f, memUsed);
    }

    fprintf(f, "W: ");
    if (!WriteSetIsEmpty(memDef)) {
        fPrintWriteSet(f, memDef);
    }

    WriteSetDestroy(memUsed);
    WriteSetDestroy(memDef);
}


char *regname(int udreg) {
    char *reg;

    switch (udreg) {
        case UD_R_AL:
            reg = "al";
            break;
        case UD_R_AH:
            reg = "ah";
            break;
        case UD_R_AX:
            reg = "ax";
            break;
        case UD_R_EAX:
            reg = "eax";
            break;
        case UD_R_CL:
            reg = "cl";
            break;
        case UD_R_CH:
            reg = "ch";
            break;
        case UD_R_CX:
            reg = "cx";
            break;
        case UD_R_ECX:
            reg = "ecx";
            break;
        case UD_R_DL:
            reg = "dl";
            break;
        case UD_R_DH:
            reg = "dh";
            break;
        case UD_R_DX:
            reg = "dx";
            break;
        case UD_R_EDX:
            reg = "edx";
            break;
        case UD_R_BL:
            reg = "bl";
            break;
        case UD_R_BH:
            reg = "bh";
            break;
        case UD_R_BX:
            reg = "bx";
            break;
        case UD_R_EBX:
            reg = "ebx";
            break;
        case UD_R_SP:
            reg = "sp";
            break;
        case UD_R_ESP:
            reg = "esp";
            break;
        case UD_R_BP:
            reg = "bp";
            break;
        case UD_R_EBP:
            reg = "ebp";
            break;
        case UD_R_SI:
            reg = "si";
            break;
        case UD_R_ESI:
            reg = "esi";
            break;
        case UD_R_DI:
            reg = "di";
            break;
        case UD_R_EDI:
            reg = "edi";
            break;

            /*    //lynx specific stuff
             case UD_LYNX_TEMP1: reg = "temp1"; break;
             case UD_LYNX_TEMP2: reg = "temp2"; break;
             case UD_LYNX_TEMP3: reg = "temp3"; break;
             case UD_LYNX_TEMP4: reg = "temp4"; break;
             case UD_LYNX_TEMP5: reg = "temp5"; break;
             case UD_LYNX_TEMP6: reg = "temp6"; break;
             case UD_LYNX_EFLAG: reg = "eflag"; break;
             case UD_LYNX_R_AHX: reg = "ahx"; break;
             case UD_LYNX_R_CHX: reg = "chx"; break;
             case UD_LYNX_R_DHX: reg = "dhx"; break;
             case UD_LYNX_R_BHX: reg = "bhx"; break;
             case UD_LYNX_R_SPHX: reg = "sphx"; break;
             case UD_LYNX_R_BPHX: reg = "bphx"; break;
             case UD_LYNX_R_SIHX: reg = "sihx"; break;
             case UD_LYNX_R_DIHX: reg = "dihx"; break;
             case UD_LYNX_MEMLOC1: reg = "memloc1"; break;
             case UD_LYNX_MEMLOC2: reg = "memloc2"; break;
             case UD_LYNX_MEMLOC3: reg = "memloc3"; break;
             case UD_LYNX_MEMLOC4: reg = "memloc4"; break;
             case UD_LYNX_MEMLOC5: reg = "memloc5"; break;
             case UD_LYNX_MEMLOC6: reg = "memloc6"; break;*/

        default:
            fprintf(stderr, "WARNING [regname]: Unrecognized register %d\n", udreg);
            reg = "r??";
            break;
    }

    return reg;
}


char *BitvectorToString(REGISTER_BITVECTOR bv) {
    char *s = malloc(88);
    strncpy(s, " ", 2);

    /* 32-bit registers */
    if (bv & LYNX_EAX) {
        strncat(s, " eax", 4);
    }
    if (bv & LYNX_ECX) {
        strncat(s, " ecx", 4);
    }
    if (bv & LYNX_EBX) {
        strncat(s, " ebx", 4);
    }
    if (bv & LYNX_EDX) {
        strncat(s, " edx", 4);
    }
    if (bv & LYNX_ESP) {
        strncat(s, " esp", 4);
    }
    if (bv & LYNX_EBP) {
        strncat(s, " ebp", 4);
    }
    if (bv & LYNX_ESI) {
        strncat(s, " esi", 4);
    }
    if (bv & LYNX_EDI) {
        strncat(s, " edi", 4);
    }

    /* 16-bit registers */
    if (bv & LYNX_AX) {
        strncat(s, " ax", 3);
    }
    if (bv & LYNX_CX) {
        strncat(s, " cx", 3);
    }
    if (bv & LYNX_DX) {
        strncat(s, " dx", 3);
    }
    if (bv & LYNX_BX) {
        strncat(s, " bx", 3);
    }
    if (bv & LYNX_SP) {
        strncat(s, " sp", 3);
    }
    if (bv & LYNX_BP) {
        strncat(s, " bp", 3);
    }
    if (bv & LYNX_SI) {
        strncat(s, " si", 3);
    }
    if (bv & LYNX_DI) {
        strncat(s, " di", 3);
    }

    /* 8-bit registers */
    if (bv & LYNX_AL) {
        strncat(s, " al", 3);
    }
    if (bv & LYNX_AH) {
        strncat(s, " ah", 3);
    }
    if (bv & LYNX_CL) {
        strncat(s, " cl", 3);
    }
    if (bv & LYNX_CH) {
        strncat(s, " ch", 3);
    }
    if (bv & LYNX_DL) {
        strncat(s, " dl", 3);
    }
    if (bv & LYNX_DH) {
        strncat(s, " dh", 3);
    }
    if (bv & LYNX_BL) {
        strncat(s, " bl", 3);
    }
    if (bv & LYNX_BH) {
        strncat(s, " bh", 3);
    }

    /* segment registers */
    if (bv & LYNX_ES) {
        strncat(s, " es", 3);
    }
    if (bv & LYNX_CS) {
        strncat(s, " cs", 3);
    }
    if (bv & LYNX_SS) {
        strncat(s, " ss", 3);
    }
    if (bv & LYNX_DS) {
        strncat(s, " ds", 3);
    }
    if (bv & LYNX_FS) {
        strncat(s, " fs", 3);
    }
    if (bv & LYNX_GS) {
        strncat(s, " gs", 3);
    }
    if (bv & LYNX_ST0) {
        strncat(s, " st0", 4);
    }
    if (bv & LYNX_ST1) {
        strncat(s, " st1", 4);
    }
    if (bv & LYNX_ST2) {
        strncat(s, " st2", 4);
    }
    if (bv & LYNX_ST3) {
        strncat(s, " st3", 4);
    }
    if (bv & LYNX_ST4) {
        strncat(s, " st4", 4);
    }
    if (bv & LYNX_ST5) {
        strncat(s, " st5", 4);
    }
    if (bv & LYNX_ST6) {
        strncat(s, " st6", 4);
    }
    if (bv & LYNX_ST7) {
        strncat(s, " st7", 4);
    }
    if (bv & LYNX_ST8) {
        strncat(s, " st8", 4);
    }

    return (s);
}

char *PSWBitVectorToString(PSW_BITVECTOR pv) {
    char buf[80];
    int i;

    for (i = 0; i < 80; i++) {
        buf[i] = 0;
    }

    if (pv == PSW_ALL) {
        sprintf(buf, "-all-");
    } else {
        if (pv & PSW_CF) {
            strncat(buf, " CF", 3);
        }

        if (pv & PSW_PF) {
            strncat(buf, " PF", 3);
        }

        if (pv & PSW_AF) {
            strncat(buf, " AF", 3);
        }

        if (pv & PSW_ZF) {
            strncat(buf, " ZF", 3);
        }

        if (pv & PSW_SF) {
            strncat(buf, " SF", 3);
        }

        if (pv & PSW_TF) {
            strncat(buf, " TF", 3);
        }

        if (pv & PSW_IF) {
            strncat(buf, " IF", 3);
        }

        if (pv & PSW_DF) {
            strncat(buf, " DF", 3);
        }

        if (pv & PSW_OF) {
            strncat(buf, " OF", 3);
        }

        if (pv & PSW_IOPL1) {
            strncat(buf, " IOPL1", 6);
        }

        if (pv & PSW_IOPL2) {
            strncat(buf, " IOPL2", 6);
        }

        if (pv & PSW_NT) {
            strncat(buf, " NT", 3);
        }

        if (pv & PSW_RF) {
            strncat(buf, " RF", 3);
        }

        if (pv & PSW_VM) {
            strncat(buf, " VM", 3);
        }

        if (pv & PSW_AC) {
            strncat(buf, " AC", 3);
        }

        if (pv & PSW_VIF) {
            strncat(buf, " VIF", 4);
        }

        if (pv & PSW_VIP) {
            strncat(buf, " VIP", 4);
        }

        if (pv & PSW_ID) {
            strncat(buf, " ID", 3);
        }
    }

    return strdup(buf);

}

char *InstrToString(Instruction *instr) {
    int length = MAX_INSTR_NAME_LEN + 22;
    char buf[length];
    int i;

    for (i = 0; i < length; i++) {
        buf[i] = 0;
    }
    sprintf(buf, "%7d: [0x%08x] %s", instr->order, instr->addr, ud_insn_asm(LoadUDisInstrObj(instr)));

    return (strdup(buf));
}

/*********************************************************************
 *                                                                   *
 *                CREATING DOT OUTPUT FOR VISUALIATION               *
 *                                                                   *
 *********************************************************************/
void PrintInstruction(FILE *f, InstrList *list, Instruction *next, bool printRegs, bool PrintSigns){
	Instruction *nextPlus1;
	if(PrintSigns){
		if (INSTR_HAS_FLAG(next, INSTR_FLAG_FW_TAINED))
			fprintf(f, "@");
		if((INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED)))
			fprintf(f, "FT");
		if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))
			fprintf(f, "BT ");
		if (INSTR_HAS_FLAG(next, INSTR_COND_CF))
			fprintf(f, "CF ");
		if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED))
			fprintf(f, "#r");
		if (INSTR_HAS_FLAG(next, INSTR_TMP0))
			fprintf(f, "+c: ");
		if (INSTR_HAS_FLAG(next, INSTR_TMP1))
			fprintf(f, "+pp: ");
		if (INSTR_HAS_FLAG(next, INSTR_TMP2))
			fprintf(f, "+op: ");
		if (INSTR_HAS_FLAG(next, INSTR_TMP3))
			fprintf(f, "+d: ");
		if (INSTR_HAS_FLAG(next, INSTR_MEM_SIMP))
			fprintf(f, "+m: ");
		if (INSTR_HAS_FLAG(next, INSTR_SIMPLIFIED_IGNORED))
			fprintf(f, "+e");
		if (INSTR_HAS_FLAG(next, INSTR_IN_SHARED_MEM))
			fprintf(f, "SH");
	}
	if (next->order < list->numInstrs - 1) {
		//nextPlus1 = GetInstruction(list, next->order + 1);
		nextPlus1 = NULL;
	}
	fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
			nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
	if (printRegs) {
		fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
				next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
				next->edi);

		fprintf(f, "  ");
		fPrintMemTargets(f, list, next);
		fprintf(f, "\n");
	}
}


void PrintSimplifiedTrace(FILE *f, InstrList *list, int full_list, int printDeadInstrs, int printSigns, uint64_t start, uint64_t end) {

    bool printRegs = true;

    Instruction *next = NULL, *nextPlus1 = NULL;
    //SetCurrentInstr(list, 0);
    SetCurrentInstr(list, start);
    while ((next = FetchNextInstr(list)) != NULL) { // = next->next) {
    	//next = GetInstruction(list, index);
    	//index = index+1;//next->next;

    	if(next->order == 435036){
    		printf("s");
    	}

    	if(next->order == end){
    	    return;
    	}

        if (!full_list && !INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE)) {
            continue;
        }
        if(!printDeadInstrs && INSTR_HAS_FLAG(next, INSTR_IS_DELETED)){
            continue;
        }

        if(printSigns){
			if (INSTR_HAS_FLAG(next, INSTR_FLAG_FW_TAINED))
				fprintf(f, "@");
			if((INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED)))
			    fprintf(f, "FT");
			if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))
				fprintf(f, "BT ");
			if (INSTR_HAS_FLAG(next, INSTR_COND_CF))
				fprintf(f, "CF ");
			if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED))
				fprintf(f, "#r");
			if (INSTR_HAS_FLAG(next, INSTR_TMP0))
				fprintf(f, "+c: ");
			if (INSTR_HAS_FLAG(next, INSTR_TMP1))
				fprintf(f, "+pp: ");
			if (INSTR_HAS_FLAG(next, INSTR_TMP2))
				fprintf(f, "+op: ");
			if (INSTR_HAS_FLAG(next, INSTR_TMP3))
				fprintf(f, "+d: ");
			if (INSTR_HAS_FLAG(next, INSTR_MEM_SIMP))
				fprintf(f, "+m: ");
			if (INSTR_HAS_FLAG(next, INSTR_SIMPLIFIED_IGNORED))
				fprintf(f, "+e");
			if (INSTR_HAS_FLAG(next, INSTR_IN_SHARED_MEM))
				fprintf(f, "SH");
        }
        if (next->order < list->numInstrs - 1) {
            nextPlus1 = GetInstruction(list, next->order + 1);
        }
        fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);

            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
            fprintf(f, "\n");
        }
    }
}


/*
 * PrintSimplifiedInstrListInternal prints
 * a simplified trace. if full_list is 0
 * only the base module instructions are printed.
 */

void PrintTrace(FILE *f, InstrList *list, int full_list, int printDeadInstrs, int printSigns, uint64_t start, uint64_t end) {

    bool printRegs = true;
    int index;
    Instruction *next = NULL, *nextPlus1 = NULL;
    //SetCurrentInstr(list, 0);
    index = start;
    while ((next = GetInstruction(list, index)) != NULL) { // = next->next) {
    	//next = GetInstruction(list, index);
    	index = next->next;
    	//index++;

    	if(next->order == 218216){
    		printf("s");
    	}

    	if(next->order == end){
    	    return;
    	}

        if (!full_list && !INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE)) {
            continue;
        }
        if(!printDeadInstrs && INSTR_HAS_FLAG(next, INSTR_IS_DELETED)){
            continue;
        }

        if(printSigns){
			if (INSTR_HAS_FLAG(next, INSTR_FLAG_FW_TAINED))
				fprintf(f, "@");
			if((INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED)))
			    fprintf(f, "FT");
			if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))
				fprintf(f, "BT ");
			if (INSTR_HAS_FLAG(next, INSTR_COND_CF))
				fprintf(f, "CF ");
			if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED))
				fprintf(f, "#r");
			if (INSTR_HAS_FLAG(next, INSTR_TMP0))
				fprintf(f, "+c: ");
			if (INSTR_HAS_FLAG(next, INSTR_TMP1))
				fprintf(f, "+pp: ");
			if (INSTR_HAS_FLAG(next, INSTR_TMP2))
				fprintf(f, "+op: ");
			if (INSTR_HAS_FLAG(next, INSTR_TMP3))
				fprintf(f, "+d: ");
			if (INSTR_HAS_FLAG(next, INSTR_MEM_SIMP))
				fprintf(f, "+m: ");
			if (INSTR_HAS_FLAG(next, INSTR_SIMPLIFIED_IGNORED))
				fprintf(f, "+e");
			if (INSTR_HAS_FLAG(next, INSTR_IN_SHARED_MEM))
				fprintf(f, "SH");
        }
        if (next->order < list->numInstrs - 1) {
            nextPlus1 = GetInstruction(list, next->order + 1);
        }
        fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);

            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
            fprintf(f, "\n");
        }
    }
}

void PrintTraceBetweenTwoInstrs(FILE *f, InstrList *list, int full_list, int start, int end) {

    bool printRegs = true;
    int index;
    Instruction *next = NULL, *nextPlus1 = NULL;


    for (index = start; index < end; index ++) { // = next->next) {
    	next = GetInstruction(list, index);


        if (next->order < list->numInstrs - 1) {
            nextPlus1 = GetInstruction(list, next->order + 1);
        }
        fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);

            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
            fprintf(f, "\n");
        }
        if (next->next > list->last) {
            Instruction *nextNew;

            int nextOrder = next->next;
            while(((nextNew = GetInstruction(list, nextOrder)) != NULL) && nextOrder > list->last){
                if(nextNew->order == 186684){
                	printf("s");
                }
                nextOrder = nextNew->next;
                fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", nextNew->order, nextNew->addr,
                        ud_insn_asm(LoadUDisInstrObj(nextNew)), "\0");
                if (printRegs) {
                    fprintf(f,
                            "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                            nextNew->eax, nextNew->ecx, nextNew->edx, nextNew->ebx, nextNew->esp,
                            nextNew->ebp, nextNew->esi, nextNew->edi);

                    fprintf(f, "  ");
                    fPrintMemTargets(f, list, nextNew);
                    fprintf(f, "\n");
                }
            }
            SetCurrentInstr(list, nextOrder);
        }
        RInstrDestroy(list, next);
    }
}

void PrintEntireMultiThreadInstrListInternal(FILE *f, InstrList *list, int full_list, int printDeadInstrs, int tid) {

    uint64_t index;
    bool printRegs = true;

    Instruction *next = NULL, *nextPlus1 = NULL;
    for(index = 0; index < list->last; index ++){ // = next->next) {
    	next = GetInstruction(list, index);

    	if(next->tid != tid){
    		continue;
    	}

        if ((!INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE) && !full_list)) {
            continue;
        }
        if(!printDeadInstrs && INSTR_HAS_FLAG(next, INSTR_IS_DELETED)){
            continue;
        }
        if(INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE)){
        	fprintf(f, "BB ");
        }

        if (INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED))
            fprintf(f, "@");
        if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))
            fprintf(f, "BT ");
        if (INSTR_HAS_FLAG(next, INSTR_COND_CF))
            fprintf(f, "CF ");
        if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED))
            fprintf(f, "#r");
        if (INSTR_HAS_FLAG(next, INSTR_TMP0))
            fprintf(f, "+c: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP1))
            fprintf(f, "+pp: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP2))
            fprintf(f, "+op: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP3))
            fprintf(f, "+d: ");
        if (INSTR_HAS_FLAG(next, INSTR_MEM_SIMP))
            fprintf(f, "+m: ");
        if (INSTR_HAS_FLAG(next, INSTR_SIMPLIFIED_IGNORED))
            fprintf(f, "+e");
        if (INSTR_HAS_FLAG(next, INSTR_IN_SHARED_MEM))
            fprintf(f, "SH");
        if (next->order < list->numInstrs - 1) {
            nextPlus1 = GetInstruction(list, next->order + 1);
        }
        fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            fprintf(f, "tid:%08d EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->tid, next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);

            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
            fprintf(f, "\n");
        }
        if (next->next > list->last) {
            Instruction *nextNew;

            int nextOrder = next->next;
            while(((nextNew = GetInstruction(list, nextOrder)) != NULL) && nextOrder > list->last){
                if(nextNew->order == 7017126){
                	printf("s");
                }
                nextOrder = nextNew->next;
                if (INSTR_HAS_FLAG(nextNew, INSTR_FORWARD_TAINTED))
                    fprintf(f, "@");
                if (INSTR_HAS_FLAG(nextNew, INSTR_BACKWARD_TAINTED))
                    fprintf(f, "BT ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_COND_CF))
                    fprintf(f, "CF ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_IS_DELETED))
                    fprintf(f, "#r");
                if (INSTR_HAS_FLAG(nextNew, INSTR_TMP0))
                    fprintf(f, "+c: ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_TMP1))
                    fprintf(f, "+pp: ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_TMP2))
                    fprintf(f, "+op: ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_TMP3))
                    fprintf(f, "+d: ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_MEM_SIMP))
                    fprintf(f, "+m: ");
                if (INSTR_HAS_FLAG(nextNew, INSTR_SIMPLIFIED_IGNORED))
                    fprintf(f, "+e");
                fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", nextNew->order, nextNew->addr,
                        ud_insn_asm(LoadUDisInstrObj(nextNew)), "\0");
                if (printRegs) {
                    fprintf(f,
                            "tid:%08d EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                            nextNew->tid, nextNew->eax, nextNew->ecx, nextNew->edx, nextNew->ebx, nextNew->esp,
                            nextNew->ebp, nextNew->esi, nextNew->edi);

                    fprintf(f, "  ");
                    fPrintMemTargets(f, list, nextNew);
                    fprintf(f, "\n");
                }
            }
            SetCurrentInstr(list, nextOrder);
        }
        RInstrDestroy(list, next);
    }
}

void PrintLiveListInternal(FILE *f, InstrList *list, int full_list) {
    //int size = list->last;
    //int index;
    bool printRegs = true;

    Instruction *next = NULL, *nextPlus1 = NULL;

    SetCurrentInstr(list, 0);
#if 0
    SetCurrentInstr(list, list->startMarker);
#endif
    while ((next = FetchNextInstr(list)) != NULL) {

        if ((!INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE) && !full_list)) {
            RInstrDestroy(list, next);
            continue;
        }

#if 0
        if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED) && !INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED))
            fprintf(f, "#r");
        if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED) && INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED))
            fprintf(f, "+TR");
        if (INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED))
            fprintf(f, "@");
        if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))
            fprintf(f, "BT");
        if (INSTR_HAS_FLAG(next, INSTR_TMP0))
            fprintf(f, "+c: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP1))
            fprintf(f, "+pp: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP2))
            fprintf(f, "+op: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP3))
            fprintf(f, "+d: ");
        if (INSTR_HAS_FLAG(next, INSTR_MEM_SIMP))
            fprintf(f, "+m: ");
        if (INSTR_HAS_FLAG(next, INSTR_SIMPLIFIED_IGNORED))
            fprintf(f, "+e");
#endif
        //if(index < size -1) {
        nextPlus1 = FetchNextInstr(list);
        //}
        fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);
            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
            fprintf(f, "\n");
        }

#if 0
       if(next->order == list->endMarker){
    	   return;
       }
#endif
        RInstrDestroy(list, next);
        next = NULL;
        if (nextPlus1) {
            SetCurrentInstr(list, nextPlus1->order);
            RInstrDestroy(list, nextPlus1);
            nextPlus1 = NULL;
        }

    }
}

/*
 * PrintSimplifiedBaseModule prints a simplified trace, no codes.
 */

void PrintSimplifiedBaseModule(FILE *f, InstrList *list){
    int size = list->last;
    int index;
    bool printRegs = true;

    Instruction *next = NULL, *nextPlus1 = NULL;

    for(index = 0; index < size; index++) {
        next = GetInstruction(list, index);

        if(INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE) &&
                !INSTR_HAS_FLAG(next, INSTR_IS_DELETED)) {
            if(index < size -1) {
                nextPlus1 = GetInstruction(list, next->order + 1);
            }
            fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s",
                    next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)), nextPlus1==NULL?"\0":GetFunctionNameById(nextPlus1->funcName));
            if(printRegs) {
                fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                        next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi, next->edi);

            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
                fprintf(f, "\n");
            }
            RInstrDestroy(list, next);
            next = NULL;
        }
    }
}

void PrintOriginalTrace(FILE *file, InstrList *iList){
    PrintTrace(file, iList, 1, 0, 1, iList->startMarker, iList->endMarker);
}

void PrintSimplifiedFullInstrList(FILE *f, InstrList *list, int simple, int printSigns, uint64_t start, uint64_t end) {
    if (simple) {
        PrintTrace(f, list, 0, 0, printSigns, start, end);
    } else {
        PrintTrace(f, list, 1, 1, printSigns, start, end);
    }
}

void PrintOneThread(FILE *f, InstrList *list, int full_list, int ThreadId) {
    int size = list->last;
    int index;
    bool printRegs = true;

    Instruction *next = NULL, *nextPlus1 = NULL;

    for (index = 0; index < size; index++) {
        next = GetInstruction(list, index);

        if ((!INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE) && !full_list) || next->tid != ThreadId) {
            RInstrDestroy(list, next);
            continue;
        }
        if (INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED))
            fprintf(f, "@");
        if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))
            fprintf(f, "BT ");
        if (INSTR_HAS_FLAG(next, INSTR_COND_CF))
            fprintf(f, "CF ");
        if (INSTR_HAS_FLAG(next, INSTR_IS_DELETED))
            fprintf(f, "#r");
        if (INSTR_HAS_FLAG(next, INSTR_TMP0))
            fprintf(f, "+c: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP1))
            fprintf(f, "+pp: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP2))
            fprintf(f, "+op: ");
        if (INSTR_HAS_FLAG(next, INSTR_TMP3))
            fprintf(f, "+d: ");
        if (INSTR_HAS_FLAG(next, INSTR_MEM_SIMP))
            fprintf(f, "+m: ");
        if (INSTR_HAS_FLAG(next, INSTR_SIMPLIFIED_IGNORED))
            fprintf(f, "+e");
        if (index < size - 1) {
            nextPlus1 = GetInstruction(list, next->order + 1);
        }
        fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", next->order, next->addr, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            fprintf(f, "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);

            fprintf(f, "  ");
            fPrintMemTargets(f, list, next);
            fprintf(f, "\n");
        }
        if (next->next > list->last) {
            Instruction *nextNew = GetInstruction(list, next->next);
            if (INSTR_HAS_FLAG(nextNew, INSTR_FORWARD_TAINTED))
                fprintf(f, "@");
            if (INSTR_HAS_FLAG(nextNew, INSTR_BACKWARD_TAINTED))
                fprintf(f, "BT ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_COND_CF))
                fprintf(f, "CF ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_IS_DELETED))
                fprintf(f, "#r");
            if (INSTR_HAS_FLAG(nextNew, INSTR_TMP0))
                fprintf(f, "+c: ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_TMP1))
                fprintf(f, "+pp: ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_TMP2))
                fprintf(f, "+op: ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_TMP3))
                fprintf(f, "+d: ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_MEM_SIMP))
                fprintf(f, "+m: ");
            if (INSTR_HAS_FLAG(nextNew, INSTR_SIMPLIFIED_IGNORED))
                fprintf(f, "+e");
            fprintf(f, "%8d: [0x%08x] %-50.50s %-40.40s", nextNew->order, nextNew->addr,
                    ud_insn_asm(LoadUDisInstrObj(nextNew)), nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
            if (printRegs) {
                fprintf(f,
                        "EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                        nextNew->eax, nextNew->ecx, nextNew->edx, nextNew->ebx, nextNew->esp,
                        nextNew->ebp, nextNew->esi, nextNew->edi);

                fprintf(f, "  ");
                fPrintMemTargets(f, list, next);
                fprintf(f, "\n");
            }
        }
        RInstrDestroy(list, next);
        next = NULL;
        if (index < size - 1) {
            RInstrDestroy(list, nextPlus1);
            nextPlus1 = NULL;
        }
    }
}

void PrintTaintedInstrs(InstrList *list, int start, int n) {
    uint64_t size = list->last;
    bool printRegs = true;

    Instruction *next = NULL, *nextPlus1 = NULL;

    SetCurrentInstr(list, start);
    if (n == 0)
        n = -1;
    while ((next = FetchNextInstr(list)) != NULL && (n--)) {

        if (next == NULL)
            break;

        if ((!INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED) && !INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED))) {
            continue;
        }
        if (INSTR_HAS_FLAG(next, INSTR_FORWARD_TAINTED)) {
            printf("FT");
        }
        if (INSTR_HAS_FLAG(next, INSTR_BACKWARD_TAINTED)) {
            printf("BT");
        }
        printf(" ");
        if (next->order < size - 1) {
            nextPlus1 = FetchNextInstr(list);
        }
        printf("%8d: [0x%08x] %d %-50.50s %-40.40s", next->order, next->addr, next->tid, ud_insn_asm(LoadUDisInstrObj(next)),
                nextPlus1 == NULL ? "\0" : GetFunctionNameById(nextPlus1->funcName));
        if (printRegs) {
            printf("EAX:%08x ECX:%08x EDX:%08x EBX:%08x ESP:%08x EBP:%08x ESI:%08x EDI:%08x",
                    next->eax, next->ecx, next->edx, next->ebx, next->esp, next->ebp, next->esi,
                    next->edi);
            printf("  ");
            PrintMemTargets(list, next);
            printf("\n");
        }

        RInstrDestroy(list, next);
        next = NULL;
        if (nextPlus1) {
            SetCurrentInstr(list, nextPlus1->order);
            RInstrDestroy(list, nextPlus1);
            nextPlus1 = NULL;
        }
    }

}
