/*
 *
 * File: read.c
 * Author: Babak Yadegari, Kevin Coogan
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#include <ctype.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include "lynx_types.h"
#include "al_helper.h"
#include "instr.h"
#include "read.h"
#include "flags.h"

bool TraceIsMultiThreaded;
ArrayList *traceThreads;

/*********************************************************************/

extern char *baseName;
extern bool InstrFileExists;

/*********************************************************************
 *                                                                   *
 * ignore_modules is a list of modules/DLLs whose instructions       *
 * should NOT be included in the code output by lynx.  The last      *
 * entry of this list should always be NULL.                         *
 *                                                                   *
 *********************************************************************/
char *ignore_modules[] = {
        "kernel32", "ntdll", "MSVCP80", "MSVCP90", "MSVCR80", "MSVCR90", "msvcrt",
        //
        "ADVAPI32", "RPCRT4", "comctl32", "SHELL32", "SHLWAPI", "USER32", "GDI32", "MPR",
        "WS2HELP", "WS2_32",
        //
        "", "dll",
        /*"unknown",*/ NULL };

/*
 * GetBaseName
 * extract module name (without suffix or extensions) from input file
 */
char *GetBaseName(char *traceFileName) {
    char *baseStart = strrchr(traceFileName, '/');

    if (baseStart != NULL) {
        //then directory information given, start at next char
        baseStart += 1;
    } else {
        //just file name given, use it
        baseStart = traceFileName;
    }

    char *suffixStart = strstr(traceFileName, "_std_trace.txt");
    if (suffixStart == NULL) {
        fprintf(stderr, "ERROR: trace filename must end with _std_trace.txt\n");
        exit(0);
    }

    int suffixLen = strlen("_std_trace.txt");

    int baseStartLen = strlen(baseStart);
    int baseNameLen = baseStartLen - suffixLen;

    char *baseName = calloc(baseNameLen + 1, 1);
    strncpy(baseName, baseStart, baseNameLen);

    return (baseName);
}

/*
 * ignoreMod(str) -- returns true if str matches the name of a module
 * that is to be ignored, false otherwise.
 */
bool ignoreMod(char *str) {
    int i;

    assert(str);

    for (i = 0; ignore_modules[i] != NULL; i++) {
        if (strcmp(str, ignore_modules[i]) == 0) {
            return true;
        }
    }

    return false;

}

void PrintInstrBytes(Instruction *instr) {
    int i;
    for (i = 0; i < instr->uInstr.length; i++) {
        printf("%02x ", instr->uInstr.bytes[i]);
    }
    printf("\n");
}

char *GetModuleName(char *fileName) {
    char *temp = malloc(255);
    char *tok = malloc(255);
    char *last = malloc(255);
    strncpy(temp, fileName, strlen(fileName));
    char *end = strstr(temp, "_std_trace.txt");
    end[0] = '\0';

    tok = strtok(temp, "/");
    while (tok != NULL) {
        strncpy(last, tok, strlen(tok) + 1);
        tok = strtok(NULL, "/");
    }
    char *modName = malloc(strlen(last) + 1);
    strncpy(modName, last, strlen(last) + 1);

    free(temp);
    free(tok);
    free(last);
    return (modName);
}

#define ENABLE_DLL_POINTER_CALCULATION 1

unsigned int lineno = 0;

bool found_first_base_instr = false;

/*
 * errGetInstrFromText() -- give error messages for GetInstrFromText()
 */
void errGetInstrFromText(int err_code, Instruction *instr) {
    int i;

    fprintf(stderr, "ERROR [line %d]: malformed trace ", lineno);

    switch (err_code) {
        case ERR_AT_ADDR:
            fprintf(stderr, "[at address]\n");
            break;
        case ERR_AT_FNAME:
            fprintf(stderr, "[at function name]\n");
            break;
        case ERR_AT_INSTRBYTES:
            fprintf(stderr, "[at instr bytes]\n");
            break;
        case ERR_AT_MNEMONIC:
            fprintf(stderr, "[at instruction mnemonic]\n");
            break;
        case ERR_AT_REGVALS:
            fprintf(stderr, "[at register values]\n");
            break;
        default:
            fprintf(stderr, " [unrecognized error code %d]\n", err_code);
            break;
    }

    if (err_code >= ERR_AT_ADDR) {
        fprintf(stderr, "       [DUMP] module: %s\n", GetModuleNameById(instr->module));
    }

    if (err_code >= ERR_AT_FNAME) {
        fprintf(stderr, "       [DUMP] addr:   0x%x\n", instr->addr);
    }

    if (err_code >= ERR_AT_INSTRBYTES) {
        fprintf(stderr, "       [DUMP] function name: %s\n", GetFunctionNameById(instr->funcName));
    }

    if (err_code >= ERR_AT_MNEMONIC) {
        fprintf(stderr, "       [DUMP] instr bytes: ");
        for (i = 0; i < instr->uInstr.length; i++) {
            fprintf(stderr, "%x ", instr->uInstr.bytes[i]);
        }
        fprintf(stderr, "\n");
    }

    if (err_code >= ERR_AT_REGVALS) {
        fprintf(stderr, "       [DUMP] mnemonic: %s\n", ud_insn_asm(LoadUDisInstrObj(instr)));
    }

    fprintf(stderr, "       ----------\n");
    abort();
}

/* GetInstrFromText
 *  Parse instruction struct from line of text
 */
int ignores = 0;
int image_base=0x400000;
int image_size=0x2af000;
Instruction *GetInstrFromText(char *s) {
    int nregs;
    int currTID = 0;
    Instruction *new = NULL;
    char *tok, *tokSave, *p;



    lineno += 1;

    // a '#' appearing as the first character of a line indicates a comment.
    if (*s == '#') {
        return NULL;
    }

    //create instr object, and initialize its udis object
    new = InstrCreate();

    //get first token, depending on version, may be module name, or may be pid
    tok = strtok_r(s, "\t", &tokSave);

    if (tok == NULL) {
        fprintf(stderr, "ERROR [line %d]: malformed trace [at entry]\n", lineno);
        abort();
    }

    //check for pid and tid.  if there, first tok will be number other than 0
    if (strtol(tok, NULL, 10) != 0) {
        //we are ignoring pid completely for now
        tok = strtok_r(NULL, "\t", &tokSave);

        //check for multithreaded program, in future may want to do more w/tid
        currTID = strtol(tok, NULL, 10);
        new->tid = currTID;

        tok = strtok_r(NULL, "\t", &tokSave);
    }

    // parse module name
    // eliminate extraneous whitespace

    for (p = tok; p && *p != '\0'; p++) {
        if (isspace(*p)) {
            *p = '\0';
        }
    }

    /*
     * If the module is one that we don't care about (e.g., kernel32, ntdll),
     * we want to just ignore the instruction and bail out.  But we can't
     * ignore _all_ such instructions, since we determine the identity of DLL
     * calls by looking at the callee instruction's function name; a similar
     * issue arises with exceptions as well.  To deal with this, we keep track
     * of whether the previous instruction was ignored, and ignore an
     * instruction only if it is from an "ignored" module AND the previous
     * instruction was ignored.
     *
     * At some point in the future we might want to have a command-line
     * switch to control this behavior.
     */
    /*
     * Now we don't wanna ignore all those instructions, we need to include
     * them to calculate the memory use/def of DLL calls (for those pointer
     * parameters)
     *
     * This enabled by setting ENABLE_DLL_POINTER_CALCULATION
     *
     * -- Gen 03/04/2011
     */
    if (ignoreMod(tok)) {
        if (found_first_base_instr)
        {

            INSTR_SET_FLAG(new, INSTR_IGNORED);
        } else { /* ignore */
        	ignores ++;
            free(new);
            return NULL;
        }
    }

    //one-way door, to initialize when we've started including instructions
    //we need this to ignore the large group of instructions at front of file
    if (found_first_base_instr == false) {
        found_first_base_instr = true;
    }

    int index;
    //strncpy(new->module, tok, strlen(tok) + 1);
    //printf("module: >>%s<<\n", new->module);



    char baseNameExe[200];
    sprintf(baseNameExe, "%s.exe", baseName);

    if(strcmp(tok, baseName) == 0 || strcmp(tok, "unknown") == 0 || strcmp(tok, baseNameExe) == 0){
        new->module = -1;
    } else {
        if((index = ModuleNameExists(tok)) != -1){
            new->module = index;
        } else {
            new->module = AddModuleNametoTable(tok);
        }
    }

    //parse address
    tok = strtok_r(NULL, "\t", &tokSave);

    if (tok == NULL) {
        errGetInstrFromText(ERR_AT_ADDR, new);
    }

    sscanf(tok, "%x", &(new->addr));
    //printf("addr:0x%08x\n", new->addr);

    //parse function name
    tok = strtok_r(NULL, "\t", &tokSave);

    if (tok == NULL) {
        errGetInstrFromText(ERR_AT_FNAME, new);
    }

    // eliminate extraneous whitespace
    for (p = tok; p && *p != '\0'; p++) {
        if (isspace(*p)) {
            *p = '\0';
        }
    }


    if(strcmp(tok, "") == 0){
        new->funcName = -1;
    } else {
        if((index = FunctionNameExists(tok)) != -1){
            new->funcName = index;
        } else {
            new->funcName = AddFunctionNametoTable(tok);
        }

    }

    //strncpy(new->funcName, tok, strlen(tok) + 1);
    //if (debugPrint && (strlen(new->funcName) > 0)) {
    //    printf("%8d:: funcName:-->%s<--\n", lineno, new->funcName);
    //}

    //parse instruction bytes
    int byte = 1;
    char *tok2 = NULL;

    tok = strtok_r(NULL, "\t", &tokSave);

    if (tok == NULL) {
        errGetInstrFromText(ERR_AT_INSTRBYTES, new);
    }

    tok2 = strtok(tok, " ");
    if (tok2 != NULL) {
        new->uInstr.bytes[0] = strtol(tok2, NULL, 16);
        while ((tok2 = strtok(NULL, " ")) != NULL) {
            new->uInstr.bytes[byte] = strtol(tok2, NULL, 16);
            byte++;
        }
        new->uInstr.length = byte;
        //PrintInstrBytes(new);
    }

    //parse instruction mnemonic
    tok = strtok_r(NULL, "\t", &tokSave);

    if (tok == NULL) {
        errGetInstrFromText(ERR_AT_MNEMONIC, new);
    }

    //strncpy(new->uInstr.name, tok, strlen(tok) + 1);
    //printf("mnemonic:%s\n", new->name);

    //parse values for each register
    tok = strtok_r(NULL, "\t", &tokSave);

    if (tok == NULL) {
        errGetInstrFromText(ERR_AT_REGVALS, new);
    }

    nregs = sscanf(tok, "EAX=%x, ECX=%x, EDX=%x, EBX=%x, ESP=%x, EBP=%x, ESI=%x, EDI=%x",
            &(new->eax), &(new->ecx), &(new->edx), &(new->ebx), &(new->esp), &(new->ebp),
            &(new->esi), &(new->edi));
    //printf("EAX=%x, ECX=%x, EDX=%x, EBX=%x, ESP=%x, EBP=%x, ESI=%x, EDI=%x\n",
    //        new->eax, new->ecx, new->edx, new->ebx,
    //        new->esp, new->ebp, new->esi, new->edi);

    if (nregs != 8) { // not all the register values were read
        errGetInstrFromText(ERR_AT_REGVALS, new);
    }


    //parse values for memory read/writes
    tok = strtok_r(NULL, "\t", &tokSave);
    if(tok != NULL){
    	int memRead = sscanf(tok, "MR:%x:%x", &(new->memUsedMin), &(new->memUsedVal));
    	if (memRead == 0){
    		sscanf(tok, "MW:%x:%x", &(new->memDefMin), &(new->memDefVal));
    	}
    }

    tok = strtok_r(NULL, "\t", &tokSave);
    if(tok != NULL){
    	sscanf(tok, "MW:%x:%x", &(new->memDefMin), &(new->memDefVal));
    }



    if (/*new->addr > image_base && new->addr < image_base + image_size*/new->module == -1) {
        INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
        ud_t *ud_ins = LoadUDisInstrObj(new);
        CopyUdins2InsStr(&(new->uInstr), ud_ins);
        free(ud_ins);
    } else {
    	//fprintf(stderr, "non base module: [0x%08x]\n", new->addr);
        ud_t *ud_ins = LoadUDisInstrObj(new);
        CopyUdins2InsStr(&(new->uInstr), ud_ins);
        free(ud_ins);
    }
    return (new);
}

uint64_t LargestThreadNumInsstrs = 0;
int LargestThreadTID = 0;
void MarkThreadBoundaries(InstrList *i_list) {
    ArrayList *threads = traceThreads;
    int totalInstrs = 0;

    int j;
    for (j = 0; j < al_size(threads); j++) {
        Thread *th = al_get(threads, j);
        //fprintf(stderr, "Thread %d, numranges: %d and with %d instrs: , %d\n\n", th->tid, th->numRanges,(int) th->numBaseModulInstrs, al_size(th->ranges));
        int threadTotal = 0;
        int k;
        for (k = 0; k < al_size(th->ranges); k++) {
            ThreadRange *thrange = al_get(th->ranges, k);
            //fprintf(stderr, "%07d --> %07d ;", (int) thrange->first, (int) thrange->last);
            threadTotal += thrange->last - thrange->first + 1;
        }
        th->numRanges = al_size(th->ranges);
        totalInstrs += th->numInstrs;
        if(LargestThreadNumInsstrs < th->numBaseModulInstrs){
        	LargestThreadNumInsstrs = th->numBaseModulInstrs;
        	LargestThreadTID = th->tid;
        }
        //fprintf(stderr, "\n");
        if (threadTotal != th->numInstrs) {
            fprintf(stderr, "Error in number of instructions: %d, %d\n", threadTotal,
                    (int) th->numInstrs);
        }

        for (k = 0; k < al_size(th->ranges) - 1; k++) {
            ThreadRange *thrange1 = al_get(th->ranges, k);
            ThreadRange *thrange2 = al_get(th->ranges, k + 1);
            Instruction *iins = GetInstruction(i_list, thrange1->last);
            Instruction *iins2 = GetInstruction(i_list, thrange2->first);
            iins->next = iins->origNext = thrange2->first;
            SaveInstrChange(i_list, iins);
            if (iins->addr == iins2->addr) { //ether's bug, in base module's thread switch, last instruction is being replicated with the effect of executing it once!
                thrange1->last = iins->prev;
                INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                SaveInstrChange(i_list, iins);
                RInstrDestroy(i_list, iins);
                iins = GetInstruction(i_list, thrange1->last);
            }
            RInstrDestroy(i_list, iins2);
            iins->next = iins->origNext = thrange2->first;
            SaveInstrChange(i_list, iins);

            iins2 = GetInstruction(i_list, thrange2->first);
            if(iins->tid != iins2->tid){
                fprintf(stderr, "Error\n");
            }
            iins2->prev = iins2->origPrev = thrange1->last;
            SaveInstrChange(i_list, iins2);
            RInstrDestroy(i_list, iins);
            RInstrDestroy(i_list, iins2);
            if (k == 0) {
                iins = GetInstruction(i_list, thrange1->first);
                iins->prev = iins->origPrev = -1;
                SaveInstrChange(i_list, iins);
                RInstrDestroy(i_list, iins);
            }
            if (k == al_size(th->ranges) - 2) {
                iins = GetInstruction(i_list, thrange2->last);
                iins->next = iins->origNext = -1;
                SaveInstrChange(i_list, iins);
                RInstrDestroy(i_list, iins);
            }
        }
    }
    if (totalInstrs != numInstrs)
        fprintf(stderr, "Error in number of Instrs: %d, %" PRIu64 "\n", totalInstrs, numInstrs);
    //exit(0);
}

InstrList *BuildIListFromFile(FILE *file, char *instrFileName) {
    int lastTID = 0, currTID = 0;
    ArrayList *threads=NULL;
    Thread *currThread = malloc(sizeof(struct Thread));;
    ThreadRange *currRange=NULL;
    InstrList *i_list = calloc(1, sizeof(InstrList));
    i_list->numInstrs = 0;
    i_list->numBaseModuleInstrs = 0;
    bool isMultiThread = false;
    currThread->numBaseModulInstrs = 0;

    lineno = 0;
    found_first_base_instr = false;

    int order = 0; //dyn trace, so add instr's in order of 1st exec.

    currThread->numBaseModulInstrs = 0;

    //we will be storing ilist either in a file or in mem
    if (STORE_INSTRS_IN_FILE) {
        i_list->iListFile = fopen(instrFileName, "w+");
        i_list = InitInstructionCache(i_list);
        i_list->nextAvailabelAddr = 0;

        fseeko(i_list->iListFile, InstrFileHeaderSize, SEEK_SET); // make room to write iList at the begining
    } else {
        i_list->iListMem = al_new();
    }

    char *s = (char *) malloc(MAX_CHARS_PER_LINE);

    while (fgets(s, MAX_CHARS_PER_LINE, file)) {
        //fprintf(stderr, "%s\n", s);
        Instruction *new = GetInstrFromText(s);


        //printf("build:instr: %s\n", ud_insn_asm(new->udInstr));
        if (new) {
            currTID = new->tid;
            new->order = order++;
            i_list->numInstrs++;
            new->next = new->origNext = order;
            new->prev = new->origPrev = new->order - 1;
            if(INSTR_HAS_FLAG(new, INSTR_IN_BASE_MODULE)){
            	currThread->numBaseModulInstrs ++;
            	i_list->numBaseModuleInstrs++;
            }

            if(INSTR_HAS_FLAG(new, INSTR_IN_BASE_MODULE)){
                i_list->lastBaseModuelInstr = new->order;
                if(i_list->nextAvailabelAddr < new->addr + MAX_INSTRLEN){
                	i_list->nextAvailabelAddr = new->addr;
                }
            }
            if(new->esp > i_list->stackBottom){
            	i_list->stackBottom = new->esp;
            }

            if (STORE_INSTRS_IN_FILE) {
                AddInstrToFile(i_list->iListFile, new);

                if (lastTID != 0 && lastTID != currTID) { // a thread switch!
                    if (!isMultiThread) {
//                        fprintf(stderr, "Trace is multithreaded! %d\n", new->order);
                        threads = al_new();
                        currThread->tid = lastTID;
                        currThread->numInstrs = 0;
                        currThread->numRanges = 0;
                        currThread->ranges = al_new();
                        currRange = malloc(sizeof(struct ThreadRange));
                        currRange->first = 0;
                        //currRange->last = new->order - 1;
                        al_add(currThread->ranges, currRange);
                        //currThread->numRanges++;
                        al_add(threads, currThread);
                        isMultiThread = true;
                        TraceIsMultiThreaded = true;

                    }
                    //currThread->tid = lastTID;
                    currRange->last = new->order - 1;
                    currThread->numInstrs += currRange->last - currRange->first + 1; // +1: inclusive!
                    currRange = malloc(sizeof(struct ThreadRange));
                    currThread->numRanges++;
                    currRange->first = new->order;

                    /* search in previous threads to see if
                     the thread already exists, otherwise
                     create a new one			*/
                    currThread = NULL;
                    int numThreads = al_size(threads);
                    int i = 0;
                    for (; i < numThreads; i++) {
                        Thread *th = al_get(threads, i);
                        if (th->tid == currTID) {
                            currThread = th;
                            break;
                        }
                    }
                    if (currThread == NULL) {
                        currThread = malloc(sizeof(struct Thread));
                        currThread->tid = currTID;
                        currThread->numInstrs = 0;
                        currThread->numBaseModulInstrs = 0;
                        currThread->numRanges = 0;
                        currThread->ranges = al_new();
                        al_add(threads, currThread);
                    }
                    al_add(currThread->ranges, currRange);

                }
                lastTID = currTID;
                free(new);

            } else {
                al_add(i_list->iListMem, new);
            }
        }
    }

    free(s);
    numInstrs = i_list->numInstrs;
    i_list->last = i_list->numInstrs - 1;

    //pad the file to handle adding instructions
    //for now, arbitrarily double size of file, though this may be excessive
    if (USE_MMAP) {
        Instruction *temp = calloc(sizeof(Instruction), numInstrs);
        fseeko(i_list->iListFile, 0, SEEK_END);
        if (fwrite(temp, sizeof(Instruction), numInstrs, i_list->iListFile) != numInstrs) {
            fprintf(stderr, "ERROR: padding file for mmap\n");
        }
        free(temp);
    }
    Instruction *newInstr = GetInstructionFromFile(i_list, i_list->numInstrs - 1);
    //Instruction *newInstr = GetInstruction(i_list, i_list->numInstrs - 1);

    newInstr->next = newInstr->origNext = -1;

    //newInstr = GetInstructionFromFile(i_list, 0);
    /*newInstr = GetInstructionFromFile(i_list, 0);
    i_list -> stackBottom = newInstr->esp;		// set bottom stack of the trace*/
    //SaveInstrChange(i_list, newInstr);

    if (WriteInstrToFile(i_list->iListFile, newInstr) == 0) {
        fprintf(stderr, "Unable to save instr changes to file\n");
        exit(0);
    }

    i_list->loop = 0;
    i_list->firstInstr = 0;

    traceThreads = threads;
    if (TraceIsMultiThreaded) {
        currRange->last = i_list->numInstrs - 1;
        currThread->numInstrs += currRange->last - currRange->first + 1;
    }
    return (i_list);
}

void InitMMAP(char *instrFileName, InstrList *iList) {

    //get file size
    fseeko(iList->iListFile, 0, SEEK_END);
    uint64_t size = ftello(iList->iListFile);

    //close the file, and reopen to get file descriptor needed by mmap
    fclose(iList->iListFile);
    free(iList->instrCache);
    iList->instrCache = NULL;

    //get page size
    uint64_t pageSize = sysconf(_SC_PAGE_SIZE);

    //calculate size of mmap
    uint64_t numPages = size / pageSize;
    uint64_t remainder = size % pageSize;

    if (remainder != 0) {
        numPages++;
    }

    //numPages += numPages;    //arbitrary num of pages for added instructions
    uint64_t len = numPages * pageSize;

    //get a file descriptor that mmap can use
    int fd = open(instrFileName, O_RDWR);

    //do the mmap
    //mmap(
    //      don't care where addr starts,
    //      size of the map,
    //      allow read and write to pages,
    //      mmap maps a file, may have semaphores, and shares memory,
    //      the file descriptor of the file
    //      offset into file (I think, though not 0 doesn't work)
    //  );
    iList->iMMap = (Instruction *) mmap((void *) 0, (size_t) len, PROT_READ | PROT_WRITE, MAP_NORESERVE | MAP_FILE
            | MAP_SHARED, fd, (off_t) 0);

    if ((int64_t) iList->iMMap < 0) {
        perror("Error mapping file");
        exit(0);
    }

    // don't need fd anymore. File is mapped.
    close(fd);
}

/*
 * InitInstrList
 * first looks for previously created instrlist for same module, and tries
 * to use that, otherwise, reads tracefile and generates new instr file.
 * Once instr file is created, creates instrList
 */
extern bool RemoveExistinginstrFile;
InstrList *InitInstrList(char *inFileName) {
    InstrList *iList = NULL;
    baseName = GetBaseName(inFileName);

#if PIN_TRACE
    //baseName = strcat(baseName, ".exe");
#endif

    char *instrFileName = calloc(1, strlen(baseName) + strlen("./") + strlen(".instrFile") + 1);
    sprintf(instrFileName, "./%s.instrFile", baseName);

    if(RemoveExistinginstrFile){
    	if( access(instrFileName, F_OK ) != -1 ){
           int status = remove(instrFileName);
           if(status != 0) {
              printf("Unable to delete .instrFile\n");
              perror("Error");
           }
    	}
    }

    char *baseInstrFileName = malloc(strlen(instrFileName) + strlen(".base")+1);
    sprintf(baseInstrFileName, "%s%s", instrFileName, ".base");

    FILE *instrFile;
    FILE *traceFile;
    FILE *baseInstrFile;
    InstrFileHeader *fileHeader;
    bool setInstrFileExist = true;

    Thread *currThread;
    ThreadRange *currRange;

    instrFile = fopen(instrFileName, "r+");
    baseInstrFile = fopen(baseInstrFileName, "r+");

    if (STORE_INSTRS_IN_FILE && (baseInstrFile != NULL) && (instrFile == NULL)) {
    	fprintf(stderr, "Making a copy of baseInstrFile...");
    	char *cpCommand = malloc(strlen("cp ") + strlen(instrFileName)+strlen(baseInstrFileName)+3);
    	sprintf(cpCommand, "cp %s %s", baseInstrFileName, instrFileName);			// copy instrFile to a base instrFile to avoid reading trace file each time!
    	if(system(cpCommand) == -1){
            fprintf(stderr, "Could not run command: %s\n", cpCommand);
            exit(-1);
    	}
    	fprintf(stderr, "done\n");
    	setInstrFileExist = false;
    	instrFile = fopen(instrFileName, "r+");
    }

    //NOTE: using previous instrFile won't work if STORE_INSTRS_IN_FILE false!!
    if (STORE_INSTRS_IN_FILE && (instrFile != NULL)) {
        /* reading file header */
        fileHeader = malloc(sizeof(InstrFileHeader));
        fseeko(instrFile, 0, SEEK_SET);
        if(fread(fileHeader, sizeof(InstrFileHeader), 1, instrFile) <= 0){
            fprintf(stderr, "Could not read the trace file!\n");
            exit(-1);
        }

        if(setInstrFileExist){
        	InstrFileExists = true;
        }

        iList = malloc(sizeof(InstrList));
        iList->iListFile = instrFile;
        iList->numInstrs = fileHeader->NumInstrs;
        iList->lastBaseModuelInstr = fileHeader->lastBaseModule;
        iList->nextAvailabelAddr = fileHeader->nextAvailAddr;
        iList->startMarker = fileHeader->startMarker;
        iList->endMarker = fileHeader->endMarker;
        strcpy(baseName, fileHeader->baseModuleName);

        numInstrs = fileHeader->NumInstrs; //setting global numInstrs


        iList->last = fileHeader->LastInstrInOriginalTrace;
        iList->loop = fileHeader->Loop;
        iList->stackBottom = fileHeader->stackButtom;
        TraceIsMultiThreaded = fileHeader->TraceIsMultiThreaded;

        /* Recover thread information if any! */
        if (fileHeader->TraceIsMultiThreaded) {
            off_t rangesOffset = fileHeader->RangeInfoOffset;
            fseeko(instrFile, rangesOffset, SEEK_SET);
            traceThreads = al_new();
            int i, j;
            for (i = 0; i < fileHeader->NumThreads; i++) {
                currThread = malloc(sizeof(Thread));

                if(fread(currThread, sizeof(Thread), 1, instrFile) <= 0){
                    fprintf(stderr, "Could not read the trace file!\n");
                    exit(-1);
                }
                if(currThread->numBaseModulInstrs > LargestThreadNumInsstrs){
                	LargestThreadNumInsstrs = currThread->numBaseModulInstrs;
                	LargestThreadTID = currThread->tid;
                }
                currThread->ranges = al_new();
                for (j = 0; j < currThread->numRanges; j++) {
                    currRange = malloc(sizeof(ThreadRange));
                    if(fread(currRange, sizeof(ThreadRange), 1, instrFile) == -1){
                        fprintf(stderr, "Could not read the trace file!\n");
                        exit(-1);
                    }
                    al_add(currThread->ranges, currRange);
                }
                al_add(traceThreads, currThread);
            }
        }
        off_t size = ftello(instrFile);

        if (USE_MMAP) {
            //NOTE This closes the open file, and replaces the file pointer
            //with mmap pointer in the iList
            InitMMAP(instrFileName, iList);
        } else {
            if (!ftruncate(fileno(instrFile), size - fileHeader->RangeInfoSize)) {
                fprintf(stderr, "Error resizing instrFile!\n");
                exit(-1);
            }
        }
    } else {
        if ((traceFile = fopen(inFileName, "r")) == NULL) {
            fprintf(stderr, "ERROR: unable to open %s\n", inFileName);
            exit(0);
        }
        iList = BuildIListFromFile(traceFile, instrFileName);

        if (USE_MMAP) {
            //NOTE This closes the open file, and replaces the file pointer
            //with mmap pointer in the iList
            InitMMAP(instrFileName, iList);
        }
        if (TraceIsMultiThreaded) {
#if MARK_THREAD_BOUNDARIES
            MarkThreadBoundaries(iList);
#endif
        }

        fprintf(stderr, "Calculate known defs and uses...\n");
        CalcKnownDefsAndUses(iList);
        fprintf(stderr, "done.\n");


        free(baseInstrFileName);
        fclose(traceFile);
    }

    free(instrFileName);
    return (iList);
}
