/*
 *
 * File: simplify.c
 * Purpose: Semantics-preserving simplifications to the code.
 * Author: Babak Yadegari
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */


#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include "lynx_types.h"
#include "instr.h"
#include "print.h"
#include "simplify.h"
#include "hash.h"
#include "taint_analysis.h"
#include "flags.h"
//#include "./udis86-1.7/libudis86/decode.h"

//extern char *BaseModule; // program file name
extern char *baseName;
extern bool TraceIsMultiThreaded;
extern ArrayList* traceThreads;


int loop = 1;
int totalIterations = 0;

//struct PageTable *PageDir;
int totalPageDirEntries;


bool PartialKnown(Instruction *iins, REGISTER_BITVECTOR KnownRegs) {
    enum ud_mnemonic_code opcode;
    ud_operand_t *op0, *op1;
    REGISTER_BITVECTOR reg1, reg2;

    opcode = iins->uInstr.mnemonic;
    op0 = &(iins->uInstr.operand[0]);
    op1 = &(iins->uInstr.operand[1]);

    if (!(OpcodeIsArith(opcode) && (op0->type == UD_OP_REG && op1->type == UD_OP_IMM))) {
        return false;
    }

    reg1 = UDRegisterToBitVector(op0->base);
    reg2 = MaskDefinedReg(UDRegisterToBitVector(op0->base));

    if ((KnownRegs & (reg2 & ~reg1))) {
        return true;
    }

    return false;

}

int SetConstsArray(int *consts, REGISTER_BITVECTOR reg, int value) {
    int result = 0, i, pos;
    result |= reg;
    for (i = 0; i < 14; i++) {
        if (((reg >> i) & 1)) { // 32 bit registers
            consts[i] = value;
            result |= (result | (1 << (i + 14)));
            consts[i + 14] = (value & 0xffff);
        }
    }
    reg = result;
    for (pos = 0; i < 18; i++, pos++) { // ax through bx registers
        if (((reg >> i) & 1)) {
            consts[i - 14] = (consts[i - 14] & 0xffff0000);
            consts[i - 14] |= (value & 0xffff);
            consts[i] = (value & 0xffff);
            result |= (result | (3 << (i + pos + 8)));
            consts[i + pos + 8] = (value & 0xff); //al
            consts[i + pos + 9] = ((value & 0xff00) >> 8); //ah
        }
    }
    return result;

}


ud_t *ChangeMemOperandToReg(Instruction *iins, int reg) {
    ud_t *udins_old;
    int i = 0;
    //fprintf(stderr, "[ConstantPropThroughMemSets] %d: %s Changed to ", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
    udins_old = malloc(sizeof(ud_t));
    InitUdis(udins_old);
    ud_set_input_buffer(udins_old, iins->uInstr.bytes, MAX_INSTRLEN);
    ud_decode(udins_old);
    if (udins_old->operand[0].type == UD_OP_MEM) {
        i = 0;
    } else if (udins_old->operand[1].type == UD_OP_MEM)
        i = 1;
    udins_old->operand[i].type = UD_OP_REG;
    udins_old->operand[i].base = reg;
    udins_old->translator(udins_old);
    //fprintf(stderr, "%s\n", ud_insn_asm(udins_old));
    return udins_old;
}

int UDSubRegister(int reg, int size) {
    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (size == 2) {
            return (reg - UD_R_EAX) + UD_R_AX;
        } else if (size == 1) {
            return (reg - UD_R_EAX) + UD_R_AL;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        if (size == 1) {
            return (reg - UD_R_AX) + UD_R_AL;
        }
    }
    return UD_NONE;
}


/*
 * BitvectorLowOrHighOrder
 * taking a register as an argument, returns 1 if
 * the register is in high order bit, 0 otherwise
 * Note: it only works for 8-bit registers.
 * returns 0 for more than 8-bit registers!
 */
int BitvectorLowOrHighOrder(REGISTER_BITVECTOR reg) {
    if (reg >= LYNX_EAX && reg <= LYNX_DI) {
        return 0;
    } else {
        switch (reg) {
        case LYNX_AH:
        case LYNX_BH:
        case LYNX_CH:
        case LYNX_DH:
            return 1;
        case LYNX_AL:
        case LYNX_BL:
        case LYNX_CL:
        case LYNX_DL:
            return 0;
        default:
            fprintf(stderr, "[BitvectorLowOrHighOrder]: Warning: Unknown register: %d!\n",
                    (int) reg);
            return 0;
        }
    }
    return 0;
}


/*
 * PropagateConstants() -- carry out constant propagation on registers
 */

#define SIMPLIFY_INDIRECT_MEM_ACCESS	1

int PropagateConstants(InstrList* iList, int first, int last) {
    Instruction *iins, *iins2;
    ud_t *udins_new = NULL;
    enum ud_mnemonic_code opcode;
    enum ud_type reg1=0, reg2, reg3;
    ud_operand_t *op0, *op1, *op2;
    int i, cx = 0, prevesp;
    unsigned int val=0, N = 0;
    REGISTER_BITVECTOR thisreg, defreg, usedreg, nonMemRegUsed;
    int consts[30];
    REGISTER_BITVECTOR KnownRegs;
    bool memIsConst;
    unsigned int regMasks[30];
    // so in case of use of any of those, we can replace the mem with reg!

    // initialize mask for registers
    for (i = 0; i < 30; i++) {
        regMasks[i] = 0xffffffff;
    }
    WriteSet *memUsed, *memDef;

    REGISTER_BITVECTOR flagsKnown = 0xffffffff; // let's start the trace with assuming all flags as known, this trick may come handy for dealing with
    int j;

    REGISTER_BITVECTOR DisallowedDef, flagsDef=0;
    int push = 0, flags = 0;
    bool looking;
    uint64_t innerloop;


    // stuff for removing useless 'add' and 'sub' esp instrs
    int espAddandSubCount;
    uint64_t lastEspAddorSubInstr;

    iins = GetInstruction(iList, 0);
    prevesp = iins->esp;

    /*
     * we assume at the beginning of the trace, all registers
     * are known and constant except esp
     *
     */

    KnownRegs = (0xffffffff & ~(LYNX_ES | LYNX_SP));
    for (i = 0; i < 8; i++) {
        int regs = (1 << i);
        //int regs = MaskDefinedReg(k);
        int val = GetRegisterValue(iins, BitVectorToUDRegister(regs));
        SetConstsArray(consts, regs, val);
    }

    ClearConstantPages(iList, 0);

    espAddandSubCount = 0; // start with unbiased stack
    SetCurrentInstr(iList, first);
    cx++;
    i = first;
    lastEspAddorSubInstr = 0;




    while ((iins = FetchNextInstr(iList)) != NULL) {
        opcode = iins->uInstr.mnemonic;
        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);
        op2 = &(iins->uInstr.operand[2]);

        defreg = RegDef(iList, iins);
        usedreg = RegUsed(iList, iins);
        nonMemRegUsed = NonMemRegUsed(iList, iins);

        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
        	Instruction *prev = GetInstruction(iList, iins->prev);
        	if(!INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE)){
        		KnownRegs &= ~MaskDefinedReg(LYNX_EAX);
        	}
        }

        udins_new = NULL;

        if (iins->uInstr.mnemonic == UD_Ijmp && iins->uInstr.operand[0].type == UD_OP_JIMM
                && (strcmp(GetFunctionNameById(iins->funcName), "") == 0)) {
            Instruction *next = GetInstruction(iList, iins->next);
            if (INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE)) {
                INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                SaveInstrChange(iList, iins);
                continue;
            }

        }

        if (opcode == UD_Ipushad){
        	int regs;
        	int order = iins->order;
            for (regs = UD_R_EAX; regs <= UD_R_EDI; regs++) {
                ud_t *pushes = UDCreatePushReg(regs);
                Instruction *new = InstrClone(iins);
                new->esp = iins->esp - 4 * (regs - UD_R_EAX);
                CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), pushes, 0);
                new->regFlagsKnown |= MemDefKnown;
                new->memDefMin = new->esp - 4;
                new->memDefSize = 4;
                new->addr = ((iins->addr + regs) | 0x0f000000);

                new = AddInstruction(iList, new, order);
                order = new->order;
                RInstrDestroy(iList, new);
            }
            INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
            SaveInstrChange(iList, iins);
            SetCurrentInstr(iList, iins->next);
            continue;
        }
        if (opcode == UD_Ipopad){
				// break popad into series of pop
				int regs;
				int order = iins->order;

	            for (regs = 7; regs >= 0; regs--) {
	                ud_t *pops;
	                if (BitVectorToUDRegister((1 << regs)) != UD_R_ESP) {
	                    pops = UDCreatePopReg(BitVectorToUDRegister((1 << regs)));
	                } else {
	                    pops = UDCreateAddImmToReg(UD_R_ESP, 4);
	                }
	                Instruction *new = InstrClone(iins);
	                new->esp = iins->esp + 4 * (7 - regs);
	                CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), pops, 0);
	                new->regFlagsKnown |= MemUsedKnown;
	                new->memUsedMin = new->esp;
	                new->memUsedSize = 4;
	                new->addr = ((iins->addr + regs) | 0x0f000000);

	                AddInstruction(iList, new, order);
	                order = new->order;
				}
				INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
				SaveInstrChange(iList, iins);
				SetCurrentInstr(iList, iins->next);
				continue;
			}

        if(IsCondJump(iins)){
        	Instruction *next;
        	while((next = FetchNextInstr(iList)) != NULL && IsCondJump(next)){
				INSTR_SET_FLAG(next, INSTR_IS_DELETED);
				INSTR_SET_FLAG(next, INSTR_TMP0);
				SaveInstrChange(iList, next);
        	}
        	SetCurrentInstr(iList, iins->next);
        }

        /*if(iins->uInstr.mnemonic == UD_Ijmp && iins->uInstr.operand[0].type == UD_OP_JIMM && strlen(iins->funcName) == 0){
         Instruction *next;
         while((next = FetchNextInstr(iList)) != NULL){
         if(next->uInstr.mnemonic == UD_Ijmp && next->uInstr.operand[0].type == UD_OP_JIMM && strlen(next->funcName) == 0){
         INSTR_SET_FLAG(next, INSTR_IS_DELETED);
         INSTR_SET_FLAG(next, INSTR_TMP0);
         SaveInstrChange(iList, next);
         } else {
         break;
         }
         }
         SetCurrentInstr(iList, iins->next);
         }*/

        /*opcode = iins->uInstr.mnemonic;
        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);
        op2 = &(iins->uInstr.operand[2]);*/

        if (opcode == UD_Imov && op0->type == UD_OP_REG && op1->type == UD_OP_REG
                && op0->base == op1->base) {
            INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
            INSTR_SET_FLAG(iins, INSTR_TMP0);
            SaveInstrChange(iList, iins);
            continue;
        }

        /*            if(opcode == UD_Imov && op0->type == UD_OP_REG && op0->base == UD_R_ESP && op1->type == UD_OP_IMM){			// mov esp, imm, weird, change to add or sub of esp
         int diff = iins->esp - op1->lval.sdword;
         if(diff == 0){
         INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
         }
         else if(diff > 0){
         udins_new = UDCreateAddImmToReg(UD_R_ESP, diff);
         } else {
         udins_new = UDCreateSubImmFromReg(UD_R_ESP, -diff);
         }
         iins->flagsUsedKnown = iins->flagsDefKnown = true;
         iins->flagsUsed = iins->flagsDef = 0;
         CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
         SaveInstrChange(iList, iins);
         }*/


        if(opcode == UD_Imov && op0->type == UD_OP_REG && op1->type == UD_OP_REG && op1->base == UD_R_ESP){
        	udins_new = UDCreateMovImmToReg(op0->base, GetRegisterValue(iins, UD_R_ESP));
        	CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
        	SaveInstrChange(iList, iins);
        	free(udins_new);
        	udins_new = NULL;
        }




        if ((defreg == 0 && usedreg == 0 && !INSTR_HAS_FLAG(iins, INSTR_READS_MEM)
                && !INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && !FlagsDef(iins) && !FlagsUsed(iins))
                || iins->uInstr.mnemonic == UD_Irdtsc) {
            RInstrDestroy(iList, iins);
            continue;
        }

        /////// esp add sub verification!

        /*if ((opcode == UD_Iadd || opcode == UD_Isub) && op0->type == UD_OP_REG
                        && op0->base == UD_R_ESP && (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST)) {
            if (opcode == UD_Iadd) {
                espAddandSubCount += op1->lval.sdword;
                lastEspAddorSubInstr = iins->order;
                INSTR_SET_FLAG(iins, INSTR_MAYBE_DELETED);
                SaveInstrChange(iList, iins);
            } else if (opcode == UD_Isub) {
                espAddandSubCount -= op1->lval.sdword;
                lastEspAddorSubInstr = iins->order;
                INSTR_SET_FLAG(iins, INSTR_MAYBE_DELETED);
                SaveInstrChange(iList, iins);
            }
        }*/


        if ((opcode == UD_Iadd || opcode == UD_Isub || opcode == UD_Imov) && op0->type == UD_OP_REG
                && op0->base == UD_R_ESP && (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST)) {
            if (!INSTR_HAS_FLAG(iins, INSTR_MAYBE_DELETED)) {
                if (opcode == UD_Iadd) {
                    espAddandSubCount += op1->lval.sdword;
                    lastEspAddorSubInstr = iins->order;
                    INSTR_SET_FLAG(iins, INSTR_MAYBE_DELETED);
                    SaveInstrChange(iList, iins);
                } else if (opcode == UD_Isub) {
                    espAddandSubCount -= op1->lval.sdword;
                    lastEspAddorSubInstr = iins->order;
                    INSTR_SET_FLAG(iins, INSTR_MAYBE_DELETED);
                    SaveInstrChange(iList, iins);
                } else {
                    espAddandSubCount = 0;
                    INSTR_SET_FLAG(iins, INSTR_MAYBE_DELETED);
                    SaveInstrChange(iList, iins);
                }
            }
        	INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
        	SaveInstrChange(iList, iins);
            continue;

        } else if ((usedreg & LYNX_ESP) || (defreg & LYNX_ESP)) { // stop removing adds and subs
            Instruction *tmp = GetInstruction(iList, lastEspAddorSubInstr);
            if (INSTR_HAS_FLAG(tmp, INSTR_MAYBE_DELETED)) {
                INSTR_REMOVE_FLAG(tmp, INSTR_MAYBE_DELETED);
                //N--;
            }
            if (espAddandSubCount) {
                if (espAddandSubCount > 0) {
                    udins_new = UDCreateAddImmToReg(UD_R_ESP, espAddandSubCount);
                } else {
                    udins_new = UDCreateSubImmFromReg(UD_R_ESP, -espAddandSubCount);
                }
                espAddandSubCount = 0;
                CarefullCopyUdins2InsStr(iList, tmp, &(tmp->uInstr), udins_new, 1);

                free(udins_new);
                udins_new = NULL;
                INSTR_SET_FLAG(tmp, INSTR_TMP0);

                SaveInstrChange(iList, tmp);
                if (INSTR_HAS_FLAG(tmp, INSTR_IN_BASE_MODULE)) {
                    //N++;
                }
                //continue;
            }
        }

        if (IsPush(iins) || IsPop(iins) || opcode == UD_Iret || opcode == UD_Icall
                || opcode == UD_Ileave) {
            if (UDOpIsRegister(op0) == UD_R_ESP) {
                /*if (iins->uInstr.mnemonic == UD_Ipop)
                 usedreg = (usedreg & ~(LYNX_ESP | LYNX_SP));
                 else if (iins->uInstr.mnemonic == UD_Ipush)
                 defreg = (defreg & ~(LYNX_ESP | LYNX_SP));*/
            } else {
                defreg = (defreg & ~(LYNX_ESP | LYNX_SP));
                usedreg = (usedreg & ~(LYNX_ESP | LYNX_SP));
            }
        }

        memIsConst = true;

        if (opcode == UD_Ior && op0->type == UD_OP_REG
                && (KnownRegs & UDRegisterToBitVector(op0->base))
                && consts[floor_log2(UDRegisterToBitVector(op0->base))] == 0) { // or [0], reg --> mov [0], reg. [0] is the reg which is known to be 0!
            if (op1->type == UD_OP_REG) {
                udins_new = UDCreateMovRegToReg(op0->base, op1->base);
            }
        }

        if (opcode == UD_Ilea && (usedreg & (~LYNX_ESP) & KnownRegs) == (usedreg & (~LYNX_ESP))) {
            Instruction *next = GetInstruction(iList, iins->next);
            if ((defreg == LYNX_ESP) && (usedreg == LYNX_ESP)) { // lea esp, [esp + val]
                val = GetRegisterValue(next, UD_R_ESP);
                int diff = val - iins->esp;
                if (diff > 0) {
                    udins_new = UDCreateAddImmToReg(UD_R_ESP, diff);
                } else {
                    udins_new = UDCreateSubImmFromReg(UD_R_ESP, -diff);
                }
            } else {

                val = GetRegisterValue(next, op0->base);
                udins_new = UDCreateMovImmToReg(op0->base, val);
            }


            CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
            iins->regFlagsKnown |= (FlagsUsedKnown | FlagsDefKnown);
            iins->flagsDef = iins->flagsUsed = 0;
            free(udins_new);
            udins_new = NULL;
        }
        /*
         * arithmetic instructions without taint can be simplified
         */
        if(OpcodeIsArith(opcode) && INSTR_HAS_FLAG(iins, INSTR_TAINT_CLEARED)){

        	if(op0->type == UD_OP_REG){
        		if(!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)){
        		//Instruction *tmp = GetInstruction(iList, iins->next);
        		val = GetRegisterValue(iins, op0->base);
        		udins_new = UDCreateMovImmToReg(op0->base, val);
        		INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
        		CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
        		SaveInstrChange(iList, iins);

                opcode = iins->uInstr.mnemonic;
                op0 = &(iins->uInstr.operand[0]);
                op1 = &(iins->uInstr.operand[1]);
                op2 = &(iins->uInstr.operand[2]);


                defreg = RegDef(iList, iins);
                usedreg = RegUsed(iList, iins);
                nonMemRegUsed = NonMemRegUsed(iList, iins);


        	} else {
        		INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
        		SaveInstrChange(iList, iins);
        	}
        	}



        }

        // simplify non-mem instrs which the operands are known!
        if (!INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && !INSTR_HAS_FLAG(iins,INSTR_WRITES_MEM)) {
            if ((usedreg & KnownRegs &~(LYNX_SP)) == (usedreg &~(LYNX_SP)) || PartialKnown(iins, KnownRegs)) { // exlude sub and add esp instrs!
                if (OpcodeIsArith(opcode) && ((usedreg & LYNX_ESP) == 0 || opcode == UD_Isbb)) {
                    reg1 = UDOpIsRegister(op0);
                    if (reg1 == 0) {
                        reg1 = BitVectorToUDRegister(defreg);
                    }
                    Instruction *tmp;
                    if (op2->type == UD_NONE) {
                        if ((reg2 = UDOpIsRegister(op1)) != UD_NONE) { // second op is a known reg
                            INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                            INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                            SaveInstrChange(iList, iins);
                            int val1 = consts[floor_log2(UDRegisterToBitVector(reg2))];
                            if (iins->uInstr.bytes[0] == 0x83 || iins->uInstr.bytes[1] == 0x83) { // arithmetics with 83 opcode, do the sign extention!
                                val1 = SignExtend(val1);
                            }
                            tmp = GetInstruction(iList, iins->next);
                            val = consts[floor_log2(UDRegisterToBitVector(reg1))] =
                                    GetRegisterValue(tmp, reg1);
                            //ArithOp(opcode, consts[floor_log2(UDRegisterToBitVector(reg1))], val1, op1->size, flagsKnown, iins);
                            udins_new = UDCreateMovImmToReg(reg1, val);
                        } else if (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST) { // second op is constant
                            INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                            SaveInstrChange(iList, iins);
                            int val1 = op1->lval.sdword;
                            if (iins->uInstr.bytes[0] == 0x83 || iins->uInstr.bytes[1] == 0x83) { // arithmetics with 83 opcode, do the sign extention!
                                val1 = SignExtend(val1);
                            }
                            tmp = GetInstruction(iList, iins->next);
                            val = consts[floor_log2(UDRegisterToBitVector(reg1))] =
                                    GetRegisterValue(tmp, reg1);
                            //ArithOp(opcode, consts[floor_log2(UDRegisterToBitVector(reg1))], val1, op0->size, flagsKnown, iins);
                            udins_new = UDCreateMovImmToReg(reg1, val);

                        } else { // one operand
                            val = consts[floor_log2(UDRegisterToBitVector(reg1))] = ArithOp(opcode,
                                    consts[floor_log2(UDRegisterToBitVector(reg1))], 1, op0->size,
                                    &flagsKnown, iins);
                            INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                            SaveInstrChange(iList, iins);
                            udins_new = UDCreateMovImmToReg(reg1, val);
                        }
                        if (!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                            udins_new = UDCreateMovImmToReg(reg1, val);
                        } else {
                        	INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
                        	SaveInstrChange(iList, iins);
                        }
                    } else { // 3 operand instr
                        reg2 = UDOpIsRegister(op1);
                        if (opcode == UD_Imul || opcode == UD_Iimul
                        			|| opcode ==UD_Ishrd || opcode == UD_Ishld) { // handling imul now! TODO: generalize for all 3 operand instrs!
                            if ((reg3 = UDOpIsRegister(op2)) != UD_NONE) {
                                val = consts[floor_log2(UDRegisterToBitVector(reg1))] = ArithOp(
                                        opcode, consts[floor_log2(UDRegisterToBitVector(reg2))],
                                        consts[floor_log2(UDRegisterToBitVector(reg3))], op2->size,
                                        &flagsKnown, iins);
                            } else { // op2 constant
                                val = consts[floor_log2(UDRegisterToBitVector(reg1))] = ArithOp(
                                        opcode, consts[floor_log2(UDRegisterToBitVector(reg2))],
                                        op2->lval.sdword, op2->size, &flagsKnown, iins);
                            }
                            if (!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                                udins_new = UDCreateMovImmToReg(reg1, val);
                            }else {
                            	INSTR_SET_FLAG(iins, INSTR_REGDEF_KNOWN);
                            	INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
                            	SaveInstrChange(iList, iins);
                            }
                        }
                    }
                    if ((usedreg & KnownRegs) == usedreg) {
                        if (opcode == UD_Icmp || opcode == UD_Itest) {
                            if ((reg1 = UDOpIsRegister(op0))
                                    && (UDRegisterToBitVector(reg1) & KnownRegs)) {
                                INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                                SaveInstrChange(iList, iins);
                            }
                            if ((reg1 = UDOpIsRegister(op1))
                                    && (UDRegisterToBitVector(reg1) & KnownRegs)) {
                                INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                                SaveInstrChange(iList, iins);
                            }
                        }
                    }
                } else if (opcode == UD_Iaaa && (((KnownRegs & LYNX_AL) != 0) || ((KnownRegs & LYNX_AX) != 0) || ((KnownRegs & LYNX_EAX) != 0))){
                	Instruction *tmp = GetInstruction(iList, iins->origNext);
                	udins_new = UDCreateMovImmToReg(UD_R_AL, GetRegisterValue(tmp, UD_R_AL));

                } else if (op0->type == UD_NONE && !FlagsDef(iins)) { // instrs without operands! cdq

                    if (opcode == UD_Icdq) {
                        udins_new = UDCreateMovImmToReg(UD_R_EDX,
                                (SignExtend(consts[0]) & 0xffffffff)); // mov sign extend eax to edx
                    } else if(opcode == UD_Ileave){
                    	udins_new = UDCreateMovImmToReg(UD_R_EBP, GetRegisterValue(iins, UD_R_EBP));
                    }
                }
            } else if (op0->type == UD_OP_REG && op1->type == UD_OP_IMM) {

            }
            if (udins_new != NULL && !INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE)) { // we recognized opcode2
                if(INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)){
                	flagsKnown = FlagsDef(iins);
                	CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);

                } else {
                	CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                }
                free(udins_new);
                udins_new = NULL;
                INSTR_SET_FLAG(iins, INSTR_TMP0);
                SaveInstrChange(iList, iins);
                if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
                    N++;
                }
                defreg = RegDef(iList, iins);
                usedreg = RegUsed(iList, iins);
                opcode = iins->uInstr.mnemonic;
                op0 = &(iins->uInstr.operand[0]);
                op1 = &(iins->uInstr.operand[1]);
                op2 = &(iins->uInstr.operand[2]);
                udins_new = NULL;

            } else {
                free(udins_new);
                udins_new = NULL;
            }

        }
        if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && !INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {

            if (opcode == UD_Ipushfd && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) {
            	INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
            	SaveInstrChange(iList, iins);
                val = -1;
            } else if (!IsStringOperation(iins) && ((nonMemRegUsed
                    && (nonMemRegUsed & KnownRegs & (~LYNX_ESP))
                            == (nonMemRegUsed & (~LYNX_ESP))) || !nonMemRegUsed)
                    && !(opcode == UD_Ipushfd)) {
                if (op2->type == UD_NONE) {
                    if (op1->type == UD_OP_REG) {
                        val = consts[floor_log2(nonMemRegUsed)];

                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                        SaveInstrChange(iList, iins);

                    } else if (op1->type == UD_OP_CONST || op1->type == UD_OP_IMM) {
                        INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                        SaveInstrChange(iList, iins);
                        val = op1->lval.sdword;
                    } else if (opcode == UD_Icall) {
                        //if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
                            val = iins->addr + iins->uInstr.length;
                        //} else {
                        //    memIsConst = false;
                        //}
                    } else if (op1->type == UD_NONE ) {
                        if (op0->type == UD_OP_REG) {
                            val = consts[floor_log2(nonMemRegUsed)];
                            if (op0->base == UD_R_ESP) {
                                val = GetRegisterValue(iins, UD_R_ESP);
                            }
                        } else {
                            val = op0->lval.sdword;
                        }
                    }
                } else {
                    // is there any? an instr with mem dest and 3 operands
                }
            } else if (opcode == UD_Icall && op0->type == UD_OP_IMM /*&& INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)*/) {
                val = iins->addr + iins->uInstr.length;
            } else if (BitVectorToUDRegisterForTest(nonMemRegUsed) && (nonMemRegUsed
                    && (MaskDefinedReg(nonMemRegUsed) & KnownRegs & (~LYNX_ESP)))) { // if a only a part of a register is known, like cl in ecx!

                /*
                 * select the write range which corresponds to the known part,
                 * lower memory addresses should go to the least significant
                 * bits!(?) Does this depend on big/little-endian?
                 */
                REGISTER_BITVECTOR tmpreg = (MaskDefinedReg(nonMemRegUsed) & KnownRegs);

                for (i = 0; i < 32; i++) {
                    REGISTER_BITVECTOR treg = (1 << i);
                    if ((tmpreg & treg)) {
                        val = consts[floor_log2(treg)];
                        int memMin, memMax;
                        memMin = iins->memDefMin
                                + BitvectorLowOrHighOrder(treg);
                        memMax = memMin + BitvectorRegisterSize(treg) - 1;
                        SetAddrAsConstWithValue(iList, memMin,
                                memMax - memMin + 1, val);
                    }
                }

            } else if (opcode == UD_Istosb || opcode == UD_Istosd
                    || opcode == UD_Istosw) {
            	if((KnownRegs & LYNX_EDI)){
            		defreg = 0;
            	}
            	ud_t *udnew;
                if (opcode == UD_Istosb && (KnownRegs & LYNX_AL)) {
                    SetAddrAsConstWithValue(iList, iins->memDefMin, 1,
                            consts[22] & 0xff);
                    val = consts[22] & 0xff;

                } else if (opcode == UD_Istosw && (KnownRegs & LYNX_AX)) {
                    SetAddrAsConstWithValue(iList, iins->memDefMin, 2,
                            consts[14] & 0xffff);
                    val = consts[14] & 0xffff;
                } else if (opcode == UD_Istosd && (KnownRegs & LYNX_EAX)) {
                    SetAddrAsConstWithValue(iList, iins->memDefMin, 4,
                            consts[1]);
                    val = consts[1];
                }


					udnew = UDCreateMovImmToOffset(iins->memDefMin, val, iins->memDefSize);
					CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udnew, 0);
					SaveInstrChange(iList, iins);
					free(udnew); udnew = NULL;


                Instruction *tmp = InstrClone(iins);
				tmp->addr += 1;
				udnew = UDCreateAddImmToReg(UD_R_EDI, 1);
				CarefullCopyUdins2InsStr(iList, tmp, &(tmp->uInstr), udnew, 0);
				AddInstruction(iList, tmp, iins->order);
				SaveInstrChange(iList, tmp);
				free(udnew);

            } else {
                // remove previous entries

                memIsConst = false;
                UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);

                if(opcode == UD_Ipushfd){           // means it is forward tainted!
                    val = FlagsUsed(iins);
                    SetAddrAsConstWithValue(iList, iins->memDefMin, iins->memDefSize, val);
                    UnsetConstAddrWithoutMemReset(iList, iins->memDefMin, iins->memDefSize);

                }

                /*
                 * if a reg is being copied to a memory, correspond the reg and memory addr
                 */
                /*if (BitVectorToUDRegisterForTest(nonMemRegUsed)) { // make sure it is one reg
                 int index = floor_log2(nonMemRegUsed);

                 if(WriteSetsOverlap(MemToRegCopy[index], memDef)){
                 INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                 SaveInstrChange(iList, iins);
                 } else{
                 MemToRegCopy[k] = MergeWriteSets(MemToRegCopy[index],
                 memDef);
                 }
                 }*/
            }

            if (memIsConst) {
            	SetAddrAsConstWithValue(iList, iins->memDefMin,
            			iins->memDefSize, val);
            } else {
            	UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
            }

        } else if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
            if (MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize)) {
                val = GetMemoryLocValue(iList, iins->memUsedMin,
                        iins->memUsedSize);
                if (INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE)) {
                    if (MemLocIsConstant(iList, iins->memDefMin,
                            iins->memDefSize)) {
                    	// simplify the instruction itself but do not propagate!
                        UnsetConstAddr(iList, iins->memUsedMin,iins->memUsedSize);
                    }
                } /*else*/

                {
                    if (OpcodeIsArith(opcode)) {
                        if (op2->type == UD_NONE) {
                            if (op1->type == UD_OP_REG) {
                                if ((KnownRegs & nonMemRegUsed) == nonMemRegUsed) {
                                    INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                                    INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                                    SaveInstrChange(iList, iins);
                                    val = GetMemoryLocValue(iList, iins->memUsedMin,
                                            iins->memUsedSize);
                                    SetAddrAsConstWithValue(iList, iins->memUsedMin,
                                            iins->memUsedSize,
                                            ArithOp(opcode, val, consts[floor_log2(nonMemRegUsed)],
                                                    4, &flagsKnown, iins));
                                } else {
                                    UnsetConstAddr(iList, iins->memDefMin,
                                            iins->memDefSize);
                                }
                            } else if (op1->type == UD_OP_CONST || op1->type == UD_OP_IMM) {
                                INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
                                SaveInstrChange(iList, iins);
                                val = GetMemoryLocValue(iList, iins->memUsedMin,
                                        iins->memUsedSize);
                                SetAddrAsConstWithValue(iList, iins->memUsedMin,
                                        iins->memUsedSize,
                                        ArithOp(opcode, val, op1->lval.sdword, 4, &flagsKnown,
                                                iins));
                            } else if (op1->type == UD_NONE) { // one operand instruction
                                if (opcode == UD_Iimul || opcode == UD_Iidiv || opcode == UD_Imul
                                        || opcode == UD_Idiv) {
                                    if ((KnownRegs & LYNX_EAX)) {
                                        val = GetMemoryLocValue(iList, iins->memUsedMin,
                                                iins->memUsedSize);
                                        SetAddrAsConstWithValue(iList, iins->memUsedMin,
                                                iins->memUsedSize,
                                                ArithOp(opcode, val,
                                                        consts[floor_log2(nonMemRegUsed)], 4,
                                                        &flagsKnown, iins));
                                        udins_new = UDCreateMovImmToReg(UD_R_EAX,
                                                ArithOp(opcode, val,
                                                        consts[floor_log2(nonMemRegUsed)], 4,
                                                        &flagsKnown, iins));
                                        INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                                    }
                                } else {
                                    INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                                    SaveInstrChange(iList, iins);
                                    SetAddrAsConstWithValue(iList, iins->memUsedMin,
                                            iins->memUsedSize,
                                            ArithOp(opcode, val, 0, 4, &flagsKnown, iins));
                                }
                            }
                        } else {
                            // is there any? an instr with mem dest and 3 operands
                        }
                    } else if (nonMemRegUsed) { // xchg, xadd, ...
                        if (opcode == UD_Ixchg) {
                            if ((nonMemRegUsed & KnownRegs)) {
                                udins_new = UDCreateMovImmToOffset(iins->memDefMin,
                                        consts[floor_log2(nonMemRegUsed)],
                                        iins->memDefSize);

                                ud_t *udins = UDCreateMovImmToReg(
                                        BitVectorToUDRegister(nonMemRegUsed), val);
                                Instruction *new = InstrClone(iins);
                                new->order = iins->order;
                                new->addr = iins->addr + 1;
                                CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
                                free(udins);

                                new = AddInstruction(iList, new, iins->order);

                                int tmp = consts[floor_log2(nonMemRegUsed)];
                                consts[floor_log2(nonMemRegUsed)] = val;

                                SetAddrAsConstWithValue(iList, iins->memDefMin,
                                        iins->memDefSize, tmp);
                            } else {
                                UnsetConstAddr(iList, iins->memDefMin,
                                        iins->memDefSize);
                                //consts[floor_log2(nonMemRegUsed)] = val;
                                //KnownRegs |= MaskDefinedReg(nonMemRegUsed);

                                // first op is mem and const, second is reg and not const OR
                                // first op is reg and not const, second is mem and const
                                // in either case:
                                // mov [mem], reg and mov reg, const
                                //udins_new = UDCreateMovOffsetToReg(
                                        //BitVectorToUDRegister(nonMemRegUsed), iins->memUsedMin, 1);

                                ud_t *udins = UDCreateMovImmToReg(
                                        BitVectorToUDRegister(nonMemRegUsed), val);
                                Instruction *new = InstrClone(iins);
                                new->addr = iins->addr + 1;
                                CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
                                free(udins);
                                new = AddInstruction(iList, new, iins->order);
                            }
                        } else if ((opcode == UD_Icmpxchg || opcode == UD_Icmpxchg8b)
                                && (KnownRegs & LYNX_EAX)) {
                            // compares eax with op2
                            bool compare = false;
                            int eax = consts[0];
                            int dest;
                            if (op0->type == UD_OP_REG && (KnownRegs & nonMemRegUsed)) {
                                dest = op0->lval.sdword;
                                compare = true;
                            } else if (op0->type == UD_OP_MEM) {
                                dest = val;
                                compare = true;
                            }

                            if (compare) {
                                if (dest == eax) { // dest <== src
                                    if (op0->type == UD_OP_MEM) {
                                        if ((KnownRegs & nonMemRegUsed)) {
                                            udins_new = UDCreateMovImmToOffset(iins->memDefMin,
                                                    consts[floor_log2(
                                                            UDRegisterToBitVector(op1->base))],
                                                    iins->memDefSize);
                                        } else {
                                            udins_new = UDCreateMovOffsetToReg(op1->base,
                                                    iins->memDefMin, 1);
                                        }
                                    } else { //op0 = reg
                                        udins_new = UDCreateMovImmToReg(op0->base, val);
                                    }

                                } else { // eax <== dest
                                    if (op0->type == UD_OP_MEM) {
                                        udins_new = UDCreateMovImmToReg(UD_R_EAX, val);
                                    } else {
                                        if ((KnownRegs & nonMemRegUsed)) {
                                            udins_new = UDCreateMovImmToReg(UD_R_EAX,
                                                    consts[floor_log2(
                                                            UDRegisterToBitVector(op0->base))]);
                                        } else {
                                            udins_new = UDCreateMovRegToReg(UD_R_EAX, op0->base);
                                        }
                                    }

                                }
                            }

                        } else if (opcode == UD_Imovsb || opcode == UD_Imovsd || opcode == UD_Imovsw
                                || opcode == UD_Imovsq) {
                            SetAddrAsConstWithValue(iList, iins->memDefMin,
                                    iins->memDefSize, val);


                        	/*Instruction *tmp = InstrClone(iins);
                        	tmp->addr += 1;
                        	ud_t *udnew = UDCreateAddImmToReg(UD_R_EDI, 1);
                        	CarefullCopyUdins2InsStr(iList, tmp, &(tmp->uInstr), udnew, 0);
                        	AddInstruction(iList, tmp, iins->order);
                        	SaveInstrChange(iList, tmp);
                        	free(udnew);*/

                        	/*tmp = InstrClone(iins);
                        	tmp->addr += 2;
                        	udnew = UDCreateAddImmToReg(UD_R_ESI, 1);
                        	CarefullCopyUdins2InsStr(iList, tmp, &(tmp->uInstr), udnew, 0);
                        	AddInstruction(iList, tmp, iins->order);
                        	SaveInstrChange(iList, tmp);
                        	free(udnew);*/
                        	defreg = 0;


                        } else if (opcode == UD_Istosb || opcode == UD_Istosd || opcode == UD_Istosw
                                || opcode == UD_Istosq) {
                            SetAddrAsConstWithValue(iList, iins->memDefMin,
                                    iins->memDefSize, val);

                        	/*Instruction *tmp = InstrClone(iins);
                        	tmp->addr += 1;
                        	ud_t *udnew = UDCreateAddImmToReg(UD_R_EDI, 1);
                        	CarefullCopyUdins2InsStr(iList, tmp, &(tmp->uInstr), udnew, 0);
                        	AddInstruction(iList, tmp, iins->order);
                        	SaveInstrChange(iList, tmp);
                        	free(udnew);*/

                        	defreg = 0;
                        }
                    } else { // mem copy: push [mem], pop [mem], cmpxchg [mem], ...
                        val = GetMemoryLocValue(iList, iins->memUsedMin,
                                iins->memUsedSize);
                        SetAddrAsConstWithValue(iList, iins->memDefMin,
                                iins->memDefSize, val);

                        // change instr to move [target], const
                        //int size = iins->memDefSize;
                        if (opcode != UD_Icall) {
                            //Range *r = al_get(memDef->ranges, 0);
                            //size = r->max - r->min + 1; // number of bytes

                            //udins_new = UDCreateMovImmToOffset(iins->memDefMin, val, iins->memDefSize);


                        }

                        if (opcode == UD_Ipop) {
                            ud_t *udins = UDCreateAddImmToReg(UD_R_ESP, iins->memUsedSize);
                            Instruction *new = InstrClone(iins);
                            new->order = iins->order;
                            new->addr = iins->addr;
                            new->flags = 0;
                            CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
                            new->regFlagsKnown |= FlagsDefKnown;
                            new->regFlagsKnown |= FlagsUsedKnown;
                            new->flagsUsed = 0;
                            new->flagsDef = 0;
                            free(udins);
                            if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                                INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
                            }

                            new = AddInstruction(iList, new, iins->order);

                        }
                        if (opcode == UD_Ipush) {
                            ud_t *udins = UDCreateSubImmFromReg(UD_R_ESP, iins->memDefSize);
                            Instruction *new = InstrClone(iins);
                            new->order = iins->order;
                            new->addr = iins->addr;
                            new->flags = 0;
                            CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
                            new->regFlagsKnown |= FlagsDefKnown;
                            new->regFlagsKnown |= FlagsUsedKnown;
                            new->flagsUsed = 0;
                            new->flagsDef = 0;
                            free(udins);
                            if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                                INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
                            }
                            new = AddInstruction(iList, new, iins->order);

                        }
                    }
                }
            } else if (opcode == UD_Ixchg) {
                if ((nonMemRegUsed & KnownRegs)) {
                    udins_new = UDCreateMovOffsetToReg(BitVectorToUDRegister(nonMemRegUsed),
                            iins->memUsedMin, OFFSET_TO_REG);

                    Instruction *tmp = InstrClone(iins);
                    ud_t *udins = UDCreateMovImmToOffset(iins->memDefMin,
                            consts[floor_log2(nonMemRegUsed)],
                            iins->memDefSize);
                    /*UDCreateMovImmToReg(BitVectorToUDRegister(nonMemRegUsed),
                     consts[floor_log2(nonMemRegUsed)]);*/
                    tmp->addr = iins->addr + 1;
                    CarefullCopyUdins2InsStr(iList, tmp, &(tmp->uInstr), udins, 1);

                    AddInstruction(iList, tmp, iins->order);
                }
            } else {
                if (op0->type == UD_OP_MEM && op1->type == UD_NONE
                        && INSTR_HAS_FLAG(iins, INSTR_OP0_CONSTANT)
                        && (opcode == UD_Icmpxchg || opcode == UD_Icmpxchg8b)
                        && (KnownRegs & LYNX_EAX)) {
                    // compares eax with op2
                    Instruction *tmp = GetInstruction(iList, iins->next);
                    int val = GetRegisterValue(tmp, UD_R_EAX);
                    udins_new = UDCreateMovImmToReg(UD_R_EAX, val);
                } else if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)
                        && opcode == UD_Icall&& INSTR_HAS_FLAG(iins, INSTR_OP0_CONSTANT)) {
                    SetAddrAsConstWithValue(iList, iins->memDefMin,
                            iins->memDefSize,
                            iins->addr + iins->uInstr.length);
                }
                else {
                    UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
                }
            }

            /*if(op0->type == UD_OP_MEM && (op1->type == UD_OP_CONST || op1->type == UD_OP_IMM)){
             for(k = 0; k < numRegs; k++){
             if (!WriteSetIsEmpty(MemToRegCopy[k]) && WriteSetSubsetOfWriteSet(MemToRegCopy[k], memUsed)){
             enum ud_type dst = BitVectorToUDRegister((1 << k));
             dst = UDRegSizeMask(dst, UDRegSize(op0->base));
             if(opcode == UD_Iadd){
             udins_new = UDCreateAddImmToReg(dst, op1->lval.sdword);
             }
             break;
             }
             }
             }*/

            //WriteSetDestroy(memDef);
            //WriteSetDestroy(memUsed);

        } else if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && !INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
            //memUsed = MemUsed(iList, iins);

            //prevknownRanges = SearchMemUsedInMemDefKnown(MemDefKnown, memUsed);
            if (MemLocIsConstant(iList, iins->memUsedMin,
                    iins->memUsedSize) || INSTR_HAS_FLAG(iins, INSTR_READS_GLOBAL_MEM)) {
                val = GetMemoryLocValue(iList, iins->memUsedMin,
                        iins->memUsedSize);
                //iins2 = GetInstruction(iList, memUsedRes->index);
                if (opcode == UD_Ipop && op0->type != UD_OP_MEM) {
                	if(op0->base != UD_R_ESP){
						udins_new = UDCreateMovImmToReg(op0->base, val);

						ud_t *udins = UDCreateAddImmToReg(UD_R_ESP, UDRegSize(op0->base));
						Instruction *new = InstrClone(iins);
						new->order = iins->order;
						new->addr = iins->addr;
						new->flags = 0;
						if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
							INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
						}
						CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
						new->regFlagsKnown |= FlagsDefKnown;
						new->regFlagsKnown |= FlagsUsedKnown;
						new->flagsUsed = 0;
						new->flagsDef = 0;
						free(udins);
						if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
							INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
						}
						new = AddInstruction(iList, new, iins->order);
                	} else {
                		int diff = GetRegisterValue(iins, UD_R_ESP);
                		Instruction *next = GetInstruction(iList, iins->order+1);
                		diff -= GetRegisterValue(next, UD_R_ESP);
                		if(diff < 0){
                			udins_new = UDCreateAddImmToReg(UD_R_ESP, -diff);
                		} else if(diff > 0) {
                			udins_new = UDCreateSubImmFromReg(UD_R_ESP, diff);
                		} else {
                			udins_new = NULL;
                			INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                			SaveInstrChange(iList, iins);
                		}
                	}


                } else if (opcode == UD_Ipopfd) {
                    INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN); //this changes a popf to nop IF the stack location is a constant. This is to get rid of the corresponding stack location!
                    udins_new = UDCreateNop();
                    iins->flagsDef = flagsDef;
                    iins->regFlagsKnown |= FlagsDefKnown;
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
                    free(udins_new);
                    udins_new = false;
                    SaveInstrChange(iList, iins);

                } else if (opcode == UD_Icmp || opcode == UD_Itest) { // Instructions of type "cmp mem/reg, mem/imm"
                    if ((op0->type == UD_OP_MEM || op0->type == UD_OP_IMM
                            || op0->type == UD_OP_CONST)
                            ) {
                        INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                        INSTR_SET_FLAG(iins, INSTR_MEM_SIMP);
                        if (op0->type == UD_OP_MEM) {
                            udins_new = ChangeMemOperandToConstant(iins, val, 0);
                        }
                        //N++;
                    }
                    if ((op1->type == UD_OP_MEM || op1->type == UD_OP_IMM
                            || op1->type == UD_OP_CONST)
                            ) {
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        INSTR_SET_FLAG(iins, INSTR_MEM_SIMP);
                        if (op1->type == UD_OP_MEM) {
                            udins_new = ChangeMemOperandToConstant(iins, val, 1);
                        }
                        //N++;
                    }
                    if (op0->type == UD_OP_REG && (KnownRegs & nonMemRegUsed)) {
                        INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                        //SaveInstrChange(iList, iins);
                    }
                    if (op1->type == UD_OP_REG && (KnownRegs & nonMemRegUsed)) {
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        //SaveInstrChange(iList, iins);
                    }

                    SaveInstrChange(iList, iins);
                } else if (iins->uInstr.bytes[0] == 0xf7) {
                    if ((KnownRegs & LYNX_AX)) {

                        val = ArithOp(opcode, val, consts[0], 1, &flagsKnown, iins);
                        consts[0] = val & 0xffffffff; // store results in edx:eax
                        consts[1] = val & 0xffffffff00000000;
                        KnownRegs |= (LYNX_EDX);
                    }
                } else if (iins->uInstr.bytes[0] == 0x66 && iins->uInstr.bytes[1] == 0xf7) {
                    if ((KnownRegs & LYNX_AX)) {

                        val = ArithOp(opcode, val, consts[floor_log2(LYNX_AX)], 1, &flagsKnown,
                                iins);
                        consts[0] = val & 0xffff; // store results in dx:ax
                        consts[1] = val & 0xffff0000;
                        KnownRegs |= (LYNX_DX);
                    }
                } else if (OpcodeIsArith(opcode)) { // it is: arith reg, [mem]
                    if ((KnownRegs & usedreg) == usedreg) {
                        val = ArithOp(opcode, consts[floor_log2(UDRegisterToBitVector(op0->base))],
                                val, op1->size, &flagsKnown, iins);
                        udins_new = UDCreateMovImmToReg(BitVectorToUDRegister(defreg), val);
                        INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                    } else {
                        udins_new = ChangeMemOperandToConstant(iins, val, 1);
                    }
                } else if (opcode == UD_Iret && INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) /*&& iins2->uInstr.mnemonic != UD_Icall*/) {
                    int size = 4;
                    if (op0->type != UD_NONE) { // number of bytes being popped off the stack
                        size += op0->lval.sdword;
                    }
                    udins_new = UDCreateJump(val - 5);
                    //INSTR_SET_FLAG(iins2, INSTR_MEM_SIMP);
                    INSTR_SET_FLAG(iins, INSTR_MEM_SIMP);

                    //INSTR_SET_FLAG(iins2, INSTR_IS_DELETED);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                    SaveInstrChange(iList, iins);
                    //SaveInstrChange(iList, iins2);
                    free(udins_new);
                    udins_new = NULL;
                    //N++;

                    ud_t *udins = UDCreateAddImmToReg(UD_R_ESP, size);
                    Instruction *new = InstrClone(iins);
                    new->order = iins->order;
                    new->addr = iins->addr;
                    new->flags = 0;
                    CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
                    new->regFlagsKnown |= FlagsDefKnown;
                    new->regFlagsKnown |= FlagsUsedKnown;
                    new->flagsUsed = 0;
                    new->flagsDef = 0;
                    free(udins);

                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                        INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
                    }

                    new = AddInstruction(iList, new, iins->prev);

                } else if (opcode == UD_Iscasb && (KnownRegs & LYNX_EAX) == 1) { // scasb
                    udins_new = UDCreateCmpOffsetWithImm(iins->memUsedMin,
                            consts[floor_log2(LYNX_AL)], iins->memUsedSize);
                    INSTR_SET_FLAG(iins, INSTR_READS_GLOBAL_MEM);
                    SaveInstrChange(iList, iins);
                } else if (defreg && !IsStringOperation(iins) && !IsMxcsrInstr(iins)) {

						udins_new = UDCreateMovImmToReg(BitVectorToUDRegister(defreg), val);
						INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);

                }
            } else if (defreg && INSTR_HAS_FLAG(iins, INSTR_READS_GLOBAL_MEM)
                    && (usedreg & KnownRegs) == usedreg) { // arithmetic with data section
                if (IsStringCompare(iins)) {
                    INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
                } else if (opcode != UD_Ixchg) {
                    iins2 = GetInstruction(iList, iins->next);
                    int val = GetRegisterValue(iins2, op0->base);
                    udins_new = UDCreateMovImmToReg(op0->base, val);
                    INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                }

            }
        }

        if (udins_new != NULL) { // we recognized opcode2

            if ((INSTR_HAS_FLAG(iins, INSTR_SIMPLIFIED_IGNORED)
                    && INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE))) { // stop simplifying
                INSTR_REMOVE_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                SaveInstrChange(iList, iins);
                free(udins_new);
                udins_new = NULL;
            } else {

                CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);

                free(udins_new);
                udins_new = NULL;
                INSTR_SET_FLAG(iins, INSTR_TMP0);
                INSTR_REMOVE_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                SaveInstrChange(iList, iins);
                if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
                    //N++;
                }
                defreg = RegDef(iList, iins);
                usedreg = RegUsed(iList, iins);
                opcode = iins->uInstr.mnemonic;
                op0 = &(iins->uInstr.operand[0]);
                op1 = &(iins->uInstr.operand[1]);
                op2 = &(iins->uInstr.operand[2]);
                udins_new = NULL;
            }
        }

        if ((usedreg & KnownRegs) != 0
                && ((defreg & usedreg) == 0
                        || ((defreg & usedreg) != 0
                                && (OpcodeIsArith(opcode)
                                        || (opcode == UD_Ilea || opcode == UD_Ixadd
                                                || opcode == UD_Ixchg || opcode == UD_Ileave
                                                || op1->type == UD_OP_MEM)
                                        || (opcode == UD_Ipush && op0->type == UD_OP_REG
                                                && (KnownRegs & UDRegisterToBitVector(op0->base))
                                                        != 0))))) { // we can propagate
            reg2 = UDOpIsRegister(op0);
            val = consts[floor_log2(UDRegisterToBitVector(UDOpIsRegister(op1)))];
            if (opcode == UD_Ixchg
                    && ((op0->type == UD_OP_REG && op1->type == UD_OP_MEM)
                            || (op0->type == UD_OP_MEM && op1->type == UD_OP_REG))) {
                if (op0->type == UD_OP_MEM) {
                    INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                } else {
                    INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                }
                SaveInstrChange(iList, iins);
            } else if (opcode == UD_Ixchg && op0->type == UD_OP_REG && op1->type == UD_OP_REG) {
                int reg0 = UDRegisterToBitVector(op0->base);
                int reg1 = UDRegisterToBitVector(op1->base);
                if ((KnownRegs & usedreg) == usedreg) { // op0 and op1 are known!
                    defreg &= ~usedreg;
                    int val0, val1;
                    val0 = consts[floor_log2(reg0)];
                    val1 = consts[floor_log2(reg1)];

                    ud_t *instrForReg0 = UDCreateMovImmToReg(op0->base, val0);
                    ud_t *instrForReg2 = UDCreateMovImmToReg(op1->base, val1);

                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), instrForReg0, 1);
                    SaveInstrChange(iList, iins);
                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;

                    Instruction *new = InstrClone(iins);
                    Instruction *next = GetInstruction(iList, iins->next);
                    SetRegisterValue(new, op0->base, GetRegisterValue(next, op0->base));
                    RInstrDestroy(iList, next);
                    CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), instrForReg2, 1);

                    new = AddInstruction(iList, new, iins->order);

                } else if ((KnownRegs & reg0)) { // op0 is known
                    int val0 = consts[floor_log2(reg0)];
                    defreg &= ~reg1;

                    ud_t *instrForReg0 = UDCreateMovRegToReg(op0->base, op1->base); //mov op0, op1
                    ud_t *instrForReg2 = UDCreateMovImmToReg(op1->base, val0);

                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), instrForReg0, 1);
                    SaveInstrChange(iList, iins);
                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;

                    Instruction *new = InstrClone(iins);
                    Instruction *next = GetInstruction(iList, iins->next);
                    SetRegisterValue(new, op0->base, GetRegisterValue(next, op0->base));
                    RInstrDestroy(iList, next);
                    CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), instrForReg2, 1);

                    new = AddInstruction(iList, new, iins->order);

                } else { // op1 is known
                    int val1 = consts[floor_log2(reg1)];
                    defreg &= ~reg0;

                    ud_t *instrForReg0 = UDCreateMovRegToReg(op1->base, op0->base); //mov op1, op0
                    ud_t *instrForReg2 = UDCreateMovImmToReg(op0->base, val1);

                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), instrForReg0, 1);
                    SaveInstrChange(iList, iins);
                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;

                    Instruction *new = InstrClone(iins);
                    Instruction *next = GetInstruction(iList, iins->next);
                    SetRegisterValue(new, op1->base, GetRegisterValue(next, op1->base));
                    RInstrDestroy(iList, next);
                    CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), instrForReg2, 1);

                    new = AddInstruction(iList, new, iins->order);
                }
            }
            /*if(opcode == UD_Ileave && (KnownRegs & LYNX_EBP)){
             udins_new = UDCreateMovImmToReg(UD_R_ESP, consts[floor_log2(LYNX_EBP)]);

             Instruction *new = InstrClone(iins);
             new->esp = consts[floor_log2(LYNX_EBP)];

             ud_t *udins = UDCreateAddImmToReg(UD_R_ESP, 4);
             new->order = iins->order;
             new->addr = iins->addr;
             new->flags = 0;
             CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
             new->flagsDefKnown = true;
             new->flagsUsedKnown = true;
             new->flagsUsed = 0;
             new->flagsDef = 0;
             free(udins);

             AddInstruction(iList, new, iins->order);

             Instruction *pop = InstrClone(new);

             udins = UDCreatePopReg(UD_R_EBP);
             pop->esp = new->esp - 4;
             pop->order = iins->order;
             pop->addr = iins->addr;
             pop->flags = 0;
             CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 1);
             pop->flagsDefKnown = true;
             pop->flagsUsedKnown = true;
             pop->flagsUsed = 0;
             pop->flagsDef = 0;
             free(udins);
             AddInstruction(iList, pop, new->order);

             }*/
            if (op0->type == UD_OP_REG && op1->type == UD_OP_REG
                    && ((UDRegisterToBitVector(UDOpIsRegister(op1)) & KnownRegs)
                    		|| (opcode ==UD_Icmp && (UDRegisterToBitVector(UDOpIsRegister(op0)) & KnownRegs)))) {
                if (opcode == UD_Imov) {
                    udins_new = UDCreateMovImmToReg(reg2, val);
                } else if (opcode == UD_Iadd) {
                    udins_new = UDCreateAddImmToReg(reg2, val);
                } else if (opcode == UD_Isub) {
                    udins_new = UDCreateSubImmFromReg(reg2, val);
                } else if (opcode == UD_Ixor) {
                    udins_new = UDCreateXorImmWithReg(reg2, val);
                } else if (opcode == UD_Iand) {
                    udins_new = UDCreateAndImmWithReg(reg2, val);
                    regMasks[floor_log2(reg2)] = ArithOp(opcode, regMasks[floor_log2(reg2)], val, 4, &flagsKnown, iins);
                } else if (opcode == UD_Ior) {
                    udins_new = UDCreateOrImmWithReg(reg2, val);
                    regMasks[floor_log2(reg2)] = ArithOp(opcode, regMasks[floor_log2(reg2)], val, 4, &flagsKnown, iins);
                } else if (opcode == UD_Icmp) {
                    INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                    if ((UDRegisterToBitVector(op0->base) & KnownRegs) && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) {
                        udins_new = UDCreateCmpImmWithReg(op1->base, val);
                        INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                    }
                    if ((UDRegisterToBitVector(op1->base) & KnownRegs) && op1->type == UD_OP_REG) {
                        if (!INSTR_HAS_FLAG(iins, INSTR_OP0_CONSTANT)) {
                            udins_new = UDCreateCmpImmWithReg(op0->base,
                                    consts[floor_log2(UDRegisterToBitVector(op1->base))]);
                        }
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                    }
                    if(udins_new){
                        if (udins_new->operand[0].type == UD_OP_REG && !(UDRegisterToBitVector(udins_new->operand[0].base) & KnownRegs)){
                            INSTR_REMOVE_FLAG(iins, INSTR_OP0_CONSTANT);
                        }
                        if (udins_new->operand[1].type == UD_OP_REG && !(UDRegisterToBitVector(udins_new->operand[1].base) & KnownRegs)){
                            INSTR_REMOVE_FLAG(iins, INSTR_OP1_CONSTANT);
                        }
                            
                    }
                    SaveInstrChange(iList, iins);
                    //udins_new = UDCreateCmpImmWithReg(reg2, val);
                } else if (opcode == UD_Itest) {
                    INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                    if ((UDRegisterToBitVector(op0->base) & KnownRegs) && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED))
                        INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                    SaveInstrChange(iList, iins);
                    //udins_new = UDCreateTestImmWithReg(reg2, val);
                } else if (opcode == UD_Imovzx) {
                    udins_new = UDCreateMovImmToReg(reg2, val);
                } else if (opcode == UD_Imovsx) {
                    int sign = (val >> (UDRegSize(reg2) * 8 - 1));
                    if (sign)
                        val |= (0xffffffff << (UDRegSize(reg2) * 8));
                    udins_new = UDCreateMovImmToReg(reg2, val);
                } else if (opcode == UD_Ishl && op1->base != UD_R_CL) {
                    udins_new = UDCreateShlImmToReg(reg2, val);
                    regMasks[floor_log2(reg2)] = ArithOp(opcode, regMasks[floor_log2(reg2)], val, 4, &flagsKnown, iins);
                } else if (opcode == UD_Ishr && op1->base != UD_R_CL) {
                    udins_new = UDCreateShrImmToReg(reg2, val);
                    regMasks[floor_log2(reg2)] = ArithOp(opcode, regMasks[floor_log2(reg2)], val, 4, &flagsKnown, iins);
                } else if (opcode == UD_Iror) {
                    udins_new = UDCreateRorImmToReg(reg2, val);
                } else if (opcode == UD_Irol) {
                    udins_new = UDCreateRolImmToReg(reg2, val);
                } else if (opcode == UD_Ixadd) {
                    if (INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        SaveInstrChange(iList, iins);
                    } else { //needs action!
                    }
                } else if (opcode == UD_Iimul) {
                    if (iins->uInstr.operand[2].type == UD_NONE) {
                        udins_new = UDCreateImulImmToReg(reg2, val);
                        INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                        SaveInstrChange(iList, iins);
                    } else if ((iins->uInstr.operand[2].type == UD_OP_IMM
                            || iins->uInstr.operand[2].type == UD_OP_CONST) && (reg2 & KnownRegs)
                            && !INSTR_HAS_FLAG(iins, INSTR_COND_CF)
                            && !INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED)) {
                        udins_new = UDCreateMovImmToReg(reg1,
                                ArithOp(opcode, consts[floor_log2(UDRegisterToBitVector(reg2))],
                                        iins->uInstr.operand[2].lval.sdword,
                                        iins->uInstr.operand[2].size, &flagsKnown, iins));
                        INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                        SaveInstrChange(iList, iins);
                    }
                } else if (opcode == UD_Iidiv) {
                    if (iins->uInstr.operand[2].type == UD_NONE) {
                        udins_new = UDCreateImulImmToReg(reg2, val);
                        INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                        SaveInstrChange(iList, iins);
                    } else if ((iins->uInstr.operand[2].type == UD_OP_IMM
                            || iins->uInstr.operand[2].type == UD_OP_CONST) && (reg2 & KnownRegs)
                            && !INSTR_HAS_FLAG(iins, INSTR_COND_CF)
                            && !INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED)) {
                        udins_new = UDCreateMovImmToReg(reg1,
                                ArithOp(opcode, consts[floor_log2(UDRegisterToBitVector(reg2))],
                                        iins->uInstr.operand[2].lval.sdword,
                                        iins->uInstr.operand[2].size, &flagsKnown, iins));
                        INSTR_SET_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
                        SaveInstrChange(iList, iins);
                    }
                }

                /* introduce code for additional operations here */
                else {
                    //fprintf(stderr, "[ConstatnProp] Warning: Unhandeled reg instr: %d: %s\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
                }
            } else if (op0->type == UD_OP_REG && op1->type == UD_NONE
                    && (UDRegisterToBitVector(UDOpIsRegister(op0)) & KnownRegs)) {
                val = consts[floor_log2(UDRegisterToBitVector(UDOpIsRegister(op0)))];
                if (opcode == UD_Ipush) {
                    udins_new = UDCreatePushImm(reg2, val);
                } else if (opcode == UD_Icall) {
                    udins_new = UDCreateCallToAddr(val - 5);
                } else if (opcode == UD_Ijmp) {
                    udins_new = UDCreateJump(val);
                } else {
                    //fprintf(stderr, "[ConstatnProp] Warning: Unhandeled 1 op instr: %d: %s\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
                }

            } else if (op0->type == UD_OP_REG
                    && (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST) && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) {
                if (opcode == UD_Itest || opcode == UD_Icmp) {
                    INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                    SaveInstrChange(iList, iins);
                }
            } else if (op1->type == UD_OP_REG
                    && (op0->type == UD_OP_IMM || op0->type == UD_OP_CONST)) {
                if (opcode == UD_Itest || opcode == UD_Icmp) {
                    INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                    SaveInstrChange(iList, iins);
                }
            } else if (op0->type == UD_OP_MEM && op1->type == UD_OP_REG
                    && (UDRegisterToBitVector(UDOpIsRegister(op1)) & KnownRegs)) {
            	if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                reg2 = UDOpIsRegister(op1);
                if (opcode == UD_Isub) {
                    udins_new = UDCreateSubImmFromMem(iins->uInstr.bytes, reg2, val);
                } else if (opcode == UD_Imov && iins->uInstr.bytes[0] != 0x64) { // prefix 0x64 is for fs register, not available in trace!
                    udins_new = UDCreateMovImmToMem(iins->uInstr.bytes, reg2, val);
                } else if (opcode == UD_Iadd) {
                    udins_new = UDCreateAddImmToMem(iins->uInstr.bytes, reg2, val);
                } else if (opcode == UD_Icmp) {

                		udins_new = UDCreateCmpMemWithImm(iins->uInstr.bytes, reg2, val);

                } else if (opcode == UD_Ixor) {
                    udins_new = UDCreateXorMemWithImm(iins->uInstr.bytes, reg2, val);
                } else if (opcode == UD_Iand) {
                    udins_new = UDCreateAndMemWithImm(iins->uInstr.bytes, reg2, val);
                } else if (opcode == UD_Ior) {
                    udins_new = UDCreateOrMemWithImm(iins->uInstr.bytes, reg2, val);
                } else if (opcode == UD_Ixchg) {
                    if (!INSTR_HAS_FLAG(iins, INSTR_OP1_CONSTANT)) {
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        SaveInstrChange(iList, iins);
                    }
                } else if (opcode == UD_Ixadd) {
                    if (INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)
                            && !INSTR_HAS_FLAG(iins, INSTR_OP1_CONSTANT)) {
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        SaveInstrChange(iList, iins);
                    } else if (!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)
                            && INSTR_HAS_FLAG(iins, INSTR_READS_GLOBAL_MEM)) {
                        Instruction *next = FetchNextInstr(iList);
                        int newVal = GetRegisterValue(next, reg2);
                        ud_t *udins = UDCreateMovImmToReg(op1->base, newVal);

                        char *bytes = malloc(iins->uInstr.length - 1);
                        int i;
                        bytes[0] = iins->uInstr.bytes[0];
                        for (i = 1; i < iins->uInstr.length - 1; i++) {
                            bytes[i] = iins->uInstr.bytes[i + 1];
                        }

                        udins_new = UDCreateMovImmToMem((unsigned char *) bytes, reg2,
                                newVal + val);
                        CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);

                        free(udins_new);
                        udins_new = NULL;
                        INSTR_SET_FLAG(iins, INSTR_TMP0);
                        SaveInstrChange(iList, iins);
                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE))
                        N++;

                        Instruction *new = InstrClone(next);
                        new->order = iins->order;
                        new->addr = iins->addr;
                        new->flags = 0;
                        CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins, 0);
                        free(udins);

                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                            INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
                        }

                        new = AddInstruction(iList, new, iins->order);

                        RInstrDestroy(iList, next);
                        SetCurrentInstr(iList, iins->next);
                    }
                } else if (opcode == UD_Icmpxchg) {
                    if (INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)
                            && !INSTR_HAS_FLAG(iins, INSTR_OP1_CONSTANT)) {
                        INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                        SaveInstrChange(iList, iins);
                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                    } else if (INSTR_HAS_FLAG(iins, INSTR_READS_GLOBAL_MEM)) {
                        Instruction *next = FetchNextInstr(iList);
                        int newVal = GetRegisterValue(next, reg2);
                        if (val == newVal) { // if reg is not changed, memory is loaded with reg
                            char *bytes = malloc(iins->uInstr.length - 1);
                            int i;
                            bytes[0] = iins->uInstr.bytes[0];
                            for (i = 1; i < iins->uInstr.length - 1; i++) {
                                bytes[i] = iins->uInstr.bytes[i + 1];
                            }
                            udins_new = UDCreateMovImmToMem((unsigned char *) bytes, reg2, val);
                        } else { //otherwise the reg is loaded with memory
                            udins_new = UDCreateMovImmToReg(reg2, newVal);
                        }
                    }
                }
                /* introduce code for additional operations here */
                else {
                    //fprintf(stderr, "[ConstatnProp] Warning: Unhandeled mem instr: %d: %s\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
                }
            	}
                /*
                 * Transforming instrs that use registers as memory pointers, replacing it with
                 * the same instr directly accessing the memory address.
                 */
            } else if (SIMPLIFY_INDIRECT_MEM_ACCESS && op0->type == UD_OP_REG && op1->type == UD_OP_MEM) {
                if (MemRegUsed(iins) != 0
                        && (KnownRegs & MemRegUsed(iins) & ~LYNX_ESP)
                                == (MemRegUsed(iins) & ~LYNX_ESP)) {
                    if (opcode == UD_Imov && INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
                        //WriteSet *memRead = MemUsed(iList, iins);

                            //Range *read = al_get(memRead->ranges, 0);
                            int offset = iins->memUsedMin;//read->min;
                            udins_new = UDCreateMovOffsetToReg(op0->base, offset, 0);

                    } else if (opcode == UD_Icmp) {
                        int type = iins->uInstr.bytes[0];
                        if(type == 0x66){
                            type = iins->uInstr.bytes[1];
                        }
                        udins_new = UDCreateCmpOffsetWithReg(iins->memUsedMin, op0->base, type);
                    } /*else if(opcode == UD_Iadd){
                        udins_new = UDCreateAddRegToOffset(iins->memUsedMin, op0->base, iins->memUsedSize);
                    }*/
                }
                if ((UDRegisterToBitVector(UDOpIsRegister(op0)) & KnownRegs)) {
                    if (opcode == UD_Icmp) {
                        if (op0->type == UD_OP_REG && !INSTR_HAS_FLAG(iins, INSTR_OP0_CONSTANT)) {
                            INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
                            SaveInstrChange(iList, iins);
                        }
                        if (op1->type == UD_OP_REG && !INSTR_HAS_FLAG(iins, INSTR_OP1_CONSTANT)) {
                            INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
                            SaveInstrChange(iList, iins);
                        }
                    }
                }
            } else if (op1->type == UD_OP_REG && op0->type == UD_OP_MEM) {
                if (MemRegUsed(iins) != 0
                        && (KnownRegs & MemRegUsed(iins) & ~LYNX_ESP)
                                == (MemRegUsed(iins) & ~LYNX_ESP)) {
                    if (opcode == UD_Icmp) {
                        int type = iins->uInstr.bytes[0];
                        if(type == 0x66){
                            type = iins->uInstr.bytes[1];
                        }
                        udins_new = UDCreateCmpOffsetWithReg(iins->memUsedMin, op0->base, type);
                    }
                }
            } else if (SIMPLIFY_INDIRECT_MEM_ACCESS && op1->type == UD_OP_IMM && op0->type == UD_OP_MEM) {
                if (MemRegUsed(iins) != 0
                        && (KnownRegs & MemRegUsed(iins) & ~LYNX_ESP)
                                == (MemRegUsed(iins) & ~LYNX_ESP)) {
                    if (opcode == UD_Icmp) {
                        udins_new = UDCreateCmpOffsetWithImm(iins->memUsedMin, op1->lval.sdword,
                                iins->memUsedSize);
                    } else if (opcode == UD_Imov) {
                        udins_new = UDCreateMovImmToOffset(iins->memDefMin, op1->lval.sdword,
                                iins->memDefSize);
                    }
                }
            } else if (op0->type == UD_OP_IMM && op1->type == UD_OP_MEM) {
                if (MemRegUsed(iins) != 0
                        && (KnownRegs & MemRegUsed(iins) & ~LYNX_ESP)
                                == (MemRegUsed(iins) & ~LYNX_ESP)) {
                    if (opcode == UD_Icmp) {
                        udins_new = UDCreateCmpOffsetWithImm(iins->memUsedMin, op0->lval.sdword,
                                iins->memUsedSize);
                    }
                }
            } else if (opcode == UD_Ijmp && (usedreg & KnownRegs) == usedreg) {
                Instruction *next = FetchNextInstr(iList);
                udins_new = UDCreateJump(next->addr - 0x5);
                RInstrDestroy(iList, next);

                INSTR_SET_FLAG(iins, INSTR_IS_DELETED); // a test, to simplify the control flow graph.
                SaveInstrChange(iList, iins);

                SetCurrentInstr(iList, iins->next);
            }

            else if (opcode == UD_Ilea && (usedreg & KnownRegs) == usedreg) {
                Instruction *next = FetchNextInstr(iList);
                udins_new = UDCreateMovImmToReg(op0->base, GetRegisterValue(next, op0->base));
                RInstrDestroy(iList, next);
                SetCurrentInstr(iList, iins->next);
            }
            if (udins_new != NULL) { // we recognized opcode2
                if (INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE)
                        && INSTR_HAS_FLAG(iins, INSTR_SIMPLIFIED_IGNORED)) {
                    free(udins_new);
                    udins_new = NULL;
                } else {
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                    free(udins_new);
                    udins_new = NULL;
                    INSTR_SET_FLAG(iins, INSTR_TMP0);
                    SaveInstrChange(iList, iins);
                    if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
                        N++;
                    }
                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);
                    op2 = &(iins->uInstr.operand[2]);

                    defreg = RegDef(iList, iins);
                    usedreg = RegUsed(iList, iins);
                }
                INSTR_REMOVE_FLAG(iins, INSTR_SIMPLIFIED_IGNORED);
            }
        }
        if (SIMPLIFY_INDIRECT_MEM_ACCESS && ((op0->type == UD_OP_MEM && op1->type == UD_OP_REG)) && (MemRegUsed(iins) != 0)) {
            if (opcode == UD_Imov) {
                if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {

                    udins_new = UDCreateMovOffsetToReg(op1->base, iins->memDefMin, 1);

                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                    free(udins_new);
                    udins_new = NULL;
                    INSTR_SET_FLAG(iins, INSTR_TMP0);
                    SaveInstrChange(iList, iins);
                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                    usedreg = RegUsed(iList, iins);
                }
            }
        } else if (SIMPLIFY_INDIRECT_MEM_ACCESS && (op1->type == UD_OP_MEM && op0->type == UD_OP_REG) && (MemRegUsed(iins) != 0)) {
            if (opcode == UD_Imov) {
                if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
                    udins_new = UDCreateMovOffsetToReg(op0->base, iins->memUsedMin, 0);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                    free(udins_new);
                    udins_new = NULL;
                    INSTR_SET_FLAG(iins, INSTR_TMP0);
                    SaveInstrChange(iList, iins);
                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                    usedreg = RegUsed(iList, iins);
                }

            }
        } else if (SIMPLIFY_INDIRECT_MEM_ACCESS && op0->type == UD_OP_MEM && op1->type == UD_NONE
                && (MemRegUsed(iins) & ~LYNX_ESP) != 0 ) { // push [ebx], pop [ebx]
            if (opcode == UD_Ipush) {
                udins_new = UDCreatePushOffset(iins->uInstr.bytes, iins->memUsedMin);
                CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                free(udins_new);
                udins_new = NULL;
                INSTR_SET_FLAG(iins, INSTR_TMP0);
                SaveInstrChange(iList, iins);
                if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE))
                    N++;
                usedreg = RegUsed(iList, iins);
            } else if (SIMPLIFY_INDIRECT_MEM_ACCESS && opcode == UD_Ipop) {
                udins_new = UDCreatePopOffset(iins->uInstr.bytes, iins->memDefMin);
                CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                free(udins_new);
                udins_new = NULL;
                INSTR_SET_FLAG(iins, INSTR_TMP0);
                SaveInstrChange(iList, iins);
                if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE))
                    N++;
                usedreg = RegUsed(iList, iins);
            }
        }

        if (opcode == UD_Ixor && (reg1 = UDOpIsRegister(op0)) != UD_NONE
                && reg1 == UDOpIsRegister(op1)) {
            ud_t *udins = UDCreateMovImmToReg(reg1, 0);
            CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 0);
            SaveInstrChange(iList, iins);
            free(udins);
            if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE))
                N++;
            opcode = iins->uInstr.mnemonic;
            op0 = &(iins->uInstr.operand[0]);
            op1 = &(iins->uInstr.operand[1]);

        }
        //int espReg = 0;
        if(InstrMasksRegister(iins) && op0->type == UD_OP_REG){
            REGISTER_BITVECTOR regist = UDRegisterToBitVector(op0->base);
            if(op1->type == UD_OP_IMM){
                if(opcode == UD_Ior){
                    ;
                } else {
                    regMasks[floor_log2(regist)] = ArithOp(opcode, regMasks[floor_log2(regist)], op1->lval.sdword, 4, &flagsKnown, iins);
                }
            } else if(op1->type == UD_OP_REG){
                //REGISTER_BITVECTOR regist2 = op1->base;
            }
        }

        if (INSTR_HAS_FLAG(iins, INSTR_REGDEF_KNOWN) || (opcode == UD_Icpuid) || (opcode == UD_Imov && (reg1 = UDOpIsRegister(op0)) != UD_NONE
                && (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST /*|| (espReg = UDOpIsRegister(op1)) == UD_R_ESP*//*|| (OpcodeIsArith(opcode) && (KnownRegs & usedreg) == usedreg)*/))) { // second part is for instrs not being simplified but operands are known!
            /*if(OpcodeIsArith(opcode) && (KnownRegs & usedreg) == usedreg){

             } else*/{
                if (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST) {
                    val = op1->lval.sdword;
                } else {
                    val = GetRegisterValue(iins, UD_R_ESP);
                }
            }

            thisreg = UDRegisterToBitVector(reg1);
            if (floor_log2(thisreg) < 18)
                KnownRegs |= SetConstsArray(consts, thisreg, val);
            else {
                KnownRegs |= thisreg;
                consts[floor_log2(thisreg)] = val;
                if (thisreg == LYNX_AL) {
                    consts[floor_log2(LYNX_AX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_AX)] |= (val & 0xff);
                    consts[floor_log2(LYNX_EAX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_EAX)] |= (val & 0xff);
                } else if (thisreg == LYNX_DL) {
                    consts[floor_log2(LYNX_DX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_DX)] |= (val & 0xff);
                    consts[floor_log2(LYNX_EDX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_EDX)] |= (val & 0xff);
                } else if (thisreg == LYNX_CL) {
                    consts[floor_log2(LYNX_CX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_CX)] |= (val & 0xff);
                    consts[floor_log2(LYNX_ECX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_ECX)] |= (val & 0xff);
                } else if (thisreg == LYNX_BL) {
                    consts[floor_log2(LYNX_BX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_BX)] |= (val & 0xff);
                    consts[floor_log2(LYNX_EBX)] &= 0xffffff00;
                    consts[floor_log2(LYNX_EBX)] |= (val & 0xff);
                } else if (thisreg == LYNX_AH) {
                    consts[floor_log2(LYNX_AX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_AX)] |= ((val & 0xff) << 8);
                    consts[floor_log2(LYNX_EAX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_EAX)] |= ((val & 0xff) << 8);
                    ;
                } else if (thisreg == LYNX_CH) {
                    consts[floor_log2(LYNX_CX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_CX)] |= ((val & 0xff) << 8);
                    consts[floor_log2(LYNX_ECX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_ECX)] |= ((val & 0xff) << 8);
                    ;
                } else if (thisreg == LYNX_DH) {
                    consts[floor_log2(LYNX_DX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_DX)] |= ((val & 0xff) << 8);
                    consts[floor_log2(LYNX_EDX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_EDX)] |= ((val & 0xff) << 8);
                    ;
                } else if (thisreg == LYNX_BH) {
                    consts[floor_log2(LYNX_BX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_BX)] |= ((val & 0xff) << 8);
                    consts[floor_log2(LYNX_EBX)] &= 0xffff00ff;
                    consts[floor_log2(LYNX_EBX)] |= ((val & 0xff) << 8);
                    ;
                }
            }
        } else if ((MaskDefinedReg(defreg) & KnownRegs)) {
            KnownRegs &= ~(UpMaskDefinedReg(defreg) | MaskDefinedReg(defreg));

            int mask=0;
            if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) { // NOTE: at this point we know that memory location has not been a constant!
                mask = GetMemoryLocValue(iList, iins->memUsedMin,
                        iins->memUsedSize);
            } else if (usedreg) {
                mask = regMasks[floor_log2(usedreg)];
            }

            int tmpReg;
            for (i = 0; i < 30; i++) {
                tmpReg = (0x1 << i);
                if (tmpReg & defreg) {
                    regMasks[i] = mask;
                }
            }
        }

        if (iins->order < iList->last
                && ((iins->esp > prevesp) /*|| opcode == UD_Icall || retChange*/)) {
        }
        prevesp = iins->esp;

        ///////////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////////////
        /*SIMPLIFYCONDITIONALJUMPA*/
        {
            flagsDef = FlagsDef(iins);
            REGISTER_BITVECTOR flagsUsed = FlagsUsed(iins);
            Instruction *iins1;


            if (flagsDef && !(INSTR_HAS_FLAG(iins, INSTR_READS_MEM) || usedreg || flagsUsed)) {
                // instrs like cld which just change flag status, not depending on anything
                INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
                SaveInstrChange(iList, iins);
            }
            if(flagsDef && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)){
                flagsKnown |= flagsDef;
                if(!defreg && !INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){
                	INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                	SaveInstrChange(iList, iins);
                }
            } else if (!(INSTR_HAS_FLAG(iins, INSTR_TMP3)
                    && (iins->uInstr.mnemonic == UD_Iadd || iins->uInstr.mnemonic == UD_Isub)
                    && iins->uInstr.operand[1].type == UD_OP_IMM)) {
                flagsKnown &= ~flagsDef;
            }
            if (flagsUsed && (flagsUsed & flagsKnown) == flagsUsed) {
                INSTR_SET_FLAG(iins, INSTR_FLAGS_KNOWN);
                SaveInstrChange(iList, iins);
            }
            if (INSTR_HAS_FLAG(iins, INSTR_FLAGS_KNOWN) && opcode == UD_Ipushfd) {
                SetAddrAsConstWithValue(iList, iins->memDefMin, 4, 0x202); // TODO: don't know the value that is being pushed!
            }
            if(iins->uInstr.mnemonic == UD_Iclc){
            	flagsKnown |= flagsDef;
            }
            if(iins->uInstr.mnemonic == UD_Icmc && (KnownRegs & flagsUsed) == flagsUsed){
                flagsKnown |= flagsDef;
            }
            if (opcode == UD_Ixor || opcode == UD_Iand || opcode == UD_Itest || opcode == UD_Ior) { // these instructions set CF and OF flags to 0!
                flagsKnown |= (PSW_CF | PSW_OF);
            }
            if (IsCondJump(iins)) {
                if ((flagsUsed & flagsKnown) == flagsUsed && !INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE)) {
                    op0 = &(iins->uInstr.operand[0]);
                    iins1 = FetchNextInstr(iList);
                    if (iins1
                            && op0->lval.sdword + iins->uInstr.length + iins->addr == iins1->addr) { // jump has been taken
                        udins_new = UDCreateJump(iins1->addr - 5);
                        CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                        if (iins->uInstr.mnemonic != UD_Ijmp) {
                            fprintf(stderr,
                                    "[SimplifyConditionalJumps] Error: in creating jmp instr. %d: %s\n",
                                    iins1->order, ud_insn_asm(LoadUDisInstrObj(iins1)));
                        }
                        INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                        SaveInstrChange(iList, iins);
                        free(udins_new);
                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                    } else {
                        INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                        SaveInstrChange(iList, iins);
                    }
                    RInstrDestroy(iList, iins1);
                    SetCurrentInstr(iList, iins->next);
                }
                /*
                 * check http://css.csail.mit.edu/6.858/2011/readings/i386/SETcc.htm
                 * there are lots of them!
                 */
            } else if (InstrIsSetOnCondition(iins->uInstr.mnemonic)
                    && (flagsUsed & flagsKnown) == flagsUsed) {
                if (iins->uInstr.operand[0].type == UD_OP_REG) {
                    enum ud_type reg;
                    reg = iins->uInstr.operand[0].base;
                    Instruction *next = FetchNextInstr(iList);
                    int val = GetRegisterValue(next, reg);
                    RInstrDestroy(iList, next);
                    SetCurrentInstr(iList, iins->next);

                    ud_t *udins = UDCreateMovImmToReg(reg, val);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 0);
                    SaveInstrChange(iList, iins);
                    free(udins);
                }
            } else if (iins->uInstr.mnemonic == UD_Isetg && (flagsUsed & flagsKnown) == flagsUsed) {
                if (iins->uInstr.operand[0].type == UD_OP_REG) { // could be memory!
                    enum ud_type reg;
                    reg = iins->uInstr.operand[0].base;
                    Instruction *next = FetchNextInstr(iList);
                    int val = GetRegisterValue(next, reg);
                    RInstrDestroy(iList, next);
                    SetCurrentInstr(iList, iins->next);

                    ud_t *udins = UDCreateMovImmToReg(reg, val);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 0);
                    SaveInstrChange(iList, iins);
                    free(udins);
                }

            } else {
                //fprintf(stderr, "SimplifyConditionalJumps: setnz m8 unhandled: %d\n", iins->order);
            }
        }
        /*SIMPLIFYCONDITIONALJUMPA*/
        ///////////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////////////


        /*REMOVEUNNECESSARYPUSHESANDPOPS*/
        ///////////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////////////
        push = flags = 0;
        bool pushpopChange = false;

        if (opcode == UD_Ipush || opcode == UD_Ipushfd) {
            if (opcode == UD_Ipush) {
                push = 1;
            } else { // (opcode1 == UD_Ipushfd) {
                flags = 1;
            }

            reg1 = UDOpIsRegister(op0);


            DisallowedDef = (MaskDefinedReg(UDRegisterToBitVector(reg1)) /*| MaskDefinedReg(
             UDRegisterToBitVector(UD_R_ESP))*/);

            memDef = MemDef(iList, iins);
            j = i + 1;
            looking = true;
            innerloop = 0;
            while (j < last && looking) {
                Instruction *iins2 = FetchNextInstr(iList);
                innerloop++;
                if (iins2->order <= iList->last){
                    j = iins2->order;
                }
                enum ud_mnemonic_code opcode2 = iins2->uInstr.mnemonic;
                op1 = &(iins2->uInstr.operand[0]);
                memUsed = MemUsed(iList, iins2);
                flagsDef = FlagsDef(iins2);
                WriteSet *ws = MemDef(iList, iins2);
                if (push && opcode2 == UD_Inot && looking
                        && WriteSetSubsetOfWriteSet(memUsed, memDef)) {
                    Instruction *tmp = FetchNextInstr(iList);
                    reg2 = UDOpIsRegister(&tmp->uInstr.operand[0]);
                    ud_t *new_ins = NULL;
                    if (tmp->uInstr.mnemonic == UD_Ipop && (reg1 == reg2) && reg1 != UD_NONE) {
                        INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                        INSTR_SET_FLAG(tmp, INSTR_IS_DELETED);
                        INSTR_SET_FLAG(iins, INSTR_TMP1);
                        INSTR_SET_FLAG(tmp, INSTR_TMP1);
                        SaveInstrChange(iList, iins);
                        SaveInstrChange(iList, tmp);

                        if (WriteSetEquals(memDef, memUsed)) {
                            new_ins = UDCreateNotOnReg(reg1);
                        } else {
                            if (iins2->memUsedSize == 2) {
                                int newReg = UDSubRegister(reg1, 2);
                                new_ins = UDCreateNotOnReg(newReg);
                            } else if (iins2->memUsedSize == 1) {
                                int newReg = UDSubRegister(reg1, 1);
                                new_ins = UDCreateNotOnReg(newReg);
                            }
                        }

                        CarefullCopyUdins2InsStr(iList, iins2, &(iins2->uInstr), new_ins, 0);
                        SaveInstrChange(iList, iins2);
                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                    }
                    SetCurrentInstr(iList, iins2->next);
                }
                if (push && opcode2 == UD_Ipop && looking && WriteSetEquals(memDef, memUsed)) {
                    reg2 = UDOpIsRegister(op1);
                    if (!(reg1 == UD_NONE || reg2 == UD_NONE)) {

                        if (reg2 != reg1) {
                            looking = false;

                            ud_t *udins = UDCreateMovRegToReg(reg2, reg1);
                            CarefullCopyUdins2InsStr(iList, iins2, &(iins2->uInstr), udins, 1);
                            INSTR_SET_FLAG(iins2, INSTR_TMP1);
                            SaveInstrChange(iList, iins2);
                            free(udins);

                            SetCurrentInstr(iList, iins->next);
                            INSTR_SET_FLAG(iins, INSTR_TMP1);
                            INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                            SaveInstrChange(iList, iins);

                            WriteSetDestroy(ws);
                            WriteSetDestroy(memUsed);
                            break;
                        } else if (iins->esp == iins2->esp + UDRegSize(reg1)) {
                            ud_t *udins = UDCreateSubImmFromReg(UD_R_ESP, UDRegSize(reg1));
                            CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 1);
                            INSTR_SET_FLAG(iins, INSTR_TMP1);
                            SaveInstrChange(iList, iins);
                            free(udins);
                            udins = UDCreateAddImmToReg(UD_R_ESP, UDRegSize(reg1));
                            CarefullCopyUdins2InsStr(iList, iins2, &(iins2->uInstr), udins, 1);
                            INSTR_SET_FLAG(iins2, INSTR_TMP1);
                            SaveInstrChange(iList, iins2);
                            free(udins);
                            if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                            //fprintf(stderr, "[RemoveUnnecessaryPushesAndPops] ESP accessed between two: %d, %d\n", iins1->order, iins2->order);
                        }
                    } else if (iins->uInstr.operand[0].type == UD_OP_MEM
                            && iins2->uInstr.operand[0].type == UD_OP_MEM) {
                        Instruction *new;
                        ud_t *udins_new;
                        enum ud_type reg;
                        int size = iins->memUsedSize;
                        if (size == 4) {
                            reg = UD_R_EAX;
                        } else if (size == 2) {
                            reg = UD_R_AX;
                        } else {
                            reg = UD_R_AL;
                        }

                        new = InstrClone(iins);
                        udins_new = UDCreatePushReg(UD_R_EAX); // save eax first!
                        CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins_new, 1);
                        new->addr = iins->addr - 1;
                        new = AddInstruction(iList, new, iins->prev); // insert it before the arithmetic instr!
                        free(udins_new);
                        iins->esp -= 4; // allocate stack!

                        // copy mem1 to eax first!
                        udins_new = UDCreateMovOffsetToReg(reg, iins->memUsedMin, 0);
                        CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
                        SaveInstrChange(iList, iins);
                        free(udins_new);

                        // copy eax to mem2!
                        new = InstrClone(iins);
                        udins_new = UDCreateMovOffsetToReg(reg, iins2->memDefMin, 1);
                        CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins_new, 1);
                        new->addr = iins->addr + 1;
                        //new->flags = 0;
                        new = AddInstruction(iList, new, iins->order);
                        free(udins_new);

                        //Instruction *tmp = GetInstruction(iList, iins->next);
                        // pop saved eax!

                        new = InstrClone(iins);
                        udins_new = UDCreatePopReg(UD_R_EAX);
                        CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins_new, 1);
                        new->addr = iins->addr + 2;
                        //new->flags = 0;
                        new = AddInstruction(iList, new, iins->next); // insert it before the arithmetic instr!
                        free(udins_new);
                        SetCurrentInstr(iList, iins->prev);

                        INSTR_SET_FLAG(iins2, INSTR_IS_DELETED);
                        SaveInstrChange(iList, iins2);
                        pushpopChange = true;

                    }
                } else if (flags && opcode2 == UD_Ipopfd && looking && iins->memDefMin == iins2->memUsedMin) {
                    INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                    INSTR_SET_FLAG(iins2, INSTR_IS_DELETED);
                    INSTR_SET_FLAG(iins, INSTR_TMP1);
                    INSTR_SET_FLAG(iins2, INSTR_TMP1);
                    SaveInstrChange(iList, iins);
                    SaveInstrChange(iList, iins2);
                    N += 2;
                }
                if ((!WriteSetIsEmpty(ws) && WriteSetsOverlap(memDef, ws))
                        || iins2->esp >= iins->esp) {
                    looking = false;
                    RInstrDestroy(iList, iins2);
                    WriteSetDestroy(ws);
                    WriteSetDestroy(memUsed);
                    break;
                }
                usedreg = MaskDefinedReg(RegUsedInclSP(iList, iins2));
                defreg = MaskDefinedReg(RegDefInclSP(iList, iins2));
                if (!flags && (defreg & DisallowedDef)
                        && !((opcode2 == UD_Iadd || opcode2 == UD_Isub) && op1->base == UD_R_ESP)) {
                    looking = false;
                    RInstrDestroy(iList, iins2);
                    WriteSetDestroy(ws);
                    WriteSetDestroy(memUsed);
                    break;
                }
                if (flags && (flagsDef)) {
                    looking = false;
                    RInstrDestroy(iList, iins2);
                    WriteSetDestroy(ws);
                    WriteSetDestroy(memUsed);
                    break;
                }
                /*
                 * For now, we [conservatively] consider _any_ write to memory as
                 * possibly writing to the stack.	This can and should be made
                 * more discriminating.
                 */
                WriteSetDestroy(memUsed);
                WriteSetDestroy(ws);
            }
            WriteSetDestroy(memDef);
        }

        if (pushpopChange == false) {
            SetCurrentInstr(iList, iins->next);
        }
        /*REMOVEUNNECESSARYPUSHESANDPOPS*/
        ///////////////////////////////////////////////////////////////////////////////////////////////
        ///////////////////////////////////////////////////////////////////////////////////////////////
    }
    fprintf(stderr, "\t[PropagateConstants]: propagated to %d instructions\n", N);
    return N;
}

int DefinitionIsGuaranteed(InstrList *iList, Instruction *iins) {
    Instruction *iins1;
    int i, window = 0;
    enum ud_mnemonic_code opcode;
    WriteSet *memUsed, *memDef, *memtmp, *memTarget;

    opcode = iins->uInstr.mnemonic;


    if (opcode == UD_Ileave
            || opcode == UD_Ipopad || opcode == UD_Ipopa || IsStringOperation(iins)) {
        return 0;
    }

    if (opcode == UD_Ipopfd) { // handle popfd differently, if it reaches all the way up to a pushfd!
        memTarget = MemUsed(iList, iins);
        for (i = iins->order - 1;
                i >= 0 && (window++ < 50) && (iins1 = FetchPrevInstr(iList)) != NULL; i--) {
            memDef = MemDef(iList, iins1);
            if (INSTR_HAS_FLAG(iins1, INSTR_WRITES_MEM) && WriteSetsOverlap(memTarget, memDef)) {
                if (INSTR_HAS_FLAG(iins1, INSTR_READS_MEM)) {
                    memUsed = MemUsed(iList, iins1);

                    memtmp = SubtractWriteSets(memTarget, memDef);
                    WriteSetDestroy(memTarget);
                    memTarget = memtmp;

                    memTarget = MergeWriteSets(memTarget, memUsed);

                    WriteSetDestroy(memUsed);
                } else {
                    if (iins1->uInstr.mnemonic == UD_Ipushfd) {
                        INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                        INSTR_SET_FLAG(iins1, INSTR_IS_DELETED);
                        SaveInstrChange(iList, iins);
                        SaveInstrChange(iList, iins1);
                        return 1;
                    }
                }
            } else {
                REGISTER_BITVECTOR flagsDef = FlagsDef(iins1);
                if (flagsDef) {
                	WriteSetDestroy(memDef);
                    return 0;
                }
            }
        }
    }
    return 0;
}

ud_t *ChangeMemOperandToConstant(Instruction *iins, int val, int operand) {
    ud_t *udins_old;
    udins_old = malloc(sizeof(ud_t));
    InitUdis(udins_old);
    ud_set_input_buffer(udins_old, iins->uInstr.bytes, MAX_INSTRLEN);
    ud_decode(udins_old);
    udins_old->operand[operand].type = UD_OP_IMM;
    udins_old->operand[operand].lval.sdword = val;
    udins_old->translator(udins_old);
    return udins_old;
}

/*
 * RemoveDead() -- perform intra-block liveness analysis on
 * registers and flags, and use this to identify and remove dead code.
 */

int GetStackSize(Instruction *instr) {
    if (instr->uInstr.bytes[0] == 0x66) {
        return 2;
    } else if (instr->uInstr.mnemonic == UD_Ipush && instr->uInstr.operand[0].type == UD_OP_IMM) {
        return instr->memDefSize;
    } else {
        return OperandSize(instr->uInstr.operand[0]);
    }

}


int RemoveDeadCode(InstrList* iList, int first, int last, uint64_t *numInstructions) {
    Instruction *iins, *next, *prev = NULL;
    enum ud_mnemonic_code opcode;
    ud_operand_t *op0, *op1;
    PSW_BITVECTOR flagsDef, flagsUsed, FlagsLive;
    REGISTER_BITVECTOR regsDef, regsUsed, RegsLive;
    int removed = 0;
    int N = 0;

    /*if (last != iList->last) { // everything is live!
        RegsLive = 0xffffffff;
        FlagsLive = 0xffffffff;
        SetMemoryMap(iList, MEMORY_LIVE);

    } else*/ {
        RegsLive = 0;
        FlagsLive = 0;
        SetMemoryMap(iList, MEMORY_DEAD);

    }
    iins = GetInstruction(iList, first);


    SetCurrentInstr(iList, last);
    while ((iins = FetchPrevInstr(iList)) != NULL) {

    	if(!INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE) && INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
				FlagsLive = 0;
				RegsLive = 0;
			}

        if (prev && prev->addr == iins->addr && InstrCompare(prev, iins)) {
            INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
            SaveInstrChange(iList, iins);
            continue;
        }
        removed = 0;
        flagsDef = FlagsDef(iins);
        flagsUsed = FlagsUsed(iins);
        regsDef = RegDef(iList, iins);
        regsUsed = RegUsed(iList, iins);

        if(iins->uInstr.mnemonic == UD_Ipushfd){
        	flagsUsed = 0x8d5;
        }

        if (!((flagsDef | flagsUsed | regsUsed | regsDef) || iins->memUsedMin || iins->memDefMin)) {
            if (iins->uInstr.mnemonic == UD_Inop || (iins->uInstr.mnemonic == UD_Ijmp && !INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL))) {
                INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                if(SaveInstrChange(iList, iins) == 0){
                	if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                }
            } else if(handle_rop || INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                (*numInstructions)++;
            }
            continue;
        }
        opcode = iins->uInstr.mnemonic;
        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);

        if ((IsPush(iins) || IsPop(iins) || opcode == UD_Icall || opcode == UD_Iret)) {
            if (UDOpIsRegister(op0) == UD_R_ESP) {
                if (opcode == UD_Ipop)
                    regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
                else if (opcode == UD_Ipush)
                    regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
            } else {
                regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
                regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
            }
        }


        if ((!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF))
                && !INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE)) { // just keep the first instruction of a function, need it when constructing cfg!
            if (!((INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)
                    && MemoryLocationIsLive(iList, iins->memDefMin,
                            iins->memDefSize)) || (flagsDef & FlagsLive) != 0
                    || (MaskDefinedReg(regsDef) & RegsLive) != 0)
                    || (OpcodeIsArithWithNoEffect(opcode) && op1 != NULL
                            && (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST)
                            && op1->lval.sdword == 0
                            && !INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED))
                    || opcode == UD_Inop || opcode == UD_Ijmp) {


                if (opcode == UD_Icall && (next = FetchNextInstr(iList)) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL)) {
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    ud_t *udins_new = UDCreateSubImmFromReg(UD_R_ESP, 4);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
                    INSTR_REMOVE_FLAG(iins, INSTR_FORWARD_TAINTED);
                    SaveInstrChange(iList, iins);
                    free(udins_new);
                    udins_new = NULL;
                    if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
                } else if ((opcode == UD_Ipush || opcode == UD_Ipushfd)) {
                    if (opcode == UD_Ipushfd) {
                        iins->flagsUsed = 0;
                        iins->flagsDef = 0;
                    }
                    ud_t* udins = NULL;
                    int stackSize = 0;
                    if (opcode == UD_Icall || opcode == UD_Ipushfd)
                        stackSize = 4;
                    else
                        stackSize = GetStackSize(iins);

                    udins = UDCreateSubImmFromReg(UD_R_ESP, stackSize);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 1);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    INSTR_REMOVE_FLAG(iins, INSTR_FORWARD_TAINTED);
                    SaveInstrChange(iList, iins);
                    free(udins);

                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);
                } else if ((opcode == UD_Ipop || opcode == UD_Ipopfd || opcode == UD_Iret) && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) {
                    if (opcode == UD_Ipopfd) {
                        iins->flagsUsed = 0;
                        iins->flagsDef = 0;
                    }
                    ud_t* udins = NULL;
                    int stackSize = 0;
                    if (opcode == UD_Ipopfd || opcode == UD_Iret)
                        stackSize = 4;
                    else
                        stackSize = GetStackSize(iins);
                    udins = UDCreateAddImmToReg(UD_R_ESP, stackSize);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 1);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    INSTR_REMOVE_FLAG(iins, INSTR_FORWARD_TAINTED);
                    SaveInstrChange(iList, iins);
                    free(udins);

                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);
                } else if (!IsControlFlowInstr(iins) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL) && !((opcode == UD_Iadd || opcode == UD_Isub) && op0->type == UD_OP_REG && op0->base == UD_R_ESP)){

                    INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    SaveInstrChange(iList, iins);
                    if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                    	if(SaveInstrChange(iList, iins) == 0){
                    		if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) N++;
						}
                    }
                    removed = 1;
                }
            } else if (opcode == UD_Ixchg) { // this instruction has two dests and one of them maybe dead!
                ud_t *udins_new = NULL;
                if (op0->type == UD_OP_MEM || op1->type == UD_OP_MEM) { // one of the operands is memory, the other is got to be reg
                    REGISTER_BITVECTOR reg = NonMemRegUsed(iList, iins);
                    if (!MemLocIsConstant(iList, iins->memDefMin,
                            iins->memDefSize)) {
                        udins_new = UDCreateMovOffsetToReg(op1->base, iins->memDefMin,
                                OFFSET_TO_REG);
                    } else if (!(reg & RegsLive)) { // so mem loc is alive
                        udins_new = UDCreateMovOffsetToReg(BitVectorToUDRegister(reg),
                                iins->memDefMin, REG_TO_OFFSET);
                    }

                } else { // both are reg
                    if (!(RegsLive & UDRegisterToBitVector(op0->base))) { // first op is dead
                        udins_new = UDCreateMovRegToReg(op1->base, op0->base);
                    } else if (!(RegsLive & UDRegisterToBitVector(op1->base))) {
                        udins_new = UDCreateMovRegToReg(op0->base, op1->base);
                    }
                }
                if (udins_new != NULL) {
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                    SaveInstrChange(iList, iins);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);

                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);
                }
            }
        }
        if (INSTR_HAS_FLAG(iins, INSTR_MAYBE_DELETED)) {
            INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
            INSTR_SET_FLAG(iins, INSTR_TMP3);
            if(SaveInstrChange(iList, iins) != -1){
            }

            removed = 1;
        }

        int garanteedResults = 0;
        if (!removed) {
            if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) && !InstrIsControlTransfer(iList, iins) && (regsDef != 0 || (opcode == UD_Imov && INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM))
                    || opcode == UD_Ipopfd
                    || (opcode == UD_Ipush && INSTR_HAS_FLAG(iins, INSTR_READS_MEM))
                    || (opcode == UD_Ipop && INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)))
                    && opcode != UD_Ixchg && opcode != UD_Irdtsc) {
                if ((op1->type == UD_NONE || op1->type != UD_OP_IMM) && (garanteedResults =
                        DefinitionIsGuaranteed(iList, iins))) {
                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);
                    SetCurrentInstr(iList, iins->prev);
                }
            }
        }

        if (handle_rop || INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
            if(!INSTR_HAS_FLAG(iins, INSTR_IS_DELETED))
                (*numInstructions)++; // add whenever want to fetch an instr!
        }
#if IGNORE_API_CALLS
        else {
            prev = iins;
            continue;
        }
#endif



        if((iins->uInstr.mnemonic == UD_Ishr || iins->uInstr.mnemonic == UD_Ishl)
                && iins->uInstr.operand[1].type == UD_OP_IMM && iins->uInstr.operand[1].lval.sdword > 1){
            flagsDef &= ~PSW_OF;
        }


        if ((flagsDef & FlagsLive)) {
            if (!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                INSTR_SET_FLAG(iins, INSTR_IN_USEDEF);
                SaveInstrChange(iList, iins);
            }
        } else {
            if (INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                INSTR_REMOVE_FLAG(iins, INSTR_IN_USEDEF);
                SaveInstrChange(iList, iins);
            }
        }
        if (!removed /*&& garanteedResults != -1*/) {
            FlagsLive = (FlagsLive & ~flagsDef) | flagsUsed;
            RegsLive = (RegsLive & ~MaskDefinedReg(regsDef)) | MaskDefinedReg(regsUsed);

            if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && iins->memDefMin > 0x100) {
                SetMemoryLocationDead(iList, iins->memDefMin, iins->memDefSize);
            }
            if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
                SetMemoryLocationLive(iList, iins->memUsedMin,
                        iins->memUsedSize);
            }

        } else {
            if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                iList->numBaseModuleInstrs--;
            }
        }
        prev = iins;
    }
    fprintf(stderr, "\t[RemoveDeadCodeIntraBbl]: removed %d instructions\n", N);
    return N;
}


int RemoveOnlyDeadCode(InstrList* iList, int first, int last, uint64_t *numInstructions) {
    Instruction *iins, *next, *prev = NULL;
    enum ud_mnemonic_code opcode;
    ud_operand_t *op0, *op1;
    PSW_BITVECTOR flagsDef, flagsUsed, FlagsLive;
    REGISTER_BITVECTOR regsDef, regsUsed, RegsLive;
    int removed = 0;
    int N = 0;

    /*if (last != iList->last) { // everything is live!
        RegsLive = 0xffffffff;
        FlagsLive = 0xffffffff;
        SetMemoryMap(iList, MEMORY_LIVE);

    } else*/ {
        RegsLive = 0;
        FlagsLive = 0;
        SetMemoryMap(iList, MEMORY_DEAD);

    }
    iins = GetInstruction(iList, first);


    SetCurrentInstr(iList, last);
    while ((iins = FetchPrevInstr(iList)) != NULL) {

    	if(!INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE) && INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
				FlagsLive = 0;
				RegsLive = 0;
			}

        if (prev && prev->addr == iins->addr && InstrCompare(prev, iins)) {
            INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
            SaveInstrChange(iList, iins);
            continue;
        }

        removed = 0;
        flagsDef = FlagsDef(iins);
        flagsUsed = FlagsUsed(iins);
        regsDef = RegDef(iList, iins);
        regsUsed = RegUsed(iList, iins);

        if (!((flagsDef | flagsUsed | regsUsed | regsDef) || iins->memUsedMin || iins->memDefMin)) {
            if (iins->uInstr.mnemonic == UD_Inop || (iins->uInstr.mnemonic == UD_Ijmp && !INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL))) {
                INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                if(SaveInstrChange(iList, iins) == 0){
                    N++;
                }
            } else if(handle_rop || INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                (*numInstructions)++;
            }
            continue;
        }
        opcode = iins->uInstr.mnemonic;
        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);

        if ((IsPush(iins) || IsPop(iins) || opcode == UD_Icall || opcode == UD_Iret)) {
            if (UDOpIsRegister(op0) == UD_R_ESP) {
                if (opcode == UD_Ipop)
                    regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
                else if (opcode == UD_Ipush)
                    regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
            } else {
                regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
                regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
            }
        }


        if ((!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF) /*|| opcode == UD_Ipopfd*/)
                && !INSTR_HAS_FLAG(iins, INSTR_UNSIMPLIFIABLE) /*&& !(strlen(iins->funcName))*/) { // just keep the first instruction of a function, need it when constructing cfg!
            if (!((INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)
                    && MemoryLocationIsLive(iList, iins->memDefMin,
                            iins->memDefSize)) || (flagsDef & FlagsLive) != 0
                    || (MaskDefinedReg(regsDef) & RegsLive) != 0)
                    || (OpcodeIsArithWithNoEffect(opcode) && op1 != NULL
                            && (op1->type == UD_OP_IMM || op1->type == UD_OP_CONST)
                            && op1->lval.sdword == 0
                            && !INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED))
                    || opcode == UD_Inop || opcode == UD_Ijmp) {

                if (opcode == UD_Icall && (next = FetchNextInstr(iList)) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL)) {

                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    ud_t *udins_new = UDCreateSubImmFromReg(UD_R_ESP, 4);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
                    //INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                    SaveInstrChange(iList, iins);
                    free(udins_new);
                    udins_new = NULL;
                    N++;

                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);

                } else if ((opcode == UD_Ipush || opcode == UD_Ipushfd)) {
                    if (opcode == UD_Ipushfd) {
                        iins->flagsUsed = 0;
                        iins->flagsDef = 0;
                    }
                    ud_t* udins = NULL;
                    int stackSize = 0;
                    if (opcode == UD_Icall || opcode == UD_Ipushfd)
                        stackSize = 4;
                    else
                        stackSize = GetStackSize(iins);

                    udins = UDCreateSubImmFromReg(UD_R_ESP, stackSize);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 1);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);

                    SaveInstrChange(iList, iins);
                    free(udins);

                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);
                } else if ((opcode == UD_Ipop || opcode == UD_Ipopfd)) {
                    if (opcode == UD_Ipopfd) {
                        iins->flagsUsed = 0;
                        iins->flagsDef = 0;
                    }
                    ud_t* udins = NULL;
                    int stackSize = 0;
                    if (opcode == UD_Ipopfd)
                        stackSize = 4;
                    else
                        stackSize = GetStackSize(iins);
                    udins = UDCreateAddImmToReg(UD_R_ESP, stackSize);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 1);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);

                    SaveInstrChange(iList, iins);
                    free(udins);

                    opcode = iins->uInstr.mnemonic;
                    op0 = &(iins->uInstr.operand[0]);
                    op1 = &(iins->uInstr.operand[1]);

                    flagsDef = FlagsDef(iins);
                    flagsUsed = FlagsUsed(iins);
                    regsDef = RegDef(iList, iins);
                    regsUsed = RegUsed(iList, iins);
                } else if (!IsControlFlowInstr(iins) && !INSTR_HAS_FLAG(iins, INSTR_USED_AS_CALL) && !((opcode == UD_Iadd || opcode == UD_Isub) && op0->type == UD_OP_REG && op0->base == UD_R_ESP)){

                    INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    SaveInstrChange(iList, iins);
                    if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                    	if(SaveInstrChange(iList, iins) == 0){
							N++;
						}
                    }
                    removed = 1;
                }
            }
        }

        ////////////////////////////////////////////////////////////
        if (handle_rop || INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
            if(!INSTR_HAS_FLAG(iins, INSTR_IS_DELETED))
                (*numInstructions)++; // add whenever want to fetch an instr!
        }
#if IGNORE_API_CALLS
        else {
            prev = iins;
            continue;
        }
#endif



        if((iins->uInstr.mnemonic == UD_Ishr || iins->uInstr.mnemonic == UD_Ishl)
                && iins->uInstr.operand[1].type == UD_OP_IMM && iins->uInstr.operand[1].lval.sdword > 1){
            flagsDef &= ~PSW_OF;
        }


        if ((flagsDef & FlagsLive)) {
            //FlagsLive &= ~flagsDef;
            if (!INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                INSTR_SET_FLAG(iins, INSTR_IN_USEDEF);
                SaveInstrChange(iList, iins);
            }
        } else {
            if (INSTR_HAS_FLAG(iins, INSTR_IN_USEDEF)) {
                INSTR_REMOVE_FLAG(iins, INSTR_IN_USEDEF);
                SaveInstrChange(iList, iins);
            }
        }
        if (!removed) {
            FlagsLive = (FlagsLive & ~flagsDef) | flagsUsed;
            RegsLive = (RegsLive & ~MaskDefinedReg(regsDef)) | MaskDefinedReg(regsUsed);

            if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && iins->memDefMin > 0x100) {
                SetMemoryLocationDead(iList, iins->memDefMin, iins->memDefSize);
            }
            if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
                SetMemoryLocationLive(iList, iins->memUsedMin,
                        iins->memUsedSize);
            }

        } else {
            if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                iList->numBaseModuleInstrs--;
            }
        }
        prev = iins;
    }
    return N;
}

int SimplifyGlobalMems(InstrList* iList) {
    Instruction *iins, *iins1;
    int N = 0;
    REGISTER_BITVECTOR regsDef, regsUsed;
    ud_t *udins_new = NULL;


    ClearConstantPages(iList, 0);

    iList->currentInstr = 0;
    while ((iins = FetchNextInstr(iList)) != NULL) {

#if IGNORE_API_CALLS
       if(!handle_rop && !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
           continue;
       }
#endif

        if (!INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && !INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
            continue;
        }


        regsUsed = NonMemRegUsed(iList, iins);
        regsDef = RegDef(iList, iins);

        if ((IsPush(iins) || IsPop(iins))) {
            if (iins->uInstr.operand[0].base == UD_R_ESP) {
                if (iins->uInstr.mnemonic == UD_Ipop)
                    regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
                else if (iins->uInstr.mnemonic == UD_Ipush)
                    regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
            } else {
                regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
                regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
            }
        }
        if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)
                && !MemLocIsConstant(iList, iins->memUsedMin, iins->memUsedSize)
                && !INSTR_HAS_FLAG(iins, INSTR_READS_TAINTED_MEM)){
            if(!INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && regsDef&& iins->uInstr.mnemonic != UD_Icall
                    && iins->uInstr.mnemonic != UD_Ipopad && iins->uInstr.mnemonic != UD_Ipopa) {
                if((regsDef & ~LYNX_ESP & ~regsUsed) != 0) {
					int memReg = GetImplicitMemRegDef(iins);
					int nonMemReg = (regsDef & ~memReg);
					iins1 = GetInstruction(iList, iins->next);
					int val=0;
					if ((iins->uInstr.mnemonic == UD_Imul || iins->uInstr.mnemonic == UD_Iimul)
					&& iins->uInstr.operand[1].type == UD_NONE) {
						val = GetRegisterValue(iins1, LYNX_EAX);
						udins_new = UDCreateMovImmToReg(UD_R_EAX, val); // change EAX reg

						Instruction *edxInstr = InstrClone(iins);
						edxInstr->eax = iins1->eax;
						ud_t *edx_udis = UDCreateMovImmToReg(UD_R_EDX, iins1->edx);// change EDX reg
						CarefullCopyUdins2InsStr(iList, edxInstr, &(edxInstr->uInstr), edx_udis, 0);
						edxInstr->flags = 0;
                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                            INSTR_SET_FLAG(edxInstr, INSTR_IN_BASE_MODULE);
                        }
						edxInstr = AddInstruction(iList, edxInstr, iins->order);
						free(edx_udis);
						RInstrDestroy(iList, edxInstr);

					} else if (nonMemReg) {
						val = GetRegisterValue(iins1, BitVectorToUDRegister(nonMemReg));
						udins_new = UDCreateMovImmToReg(BitVectorToUDRegister(nonMemReg), val);
					}
					if(udins_new){
                        CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 0);
                        SaveInstrChange(iList, iins);
                        free(udins_new);
                        udins_new = NULL;
					}

					if (memReg != 0) {
						udins_new = UDCreateAddImmToReg(BitVectorToUDRegister(memReg),
						GetRegisterValue(iins1, BitVectorToUDRegister(memReg))
						- GetRegisterValue(iins, BitVectorToUDRegister(memReg)));
						Instruction *new = InstrClone(iins1);
						SetRegisterValue(new, BitVectorToUDRegister(nonMemReg), val);
						CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins_new, 0);
						new->flags = 0;
                        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
                            INSTR_SET_FLAG(new, INSTR_IN_BASE_MODULE);
                        }
						new = AddInstruction(iList, new, iins->order);
						free(udins_new);
						udins_new = NULL;
						//RInstrDestroy(iList, new);
					}

					N++;

				} else if((regsDef & ~LYNX_ESP) != 0) {
					INSTR_SET_FLAG(iins, INSTR_READS_GLOBAL_MEM);
				}

			} else if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
				if (NonMemRegUsed(iList, iins) == 0 /*&& !INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED)*/) { // like push [global_mem]
					if(iins->uInstr.mnemonic == UD_Ipush){
						udins_new = UDCreatePushImm(UD_R_EAX, 0);
					} else {
						udins_new = UDCreateMovImmToOffset(iins->memDefMin, 0, iins->memDefSize);
					}
					CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
					SaveInstrChange(iList, iins);
					free(udins_new);
				} else { //like xadd, cmpxchg, add [global_mem], reg, inc [global_mem] etc.
					if (iins->uInstr.mnemonic == UD_Ixadd || iins->uInstr.mnemonic == UD_Icmpxchg || iins->uInstr.mnemonic == UD_Icmpxchg8b) {
						INSTR_SET_FLAG(iins, INSTR_READS_GLOBAL_MEM);
						INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
						SaveInstrChange(iList, iins);
					} else if(iins->uInstr.mnemonic == UD_Ixchg) {
						enum ud_type dest = (iins->uInstr.operand[0].type == UD_OP_REG ? iins->uInstr.operand[0].base : iins->uInstr.operand[1].base);
						udins_new = UDCreateMovOffsetToReg(dest, iins->memDefMin, 1);
						CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
						SaveInstrChange(iList, iins);
						free(udins_new);

						Instruction *new = InstrClone(iins);
						Instruction *next = GetInstruction(iList, iins->next);
						udins_new = UDCreateMovImmToReg(dest, GetRegisterValue(next, dest));

						CarefullCopyUdins2InsStr(iList, new, &(new->uInstr), udins_new, 1);
						new = AddInstruction(iList, new, iins->order);
						free(udins_new);
						udins_new = NULL;
						UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
						N++;
						continue;
					} else {
						if (iins->uInstr.operand[0].type == UD_OP_MEM) {
							INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
						} else if (iins->uInstr.operand[1].type == UD_OP_MEM) {
							INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
						}
						SaveInstrChange(iList, iins);
					}
				}
			} else if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) && (iins->uInstr.mnemonic == UD_Iret || iins->uInstr.mnemonic == UD_Icall) && !INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) {

				Instruction *next = GetInstruction(iList, iins->next);
				uint64_t addr = next->addr;
				udins_new = UDCreateJump(addr - 5);

				CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
				SaveInstrChange(iList, iins);
				free(udins_new);
				udins_new = NULL;
			}
			else {
				if (iins->uInstr.operand[0].type == UD_OP_MEM) {
					INSTR_SET_FLAG(iins, INSTR_OP0_CONSTANT);
				} else if (iins->uInstr.operand[1].type == UD_OP_MEM) {
					INSTR_SET_FLAG(iins, INSTR_OP1_CONSTANT);
				}
				SaveInstrChange(iList, iins);
			}
		}
        if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)
                && !MemLocIsConstant(iList, iins->memDefMin,
                        iins->memDefSize)) {
            if (!INSTR_HAS_FLAG(iins, INSTR_READS_MEM))
                SetAddrAsConst(iList, iins->memDefMin, iins->memDefSize);
            else {
                if (MemLocIsConstant(iList, iins->memUsedMin,
                        iins->memUsedSize))
                    SetAddrAsConst(iList, iins->memDefMin, iins->memDefSize);
            }
        }
    }
    return N;
}



void WriteInstrFileHeaderIter(InstrList *iList, FILE *f) {
    InstrFileHeader fileHeader;
    Thread *currThread;
    ThreadRange *currRange;
    int i, j, headerSize = 0;

    if (TraceIsMultiThreaded) {
        fseeko(f, 0, SEEK_END);
        fileHeader.RangeInfoOffset = ftello(f);
        for (i = 0; i < al_size(traceThreads); i++) {
            currThread = al_get(traceThreads, i);
            fwrite(currThread, sizeof(Thread), 1, f);
            headerSize += sizeof(Thread);
            for (j = 0; j < al_size(currThread->ranges); j++) {
                currRange = al_get(currThread->ranges, j);
                fwrite(currRange, sizeof(ThreadRange), 1, f);
                headerSize += sizeof(ThreadRange);
            }
        }
    }
    fileHeader.TraceIsMultiThreaded = TraceIsMultiThreaded;
    fileHeader.LastInstrInOriginalTrace = iList->last;
    fileHeader.NumInstrs = numInstrs; // set to global numInstr variable since iList might be shared between threads!
    fileHeader.Loop = iList->loop;
    fileHeader.NumThreads = (TraceIsMultiThreaded ? al_size(traceThreads) : 0);
    fileHeader.RangeInfoSize = headerSize;
    fseeko(f, 0, SEEK_SET);
    fwrite(&fileHeader, sizeof(InstrFileHeader), 1, f);

}

int WriteInstrFile(InstrList *iList, int loop) {
    FILE *instrFile;
    char *instrFileName = calloc(1, strlen(baseName) + strlen("./") + strlen(".instrFile") + 5);
    sprintf(instrFileName, "%s%s%s%04d", "./", baseName, ".instrFile", loop);

    Instruction *next;
    if (STORE_INSTRS_IN_FILE && ((instrFile = fopen(instrFileName, "w+")) != NULL)) {
        //fwrite(iList, sizeof(struct InstrList), 1, instrFile);
        WriteInstrFileHeaderIter(iList, instrFile);
        int i;
        for (i = 0; i < numInstrs; i++) {
            next = GetInstruction(iList, i);
            // next->self = NULL;
            AddInstrToFile(instrFile, next);
            RInstrDestroy(iList, next);
        }
    } else {
        return -1; // fail to open file(?)
    }
    fclose(instrFile);
    return 0;

}

void SimplifyConstantsSingleThread(InstrList *iList) {
    uint64_t nelim = 0;
    uint64_t n;
    bool chg;
    int i = 0;
    uint64_t numProcessed = 0, prevNumProcessed = 0;

    int first = 0;
    int last;

    last = iList->lastInstr;
    first = iList->firstInstr;

    loop = iList->loop + 1;

    RemoveOnlyDeadCode(iList, first, last, &numProcessed);

    do {
        chg = false;

        n = RemoveDeadCode(iList, first, last, &numProcessed);
        nelim += n;
        chg |= (n > 0);

        n = PropagateConstants(iList, first, last);
        nelim += n;
        chg |= (n > 0);
        numProcessed += n;

        iList->loop = loop;
        loop++;
        if(numProcessed == prevNumProcessed){
        	break;
        }

        prevNumProcessed = numProcessed;

        if (numProcessed < 2)
            break;
    } while (chg);

    for (i = 0; i < totalPageDirEntries; i++) {
        free(iList->pageDir[i].pageAddr);
    }
}
