/*
 *
 * File: taint_analysis.c
 * Author: Babak Yadegari
 */

/*
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */

#include "taint_analysis.h"
#include <iostream>
#include <fstream>

TaintBlock *GetMemAddrTaint(InstrList *iList, int addr, uint size){
	TaintBlock *block = new TaintBlock();
	uint i;
	for(i = 0; i < size; i++){
		TByte *tmp = (TByte *)GetMemoryLocPointer(iList, addr+i);
		if (!tmp){
			tmp = new TByte(DATAFLAG_CONST, 0);
			SetMemoryLocPointer(iList, addr+i, reinterpret_cast<void *>(tmp));
		}
		block->AddByte(tmp);
	}
	return block;
}

void ClearMemAddrTaint(InstrList *iList, int addr, int size){
    UntaintMemoryLoc(iList, addr, size);
    SetMemoryLocPointer(iList, addr, 0);
}

void AddMemAddrTaint(InstrList *iList, int addr, TaintBlock *b){
	for (uint i=0; i < b->Size(); i++){
		TaintMemoryLoc(iList, addr+i, 1, 0);
		SetMemoryLocPointer(iList, addr+i, reinterpret_cast<void *>(b->GetTByte(i)));
	}
}

void BackwardTaintAnnotationNew(InstrList *iList) {
    Instruction *iins;//, *iinsNext;
    ud_mnemonic_code_t opcode;
    //int esp = 0xffffffff, offset;
    //uint64_t calleeOrder;
    //bool inFunction = false, inInputFunction = false;
    //WriteSet *params = WriteSetCreate(), *memUsed, *memDef;

    SetMemoryMap(iList, MEMORY_DEAD);
    SetCurrentInstr(iList, iList->lastInstr);
    iins = FetchPrevInstr(iList);
    while(iins){
        while(iins && !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
            opcode = iins->uInstr.mnemonic;
            if(opcode != UD_Iret && INSTR_HAS_FLAG(iins, INSTR_READS_MEM)){
                SetMemoryLocationLive(iList, iins->memUsedMin, iins->memUsedSize);
                /*if (iins->memUsedMin == 0x0012e764){
                	fprintf(stderr, "============== %d\n", iins->order);
                }*/
            }

            if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){
            	SetMemoryLocationDead(iList, iins->memDefMin, iins->memDefSize);
            }
            iins = FetchPrevInstr(iList);
        }
        while(iins && INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
            if(iins->uInstr.mnemonic != UD_Ixchg && INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && MemoryLocationIsLive(iList, iins->memDefMin, iins->memDefSize)){
                INSTR_SET_FLAG(iins, INSTR_BACKWARD_TAINTED);
                INSTR_SET_FLAG(iins, INSTR_UNSIMPLIFIABLE);
                SaveInstrChange(iList, iins);
#if DEBUG_MESSAGES
                printf("Backward: %d, %s\n", iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));
#endif
                SetMemoryLocationDead(iList, iins->memDefMin, iins->memDefSize);
            }
            iins = FetchPrevInstr(iList);
        }

        ClearMemoryPages(iList, MEMORY_DEAD);
    }

}

bool SkipInputFuncs(Instruction *iins){
    char *skipFuncs[]={(char*)"lstrcpy", (char*)"strcpy", (char*)"malloc", (char*)"toupper", NULL};
    int i = 0;
    for (i = 0; skipFuncs[i] != NULL; i++){
        if(strcmp(GetFunctionNameById(iins->funcName), skipFuncs[i]) == 0){
            return true;
        }
    }
    return false;
}


void ForwardTaintAnnotation(InstrList *iList) {
	Instruction *iins;
	REGISTER_BITVECTOR regsDef, regsUsed;

	ClearConstantPages(iList, 0);

	SetCurrentInstr(iList, iList->lastBaseModuelInstr);
	while ((iins = FetchPrevInstr(iList)) != NULL && iins->order > 0) {
		//printf("%d\n", iins->order);


		if (!INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {

		    /* first check the function! we need to exlude certain functions
		     * which do not interact with the out world such as string copy functions!
		     */

		    int afterFunc = iins->order;

		    // find the next non-base module
		    Instruction *prev = NULL;
		    while(iins && !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
		        prev = iins;
		        iins = FetchPrevInstr(iList);
		    }
		    if(strcmp(GetFunctionNameById(prev->funcName), "") != 0){
                if(SkipInputFuncs(prev)){
                    ClearConstantPages(iList, 0);
                    break;
                }
            }

            SetCurrentInstr(iList, afterFunc);
		    iins = GetInstruction(iList, afterFunc);        //restore state
		    bool eaxMarked = false;
			do {
				regsDef = (RegDef(iList, iins) & ~(LYNX_ESP | LYNX_SP));
				regsUsed = (RegUsed(iList, iins) & ~(LYNX_ESP | LYNX_SP));

				if (!((regsUsed | regsDef) || iins->memUsedMin
						|| iins->memDefMin)) {
					if (iins->uInstr.mnemonic == UD_Inop) {
						INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
						SaveInstrChange(iList, iins);
					}
					iins = FetchPrevInstr(iList);
					continue;
				}

				regsDef = RegDef(iList, iins);
				if((INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && MemLocIsConstant(iList, iins->memDefMin, iins->memDefSize))
						|| (!eaxMarked && (regsDef & MaskDefinedReg(LYNX_EAX)))) {
					INSTR_SET_FLAG(iins, INSTR_FORWARD_TAINTED);
					INSTR_SET_FLAG(iins, INSTR_INPUT_SYMBOLIC);
					INSTR_SET_FLAG(iins, INSTR_UNSIMPLIFIABLE);
					SaveInstrChange(iList, iins);
					if(regsDef & MaskDefinedReg(LYNX_EAX)){
						eaxMarked = true;
					}
				}

				if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
					UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
				}

				iins = FetchPrevInstr(iList);
			} while (iins && !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE));
		}
		if(!iins){
		    return;
		}
		if(iins->order == iList->startMarker + 2){
		    SetCurrentInstr(iList, iins->order-3);
		    continue;
		}


		regsDef = (RegDef(iList, iins) & ~(LYNX_ESP | LYNX_SP));

		if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && (regsDef)) {
			SetAddrAsConst(iList, iins->memUsedMin, iins->memUsedSize);
		}

		if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
			UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
		}
	}
}


/*
 * InitializeTaints
 * Initializes a given TaintedRes structure to
 * zeroes and returns the initialized structure.
 */
TaintedRes *InitializeTaintedResStruct(TaintedRes *taints){
	if(!taints){
		return NULL;
	}
	taints->regs = taints->flags = 0;
	int i;
	for(i = 0; i < 8; i++){
		taints->regsMask[i] = 0;
	}
	/*taints->eaxMask = 0;
	taints->ebpMask = 0;
	taints->ebxMask = 0;
	taints->ecxMask = 0;
	taints->ediMask = 0;
	taints->edxMask = 0;
	taints->esiMask = 0;
	taints->espMask = 0;*/

	return taints;
}

TaintedNoFlagRes *InitializeTaintedNonFlagResStruct(TaintedNoFlagRes *taints){
    if(!taints){
        return NULL;
    }
    taints->regs = 0;
    int i;
    for(i = 0; i < 8; i++){
        taints->regsMask[i] = 0;
    }
    /*taints->eaxMask = 0;
    taints->ebpMask = 0;
    taints->ebxMask = 0;
    taints->ecxMask = 0;
    taints->ediMask = 0;
    taints->edxMask = 0;
    taints->esiMask = 0;
    taints->espMask = 0;*/

    return taints;
}

int RegIndex(REGISTER_BITVECTOR reg){
	reg = BitVectorToUDRegister(reg);
    switch (reg) {
        case UD_R_EAX:
        case UD_R_AX:
        case UD_R_AH:
        case UD_R_AL:
            return 0;
        case UD_R_ECX:
        case UD_R_CX:
        case UD_R_CL:
        case UD_R_CH:
        	return 1;
        case UD_R_EDX:
        case UD_R_DX:
        case UD_R_DL:
        case UD_R_DH:
        	return 2;
        case UD_R_EBX:
        case UD_R_BX:
        case UD_R_BL:
        case UD_R_BH:
        	return 3;
        case UD_R_ESP:
        case UD_R_SP:
        	return 4;
        case UD_R_EBP:
        case UD_R_BP:
        	return 5;
        case UD_R_ESI:
        case UD_R_SI:
        	return 6;
        case UD_R_EDI:
        case UD_R_DI:
            return 7;
        default:
            fprintf(stderr, "[RegIndex]: Invalid or unhandled register: %d\n", (int)reg);
            break;
    }
    return -1;
}




/*
 *
 *  ------------------------------
 * |        EAX       |
 * |    AX   | ...    |
 * | AL | AH |        |
 *  ------------------------------
 * ^
 * |
 * returns this for example for al!
 */

int RegIndexInArray2(REGISTER_BITVECTOR reg){
    //reg = BitVectorToUDRegister(reg);
    switch (reg) {
        case LYNX_EAX:
        case LYNX_AX:
        case LYNX_AL:
            return 0;
        case LYNX_AH:
            return 1;
        case LYNX_ECX:
        case LYNX_CX:
        case LYNX_CL:
            return 4;
        case LYNX_CH:
            return 5;
        case LYNX_EDX:
        case LYNX_DX:
        case LYNX_DL:
            return 8;
        case LYNX_DH:
            return 9;
        case LYNX_EBX:
        case LYNX_BX:
        case LYNX_BL:
            return 12;
        case LYNX_BH:
            return 13;
        case LYNX_ESP:
        case LYNX_SP:
            return 16;
        case LYNX_EBP:
        case LYNX_BP:
            return 20;
        case LYNX_ESI:
        case LYNX_SI:
            return 24;
        case LYNX_EDI:
        case LYNX_DI:
            return 28;
        default:
            fprintf(stderr, "[RegIndexInArray2]: Invalid or unhandled register: %d\n", (int)reg);
            //assert(0);
            break;
    }
    return -1;
}

/*
 * returns a pointer to the last byte's
 * variables in the list. So if we are
 * dealing with a 4 byte register, it
 * should iterate back from the given
 * pointer.
 */
TaintBlock *GetRegBitVars(BitDataDep *dataDep, REGISTER_BITVECTOR reg){
	TaintBlock *tmp = new TaintBlock();
    int regIndex = RegIndexInArray2(reg);
    if(regIndex == -1){
        return NULL;
    }
    uint size = BitvectorRegisterSize(reg);
    for(uint i=0; i < size; i++){
    	tmp->AddByte(&(dataDep->regsTaint[regIndex+i]));
    }
    return tmp;
}

void ApplyRegTaintWithMask(BitDataDep *taints, int taint, REGISTER_BITVECTOR reg, uint order){
	int i, tmpReg;
	for (i = 0; i < 30 && reg; i++){
		tmpReg = (0x1 << i);
		if(tmpReg & reg){
			TaintBlock *regvars = GetRegBitVars(taints,  tmpReg);

			uint j;
			for(j=0; j < BitvectorRegisterSize(tmpReg)*8; j++ ){
				(*regvars)[j].tag = taint;
				(*regvars)[j].src_instr = order;
			}

			if(taint == DATAFLAG_CONST){
			    taints->taintedRegs &= ~MaskDefinedReg(tmpReg);
			} else {
				taints->taintedRegs |= MaskDefinedReg(tmpReg);
			}
		}
	}
}


void BackwardTaintPropagationNew(InstrList *iList, int StartIndex) {

	Instruction *iins;
	int i;
	ud_operand_t *op0;
	enum ud_mnemonic_code opcode;
	REGISTER_BITVECTOR regsDef, regsUsed, regsTainted = 0, flagsUsed, flagsDef, flagsTainted = 0,
			nonMemRegUsed;
	//WriteSet *memUsed, *memDef, *memtmp, *memTainted;
#if !CHECKSUM_DETECT
	ClearConstantPages(iList, 0);
#endif

	i = StartIndex;
	SetCurrentInstr(iList, i);
	//SetAddrAsConst(iList, 0x401002, 4);


	/*iins = GetInstruction(iList, 217576);
	INSTR_SET_FLAG(iins, INSTR_BACKWARD_TAINTED);
	SaveInstrChange(iList, iins);*/

	//i = 65425;
	//SetCurrentInstr(iList, 65425);



	while ((iins = FetchPrevInstr(iList)) != NULL) {
		opcode = iins->uInstr.mnemonic;

		regsDef = MaskDefinedReg(RegDef(iList, iins));
		regsUsed = MaskDefinedReg(RegUsed(iList, iins));
		nonMemRegUsed = MaskDefinedReg(NonMemRegUsed(iList, iins));
		//memDef = MemDef(iList, iins);
		//memUsed = MemUsed(iList, iins);
		flagsUsed = FlagsUsed(iins);
		flagsDef = FlagsDef(iins);
		op0 = &(iins->uInstr.operand[0]);

		if(OpcodeIsArith(opcode) && iins->uInstr.operand[1].type == UD_OP_REG && GetRegisterValue(iins, iins->uInstr.operand[1].base) == 0){
		    continue;
		}

		if ((IsPush(iins) || IsPop(iins))) {
			if (UDOpIsRegister(op0) == UD_R_ESP) {
				if (iins->uInstr.mnemonic == UD_Ipop)
					regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
				else if (iins->uInstr.mnemonic == UD_Ipush)
					regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
			} else {
				regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
				regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
			}
		}

		if (INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED)) {
			if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
				SetAddrAsConst(iList, iins->memUsedMin, iins->memUsedSize);
				//memTainted = MergeWriteSets(memTainted, memUsed);
			}
			if (nonMemRegUsed != 0)
				regsTainted |= nonMemRegUsed;
			if (flagsUsed){
				flagsTainted |= flagsUsed;
			}
			if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && MemLocIsConstant(iList, iins->memDefMin, iins->memDefSize)){
				UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
			}
		} else {
			if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)
					&&
#if CHECKSUM_DETECT
					PartialMemLocIsConstant(iList, iins->memDefMin, iins->memDefSize)
#else
					MemLocIsConstant(iList, iins->memDefMin, iins->memDefSize)
#endif

			) {
				INSTR_SET_FLAG(iins, INSTR_BACKWARD_TAINTED);
			}
			if ((regsDef & regsTainted) != 0) {
				INSTR_SET_FLAG(iins, INSTR_BACKWARD_TAINTED);
			}
			if ((flagsDef & flagsTainted) != 0) {
				INSTR_SET_FLAG(iins, INSTR_BACKWARD_TAINTED);
			}
			if (INSTR_HAS_FLAG(iins, INSTR_BACKWARD_TAINTED)
					&& !(opcode == UD_Ixor && op0->base == iins->uInstr.operand[1].base)) {
				//fprintf(stderr, "# %s %d\n", ud_insn_asm(LoadUDisInstrObj(iins)), iins->order);
				SaveInstrChange(iList, iins);
				if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
					if (opcode == UD_Ipopad) {
						int taintRegs = (regsTainted & 0xff);
						//WriteSet *taintMems = WriteSetCreate();
						int k;
						for (k = 0; k < 8 && (taintRegs); k++) {
							if ((taintRegs & 0x80)) {
								Range *r = RangeCreate(iins->esp + (k) * 4,
										iins->esp + (k + 1) * 4 - 1);
								SetAddrAsConst(iList, r->min, r->max - r->min + 1);
							}
							taintRegs = (taintRegs << 1);
							taintRegs &= 0xff;
						}
					}
					else if (opcode != UD_Ixchg) {
						regsTainted &= (~UpToAllRegs(regsDef));
						if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
							UnsetConstAddr(iList, iins->memDefMin,
									iins->memDefSize);
						}
						SetAddrAsConst(iList, iins->memUsedMin,
								iins->memUsedSize);

					}
				}
				if (nonMemRegUsed != 0) {
					if (opcode == UD_Ipushad) {
						int k;
						for (k = 0; k < 8; k++) {
							Range *r = RangeCreate(iins->esp - (k + 1) * 4,
									iins->esp - (k) * 4 - 1);
							if (MemLocIsConstant(iList, r->min, r->max - r->min + 1)) {
								regsTainted |= MaskDefinedReg((0x1 << k));
							}
						}
					} else if (opcode == UD_Ixchg) {
						bool regIsTainted = false;
						if (op0->type == UD_OP_MEM) { //op0 is memory
							if (MemLocIsConstant(iList, iins->memUsedMin,
									iins->memUsedSize))
								regIsTainted = true;
							if ((regsTainted & regsDef) == 0) {
								UnsetConstAddr(iList, iins->memDefMin,
										iins->memUsedSize);
							} else {
								SetAddrAsConst(iList, iins->memDefMin,
										iins->memDefSize);
								regsTainted &= ~(UpToAllRegs(regsDef));
							}
							if (regIsTainted)
								regsTainted |= regsDef;
						} else if (iins->uInstr.operand[1].type == UD_OP_MEM) { //op0 is reg
							if (MemLocIsConstant(iList, iins->memUsedMin,
									iins->memUsedSize))
								regIsTainted = true;
							if ((regsTainted & regsDef) == 0) {
								UnsetConstAddr(iList, iins->memDefMin,
										iins->memDefSize);
							} else {
								SetAddrAsConst(iList, iins->memDefMin,
										iins->memDefSize);
								regsTainted &= ~(UpToAllRegs(regsDef));
							}
							if (regIsTainted)
								regsTainted |= regsDef;
						} else { // op0 and op1 are regs
							if ((regsTainted & UDRegisterToBitVector(op0->base)) != 0
									&& (regsTainted
											& UDRegisterToBitVector(iins->uInstr.operand[1].base))
											== 0) {
								regsTainted |= (MaskDefinedReg(
										UDRegisterToBitVector(iins->uInstr.operand[1].base))
										| UpToAllRegs(UDRegisterToBitVector(op0->base)));
								regsTainted &= ~(MaskDefinedReg(UDRegisterToBitVector(op0->base))
										| UpToAllRegs(UDRegisterToBitVector(op0->base)));
							} else if ((regsTainted
									& MaskDefinedReg(
											UDRegisterToBitVector(iins->uInstr.operand[1].base)))
									!= 0
									&& (regsTainted
											& MaskDefinedReg(UDRegisterToBitVector(op0->base)))
											== 0) {
								regsTainted |= (MaskDefinedReg(UDRegisterToBitVector(op0->base))
										| UpToAllRegs(UDRegisterToBitVector(op0->base)));
								regsTainted &=
										~(MaskDefinedReg(
												UDRegisterToBitVector(iins->uInstr.operand[1].base))
												| UpToAllRegs(
														UDRegisterToBitVector(
																iins->uInstr.operand[1].base)));
							}
						}
					} else {
						if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
							UnsetConstAddr(iList, iins->memDefMin,
									iins->memDefSize);
						}
						regsTainted &= ~(UpToAllRegs(regsDef));
						regsTainted |= nonMemRegUsed;
					}
				}
				if (nonMemRegUsed == 0 && regsDef) {
					regsTainted &= ~(UpToAllRegs(regsDef));
				}
				if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)
						&& MemLocIsConstant(iList, iins->memDefMin,
								iins->memDefSize) && opcode != UD_Ixchg) {
					UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
				}
				if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)
						&& !(opcode == UD_Ipopad || opcode == UD_Ixchg)) {
					SetAddrAsConst(iList, iins->memUsedMin,
							iins->memUsedSize);
				}
				if (flagsDef || flagsUsed) {
					flagsTainted |= flagsUsed;
					flagsTainted &= ~flagsDef;
				}
			} else {
				if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)
						&& MemLocIsConstant(iList, iins->memDefMin,
								iins->memDefSize)) {
					UnsetConstAddr(iList, iins->memDefMin, iins->memDefSize);
				}
				if ((regsDef & regsTainted) != 0) {
					regsTainted &= ~(UpToAllRegs(regsDef));
				}
				if ((flagsDef & flagsTainted) != 0) {
					flagsTainted &= ~flagsDef;
				}
			}
		}
	}
	return;
}

void ApplyFlagsToMemBits(BitDataDep *dataDep, int* vars){
	//dataDep->regBits[regIndex] = PSW_ALL;
	//int i = 0;

	vars[31 - PSW_CF_INDEX] = PSW_CF;
	vars[31 - PSW_PF_INDEX] = PSW_PF;
	vars[31 - PSW_AF_INDEX] = PSW_AF;
	vars[31 - PSW_ZF_INDEX] = PSW_ZF;
	vars[31 - PSW_SF_INDEX] = PSW_SF;
	vars[31 - PSW_TF_INDEX] = DATAFLAG_CONST; //PSW_TF;

	vars[31 - PSW_IF_INDEX] = PSW_IF;
	vars[31 - PSW_DF_INDEX] = PSW_DF;
	vars[31 - PSW_OF_INDEX] = PSW_OF;
}


void InitializeBitVars(BitDataDep *dataDep){
	dataDep->taintedRegs = dataDep->taintedFlags = 0x0;
	dataDep->flags = new TaintBlock(4, DATAFLAG_CONST, 0);
	return;

}


bool HandleArithmeticBitsWithMapping(InstrList *iList, Instruction *iins, BitDataDep *dataDep){
    ud_mnemonic_code_t opcode = iins->uInstr.mnemonic;
    ud_operand_t *op0 = &(iins->uInstr.operand[0]);
    ud_operand_t *op1 = &(iins->uInstr.operand[1]);
    bool taintsClear = true;

    TaintBlock *op0Val=NULL, *op1Val=NULL;
    uint op1ConstVal = 0, op0ConstVal = 0;


    bool op0Mem, op1Mem, op1Const = false, op0Const = false;
    op0Mem = (op0->type == UD_OP_MEM) ? true : false;
    op1Mem = (op1->type == UD_OP_MEM) ? true : false;


    if(op0Mem){
        op0Val = GetMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
        op0ConstVal = GetMemoryLocValue(iList, iins->memDefMin, iins->memDefSize);
        if(!MemoryLocIsTainted(iList, iins->memDefMin, iins->memDefSize)){
        	op0Const = true;
        }
    } else if(op0->type == UD_OP_REG){
        op0Val = GetRegBitVars(dataDep, UDRegisterToBitVector(op0->base));
        op0ConstVal = GetRegisterValue(iins, op0->base);
        if(!(dataDep->taintedRegs & UDRegisterToBitVector(op0->base))){
        	op0Const = true;
        }
    }


    if(op1Mem){
    	op1Val = GetMemAddrTaint(iList, iins->memUsedMin, iins->memUsedSize);
        op1ConstVal = GetMemoryLocValue(iList, iins->memUsedMin, iins->memUsedSize);
    	if(!MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize)){
    		op1Const = true;
    	}
    } else if(op1->type == UD_OP_REG){
            op1Val = GetRegBitVars(dataDep, UDRegisterToBitVector(op1->base));
            op1ConstVal = GetRegisterValue(iins, op1->base);
            if(!(dataDep->taintedRegs & UDRegisterToBitVector(op1->base))){
            	op1Const = true;
            }
    } else if(op1->type != UD_NONE){ //constant
    	op1Const = true;
		op1ConstVal = op1->lval.sdword;         // the const value itself if the operand is constant
    }
    switch(opcode){

    case UD_Iadc:
    	if(op1Const){
    		op0Val->Adc(op0Const, op1ConstVal, *dataDep->flags);
    	} else if (op0Const) {
    		(*op0Val) = op1Val->Adc(*op0Val, op0ConstVal, op1ConstVal, *dataDep->flags);
    	} else {
    		op0Val->Adc(*op1Val, op0ConstVal, op1ConstVal, *dataDep->flags);
    	}

    	break;
    case UD_Inot:
    	(*op0Val) = !(*op0Val);
    	return op0Val->Tainted();
    case UD_Ishr:

    	// not really possible if the second
    	// operand is not constant
    	(*op0Val) >> op1ConstVal;
    	return op0Val->Tainted();
    case UD_Ishl:
     	(*op0Val) << op1ConstVal;

    	return op0Val->Tainted();
    case UD_Iror:
    case UD_Ircr:

    	(*op0Val).Ror(op1ConstVal);
        break;
    case UD_Irol:
    case UD_Ircl:
    	(*op0Val).Rol(op1ConstVal);
    	if(!(*op0Val)[0].Tainted()){
    		dataDep->taintedFlags &= ~PSW_CF;
    		iins->flagsDef &= ~PSW_CF;
    	}
        break;
    case UD_Iadd:
    case UD_Isub:
    case UD_Imul:
    	if(!op1Val){
    		int size = op0Val->Size();
    		TaintBlock *tmp=NULL;
    		if(size == 1){
    			tmp = GetRegBitVars(dataDep, LYNX_AL);
    		} else if(size == 2){
    			tmp = GetRegBitVars(dataDep, LYNX_AX);
    		} else {
    			tmp = GetRegBitVars(dataDep, LYNX_EAX);
    		}
    		op1Val = op0Val;
    		op0Val = tmp;
    	}

    	if(op1Const){
    		op0Val->Add(op0ConstVal, op1ConstVal);
    	} else {
    		op0Val->Add(*op1Val, op0ConstVal, op1ConstVal);
    	}
    	return op0Val->Tainted();
    case UD_Ixor:
    	if(op1Const){
    		(*op0Val) =  (*op0Val) ^ op1ConstVal;
    	} else if (op0Const) {
    		(*op0Val) = (*op1Val) ^ op0ConstVal;
    	} else {
    		op0Val->Xor(*op1Val, op0ConstVal, op1ConstVal);
    	}
    	return  op0Val->Tainted();
    case UD_Iand:
    	if(op1Const){
    		(*op0Val) =  (*op0Val) & op1ConstVal;
    	} else if (op0Const) {
    		(*op0Val) = (*op1Val) & op0ConstVal;
    	} else {
    		(*op0Val) &= (*op1Val);
    	}
    	return op0Val->Tainted();
    case UD_Ior:
    	if(op1Const){
    		(*op0Val) =  (*op0Val) | op1ConstVal;
    	} else if (op0Const) {
    		(*op0Val) = (*op1Val) | op0ConstVal;
    	} else {
    		(*op0Val) = (*op0Val) | (*op1Val);
    	}
    	return op0Val->Tainted();
    default:
        //fprintf(stderr, "[HandleArithmeticBits] Warning: arithmetic opcode %s not handled: %d\n", ud_mnemonics_str[opcode], iins->order);
        break;

    }
    return taintsClear;
}

uint64_t lastPushfd = 0;
void ForwardCombinedBitTaintPropagation(InstrList *iList, int StartIndex, uint64_t n) {
    Instruction *iins, *prev = NULL;
    int i;
    ud_operand_t *op0, *op1;
    enum ud_mnemonic_code opcode;
    REGISTER_BITVECTOR regsDef, regsUsed;
    uint64_t numObfInstrs = 0;

    REGISTER_BITVECTOR flagsDef, flagsUsed, nonMemRegUsed;
    BitDataDep *dataBitLevel = new BitDataDep;
    TaintBlock *op0Val=NULL, *op1Val=NULL;
    InitializeBitVars(dataBitLevel);
    bool forwardTainted = false;


    ClearConstantPages(iList, 0);

    i = StartIndex;
    SetCurrentInstr(iList, i);
    if (n == 0)
        n = -1;

    while ((iins = FetchNextInstr(iList)) != NULL && (n--)) {
        if (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)) {
            numObfInstrs++;
        }
        if (prev
                && (INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE)
                        && !INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE))) {
            INSTR_SET_FLAG(prev, INSTR_USED_AS_CALL);
            SaveInstrChange(iList, prev);
        }

        if(iins->order >= iList->startMarker && iins->order <= iList->startMarker + 2){
            continue;
        }


        //fprintf(stderr, "%d\n", iins->order);
        opcode = iins->uInstr.mnemonic;
        regsDef = RegDef(iList, iins);
        regsUsed = RegUsed(iList, iins);
        nonMemRegUsed = NonMemRegUsed(iList, iins);


        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);


        if (opcode == UD_Ixor && InstrXorAsMov(iins)) {
            ud_t *udins_new = UDCreateMovImmToReg(op0->base, 0);

            CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
            SaveInstrChange(iList, iins);

            opcode = iins->uInstr.mnemonic;
            regsDef = (RegDef(iList, iins));
            regsUsed = (RegUsed(iList, iins));
            op0 = &(iins->uInstr.operand[0]);
            op1 = &(iins->uInstr.operand[1]);
        }

        /* Transform 3 xors to an xchg instruction */
        if(opcode == UD_Ixor){
            Instruction *tmp = NULL;
            if((tmp = ThreeXorsToXchg(iList, iins)) != NULL){
                regsDef = RegDef(iList, iins);
                regsUsed = (RegUsed(iList, iins));

                opcode = iins->uInstr.mnemonic;
                op0 = &(iins->uInstr.operand[0]);
                op1 = &(iins->uInstr.operand[1]);
            } else {
                SetCurrentInstr(iList, iins->next);
            }
        }

        // transform xchg op1,op2, not op2, xchg op1,op2 to not op1!
        if (opcode == UD_Ixchg) {
            Instruction *tmp = NULL;
            if((tmp = TransformXchgNotXchgToNot(iList, iins)) != NULL){
                regsDef = (RegDef(iList, iins));
                regsUsed = (RegUsed(iList, iins));

                opcode = iins->uInstr.mnemonic;
                op0 = &(iins->uInstr.operand[0]);
                op1 = &(iins->uInstr.operand[1]);
            } else {
                SetCurrentInstr(iList, iins->next);
            }
        }


        if ((IsPush(iins) || IsPop(iins))) {
            if (UDOpIsRegister(op0) == UD_R_ESP) {
                if (iins->uInstr.mnemonic == UD_Ipop)
                    regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
                else if (iins->uInstr.mnemonic == UD_Ipush)
                    regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
            } else {
                regsDef = (regsDef & ~(LYNX_ESP | LYNX_SP));
                regsUsed = (regsUsed & ~(LYNX_ESP | LYNX_SP));
            }
        }

        // remove tainted registers when returning from an API except EAX!
        if(prev && (INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) && !INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE))){
        	dataBitLevel->taintedRegs = MaskDefinedReg(LYNX_EAX);
        	ApplyRegTaintWithMask(dataBitLevel, 0xffffffff, MaskDefinedReg(LYNX_EAX), iins->order);
        }

        // remove tainted registers when entering an API!
		if(prev && (!INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE) && INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE))){
			dataBitLevel->taintedRegs = 0;
		}



        forwardTainted = false;
        opcode = iins->uInstr.mnemonic;
        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);

        flagsDef = FlagsDef(iins);
        flagsUsed = FlagsUsed(iins);

        if((IsCondJump(iins) && !((flagsUsed & dataBitLevel->taintedFlags) == flagsUsed))){
        	INSTR_SET_FLAG(iins, INSTR_IS_DELETED);
        	SaveInstrChange(iList, iins);
        	continue;
        }
        if(opcode == UD_Ior || opcode == UD_Iand || opcode == UD_Itest || opcode == UD_Ixor){
            flagsDef &= ~(PSW_CF | PSW_OF);
        }

        if (INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)) {
            if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
                //memTainted = MergeWriteSets(memTainted, memDef);
            	TaintBlock taint;
            	for(i=0; i < iins->memDefSize; i++){
            		taint.AddByte(new TByte(0xffffffffff, iins->order));
            	}
            	AddMemAddrTaint(iList, iins->memDefMin, &taint);
                //TaintMemoryLoc(iList, iins->memDefMin, iins->memDefSize, 0xffffffff);
            }
            if (regsDef != 0){
            	dataBitLevel->taintedRegs |= MaskDefinedReg(regsDef);

                ApplyRegTaintWithMask(dataBitLevel, 0xffffffff, regsDef, iins->order);
                //taints->regsMask[floor_log2(regsDef)] = 0xffffffff;
            }
        } else {
        	if (!(opcode == UD_Ixor && op0->base == iins->uInstr.operand[1].base)) {
				if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)
						&& (MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize))) {
					INSTR_SET_FLAG(iins, INSTR_FORWARD_TAINTED);
				}
				if (((regsUsed) & dataBitLevel->taintedRegs) != 0) {
					INSTR_SET_FLAG(iins, INSTR_FORWARD_TAINTED);
				}
				if(flagsUsed & dataBitLevel->taintedFlags){
					INSTR_SET_FLAG(iins, INSTR_FORWARD_TAINTED);
				}

			}
			if (INSTR_HAS_FLAG(iins, INSTR_FORWARD_TAINTED)
					&& !(opcode == UD_Ixor && op0->base == iins->uInstr.operand[1].base)) {
				SaveInstrChange(iList, iins);


				if(opcode == UD_Ixchg){
					//TaintBlock *op;
					bool op0Tainted = false;
					bool op1Tainted = false;
					int opSize=0;
					if(op0->type == UD_OP_REG){
						op0Val = GetRegBitVars(dataBitLevel, UDRegisterToBitVector(op0->base));
						opSize = UDRegSize(op0->base);
						if(UDRegisterToBitVector(op0->base) & dataBitLevel->taintedRegs){
							op0Tainted = true;
						}
					} else {
						op0Val = GetMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
						op0Tainted = MemoryLocIsTainted(iList, iins->memDefMin, iins->memDefSize);
						opSize = iins->memDefSize;
					}
					if(op1->type == UD_OP_REG){
						op1Val = GetRegBitVars(dataBitLevel, UDRegisterToBitVector(op1->base));
						if(UDRegisterToBitVector(op1->base) & dataBitLevel->taintedRegs){
							op1Tainted = true;
						}
					} else {
						op1Val = GetMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
						op1Tainted = MemoryLocIsTainted(iList, iins->memDefMin, iins->memDefSize);
					}

					if(op1Tainted || op0Tainted){

						TaintBlock tmp(opSize, 0, 0);
						tmp = (*op0Val);
						(*op0Val) = (*op1Val);
						(*op1Val) = tmp;


						if (op0->type == UD_OP_MEM){
							if(op1Tainted){
								SetMemoryLocTainted(iList, iins->memDefMin, iins->memDefSize);
							} else {
								ClearMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
							}
						}
						if(op1->type == UD_OP_MEM){
							if(op0Tainted){
								SetMemoryLocTainted(iList, iins->memDefMin, iins->memDefSize);
							} else {
								ClearMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
							}
						}

						if(op0->type == UD_OP_REG){
							if(op1Tainted){
								dataBitLevel->taintedRegs |= MaskDefinedReg(UDRegisterToBitVector(op0->base));
							} else {
								dataBitLevel->taintedRegs &= ~MaskDefinedReg(UDRegisterToBitVector(op0->base));
							}
						}
						if(op1->type == UD_OP_REG){
							if(op0Tainted){
								dataBitLevel->taintedRegs |= MaskDefinedReg(UDRegisterToBitVector(op1->base));
							} else {
								dataBitLevel->taintedRegs &= ~MaskDefinedReg(UDRegisterToBitVector(op1->base));
							}
						}
					}
					prev = iins;
					continue;

				} else if (opcode == UD_Ipushad){
                    int taintRegs = (dataBitLevel->taintedRegs & 0xff);
                    int k;
                    for (k = 0; k < 8 && (taintRegs); k++) {
                        if ((taintRegs & 0x1)) {
                            Range *r = RangeCreate(iins->esp - (k + 1) * 4, iins->esp - (k) * 4 - 1);
                            TaintBlock *b = GetRegBitVars(dataBitLevel, (taintRegs & 0x1));
                            AddMemAddrTaint(iList, r->min, b);
                            forwardTainted = true;
                        }
                        taintRegs = (taintRegs >> 1);
                    }
                } else if(opcode == UD_Ipopad){
                    int k;
                    for (k = 0; k < 8; k++) {
                    	if(k == 3) continue; 		// skip esp!
                        Range *r = RangeCreate(iins->esp + (k) * 4,
                                iins->esp + (k + 1) * 4 - 1);
                        if (MemoryLocIsTainted(iList, r->min, r->max - r->min + 1)) {
                        	forwardTainted = true;
                        	dataBitLevel->taintedRegs |= MaskDefinedReg((0x1 << (7 - k)));
                        	TaintBlock *regTaint = GetRegBitVars(dataBitLevel, (0x1 << (7 - k)));
                        	TaintBlock *memTaint = GetMemAddrTaint(iList, r->min, r->max - r->min + 1);
                        	(*regTaint) = (*memTaint);
                            //taints->regsMask[k*4] = GetMemoryLocTaint(iList, r->min, 4);
                        }
                    }

                } else if(opcode == UD_Ipopfd){
                	TaintBlock *flags;
                	flags = GetMemAddrTaint(iList, iins->memUsedMin, iins->memUsedSize);
                	dataBitLevel->taintedFlags = 0;
					for (i = 0; i < 32; i++){
						if((*flags)[i].Tainted()){
							dataBitLevel->taintedFlags |=  (1 << i);
							forwardTainted = true;
						}
					}
					(*dataBitLevel->flags) = (*flags);
                } else if (opcode == UD_Ipushfd){

					TaintBlock *write;
					write = GetMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
					forwardTainted = true;
					for (i = 0; i < 32; i++){
						if(dataBitLevel->taintedFlags & (1 << i)){
							(*write)[i] = (*dataBitLevel->flags)[i];
							forwardTainted = true;
						} else {
							(*write)[i].tag = DATAFLAG_CONST;
						}
					}
					if(forwardTainted){
						SetMemoryLocTainted(iList, iins->memDefMin, iins->memDefSize);
					}

                } else if((opcode == UD_Itest || opcode == UD_Icmp || opcode == UD_Icmpxchg)
                		&& (((op0->type == UD_OP_IMM || (op0->type == UD_OP_REG && (dataBitLevel->taintedRegs & UDRegisterToBitVector(op0->base)) == 0) || (op0->type==UD_OP_MEM && !MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize)))
                		|| (op1->type == UD_OP_IMM  || (op1->type == UD_OP_REG && (dataBitLevel->taintedRegs & UDRegisterToBitVector(op1->base)) == 0) || (op1->type==UD_OP_MEM && !MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize)))))){
                	TaintBlock *t = NULL;
                	uint c = 0, tMask=0;
                	int opSize=0;
                	if(op0->type == UD_OP_REG){
                		opSize = UDRegSize(op0->base);
                		if((dataBitLevel->taintedRegs & UDRegisterToBitVector(op0->base))!=0){
                			t = GetRegBitVars(dataBitLevel, UDRegisterToBitVector(op0->base));
                		} else {
                			c = GetRegisterValue(iins, op0->base);
                		}
                	} else if (op0->type == UD_OP_MEM){
                		opSize = iins->memUsedSize;
                		if(MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize)){
                			t = GetMemAddrTaint(iList, iins->memUsedMin, iins->memUsedSize);
                		} else {
                			c = GetMemoryLocValue(iList, iins->memUsedMin, iins->memUsedSize);
                		}
                	} else {
                		opSize = op0->size;
                		c = op0->lval.sdword;
                	}

                	if(op1->type == UD_OP_REG){
                		if((dataBitLevel->taintedRegs & UDRegisterToBitVector(op1->base))!=0){
                			t = GetRegBitVars(dataBitLevel, UDRegisterToBitVector(op1->base));
                		} else {
                			c = GetRegisterValue(iins, op1->base);
                		}
                	} else if (op1->type == UD_OP_MEM){
                		if(MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize)){
                			t = GetMemAddrTaint(iList, iins->memUsedMin, iins->memUsedSize);
                		} else {
                			c = GetMemoryLocValue(iList, iins->memUsedMin, iins->memUsedSize);
                		}
                	} else {
                		c = op1->lval.sdword;
                	}
                	if(t){
						TaintBlock tmp1(opSize, 0, 0);
						tmp1 = (*t);
						for(i=0; i < opSize*8; i++){
							if(tmp1[i].Tainted()){
								tMask |= (1 << i);
							}
						}
						if(opcode == UD_Itest){
							forwardTainted = (tmp1&c).Tainted();
						} else {
							forwardTainted = (tMask > c);
						}
                	} else {
                		forwardTainted = false;
                	}

				} else if(OpcodeIsArith(opcode) && ((regsUsed & dataBitLevel->taintedRegs) ||
						MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize) || (flagsUsed & dataBitLevel->taintedFlags))){
					forwardTainted = HandleArithmeticBitsWithMapping(iList, iins, dataBitLevel);
					if(!forwardTainted){
						INSTR_SET_FLAG(iins, INSTR_TAINT_CLEARED);
						Instruction *nextTmp = GetInstruction(iList, iins->next);
						if(op0->type == UD_OP_MEM){
							int val = GetMemoryLocValue(iList, iins->memDefMin, iins->memDefSize);
							ud_t *udnew = UDCreateMovImmToOffset(iins->memDefMin, val, iins->memDefSize);
							CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udnew, 0);
							free(udnew);
						} else {
							int val = GetRegisterValue(nextTmp, op0->base);
							ud_t *udnew = UDCreateMovImmToReg(op0->base, val);
							CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udnew, 0);
							free(udnew);
						}
						SaveInstrChange(iList, iins);
					}
				} else if (opcode == UD_Ilea && (regsUsed & dataBitLevel->taintedRegs) != 0){
					forwardTainted = true;
				} else if (opcode == UD_Ibtc){		// sets cf based on the first operand
					if(!(dataBitLevel->taintedRegs & UDRegisterToBitVector(op0->base))){
						dataBitLevel->taintedFlags &= ~PSW_CF;
						flagsDef &= ~PSW_CF;
					}
					forwardTainted = true;
				} else if (opcode == UD_Ibswap){

                } else if (IsStringOperation(iins)){

                } else if (opcode == UD_Icwd || opcode == UD_Icdq){

                } else if (InstrIsSetOnCondition(opcode)){
                	TaintBlock *bitVars = GetRegBitVars(dataBitLevel, regsDef);
                	for(uint i=0; i < BitvectorRegisterSize(regsDef)*8; i++){
                		if(dataBitLevel->taintedFlags & (1 << i)){
                			(*bitVars)[i].tag = (1 << i);
                			(*bitVars)[i].src_instr = iins->order;
                			forwardTainted = true;
                		}
                	}
                } else if (opcode == UD_Imovzx){
					op0Val = GetRegBitVars(dataBitLevel, UDRegisterToBitVector(op0->base));
					if (iins->uInstr.operand[1].type == UD_OP_REG){
						op1Val = GetRegBitVars(dataBitLevel, UDRegisterToBitVector(op1->base));
					} else if(iins->memUsedSize) {
						op1Val = GetMemAddrTaint(iList, iins->memUsedMin, iins->memUsedSize);
					}
					(*op0Val) = (*op1Val);
					forwardTainted = op0Val->Tainted();
				} else {
					// extract the taint by looking at things an instruction uses

					//TaintBlock regVars;

					if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) || regsDef){
						TaintBlock *source = NULL;
						if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM)){
							if(MemoryLocIsTainted(iList, iins->memUsedMin, iins->memUsedSize)){
								source = GetMemAddrTaint(iList, iins->memUsedMin, iins->memUsedSize);
							} else if(MemRegUsed(iins) & dataBitLevel->taintedRegs){
								source = new TaintBlock(UDRegSize(op0->base), 0xffffffff, iins->order);
							}
						}
						if(nonMemRegUsed & dataBitLevel->taintedRegs){
							source = GetRegBitVars(dataBitLevel, nonMemRegUsed&dataBitLevel->taintedRegs);
						}

						if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && source){
							TaintBlock *memWriteVars;
							memWriteVars = GetMemAddrTaint(iList, iins->memDefMin, iins->memDefSize);
							(*memWriteVars) = (*source);
							if((forwardTainted = memWriteVars->Tainted())){
								SetMemoryLocTainted(iList, iins->memDefMin, iins->memDefSize);
							}

						} else if (regsDef && source){
							assert(source);
							TaintBlock *regWriteVars = GetRegBitVars(dataBitLevel, regsDef) ;
							(*regWriteVars) = (*source);
							forwardTainted = regWriteVars->Tainted();
							if(!forwardTainted){
								int val = 0;
								if(op0->type == UD_OP_REG){
									val = GetRegisterValue(iins, op1->base);
								} else if(op1->type == UD_OP_MEM){
									val = GetMemoryLocValue(iList, iins->memUsedMin, iins->memUsedSize);
								}
								ud_t *newud = UDCreateMovImmToReg(BitVectorToUDRegister(regsDef), val);
								CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), newud, 0);
							}
						}
					} else {			// writes flags
						if(flagsDef){
							forwardTainted = true;
							for(uint i=0; i<32; i++){
								uint mask = (1 << i)&flagsDef;
								if(mask){
							 		(*dataBitLevel->flags)[i] = *(new Taint(mask, iins->order));
								}
							}
							dataBitLevel->taintedFlags |= flagsDef;

							forwardTainted = true;
						}
					}

					if (flagsUsed & dataBitLevel->taintedFlags){
						forwardTainted = true;
					}
					if(IsControlFlowInstr(iins)){
						forwardTainted = true;
					}
				}
			}

			if (forwardTainted && opcode != UD_Ipopad) {
				//INSTR_SET_FLAG(iins, INSTR_FLAG_FW_TAINED);
				INSTR_SET_FLAG(iins, INSTR_FORWARD_TAINTED);
				SaveInstrChange(iList, iins);
				forwardTainted = false;
				dataBitLevel->taintedRegs |= UpToAllRegs(MaskDefinedReg(regsDef));

				if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)){
					SetMemoryLocTainted(iList, iins->memDefMin, iins->memDefSize);
				}

				if(flagsDef && opcode != UD_Ipopfd){
					dataBitLevel->taintedFlags |= iins->flagsDef;
				}

				if(IsCondJump(iins)){
					/*
					 * exploit the fact that after a conditional jump, used
					 * flags by conditional are known (e.g., a jnz after a jz
					 * acts like an opaque conditional!)
					 */
					dataBitLevel->taintedFlags &= ~flagsUsed;

				}
			} else {
				dataBitLevel->taintedRegs &= (~MaskDefinedReg(regsDef) & ~UpToAllRegs(regsDef));
				INSTR_REMOVE_FLAG(iins, INSTR_FORWARD_TAINTED);
				SaveInstrChange(iList, iins);
				if(INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && MemoryLocIsTainted(iList, iins->memDefMin, iins->memDefSize)){
					UntaintMemoryLoc(iList, iins->memDefMin, iins->memDefSize);
				}
				if(flagsDef & dataBitLevel->taintedFlags){
					dataBitLevel->taintedFlags &= ~flagsDef;
				}
			}
        }
        prev = iins;
    }
    return;
}


void TaintAnalysis(InstrList *iList, int method, int doBackward) {
	int forwardStart = 0;
	uint64_t backwardStart = iList->lastBaseModuelInstr;

	ForwardTaintAnnotation(iList);

	if(doBackward){
		BackwardTaintAnnotationNew(iList); // initialize starting point of taint analysis
	}
    ForwardCombinedBitTaintPropagation(iList, forwardStart, 0);

	if(doBackward){
	    BackwardTaintPropagationNew(iList, backwardStart);
	}
}
