/*
 *
 * File: instr.c
 * functions related to instructions, their manipulation, or
 *      their characteristics
 * Author: Babak Yadegari, Kevin Coogan, Saumya Debray
 */

/* 
 * Copyright 2015 Arizona Board of Regents on behalf of the University of Arizona.
 */


#include <fcntl.h>
#include <sys/mman.h>
#include <limits.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include "lynx_types.h"
#include "instr.h"
#include "read.h"
#include "al_helper.h"
#include "print.h"
#include "hash.h"
#include "mem_maps.h"
#include "flags.h"

extern int clear_insn(register struct ud* u);
extern int get_prefixes(struct ud* u);



/*
 * BitvectorRegisterSize
 * returns the register size of the argument
 * argument is of type lynx registers (not UD)
 */
uint BitvectorRegisterSize(REGISTER_BITVECTOR reg) {
    if (reg >= LYNX_EAX && reg <= LYNX_EDI) {
        return 4;
    } else if (reg >= LYNX_AX && reg <= LYNX_DI) {
        return 2;
    } else {
        return 1;
    }
    fprintf(stderr, "[BitvectorRegisterSize]: Warning: Unknown register: %d!\n", reg);
    return 0;
}


int FindChangedRegs(InstrList *iList, Instruction *iins1) {
    Instruction *iins2 = FetchNextInstr(iList);
    int changedReg = UD_NONE;
    if (iins1->eax != iins2->eax)
        changedReg |= LYNX_EAX;
    if (iins1->ecx != iins2->ecx)
        changedReg |= LYNX_ECX;
    if (iins1->edx != iins2->edx)
        changedReg |= LYNX_EDX;
    if (iins1->ebx != iins2->ebx)
        changedReg |= LYNX_EBX;
    if (iins1->esi != iins2->esi)
        changedReg |= LYNX_ESI;
    if (iins1->edi != iins2->edi)
        changedReg |= LYNX_EDI;
    if (iins1->esp != iins2->esp)
        changedReg |= LYNX_ESP;
    if (iins1->ebp != iins2->ebp)
        changedReg |= LYNX_EBP;
    SetCurrentInstr(iList, iins2->order);
    free(iins2);
    return changedReg;

}

/* adding a header to the .instrFile
 */
extern bool TraceIsMultiThreaded;
extern ArrayList* traceThreads;
uint64_t numInstrs; // global numInstrs for sharing between threads

bool InstrIsIndirectJump(Instruction *instr){
    assert(instr);
    if((instr->uInstr.mnemonic == UD_Ijmp && instr->uInstr.operand[0].type != UD_OP_JIMM) || instr->uInstr.mnemonic == UD_Iret){
        return true;
    }
    return false;

}

ins_structure *InstrNeedsToBeModified(InstrList *iList, Instruction *iins){
	ud_t *udins_new = NULL;
	if(iins->uInstr.mnemonic == UD_Ixadd || iins->uInstr.mnemonic == UD_Icmpxchg){
		Instruction *tmp = GetInstruction(iList, iins->next);
		int val = GetRegisterValue(tmp, iins->uInstr.operand[1].base);
		udins_new = UDCreateMovImmToReg(iins->uInstr.operand[1].base, val);
	} else if(iins->uInstr.bytes[0] == 0xa9){
		udins_new = UDCreateNop();
	} else if(InstrIsControlTransfer(iList, iins)){
		Instruction *tmp = GetInstruction(iList, iins->next);
		int diff = iins->esp - tmp->esp;			// for call and ret instrs!
		if(iins->uInstr.mnemonic == UD_Icall){
			udins_new = UDCreatePushImm(UD_R_EAX, iins->addr+iins->uInstr.length);
		}
		else if (diff == 0){
			udins_new = UDCreateNop();
		}
		else if(diff > 0){
			udins_new = UDCreateMovImmToReg(UD_R_ESP, tmp->esp);
		} else{
		    udins_new = UDCreateMovImmToReg(UD_R_ESP, tmp->esp);
		}
	}

	if(INSTR_HAS_FLAG(iins, INSTR_READS_MEM) && MemoryLocationIsLive(iList, iins->memUsedMin, iins->memUsedSize)){
	    REGISTER_BITVECTOR regdef = RegDef(iList, iins);
	    if(regdef){
	        Instruction *tmp = GetInstruction(iList, iins->order + 1);
	        int val = GetRegisterValue(tmp, BitVectorToUDRegister(regdef));
	        udins_new = UDCreateMovImmToReg(BitVectorToUDRegister(regdef), val);
	    }
	}

	if(udins_new != NULL){
		ins_structure *tmp_ins = malloc(sizeof(ins_structure));
		CopyUdins2InsStr(tmp_ins, udins_new);
		free(udins_new);
		return tmp_ins;
	}
	return NULL;
}


/*
 * InstrIsSetOnCondition
 * returns true if opcode is one of
 * set on condition instructions
 */

bool InstrIsSetOnCondition(enum ud_mnemonic_code opcode){
	switch (opcode){
		case UD_Iseta:
		case UD_Isetb:
		case UD_Isetbe:
		case UD_Isetg:
		case UD_Isetge:
		case UD_Isetl:
		case UD_Isetle:
		case UD_Isetnb:
		case UD_Isetno:
		case UD_Isetnp:
		case UD_Isetns:
		case UD_Isetnz:
		case UD_Iseto:
		case UD_Isetp:
		case UD_Isets:
		case UD_Isetz:
			return true;
		default:
			return false;
	}
	return false;
}

void PrintFileHedear(InstrFileHeader h){
	fprintf(stderr, "h.TraceIsMultiThreaded: %d\n", h.TraceIsMultiThreaded);
	fprintf(stderr, "h.LastInstrInOriginalTrace: %" PRIu64 "\n", h.LastInstrInOriginalTrace);
	fprintf(stderr, "h.NumInstrs: %" PRIu64 "\n", h.NumInstrs);
	fprintf(stderr, "h.NumThreads: %d\n", h.NumThreads);
	fprintf(stderr, "h.RangeInfoSize: %zu\n", h.RangeInfoSize);
	fprintf(stderr, "h.lastBaseModule: %" PRIu64 "\n", h.lastBaseModule );
	//fprintf(stderr, "%lld\n", );

}

extern char *baseName;
void WriteInstrFileHeader(InstrList *iList, FILE *file, bool mmap) {
    InstrFileHeader fileHeader;
    Thread *currThread;
    ThreadRange *currRange;
    int i, j; size_t headerSize = 0;

    if(file == NULL){
    	return;
    }

    //	uint64_t size = ftello(instrFile) - InstrFileHeaderSize;
    /* Write thread information first
     */
    if (USE_MMAP && mmap && TraceIsMultiThreaded) {
        fileHeader.RangeInfoOffset = (uint64_t)InstrFileHeaderSize +
                (uint64_t)iList->numInstrs*sizeof(Instruction);

        for (i = 0; i < al_size(traceThreads); i++) {
            currThread = al_get(traceThreads, i);
            memcpy((void *)((uint64_t)iList->iMMap + (uint64_t)fileHeader.RangeInfoOffset + headerSize),
                    currThread, sizeof(Thread));
            headerSize += sizeof(Thread);
            for (j = 0; j < al_size(currThread->ranges); j++) {
                currRange = al_get(currThread->ranges, j);
                memcpy((void *)((uint64_t)iList->iMMap + fileHeader.RangeInfoOffset + headerSize),
                        currRange, sizeof(ThreadRange));
                headerSize += sizeof(ThreadRange);
            }
        }
    } else if (TraceIsMultiThreaded) {
        fseeko(file, 0, SEEK_END);
        fileHeader.RangeInfoOffset = ftello(file);
        for (i = 0; i < al_size(traceThreads); i++) {
            currThread = al_get(traceThreads, i);
            fwrite(currThread, sizeof(Thread), 1, file);
            headerSize += sizeof(Thread);
            for (j = 0; j < al_size(currThread->ranges); j++) {
                currRange = al_get(currThread->ranges, j);
                fwrite(currRange, sizeof(ThreadRange), 1, file);
                headerSize += sizeof(ThreadRange);
            }
        }
    }
    strcpy(fileHeader.baseModuleName, baseName);
    fileHeader.TraceIsMultiThreaded = TraceIsMultiThreaded;
    fileHeader.LastInstrInOriginalTrace = iList->last;
    fileHeader.NumInstrs = numInstrs; // set to global numInstr variable since iList might be shared between threads!
    fileHeader.Loop = iList->loop;
    fileHeader.NumThreads = (TraceIsMultiThreaded ? al_size(traceThreads) : 0);
    fileHeader.RangeInfoSize = headerSize;
    fileHeader.lastBaseModule = iList->lastBaseModuelInstr;
    fileHeader.stackButtom = iList->stackBottom;
    fileHeader.nextAvailAddr = iList->nextAvailabelAddr;
    fileHeader.startMarker = iList->startMarker;
    fileHeader.endMarker = iList->endMarker;

    //PrintFileHedear(fileHeader);

    if(USE_MMAP && mmap) {
        //write to the beginning of the mmap itself
        uint64_t fileHdrAddr = (uint64_t)iList->iMMap;
        memcpy((void *)fileHdrAddr, &fileHeader, sizeof(InstrFileHeader));

    } else {
        //write it to the beginning of the file
        fseeko(file, 0, SEEK_SET);
        fwrite(&fileHeader, sizeof(InstrFileHeader), 1, file);
        fprintf(stderr, "%d bytes was written to the InstrFile\n", (int)headerSize);
    }
}
/* write instr to current file pointer
 *
 */
void AddInstrToFile(FILE *f, Instruction *instr) {
    if (!fwrite(instr, sizeof(Instruction), 1, f)) {
        fprintf(stderr, "Could not write instr to file\n");
        exit(0);
    }
    //printf("ftello(f) = %lld\n", (long long) ftello(f));
}

void InitUdis(ud_t *ud_obj) {
    //udis86 set up and initialization
    ud_init(ud_obj);
    ud_set_mode(ud_obj, 32);
    ud_set_syntax(ud_obj, UD_SYN_INTEL);
    ud_set_vendor(ud_obj, UD_VENDOR_INTEL);
}

/*
 * InstrCompare
 * compare two instrs byte to byte
 * return false if not equal.
 */
bool InstrCompare(Instruction *iins1, Instruction *iins2){
	/*ud_operand_t op0, op1, op2;

	if (iins1->uInstr.mnemonic == iins2->uInstr.mnemonic &&
			iins1->uInstr.operand[0].type == iins2->uInstr.operand[0].type &&
			iins1->uInstr.operand[1].type == iins2->uInstr.operand[1].type &&
			iins1->uInstr.operand[2].type == iins2->uInstr.operand[2].type &&

		){
		return false;
	}*/
	int i;
	if(iins1->uInstr.length != iins2->uInstr.length){
		return false;
	}

	for(i = 0; i < iins1->uInstr.length; i++){
		if(iins1->uInstr.bytes[i] != iins2->uInstr.bytes[i]){
			return false;
		}
	}
	return true;
}

Instruction *InstrCreate() {
    Instruction *new = calloc(1, sizeof(Instruction));
    //    new->udInstr = NULL;
    //new->count = 1;
    new->flags = 0;
    //new->newInstrAddedAfter = -1;
    //new->outOfRangeInstr = -1;

    new->regFlagsKnown &= ~FlagsDefKnown;
    new->regFlagsKnown &= ~FlagsUsedKnown;
    new->regFlagsKnown &= ~RegDefKnown;
    new->regFlagsKnown &= ~MemRegUsedKnown;
    new->regFlagsKnown &= ~NonMemRegUsedKnown;
    new->regFlagsKnown &= ~MemDefKnown;
    new->regFlagsKnown &= ~MemUsedKnown;

    return (new);
}


void RInstrDestroy(InstrList *iList, Instruction *instr) {

}
Instruction *InstrClone(Instruction *instr) {
    Instruction *new = (Instruction *) alloc(sizeof(Instruction));
    memcpy(new, instr, sizeof(Instruction));
    return (new);
}

InstrList *InitInstructionCache(InstrList *iList) {
    iList->instrCache = malloc((uint64_t)CACHE_SIZE);
    memset((iList->instrCache), 0, (uint64_t)CACHE_SIZE);
    iList->firstInstr = NO_INSTR;
    iList->lastInstr = NO_INSTR;

    iList->newInstrCache = malloc((uint64_t)CACHE_SIZE);
    memset((iList->newInstrCache), 0, (uint64_t)CACHE_SIZE);
    iList->newFirstInstr = NO_INSTR;
    iList->newLastInstr = NO_INSTR;

    return iList;
}

ud_t *LoadUDisInstrObj(Instruction *instr) {
    ud_t *udins;
    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_pc(udins, instr->addr);
    ud_set_input_buffer(udins, instr->uInstr.bytes, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

void CopyUdins2InsStr(ins_structure *uIns, ud_t *udIns) {
    int i;
    unsigned char *ptr;

    uIns->mnemonic = udIns->mnemonic;

    // copy operands
    for (i = 0; i < 3; i++) {
        memcpy(&(uIns->operand[i]), &(udIns->operand[i]), sizeof(struct ud_operand));
        /*
         * udis86 objects have the unexpected behavior that, if the scale is not
         * explicitly mentioned in the instruction, the scale value in the udis86
         * object is set to 0.  We fix this here.
         */
        if (uIns->operand[i].type == UD_OP_MEM && uIns->operand[i].scale == 0) {
            uIns->operand[i].scale = 1;
        }
    }

    uIns->length = ud_insn_len(udIns);

    // copy instruction bytes
    for (i = 0, ptr = ud_insn_ptr(udIns); i < uIns->length; i++, ptr++) {
        uIns->bytes[i] = *ptr;
    }

    // copy instruction name
    //snprintf(uIns->name, MAX_INSTR_NAME_LEN, "%-51s", ud_insn_asm(udIns));

    uIns->pfx_seg = udIns->pfx_seg;
}

/* FillCacheFromFile
 *  This method is designed to fill a cache of instructions from a file
 *  the cache is, by definition, two pages worth of instructions with
 *  no partial instructions included.
 */
pthread_mutex_t filelock = PTHREAD_MUTEX_INITIALIZER;
InstrList *FillCacheFromFile(InstrList *iList, int index) {
    //let's blank out the old cache first
    if (index <= iList->last) {
        memset((iList->instrCache), 0, CACHE_SIZE);
        //fprintf(stderr, "[FillCacheFromFile]: was called for %d **********************************************************\n", index);
        //calculate what page we need, we will retrieve this page and page + 1
        uint64_t instrSz = sizeof(Instruction);
        uint64_t offset = instrSz * index;
        uint64_t instrPg = offset / PAGE_SIZE;

        //get addr of first page to retrieve
        uint64_t pgAddr = instrPg * PAGE_SIZE; //start addr of page

        //and calculate first full instruction on that page
        iList->firstInstr = pgAddr / instrSz;
        if (pgAddr % instrSz != 0) {
            iList->firstInstr++;
        }

        //Get addr of page after last page to retrieve
        uint64_t endAddr = (instrPg + 2) * PAGE_SIZE;

        //and calculate last full instruction on previous page
        iList->lastInstr = endAddr / instrSz - 1;
        if (iList->lastInstr > iList->last)
            iList->lastInstr = iList->last;
        pthread_mutex_lock(&filelock);
        //read first to last instructions into cache
        int status = fseeko(iList->iListFile, (iList->firstInstr * instrSz) + InstrFileHeaderSize,
                SEEK_SET);
        if (status != 0) {
            perror("FillCacheFromFile");
        }

        if (fread(iList->instrCache, instrSz, iList->lastInstr - iList->firstInstr + 1,
                iList->iListFile) == 0) {
            perror("Failed to read instructions from file");
        }
        pthread_mutex_unlock(&filelock);
    } else {
        memset((iList->newInstrCache), 0, CACHE_SIZE);
        //fprintf(stderr, "[FillCacheFromFile]: was called for %d **********************************************************\n", index);
        //calculate what page we need, we will retrieve this page and page + 1
        uint64_t instrSz = sizeof(Instruction);
        uint64_t offset = instrSz * index;
        uint64_t instrPg = offset / PAGE_SIZE;

        //get addr of first page to retrieve
        uint64_t pgAddr = instrPg * PAGE_SIZE; //start addr of page

        //and calculate first full instruction on that page
        iList->newFirstInstr = pgAddr / instrSz;
        if (pgAddr % instrSz != 0) {
            iList->newFirstInstr++;
        }
        if (iList->newFirstInstr <= iList->last)
            iList->newFirstInstr = iList->last + 1;
        //Get addr of page after last page to retrieve
        uint64_t endAddr = (instrPg + 2) * PAGE_SIZE;

        //and calculate last full instruction on previous page
        iList->newLastInstr = endAddr / instrSz - 1;

        if (iList->newLastInstr >= numInstrs)
            iList->newLastInstr = numInstrs - 1;

        pthread_mutex_lock(&filelock);
        //read first to last instructions into cache
        fseeko(iList->iListFile, (iList->newFirstInstr * instrSz) + InstrFileHeaderSize, SEEK_SET);
        if (fread(iList->newInstrCache, instrSz, iList->newLastInstr - iList->newFirstInstr + 1,
                iList->iListFile) == 0) {
            perror("Failed to read instructions from file");
        }
        pthread_mutex_unlock(&filelock);

    }
    return (iList);
}

size_t WriteInstrToFile(FILE *f, Instruction *instr) {
    assert(instr);
    size_t instrsWritten = 0;
    uint64_t offset = instr->order * sizeof(Instruction) + InstrFileHeaderSize;
    pthread_mutex_lock(&filelock);
    fseeko(f, offset, SEEK_SET);

    instrsWritten = fwrite(instr, sizeof(Instruction), 1, f);

    pthread_mutex_unlock(&filelock);
    return (instrsWritten);
}


/*
 * InstrXorAsMov
 * returns true if instr is xor reg, reg
 * which is equal to mov reg, 0
 */

bool InstrXorAsMov(Instruction *instr){
	if(instr->uInstr.operand[0].type == UD_OP_REG && instr->uInstr.operand[0].type == instr->uInstr.operand[1].type &&
			instr->uInstr.operand[0].base == instr->uInstr.operand[1].base){
		return true;
	}
	return false;
}

/*
 * look at the opcode and determine if instr
 * masks out a part of the operand register
 * (Or the operand could be register which
 * does not make a difference)
 */
bool InstrMasksRegister(Instruction *instr){
    enum ud_mnemonic_code opcode = instr->uInstr.mnemonic;

    switch (opcode){
    case UD_Iand:
    case UD_Ior:
    case UD_Ishr:
    case UD_Ishl:
        return true;
    default:
        return false;
    }

    return false;
}

/*
 * InstrIsaCallToOutputFunction
 * determines if the instr is jump to an output
 * function call, functions are defined bellow
 */

bool InstrIsaCallToOutputFunction(char *str){
    char *output_funcs[] = {"printf", "_printf", "??6?$basic_ostream@DU?$char_traits@D@std", "_puts", "puts", "putchar", NULL};
    int i;
    if(str == NULL){
        return false;
    }
    for(i=0; output_funcs[i] != NULL; i++){
        if(strcmp(str, output_funcs[i]) == 0){
            return true;
        }
    }
    return false;
}

bool InstrIsACallToIgnoreFuncs(Instruction *iins){
    char *output_funcs[] = {"VirtualAlloc", "GlobalAlloc", NULL};
    char *str = GetFunctionNameById(iins->funcName);
    int i;
    if(str == NULL || strcmp(str, "") == 0){
        return true;
    }
    for(i=0; output_funcs[i] != NULL; i++){
        if(strcmp(str, output_funcs[i]) == 0){
            return true;
        }
    }
    return false;
}

int InstrIsaCallToInputFunction(char *str){
    char *input_funcs[] = {"atoi", "strlen", "_strlen", "_atoi", NULL};
    int input_function_offsets[] = {
#if CRISTIAN
    		13
#else
    		/*31*/ 11
#endif
    		, 1, 4, 28, -1};			// distance from the end of function call
    int i;
    if(str == NULL){
        return false;
    }
    for(i=0; input_funcs[i] != NULL; i++){
        if(strcmp(str, input_funcs[i]) == 0){
            return input_function_offsets[i];
        }
    }
    return -1;
}


void MakeInstrAlive(InstrList *iList, Instruction *instr){
	Instruction *prev, *next;


	prev = GetInstruction(iList, instr->origPrev);
	next = GetInstruction(iList, instr->origNext);
	while(INSTR_HAS_FLAG(prev, INSTR_IS_DELETED)){
		prev = GetInstruction(iList, prev->origPrev);
	}
	while(INSTR_HAS_FLAG(next, INSTR_IS_DELETED)){
		next = GetInstruction(iList, next->origNext);
	}



	INSTR_REMOVE_FLAG(instr, INSTR_IS_DELETED);
	if(prev){
	instr->prev = prev->order;
	prev->next = instr->order;
	SaveInstrChange(iList, prev);

	}
	if (next){
	instr->next = next->order;
	next->prev = instr->order;
	SaveInstrChange(iList, next);
	}


	SaveInstrChange(iList, instr);


}

int SaveInstrChange(InstrList *iList, Instruction *instr) {
    assert(instr);


    if (INSTR_HAS_FLAG(instr, INSTR_IS_DELETED) && INSTR_HAS_FLAG(instr, INSTR_USED_AS_CALL)){
    	INSTR_REMOVE_FLAG(instr, INSTR_IS_DELETED);
    	return -1;
    }
    if (instr->next == -1 || instr->prev == -1){
    	return -1;
    }

    if (INSTR_HAS_FLAG(instr, INSTR_IS_DELETED)) {

        if(strlen(GetFunctionNameById(instr->funcName)) && strcmp(GetFunctionNameById(instr->funcName), "unknown") !=0){
            INSTR_REMOVE_FLAG(instr, INSTR_IS_DELETED);
            INSTR_SET_FLAG(instr, INSTR_UNSIMPLIFIABLE);
            return -1;
        }

        Instruction *NextLive = GetInstruction(iList, instr->next);;
        Instruction *PrevLive = GetInstruction(iList, instr->prev);
        if(PrevLive && NextLive){
        	PrevLive->next = NextLive->order;
        	NextLive->prev = PrevLive->order;
        }
    }
    return 0;       // 0 if instruction was removed successfully!
}


int CountInstructionsBetweenTwoInstrs2(InstrList *iList, Instruction *iins, Instruction *jins){
	int count = 0;
	Instruction *tmp;
	tmp = iins;
	do{
		count ++;
		tmp = GetInstruction(iList, tmp->origNext);
	} while(tmp && tmp != jins);
	return count;
}

Instruction *AddInstruction(InstrList *i_list, Instruction *newInstr, int after) {

    Instruction *prev = GetInstruction(i_list, after);
    Instruction *next = GetInstruction(i_list, prev->next);
    uint64_t addr;

    //update number of instructions carefully


    pthread_mutex_lock(&filelock);

    newInstr->order = numInstrs;

    if(next){
    	newInstr->next = newInstr->origNext = next->order;
    	if(i_list->last > after){
    		/*
    		 * in a case an instr after 'after'
    		 * is removed.
    		 */
    	    newInstr->origNext = after+1;
    	}
    	next->prev = next->origPrev = newInstr->order;

    	SaveInstrChange(i_list, next);
    }
    newInstr->prev = newInstr->origPrev = prev->order;
    prev->next = prev->origNext = newInstr->order;

	addr = (uint64_t)i_list->iMMap +
			numInstrs*sizeof(Instruction) + InstrFileHeaderSize;//sizeof(InstrList);
	memcpy((void *)addr, newInstr, sizeof(Instruction));
    numInstrs++; // increment number of instruction only after it's written to file!
    pthread_mutex_unlock(&filelock);


    SaveInstrChange(i_list, prev);
    return (Instruction *)addr;
}

void SetCurrentInstr(InstrList *iList, uint64_t order) {
    iList->currentInstr = order;
}

Instruction *FetchPrevInstr(InstrList *iList) {
    if (iList->currentInstr == -1){
        return NULL;
    }
    uint64_t addr;

    addr = (uint64_t)(iList->iMMap) + iList->currentInstr * sizeof(Instruction) +
    		InstrFileHeaderSize;
    //just use the mmap'd location, set self and AllocCount to be safe (prob not nec.)
    iList->currentInstr = ((Instruction *) addr)->prev;
    return (Instruction *)addr;
}

Instruction *GetNextLiveInstr(InstrList *iList) {
    Instruction *next;
    while ((next = FetchNextInstr(iList)) != NULL) {
        if (!INSTR_HAS_FLAG(next, INSTR_IS_DELETED)) {
            break;
        }
    }
    return next;

}

Instruction *GetPrevLiveInstr(InstrList *iList) {
    Instruction *prev;
    while ((prev = FetchPrevInstr(iList)) != NULL) {
        if (INSTR_HAS_FLAG(prev, INSTR_IS_DELETED)) {
            free(prev);
        } else
            break;
    }
    return prev;

}
Instruction *FetchNextInstr(InstrList *iList) {
    if (iList->currentInstr == -1/* || iList->currentInstr >= numInstrs*/){
        return NULL;
    }
    uint64_t addr;
    addr = (uint64_t)(iList->iMMap) + iList->currentInstr * sizeof(Instruction) +
    		InstrFileHeaderSize;//sizeof(InstrList);
    //just use the mmap'd location, set self and AllocCount to be safe (prob not nec.)
    iList->currentInstr = ((Instruction *) addr)->next;
    return (Instruction *)addr;
}

Instruction *GetInstructionFromFile(InstrList *iList, int index) {
    Instruction *instr = NULL;
    if ((index < iList->firstInstr) || (index > iList->lastInstr && index < iList->newFirstInstr)
            || index > iList->newLastInstr) {
        //need to fill cache with appropriate pages

        iList = FillCacheFromFile(iList, index);
    }

    if ((index >= iList->firstInstr) && (index <= iList->lastInstr)) {
        instr = ((Instruction *) (iList->instrCache) + (index - iList->firstInstr));
    } else if ((index >= iList->newFirstInstr) && (index <= iList->newLastInstr)) {
        instr = ((Instruction *) (iList->newInstrCache) + (index - iList->newFirstInstr));
    } else {
        fprintf(stderr, "index: %d NOT FOUND in cache or file\n", index);
    }
    return (instr);
}

/*
 * GetInstruction
 *  Given instruction index, get the instruction object
 *
 *  NOTE: since file instructions are stored in a cache, and the
 *  cache may change while the instr is still needed, we malloc
 *  space for a new instruction and copy the found instruction into
 *  it.  This means that calls to GetInstruction must be paired
 *  with a free.
 */
Instruction *GetInstruction(InstrList *iList, uint64_t index) {
    if (index >= numInstrs || index == -1) {
        return (NULL);
    }


    uint64_t addr;

	//get addr in mmap of the instr, remember to offset
	addr = (uint64_t)(iList->iMMap) + index * sizeof(Instruction) +
			InstrFileHeaderSize;//sizeof(InstrList);

	//just use the mmap'd location, set self and AllocCount to be safe (prob not nec.)
	return (Instruction *)addr;
}


Instruction *GetNextInstrAtAddr(InstrList *iList, int index, ADDRESS next_addr) {
    int size = iList->numInstrs;
    bool found = false;
    Instruction *instr = NULL;

    while ((index < size) && !found) {
        instr = GetInstruction(iList, index);
        if (instr->addr == next_addr) {
            found = true;
        } else {
            index++;
            free(instr);
            instr = NULL;
        }
    }

    return (instr);
}


bool IsCondJump(Instruction *instr) {
    bool isJCC = false;

    // if(!INSTR_HAS_FLAG(instr, INSTR_IN_BASE_MODULE)) {
    //   return isJCC;
    //}

    switch (instr->uInstr.mnemonic) {
        case UD_Ijo:
        case UD_Ijno:
        case UD_Ijb:
        case UD_Ijae:
        case UD_Ijz:
        case UD_Ijnz:
        case UD_Ijbe:
        case UD_Ija:
        case UD_Ijs:
        case UD_Ijns:
        case UD_Ijp:
        case UD_Ijnp:
        case UD_Ijl:
        case UD_Ijge:
        case UD_Ijle:
        case UD_Ijg:
        case UD_Ijcxz:
        case UD_Ijecxz:
        case UD_Ijrcxz:
            isJCC = true;
            break;
        default:
            break;
    }

    return (isJCC);
}

bool InstrIsDirectJmp(Instruction *iins){
	return (iins && iins->uInstr.mnemonic == UD_Ijmp && iins->uInstr.operand[0].type == UD_OP_JIMM);
}

bool IsControlFlowInstr(Instruction *instr) {
    bool isCF = false;
    //if(!INSTR_HAS_FLAG(instr, INSTR_IN_BASE_MODULE)) {
    //  return (isCF);
    //}

    switch (instr->uInstr.mnemonic) {
        case UD_Ijmp:
            //only looking at indirect jumps for now
            //if(instr->uInstr.operand[0].type == UD_OP_MEM) {
            isCF = true;
            //}
            break;
        case UD_Icall:
            //looking at indirect calls
            // if(instr->uInstr.operand[0].type == UD_OP_MEM) {
            isCF = true;
            //}
            break;
        case UD_Iret:
            isCF = true;
            break;
        case UD_Ijo:
        case UD_Ijno:
        case UD_Ijb:
        case UD_Ijae:
        case UD_Ijz:
        case UD_Ijnz:
        case UD_Ijbe:
        case UD_Ija:
        case UD_Ijs:
        case UD_Ijns:
        case UD_Ijp:
        case UD_Ijnp:
        case UD_Ijl:
        case UD_Ijge:
        case UD_Ijle:
        case UD_Ijg:
        case UD_Ijcxz:
        case UD_Ijecxz:
        case UD_Ijrcxz:
            //conditional jumps
            isCF = true;
            break;

        default:
            isCF = false;
            break;
    }

    return (isCF);
}

bool IsPush(Instruction *instr) {
    bool isPush = false;

        /*if ((instr->uInstr.mnemonic == UD_Ipush) || (instr->uInstr.mnemonic == UD_Ipusha)
                || (instr->uInstr.mnemonic == UD_Ipushad) || (instr->uInstr.mnemonic == UD_Ipushfw)
                || (instr->uInstr.mnemonic == UD_Ipushfd) || (instr->uInstr.mnemonic == UD_Ipushfq)) {*/
    if(instr->uInstr.mnemonic >= UD_Ipush && instr->uInstr.mnemonic <= UD_Ipushfq){			// according to itab.h!
		isPush = true;
	}

    return (isPush);
}

/*
 * IsMov currently only handling mov, push, and pop as mov instructions
 */
bool IsMov(Instruction *instr) {
    bool isMov = false;

    switch (instr->uInstr.mnemonic) {
        case UD_Imov:
        case UD_Ipush:
        case UD_Ipop:
            isMov = true;
            break;
        default:
            break;
    }

    return (isMov);
}

bool IsPushAll(Instruction *instr) {
    bool isPushAll = false;
	if ((instr->uInstr.mnemonic == UD_Ipusha) || (instr->uInstr.mnemonic == UD_Ipushad)) {
		isPushAll = true;
	}
    return (isPushAll);
}

bool IsPop(Instruction *instr) {
    bool isPop = false;

	if ((instr->uInstr.mnemonic == UD_Ipop) || (instr->uInstr.mnemonic == UD_Ipopa)
			|| (instr->uInstr.mnemonic == UD_Ipopad) || (instr->uInstr.mnemonic == UD_Ipopfw)
			|| (instr->uInstr.mnemonic == UD_Ipopfd) || (instr->uInstr.mnemonic == UD_Ipopfq)) {
		isPop = true;
	}

    return (isPop);
}

bool IsPopAll(Instruction *instr) {
    bool isPopAll = false;
        if ((instr->uInstr.mnemonic == UD_Ipopa) || (instr->uInstr.mnemonic == UD_Ipopad)) {
            isPopAll = true;
        }


    return (isPopAll);
}

bool isEscapePrefixCode(uint8_t byte) {
    return (byte == 0x0f);
}

int isNonEscapePrefixCode(uint8_t byte) {
    int isPref = 0;
    switch (byte) {
        case 0x26: // group 2
        case 0x2e: // group 2
        case 0x36: // group 2
        case 0x3e: // group 2
        case 0x64: // group 2
        case 0x65: // group 2
        case 0x66: // group 3
        case 0x67: // group 3
        case 0xf0: // group 1
        case 0xf2: // group 1
        case 0xf3: // group 1
            isPref = byte;
            break;
        default:
            break;
    }
    return (isPref);
}

X86op GetTableEntry(Instruction *instr) {
    int grp;
    uint8_t *bytes;
    uint8_t /*prefix = 0,*/ opCode1 = 0, opCode2 = 0, opCode3 = 0;
    X86op tableEntry;

    bytes = instr->uInstr.bytes;

    // check for prefixes, and set first op code
    opCode1 = bytes[0];
    opCode2 = bytes[1];
    opCode3 = bytes[2];
    if (isNonEscapePrefixCode(opCode1)) {
        /*prefix = opCode1;*/
        opCode1 = bytes[1];
        opCode2 = bytes[2];
        opCode3 = bytes[3];
    }

    // if it is one byte opcode
    if (!isEscapePrefixCode(opCode1)) {
        tableEntry = dis386[opCode1];

        // has name string indicates normal instr
        if (tableEntry.x_name == NULL) {
            grp = tableEntry.x_bytemode1;
            //opCode2 holds subgrp entry in bytes 3,4, and 5
            int subgrp = opCode2 >> 3;
            subgrp &= 0x00000007;

            tableEntry = grps[grp][subgrp];
        }
        // else its a 2 or 3 byte opcode
    } else {
        tableEntry = dis386_twobyte[opCode2];
        if (tableEntry.x_type == TYPE_ERROR) {
            //fprintf(stderr, "WARNING: Could not disassemble 2 or 3 byte opcode: %s\n",
            //        InstrToString(instr));
            //exit(0);
        } else {
            if (tableEntry.x_name == NULL) {
                //fprintf(stderr,"NEED TO CONFIRM: opcode3 holds grp num for 2byte opcode\n");
                grp = tableEntry.x_bytemode1;
                //opCode3 holds subgrp entry in bytes 3,4, and 5
                int subgrp = opCode3 >> 3;
                subgrp &= 0x00000007;

                tableEntry = grps[grp][subgrp];
            }
        }
    }

    return (tableEntry);
}

uint SetRegisterValue(Instruction *instr, int reg, int Val) {
    switch (reg) {
        case UD_R_EAX:
        case UD_R_AX:
        case UD_R_AH:
        case UD_R_AL:
            instr->eax = Val;
            break;
        case UD_R_ECX:
        case UD_R_CX:
        case UD_R_CL:
        case UD_R_CH:
            instr->ecx = Val;
            break;
        case UD_R_EDX:
        case UD_R_DX:
        case UD_R_DL:
        case UD_R_DH:
            instr->edx = Val;
            break;
        case UD_R_EBX:
        case UD_R_BX:
        case UD_R_BL:
        case UD_R_BH:
            instr->ebx = Val;
            break;
        case UD_R_ESP:
        case UD_R_SP:
            instr->esp = Val;
            break;
        case UD_R_EBP:
        case UD_R_BP:
            instr->ebp = Val;
            break;
        case UD_R_ESI:
        case UD_R_SI:
            instr->esi = Val;
            break;
        case UD_R_EDI:
        case UD_R_DI:
            instr->edi = Val;
            break;
        case UD_NONE:
            break;
        default:
            fprintf(stderr, "[SetRegisterValue]: Invalid or unhandled register: %d\n", reg);
            break;
    }
    return Val;
}

uint GetRegisterValue(Instruction *instr, enum ud_type reg) {
	if(!instr){
		return 0;
	}
    uint retVal = 0;
    switch (reg) {
        case UD_R_EAX:
        case UD_R_AX:
        case UD_R_AH:
        case UD_R_AL:
            retVal = instr->eax;
            break;
        case UD_R_ECX:
        case UD_R_CX:
        case UD_R_CL:
        case UD_R_CH:
            retVal = instr->ecx;
            break;
        case UD_R_EDX:
        case UD_R_DX:
        case UD_R_DL:
        case UD_R_DH:
            retVal = instr->edx;
            break;
        case UD_R_EBX:
        case UD_R_BX:
        case UD_R_BL:
        case UD_R_BH:
            retVal = instr->ebx;
            break;
        case UD_R_ESP:
        case UD_R_SP:
            retVal = instr->esp;
            break;
        case UD_R_EBP:
        case UD_R_BP:
            retVal = instr->ebp;
            break;
        case UD_R_ESI:
        case UD_R_SI:
            retVal = instr->esi;
            break;
        case UD_R_EDI:
        case UD_R_DI:
            retVal = instr->edi;
            break;
        case UD_NONE:
            break;
        default:
            fprintf(stderr, "[GetRegisterValue]: Invalid or unhandled register: %d\n", reg);
            break;
    }

    switch (reg) {
        case UD_R_AX:
        case UD_R_CX:
        case UD_R_DX:
        case UD_R_BX:
        case UD_R_SP:
        case UD_R_BP:
        case UD_R_SI:
        case UD_R_DI:
            retVal &= 0xffff;
            break;
        case UD_R_AL:
        case UD_R_CL:
        case UD_R_DL:
        case UD_R_BL:
            retVal &= 0xff;
            break;
        case UD_R_AH:
        case UD_R_CH:
        case UD_R_DH:
        case UD_R_BH:
            retVal &= 0xff00;
            retVal = retVal >> 8;
            break;
        default:
            break;
    }

    return (retVal);
}

/*
 * GetWSOverlapMask
 * this is a specific helper function that returns a bit mask representing
 * the overlap between two write sets.  It makes the following assumption
 *      -- That both ws1 and ws2 contain exactly one range each
 *      -- That the maximum size of the range is 4, ie, a 32-bit addr
 *
 * Thus, for write sets that don't follow these assumptions, this function is
 * meaningless
 *
 * for write sets that don't overlap, 0x00000000 is returned.
 *
 * for ranges of less than size 4, values are assumed at least significant
 *      end of 4 byte value.  ie, (1,2) and (1,2) overlap is 0x0000ffff
 *
 * ws2 determines the location of the overlap,
 *          eg ws1 = (2,3) and ws2 = (1,4), overlap is 0x00ffff00
 *          eg ws1 = (1,4) and ws2 = (2,3), overlap is 0x0000ffff
 */
uint32_t GetWSOverlapMask(WriteSet *ws1, WriteSet *ws2) {
    uint32_t overlap = 0x0;

    //start with some checks on our assumptions
    if ((ws1 == NULL) || (ws2 == NULL)) {
        return 0x0;
    }

    if ((al_size(ws1->ranges) != 1) || (al_size(ws2->ranges) != 1)) {
        fprintf(stderr, "WARNING[GetWSOverlapMask]: WriteSet of size > 1\n");
        return 0x0;
    }

    Range *r1 = al_get(ws1->ranges, 0);
    Range *r2 = al_get(ws2->ranges, 0);

    if ((r2->min <= r2->max) && //in r2
            (r2->min >= r1->min) && (r2->min <= r1->max)) { //and in r1
        overlap |= 0x000000ff;
    }

    if (((r2->min + 1) <= r2->max) && //in r2
            ((r2->min + 1) >= r1->min) && ((r2->min + 1) <= r1->max)) { //and in r1
        overlap |= 0x0000ff00;
    }

    if (((r2->min + 2) <= r2->max) && //in r2
            ((r2->min + 2) >= r1->min) && ((r2->min + 2) <= r1->max)) { //and in r1
        overlap |= 0x00ff0000;
    }

    if (((r2->min + 3) <= r2->max) && //in r2
            ((r2->min + 3) >= r1->min) && ((r2->min + 3) <= r1->max)) { //and in r1
        overlap |= 0xff000000;
    }

    return (overlap);
}

/*
 * starts at instr, and searches forward looking for the use of ws, so it can
 * figure out it's value.  Stops if ws is defined before the value can be
 * found, and returns false.  Returns true if val was found and set.
 *
 * TODO: This is a function that was repurposed, and I think that it is
 * impossible to find a usage if it occurs on instr 0.  Needs some thought.
 */
bool GetMemoryValueFWD(InstrList *iList, Instruction *instr, WriteSet *ws, uint32_t *val) {
    int i = instr->order + 1;
    int size = iList->numInstrs;
    int wsSize = 0;
    bool found = false;
    bool done = false;

    //calculate size of write set we are looking for
    Range *r = al_get(ws->ranges, 0);
    wsSize = r->max - r->min + 1;
    if (wsSize <= 0) {
        fprintf(stderr, "ERROR: something very unexpected just happened\n");
        exit(0);
    }

    while ((i < size) && !found && !done) {
        Instruction *next = GetInstruction(iList, i);
        WriteSet *write = MemDef(iList, next);
        if (WriteSetsOverlap(ws, write)) {
            //then we are defining a part of ws, before we could find its val
            done = true;
            found = false;
        } else {
            WriteSet *read = MemUsed(iList, next);

            if ((instr->uInstr.mnemonic == UD_Imov) && WriteSetSubsetOfWriteSet(ws, read)) {

                //then we are moving ws, or some larger version, into reg
                uint32_t mask = GetWSOverlapMask(ws, read);
                if (mask != 0x0) {
                    Instruction *afterNext = GetInstruction(iList, next->order + 1);
                    uint32_t regVal = GetRegisterValue(afterNext, instr->uInstr.operand[0].base);
                    free(afterNext);

                    *val = regVal & mask;

                    //if ws and read don't line up at least sig bits, then need to
                    //shift value to right to get real val
                    if ((mask & 0x000000ff) == 0) {
                        *val = *val >> 8;
                        if ((mask & 0x0000ff00) == 0) {
                            *val = *val >> 8;
                            if ((mask & 0x00ff0000) == 0) {
                                *val = *val >> 8;
                                //if mask & 0xff000000 is 0, we can't be here
                            }
                        }
                    }
                }
            }
            WriteSetDestroy(read);
        }
        free(next);
        WriteSetDestroy(write);
        i++;
    }

    return (found);
}

/* GetTarget
 *  Takes instr and operand, and returns range of memory addresses pointed
 *  to by that operand.
 *
 *  assumes that op has already been determined to be memory access operand
 */
Range *GetTarget(Instruction *instr, ud_operand_t op) {
    int target = 0;
    int base = GetRegisterValue(instr, op.base);
    int index = GetRegisterValue(instr, op.index);
    int scale = op.scale;
    int offset = op.offset;
    int displacement = op.lval.udword;

    if(op.base == UD_NONE && op.index == UD_NONE && op.lval.udword == 0){     // added by Babak. For instrs using fs register, which is not available in our trace!
        return NULL;
    }

    int min = 1, max = 0;
    Range *r = NULL;

    //HACK: for src operand of the instr: mov edx, [edi+eax]
    //udis86 will give base reg = edi, index reg = eax, scale = 0!!
    if ((index > 0) && (scale == 0)) {
        scale = 1;
    }

    if (op.base != UD_NONE) {
        /*
         * An interesting corner case occurs when we pop into the location on top
         * of the stack: "pop [esp]".  According to the Intel reference manual,
         * the destination location will be updated before ESP is updated,
         * which would mean that the location written is popped off ths stack;
         * however, dynamic traces indicate that ESP is updated before the
         * destination is written.  Possibly the reason is that the actual
         * hardware implementation is along the lines of:
         *
         *     _tmpReg := pop(esp); dest := _tmpReg
         *
         * This only affects instructions where we use esp as a base or index
         * register in the pop; such instructions have to be special-cased.
         *
         */
        if (instr->uInstr.mnemonic == UD_Ipop) {
            if (op.base == UD_R_ESP) {
                base += OperandSize(op);
            }
            if (op.index == UD_R_ESP) {
                base += OperandSize(op);
            }
        }

        target = base + (index * scale);

        switch (offset) {
            case 8:
                target += op.lval.sbyte;
                break;
            case 16:
                target += op.lval.sword;
                break;
            case 32:
                target += op.lval.sdword;
                break;
            case 64:
                fprintf(stderr, "WARNING: 64 bit offsets not implemented\n");
                break;
            default:
                break;
        }
    } else {
        target = displacement;
        if (op.index != UD_NONE) {
            //we have the form [index * scale + immediate_dispacement]
            target += index * scale;
        }
    }
    min = target;
    max = target + ((op.size / 8) - 1);

    if (min <= max) {
        r = RangeCreate(min, max);
    }
    return (r);
}

bool InstrHasREPPrefix(Instruction *instr) {
    return ((instr->uInstr.bytes[0] == 0xf2) || (instr->uInstr.bytes[0] == 0xf3));
}



/*
 * Returns the number of bytes occupied by an operand.
 * This code fixes a problem (bug?) in udis86 where
 * the value of the "size" field of a register operand
 * is returned as 0.
 */
int OperandSize(ud_operand_t op) {
    int size;

    if (op.type == UD_OP_REG) {
        if (op.base >= UD_R_AL && op.base <= UD_R_BH) {
            size = 1;
        } else if (op.base >= UD_R_AX && op.base <= UD_R_DI) {
            size = 2;
        } else if (op.base >= UD_R_EAX && op.base <= UD_R_EDI) {
            size = 4;
        } else if(op.base >= UD_R_ES && op.base <= UD_R_GS){
        	size = 2;
        }
        else {
            fprintf(stderr, "WARNING: 64-bit registers not implemented\n");
            size = -1;
        }
    } else {
        size = op.size / 8;
    }

    assert(size > 0 && size <= 4); // in case udis86 has other surprises
    return size;
}

/* handling these as special cases, need to add more as we find them
 */
Range *GetImplicitTarget(InstrList *iList, Instruction *instr) {
    int addrSize = 0;
    int destVal = 0;
    int start = 1;
    int end = 0;
    int opsize;

    int min = 1, max = 0;
    Range *r = NULL;

    switch (instr->uInstr.mnemonic) {
        case UD_Imovsb:
        case UD_Istosb:
        case UD_Imovsw:
        case UD_Istosw:
        case UD_Imovsd:
        case UD_Istosd:
        case UD_Imovsq:
        case UD_Istosq:
            switch (instr->uInstr.mnemonic) {
                case UD_Imovsb:
                case UD_Istosb:
                    addrSize = 1;
                    break;

                case UD_Imovsw:
                case UD_Istosw:
                    addrSize = 2;
                    break;

                case UD_Imovsd:
                case UD_Istosd:
                    addrSize = 4;
                    break;

                case UD_Imovsq:
                case UD_Istosq:
                    addrSize = 8;
                    break;

                default:
                    addrSize = 0;
                    break;
            }
            Instruction *next = GetInstruction(iList, instr->order + 1);
            destVal = GetRegisterValue(instr, UD_R_EDI);

            //if instr has repeat prefix, need to calculate how many iterations
            if (InstrHasREPPrefix(instr)) {
                start = GetRegisterValue(instr, UD_R_ECX);
                if (next) {
                    end = GetRegisterValue(next, UD_R_ECX);
                }
            }

            //if no repeat, then it was just one by definition.
            min = destVal;
            max = destVal + ((start - end) * addrSize - 1);
            //printf("implicit target: 0x%08x -- 0x%08x\n",r->min, r->max);

            RInstrDestroy(iList, next);
            break;

        case UD_Ipush:
            opsize = instr->uInstr.operand[0].type == UD_OP_IMM ? PTRSIZE : OperandSize(
                    instr->uInstr.operand[0]);
            min = GetRegisterValue(instr, UD_R_ESP) - opsize;
            max = min + opsize - 1;
            break;

        case UD_Icall:
        case UD_Ipushfw:
        case UD_Ipushfd:
        case UD_Ipushfq:
            //call is implicit push, pushes addr of next instr for ret
            min = GetRegisterValue(instr, UD_R_ESP) - PTRSIZE;
            max = min + PTRSIZE - 1;
            break;
        case UD_Ipusha:
        case UD_Ipushad:
            //push all's push 8 user registers.
            //TODO: Push all is illegal instruction in 64 bit.  Need to handle this.
            min = GetRegisterValue(instr, UD_R_ESP) - 8 * PTRSIZE;
            max = min + 8 * PTRSIZE - 1;
            break;
        default:
            break;
    }

    if (min <= max) {
        r = RangeCreate(min, max);
    }

    return (r);
}

/* handling these as special cases, need to add more as we find them
 */
Range *GetImplicitMemSource(InstrList *iList, Instruction *instr) {
    int addrSize = 0;
    int srcVal = 0;
    int start = 1;
    int end = 0;
    int srcReg;
    int opsize;

    int min = 1, max = 0;
    Range *r = NULL;

    switch (instr->uInstr.mnemonic) {
        case UD_Ilodsb:
        case UD_Imovsb:
        case UD_Iscasb:
        case UD_Ilodsw:
        case UD_Imovsw:
        case UD_Iscasw:
        case UD_Ilodsd:
        case UD_Imovsd:
        case UD_Iscasd:
        case UD_Ilodsq:
        case UD_Imovsq:
        case UD_Iscasq:
            switch (instr->uInstr.mnemonic) {
                case UD_Ilodsb:
                case UD_Imovsb:
                    addrSize = 1;
                    srcReg = UD_R_ESI;
                    break;
                case UD_Iscasb:
                    addrSize = 1;
                    srcReg = UD_R_EDI;
                    break;
                case UD_Ilodsw:
                case UD_Imovsw:
                    addrSize = 2;
                    srcReg = UD_R_ESI;
                    break;
                case UD_Iscasw:
                    addrSize = 2;
                    srcReg = UD_R_EDI;
                    break;
                case UD_Ilodsd:
                case UD_Imovsd:
                    addrSize = 4;
                    srcReg = UD_R_ESI;
                    break;
                case UD_Iscasd:
                    addrSize = 4;
                    srcReg = UD_R_EDI;
                    break;
                case UD_Ilodsq:
                case UD_Imovsq:
                    addrSize = 8;
                    srcReg = UD_R_ESI;
                    break;
                case UD_Iscasq:
                    addrSize = 8;
                    srcReg = UD_R_EDI;
                    break;
                default:
                    addrSize = 0;
                    srcReg = 0;
                    break;
            }
            Instruction *next = GetInstruction(iList, instr->order + 1);
            srcVal = GetRegisterValue(instr, srcReg);

            //if instr has repeat prefix, need to calculate how many iterations
            if (InstrHasREPPrefix(instr)) {
                start = GetRegisterValue(instr, UD_R_ECX);
                if (next) {
                    end = GetRegisterValue(next, UD_R_ECX);
                }
            }

            //if no repeat, then it was just one by definition.
            min = srcVal;
            max = srcVal + ((start - end) * addrSize - 1);

            RInstrDestroy(iList, next);
            break;

        case UD_Ipop:
            opsize = OperandSize(instr->uInstr.operand[0]);
            min = GetRegisterValue(instr, UD_R_ESP);
            max = min + opsize - 1;
            break;

        case UD_Iret:
        case UD_Ipopfw:
        case UD_Ipopfd:
        case UD_Ipopfq:
            //return instruction is implicit pop, pops addr to jump to
            min = GetRegisterValue(instr, UD_R_ESP);
            max = min + PTRSIZE - 1;
            break;
        case UD_Ipopa:
        case UD_Ipopad:
            min = GetRegisterValue(instr, UD_R_ESP);
            max = min + 8 * PTRSIZE - 1;
            break;
        default:
            break;
    }

    if (min <= max) {
        r = RangeCreate(min, max);
    }

    return (r);
}

bool IsMxcsrInstr(Instruction *instr){
	bool isMxcsr = false;
	switch(instr->uInstr.mnemonic){
	case UD_Istmxcsr:
	case UD_Ildmxcsr:
		isMxcsr = true;
		break;
	default:
		break;
	}
	return isMxcsr;
}

bool IsStringOperation(Instruction *instr) {
    bool isString = false;
    switch (instr->uInstr.mnemonic) {
        case UD_Iscasb:
        case UD_Iscasw:
        case UD_Iscasd:
        case UD_Iscasq:

        case UD_Istosb:
        case UD_Istosw:
        case UD_Istosd:
        case UD_Istosq:

        case UD_Icmpsb:
        case UD_Icmpsw:
        case UD_Icmpsd:
        case UD_Icmpsq:

        case UD_Imovsb:
        case UD_Imovsw:
        case UD_Imovsd:
        case UD_Imovsq:

        case UD_Ilodsb:
        case UD_Ilodsw:
        case UD_Ilodsd:
        case UD_Ilodsq:
            isString = true;
            break;
        default:
            break;
    }

    return (isString);
}


bool IsStringCompare(Instruction *instr){
    bool isStringComp = false;
    switch (instr->uInstr.mnemonic) {
        case UD_Iscasb:
        case UD_Iscasw:
        case UD_Iscasd:
        case UD_Iscasq:
            isStringComp = true;
            break;
        default:
            break;
    }

    return (isStringComp);
}

REGISTER_BITVECTOR GetImplicitMemRegUsed(Instruction *instr) {
    REGISTER_BITVECTOR bv = 0x0;
    if (IsStringOperation(instr) && (instr->uInstr.operand[0].type != UD_NONE)) {
        fprintf(stderr, "unhandled explicit string operation\n");
        return (bv);
    }
    switch (instr->uInstr.mnemonic) {
        case UD_Istosb:
        case UD_Iscasb:
        case UD_Istosw:
        case UD_Iscasw:
        case UD_Istosd:
        case UD_Iscasd:
        case UD_Istosq:
        case UD_Iscasq:
            bv |= LYNX_EDI;
            break;
        case UD_Icmpsb:
        case UD_Icmpsw:
        case UD_Icmpsd:
        case UD_Icmpsq:

        case UD_Imovsb:
        case UD_Imovsw:
        case UD_Imovsd:
        case UD_Imovsq:
            bv = LYNX_ESI | LYNX_EDI;
            break;
        case UD_Ilodsb:
        case UD_Ilodsw:
        case UD_Ilodsd:
        case UD_Ilodsq:
            bv |= LYNX_ESI;
            break;
        default:
            break;
    }

    return (bv);
}

REGISTER_BITVECTOR GetImplicitNonMemRegUsed(Instruction *instr) {
    REGISTER_BITVECTOR bv = 0x0;
    if (IsStringOperation(instr) && (instr->uInstr.operand[0].type != UD_NONE)) {
        fprintf(stderr, "unhandled explicit string operation\n");
        return (bv);
    }
    if (instr->uInstr.bytes[0] == 0xf2 || instr->uInstr.bytes[0] == 0xf3) { //rep prefix which defines ecx register. Added by Babak.
        bv |= LYNX_ECX;
    }
    switch (instr->uInstr.mnemonic) {
        case UD_Istosb:
        case UD_Iscasb:
            bv |= LYNX_AL;
            break;
        case UD_Istosw:
        case UD_Iscasw:
            bv |= LYNX_AX;
            break;
        case UD_Istosd:
        case UD_Iscasd:
        case UD_Istosq:
        case UD_Iscasq:
            bv |= LYNX_EAX;
            break;
        case UD_Icmpsb:
        case UD_Icmpsw:
        case UD_Icmpsd:
        case UD_Icmpsq:
        case UD_Imovsb:
        case UD_Imovsw:
        case UD_Imovsd:
        case UD_Imovsq:
            break;
        case UD_Ilodsb:
        case UD_Ilodsw:
        case UD_Ilodsd:
        case UD_Ilodsq:
            break;
        default:
            break;
    }

    return (bv);
}

REGISTER_BITVECTOR GetImplicitRegUsed(Instruction *instr) {
    return (GetImplicitNonMemRegUsed(instr) | GetImplicitMemRegUsed(instr));
}

/*
 * returns size of a register in bytes
 */
int UDRegSize(REGISTER_BITVECTOR reg) {
    if (reg >= UD_R_EAX && reg <= UD_R_EDI)
        return 4;
    else if (reg >= UD_R_AX && reg <= UD_R_DI)
        return 2;
    else
        return 1;

}
/*
 * returns the corresponding register according
 * to src which matches in the size.
 * (EAX, 16)	=>	ax
 */
enum ud_type UDRegSizeMask(enum ud_type src, int size){
	int diff;
    if (src >= UD_R_EAX && src <= UD_R_EDI){
        diff = src - UD_R_EAX;
    } else if (src >= UD_R_AX && src <= UD_R_DI){
        diff = src - UD_R_AX;
    } else {
        diff = src - UD_R_AL;
    }

    if(size == 4){
    	return diff + UD_R_EAX;
    } else if(size == 2){
    	return diff + UD_R_AX;
    } else {
    	return diff + UD_R_AL;
    }

}

/* LynxRegisterToBitVector
 *  takes a register value as used in the table.c, and converts to
 *  our own Lynx bit-vector representation.
 *
 *  Note: We treat 8 bit, 16bit, and 32 bit versions of same actual register
 *  on cpu as the same register.  For example: al, ah, ax, and eax all
 *  return same bit vector.  This is a conversative estimate for u-d chains,
 *  dynamic slicing, etc...
 */
REGISTER_BITVECTOR LynxRegisterToBitVector(REGISTER_BITVECTOR reg) {
    REGISTER_BITVECTOR bv = 0x0000000000000000LLU;

    switch (reg) {
        case eAX_reg:
            bv = LYNX_EAX;
            break;
        case al_reg:
            bv = LYNX_AL;
            break;
        case ah_reg:
            bv = LYNX_AH;
            break;
        case ax_reg:
            bv = LYNX_AX;
            break;

        case eCX_reg:
            bv = LYNX_ECX;
            break;
        case cl_reg:
            bv = LYNX_CL;
            break;
        case ch_reg:
            bv = LYNX_CH;
            break;
        case cx_reg:
            bv = LYNX_CX;
            break;

        case eBX_reg:
            bv = LYNX_EBX;
            break;
        case bl_reg:
            bv = LYNX_BL;
            break;
        case bh_reg:
            bv = LYNX_BH;
            break;
        case bx_reg:
            bv = LYNX_BX;
            break;

        case eDX_reg:
            bv = LYNX_EDX;
            break;
        case dl_reg:
            bv = LYNX_DL;
            break;
        case dh_reg:
            bv = LYNX_DH;
            break;
        case dx_reg:
            bv = LYNX_DX;
            break;

        case eSP_reg:
            bv = LYNX_ESP;
            break;
        case sp_reg:
            bv = LYNX_SP;
            break;

        case eBP_reg:
            bv = LYNX_EBP;
            break;
        case bp_reg:
            bv = LYNX_BP;
            break;

        case eSI_reg:
            bv = LYNX_ESI;
            break;
        case si_reg:
            bv = LYNX_SI;
            break;

        case eDI_reg:
            bv = LYNX_EDI;
            break;
        case di_reg:
            bv = LYNX_DI;
            break;

        case es_reg:
            bv = LYNX_ES;
            break;
        case cs_reg:
            bv = LYNX_CS;
            break;
        case ss_reg:
            bv = LYNX_SS;
            break;
        case ds_reg:
            bv = LYNX_DS;
            break;
        case fs_reg:
            bv = LYNX_FS;
            break;
        case gs_reg:
            bv = LYNX_GS;
            break;
        case all_regs:
            bv = LYNX_ALL;
            break;
        case no_reg:
            bv = LYNX_NONE;
            break;
        case lptr:
            fprintf(stderr, "Unhandled register type: lptr\n");
            break;
        case indir_dx_reg:
            fprintf(stderr, "Unhandled register type: lptr\n");
            break;
        case st0_reg:
            bv = LYNX_ST0;
            break;
        case st1_reg:
            bv = LYNX_ST1;
            break;
        case st2_reg:
            bv = LYNX_ST2;
            break;
        case st3_reg:
            bv = LYNX_ST3;
            break;
        case st4_reg:
            bv = LYNX_ST4;
            break;
        case st5_reg:
            bv = LYNX_ST5;
            break;
        case st6_reg:
            bv = LYNX_ST6;
            break;
        case st7_reg:
            bv = LYNX_ST7;
            break;
        case st8_reg:
            bv = LYNX_ST8;
            break;
        case fpu_regs:
            bv = LYNX_ALL_FPU;
            break;
        default:
            break;
    }

    return (bv);
}

REGISTER_BITVECTOR UDRegisterToBitVector(enum ud_type reg) {
    REGISTER_BITVECTOR bv = 0;

    switch (reg) {
        case UD_R_RAX:
        case UD_R_EAX:
            bv = LYNX_EAX;
            break;
        case UD_R_AX:
            bv = LYNX_AX;
            break;
        case UD_R_AL:
            bv = LYNX_AL;
            break;
        case UD_R_AH:
            bv = LYNX_AH;
            break;
        case UD_R_RCX:
        case UD_R_ECX:
            bv = LYNX_ECX;
            break;
        case UD_R_CX:
            bv = LYNX_CX;
            break;
        case UD_R_CL:
            bv = LYNX_CL;
            break;
        case UD_R_CH:
            bv = LYNX_CH;
            break;

        case UD_R_RBX:
        case UD_R_EBX:
            bv = LYNX_EBX;
            break;
        case UD_R_BX:
            bv = LYNX_BX;
            break;
        case UD_R_BL:
            bv = LYNX_BL;
            break;
        case UD_R_BH:
            bv = LYNX_BH;
            break;

        case UD_R_RDX:
        case UD_R_EDX:
            bv = LYNX_EDX;
            break;
        case UD_R_DX:
            bv = LYNX_DX;
            break;
        case UD_R_DL:
            bv = LYNX_DL;
            break;
        case UD_R_DH:
            bv = LYNX_DH;
            break;

        case UD_R_RSP:
        case UD_R_ESP:
            bv = LYNX_ESP;
            break;
        case UD_R_SP:
        case UD_R_SPL:
            bv = LYNX_SP;
            break;
        case UD_R_RBP:
        case UD_R_EBP:
            bv = LYNX_EBP;
            break;
        case UD_R_BP:
        case UD_R_BPL:
            bv = LYNX_BP;
            break;
        case UD_R_RSI:
        case UD_R_ESI:
            bv = LYNX_ESI;
            break;
        case UD_R_SI:
        case UD_R_SIL:
            bv = LYNX_SI;
            break;
        case UD_R_RDI:
        case UD_R_EDI:
            bv = LYNX_EDI;
            break;
        case UD_R_DI:
        case UD_R_DIL:
            bv = LYNX_DI;
            break;
        case UD_R_ES:
            bv = LYNX_ES;
            break;
        case UD_R_CS:
            bv = LYNX_CS;
            break;
        case UD_R_SS:
            bv = LYNX_SS;
            break;
        case UD_R_DS:
            bv = LYNX_DS;
            break;
        case UD_R_FS:
            bv = LYNX_FS;
            break;
        case UD_R_GS:
            bv = LYNX_GS;
            break;
        case UD_NONE:
            bv = LYNX_NONE;
            break;
        case UD_R_ST0:
            bv = LYNX_ST0;
            break;
        case UD_R_ST1:
            bv = LYNX_ST1;
            break;
        case UD_R_ST2:
            bv = LYNX_ST2;
            break;
        case UD_R_ST3:
            bv = LYNX_ST3;
            break;
        case UD_R_ST4:
            bv = LYNX_ST4;
            break;
        case UD_R_ST5:
            bv = LYNX_ST5;
            break;
        case UD_R_ST6:
            bv = LYNX_ST6;
            break;
        case UD_R_ST7:
            bv = LYNX_ST7;
            break;
        default:
            //fprintf(stderr, "unhandled ud register type\n");
            break;
    }

    return (bv);
}

enum ud_type BitVectorToUDRegister(REGISTER_BITVECTOR reg) {
    enum ud_type udReg = 0;

    switch (reg) {
        case LYNX_EAX:
            udReg = UD_R_EAX;
            break;
        case LYNX_AX:
            udReg = UD_R_AX;
            break;
        case LYNX_AH:
            udReg = UD_R_AH;
            break;
        case LYNX_AL:
            udReg = UD_R_AL;
            break;
        case LYNX_ECX:
            udReg = UD_R_ECX;
            break;
        case LYNX_CX:
            udReg = UD_R_CX;
            break;
        case LYNX_CH:
            udReg = UD_R_CH;
            break;
        case LYNX_CL:
            udReg = UD_R_CL;
            break;
        case LYNX_EDX:
            udReg = UD_R_EDX;
            break;
        case LYNX_DX:
            udReg = UD_R_DX;
            break;
        case LYNX_DH:
            udReg = UD_R_DH;
            break;
        case LYNX_DL:
            udReg = UD_R_DL;
            break;
        case LYNX_EBX:
            udReg = UD_R_EBX;
            break;
        case LYNX_BX:
            udReg = UD_R_BX;
            break;
        case LYNX_BH:
            udReg = UD_R_BH;
            break;
        case LYNX_BL:
            udReg = UD_R_BL;
            break;
        case LYNX_ESP:
            udReg = UD_R_ESP;
            break;
        case LYNX_SP:
            udReg = UD_R_SP;
            break;
        case LYNX_EBP:
            udReg = UD_R_EBP;
            break;
        case LYNX_BP:
            udReg = UD_R_BP;
            break;
        case LYNX_ESI:
            udReg = UD_R_ESI;
            break;
        case LYNX_SI:
            udReg = UD_R_SI;
            break;
        case LYNX_EDI:
            udReg = UD_R_EDI;
            break;
        case LYNX_DI:
            udReg = UD_R_DI;
            break;
        default:
        	if(DEBUG_MESSAGES){
        	//	fprintf(stderr, "[BitVectorToUDRegister] unhandled bitvector register type\n");
        	}
            //fprintf(stderr, "[BitVectorToUDRegister] unhandled bitvector register type\n");
            break;
    }

    return (udReg);
}

/*
 * this function identifies if the argument
 * is one register or not witout printing error!
 */
enum ud_type BitVectorToUDRegisterForTest(REGISTER_BITVECTOR reg) {
    enum ud_type udReg = 0;

    switch (reg) {
        case LYNX_EAX:
            udReg = UD_R_EAX;
            break;
        case LYNX_AX:
            udReg = UD_R_AX;
            break;
        case LYNX_AH:
            udReg = UD_R_AH;
            break;
        case LYNX_AL:
            udReg = UD_R_AL;
            break;
        case LYNX_ECX:
            udReg = UD_R_ECX;
            break;
        case LYNX_CX:
            udReg = UD_R_CX;
            break;
        case LYNX_CH:
            udReg = UD_R_CH;
            break;
        case LYNX_CL:
            udReg = UD_R_CL;
            break;
        case LYNX_EDX:
            udReg = UD_R_EDX;
            break;
        case LYNX_DX:
            udReg = UD_R_DX;
            break;
        case LYNX_DH:
            udReg = UD_R_DH;
            break;
        case LYNX_DL:
            udReg = UD_R_DL;
            break;
        case LYNX_EBX:
            udReg = UD_R_EBX;
            break;
        case LYNX_BX:
            udReg = UD_R_BX;
            break;
        case LYNX_BH:
            udReg = UD_R_BH;
            break;
        case LYNX_BL:
            udReg = UD_R_BL;
            break;
        case LYNX_ESP:
            udReg = UD_R_ESP;
            break;
        case LYNX_SP:
            udReg = UD_R_SP;
            break;
        case LYNX_EBP:
            udReg = UD_R_EBP;
            break;
        case LYNX_BP:
            udReg = UD_R_BP;
            break;
        case LYNX_ESI:
            udReg = UD_R_ESI;
            break;
        case LYNX_SI:
            udReg = UD_R_SI;
            break;
        case LYNX_EDI:
            udReg = UD_R_EDI;
            break;
        case LYNX_DI:
            udReg = UD_R_DI;
            break;
        default:
            break;
    }

    return (udReg);
}

REGISTER_BITVECTOR GetImplicitNonMemRegDef(Instruction *instr) {
    REGISTER_BITVECTOR bv = 0x0;
    if (IsStringOperation(instr) && (instr->uInstr.operand[0].type != UD_NONE)) {
        fprintf(stderr, "unhandled explicit string operation\n");
        return (bv);
    }
    if (instr->uInstr.bytes[0] == 0xf2 || instr->uInstr.bytes[0] == 0xf3) { //rep prefix which defines ecx register. Added by Babak.
        bv |= LYNX_ECX;
    }
    switch (instr->uInstr.mnemonic) {
        case UD_Istosb:
        case UD_Iscasb:
        case UD_Istosw:
        case UD_Iscasw:
        case UD_Istosd:
        case UD_Iscasd:
        case UD_Istosq:
        case UD_Iscasq:
            break;
        case UD_Imovsb:
        case UD_Imovsw:
        case UD_Imovsd:
        case UD_Imovsq:
            break;
        case UD_Ilodsb:
            bv = LYNX_AL;
            break;
        case UD_Ilodsw:
            bv = LYNX_AX;
            break;
        case UD_Ilodsd:
        case UD_Ilodsq:
            bv = LYNX_EAX;
            break;
        default:
            break;
    }

    return (bv);
}

REGISTER_BITVECTOR GetImplicitMemRegDef(Instruction *instr) {
    REGISTER_BITVECTOR bv = 0x0;
    if (IsStringOperation(instr) && (instr->uInstr.operand[0].type != UD_NONE)) {
        fprintf(stderr, "unhandled explicit string operation\n");
        return (bv);
    }
    switch (instr->uInstr.mnemonic) {
        case UD_Istosb:
        case UD_Iscasb:
        case UD_Istosw:
        case UD_Iscasw:
        case UD_Istosd:
        case UD_Iscasd:
        case UD_Istosq:
        case UD_Iscasq:
            bv = LYNX_EDI;
            break;
        case UD_Imovsb:
        case UD_Imovsw:
        case UD_Imovsd:
        case UD_Imovsq:
            bv = LYNX_ESI | LYNX_EDI;
            break;
        case UD_Ilodsb:
        case UD_Ilodsw:
        case UD_Ilodsd:
        case UD_Ilodsq:
            bv = LYNX_ESI;
            break;
        default:
            break;
    }

    return (bv);
}

REGISTER_BITVECTOR GetImplicitRegDef(Instruction *instr) {
    return (GetImplicitNonMemRegDef(instr) | GetImplicitMemRegDef(instr));
}

bool InstrHasDest(Instruction *instr) {
    bool hasDest = false;

    X86op tableEntry = GetTableEntry(instr);
    if (tableEntry.x_op1 != NULL) {
        hasDest = true;
    }

    return (hasDest);
}

/* GetWriteTargets
 *  For a given instruction, returns a write set that contains all
 *  possible target addresses that could be written to.
 */
WriteSet *GetWriteTargets(InstrList *iList, Instruction *instr) {
    Range *target;
    WriteSet *ws = WriteSetCreate();
    ud_operand_t op0 = instr->uInstr.operand[0];
    ud_operand_t op1 = instr->uInstr.operand[1];

    if (InstrHasDest(instr) && op0.type == UD_OP_MEM) {
        // get explicit targets -- xchg is only instr I know of that can
        // have second op as mem dest.  xchg cannot have two mem operands.
        if (strstr("xchg", ud_insn_asm(LoadUDisInstrObj(instr))) && op1.type == UD_OP_MEM) {
            target = GetTarget(instr, op1);
        } else {
            target = GetTarget(instr, op0);
        }
        AddRangeToWriteSet(ws, target);
    }

    // get implicit targets
    target = GetImplicitTarget(iList, instr);
    AddRangeToWriteSet(ws, target);

    return (ws);
}

/* udis86 does not tell us if operands are sources or destinations.  Some can
 * be both. For example, add eax, $4.
 * This function uses two methods to determine if the first operand
 * is a source.  First, if x_op1 for the table entry is null, then the first
 * operand is NOT a source.  Next, op1 for mov instructions are always dest
 * only.  Also lea instructions.
 */
bool FirstOpIsSrc(Instruction *instr) {
    bool isSrc = true;
    //    X86op tableEntry = GetTableEntry(instr);
    //    if(!tableEntry.x_op1) {
    //        isSrc = false;
    //    }

    switch (instr->uInstr.mnemonic) {
        case UD_Imovsxd:
        case UD_Icmovo:
        case UD_Icmovno:
        case UD_Icmovb:
        case UD_Icmovae:
        case UD_Icmovz:
        case UD_Icmovnz:
        case UD_Icmovbe:
        case UD_Icmova:
        case UD_Icmovs:
        case UD_Icmovns:
        case UD_Icmovp:
        case UD_Icmovnp:
        case UD_Icmovl:
        case UD_Icmovge:
        case UD_Icmovle:
        case UD_Icmovg:
        case UD_Ifcmovb:
        case UD_Ifcmove:
        case UD_Ifcmovbe:
        case UD_Ifcmovu:
        case UD_Ifcmovnb:
        case UD_Ifcmovne:
        case UD_Ifcmovnbe:
        case UD_Ifcmovnu:
        case UD_Imaskmovq:
        case UD_Imov:
        case UD_Imovapd:
        case UD_Imovaps:
        case UD_Imovd:
        case UD_Imovddup:
        case UD_Imovdqa:
        case UD_Imovdqu:
        case UD_Imovdq2q:
        case UD_Imovhpd:
        case UD_Imovhps:
        case UD_Imovlhps:
        case UD_Imovlpd:
        case UD_Imovlps:
        case UD_Imovhlps:
        case UD_Imovmskpd:
        case UD_Imovmskps:
        case UD_Imovntdq:
        case UD_Imovnti:
        case UD_Imovntpd:
        case UD_Imovntps:
        case UD_Imovntq:
        case UD_Imovq:
        case UD_Imovqa:
        case UD_Imovq2dq:
        case UD_Imovsb:
        case UD_Imovsw:
        case UD_Imovsd:
        case UD_Imovsq:
        case UD_Imovsldup:
        case UD_Imovshdup:
        case UD_Imovss:
        case UD_Imovsx:
        case UD_Imovupd:
        case UD_Imovups:
        case UD_Imovzx:
        case UD_Ipmovmskb:
        case UD_Ipop:
        case UD_Ipopa:
        case UD_Ipopad:
        case UD_Ipopfw:
        case UD_Ipopfd:
        case UD_Ipopfq:
        case UD_Ilds:
        case UD_Ilea:
        case UD_Iles:
        case UD_Ilfs:
        case UD_Ilgs:
        case UD_Ilss:
        case UD_Ilodsb: // Added by Babak. al/ax/eax is just defined in these instruction.
        case UD_Ilodsw:
        case UD_Ilodsd:
        case UD_Ilodsq:

        case UD_Iseto:
        case UD_Isetno:
        case UD_Isetb:
        case UD_Isetnb:
        case UD_Isetz:
        case UD_Isetnz:
        case UD_Isetbe:
        case UD_Iseta:
        case UD_Isets:
        case UD_Isetns:
        case UD_Isetp:
        case UD_Isetnp:
        case UD_Isetl:
        case UD_Isetge:
        case UD_Isetle:
        case UD_Isetg:

            isSrc = false;
            break;
        default:
            break;
    }

    return (isSrc);
}

REGISTER_BITVECTOR Thirty2To16BitRegs(int reg) {
    //return reg;                 //TODO: fix the problem with vm traces!
    switch (reg) {
        case eAX_reg:
            return ax_reg;
        case eCX_reg:
            return cx_reg;
        case eDX_reg:
            return dx_reg;
        case eBX_reg:
            return bx_reg;
        case eSI_reg:
            return si_reg;
        case eBP_reg:
            return bp_reg;
        case eSP_reg:
            return sp_reg;
        case eDI_reg:
            return di_reg;
        case no_reg:
            return no_reg;
        default:
            return reg;             // in case of error, return the reg itself, does'nt hurt anything!
            //fprintf(stderr, "[Thirty2To16BitRegs] Unhandled register value: %d\n", reg);
            //break;
    }
    return 0;
}

void InstrRegsFalgsUseDef(InstrList *iList, Instruction *iins){
    
    
    /////////////////// RegDef ///////////////////////
    REGISTER_BITVECTOR def = 0x0;
    X86op tableEntry = GetTableEntry(iins);
    ud_operand_t op0 = iins->uInstr.operand[0];
    ud_operand_t op1 = iins->uInstr.operand[1];
    
    /*
     printf("%6d: [0x%08x] <0x%08x> %3d:: <0x%08x> %3d:: <0x%08x> %3d:: %s\n",
     instr->order, instr->addr,
     (unsigned int)tableEntry.x_op1, tableEntry.x_bytemode1,
     (unsigned int)tableEntry.x_op2, tableEntry.x_bytemode2,
     (unsigned int)tableEntry.x_op3, tableEntry.x_bytemode3,
     instr->uInstr.name);
     */
    
    //if x_op1 is not null, it represents a destination
    if (tableEntry.x_op1) {
        if (tableEntry.x_bytemode1 >= 98) {
            if (isNonEscapePrefixCode(iins->uInstr.bytes[0]) == 0x66) {
                def |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode1));
            } else
                def |= LynxRegisterToBitVector(tableEntry.x_bytemode1);
        } else if (op0.type == UD_OP_REG) {
            def |= UDRegisterToBitVector(op0.base);
        }
    }
    
    //need to handle xchng instr as special case, it has 2 dest's
    //xadd also has the second operand as destination. added by Babak.
    if (iins->uInstr.mnemonic == UD_Ixchg || iins->uInstr.mnemonic == UD_Ixadd) {
        if (op1.type == UD_OP_REG) {
            def |= UDRegisterToBitVector(op1.base);
        }
    }
    
    // else if "internal" call, then return value calculation should be
    // visible, no need to guess or set it
    
    //op2 will never have dest registers, so skip it
    
    //need to include implicit destination information, if any
    if (isNonEscapePrefixCode(iins->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
        def |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitDest1));
        def |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitDest2));
    } else {
        def |= LynxRegisterToBitVector(tableEntry.x_implicitDest1);
        def |= LynxRegisterToBitVector(tableEntry.x_implicitDest2);
    }
    
    
    /*
     * mov ax,[ss:eax]
     * above code does not handle this!
     * so add it manually
     */
    if(iins->uInstr.mnemonic == UD_Imov && iins->uInstr.operand[0].type == UD_OP_REG && def == 0){
    	def = UDRegisterToBitVector(iins->uInstr.operand[0].base);
    }
    
    //handle special cases (e.g. string operations)
    def |= GetImplicitRegDef(iins);
    
    //if(isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66){	//chnage all 32 bit regs to 16 bits
    //fprintf(stderr, "[RegDef]: %d: %s regsDef: %d\n", instr->order, instr->uInstr.name, def);
    
    //}
    
    /*
     * popad doesn't pop ESP
     */
    if (iins->uInstr.mnemonic == UD_Ipopa || iins->uInstr.mnemonic == UD_Ipopad) {
        def |= LYNX_ALL;
    }
    /*
     * 1. CMPXCHG
     * The reg-def depends on the ZF flag of next instruction in dynamic trace
     * (OP2 must be a register)
     */
    if (iins->uInstr.mnemonic == UD_Icmpxchg) {
        unsigned int eax1, eax2;
        eax1 = iins->eax;
        Instruction *temp_ins = GetInstruction(iList, (iins->order) + 1);
        eax2 = temp_ins->eax;
        if (eax1 == eax2) {
            if (iins->uInstr.operand[0].type == UD_OP_REG) {
                //if OP1 is EAX
                if (UDRegisterToBitVector(iins->uInstr.operand[0].base) == LYNX_EAX) {
                    ; //do nothing
                }
                // if OP1 is not EAX, remove EAX from def
                else {
                    def &= ~LYNX_EAX;
                }
            }
            //instr->uInstr.operand[0].type == UD_OP_MEM
            else {
                def &= ~LYNX_EAX;
            }
        }
        //ZF is cleared in next flag, i.e. OP1 != AL/AX/EAX/RAX
        else {
            if (iins->uInstr.operand[0].type == UD_OP_REG)
                def &= ~(UDRegisterToBitVector(iins->uInstr.operand[0].base));
        }
    }
    if(iins->uInstr.mnemonic == UD_Icall){
    	def |= LYNX_ESP;
    }
    iins->regDef = def;
    iins->regFlagsKnown |= RegDefKnown;
    
    
    /////////////////////// MemRegUsed /////////////////////
    int i;
    REGISTER_BITVECTOR use = 0x0;
    //X86op tableEntry = GetTableEntry(instr);
    
    for (i = 0; i < 3; i++) {
        if ((iins->uInstr.mnemonic != UD_Ilea) && (iins->uInstr.operand[i].type == UD_OP_MEM)) {
            //handle lea instructions as non-memory
            use |= UDRegisterToBitVector(iins->uInstr.operand[i].base);
            use |= UDRegisterToBitVector(iins->uInstr.operand[i].index);
        }
    }
    //handle implicit memory reg sources TODO: others?
    if (IsPush(iins) || IsPop(iins) || (iins->uInstr.mnemonic == UD_Iret)
        || (iins->uInstr.mnemonic == UD_Ileave) || (iins->uInstr.mnemonic == UD_Icall)) {
        use |= LynxRegisterToBitVector(tableEntry.x_implicitSource1);
        use |= LynxRegisterToBitVector(tableEntry.x_implicitSource2);
    }
    
    //handle special cases (e.g. string operations)
    use |= GetImplicitMemRegUsed(iins);
    iins->memRegUsed = use;
    iins->regFlagsKnown |= MemRegUsedKnown;
    
    /////////////////////// nonMemRegUsed /////////////////////
    
    
    //int i;
    use = 0x0;
    //X86op tableEntry = GetTableEntry(instr);
    //ud_operand_t op0 = iins->uInstr.operand[0];
    
    //first (0th) operand may be src, dest, or both. Handle as special case
    if (FirstOpIsSrc(iins)) {
        if (tableEntry.x_bytemode1 > 98) {
            if (isNonEscapePrefixCode(iins->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
                use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode1));
            } else {
                use |= LynxRegisterToBitVector(tableEntry.x_bytemode1);
            }
        }
        if (op0.type == UD_OP_REG) {
            use |= UDRegisterToBitVector(op0.base);
        }
    }
    
    //use bytemode info
    if (tableEntry.x_bytemode2 > 98) {
        if (isNonEscapePrefixCode(iins->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode2));
        } else {
            use |= LynxRegisterToBitVector(tableEntry.x_bytemode2);
        }
    }
    if (tableEntry.x_bytemode3 > 98) {
        if (isNonEscapePrefixCode(iins->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode3));
        } else {
            use |= LynxRegisterToBitVector(tableEntry.x_bytemode3);
        }
    }
    
    for (i = 1; i < 3; i++) {
        if (iins->uInstr.operand[i].type == UD_OP_REG) {
            use |= UDRegisterToBitVector(iins->uInstr.operand[i].base);
        } else if ((iins->uInstr.mnemonic == UD_Ilea) && (iins->uInstr.operand[i].type
                                                           == UD_OP_MEM)) {
            //handle lea as non-memory use of registers
            use |= UDRegisterToBitVector(iins->uInstr.operand[i].base);
            use |= UDRegisterToBitVector(iins->uInstr.operand[i].index);
        }
    }
    
    //handle implicit sources TODO: Other exceptions besides push/pop?
    if (!IsPush(iins) && !IsPop(iins) && (iins->uInstr.mnemonic != UD_Iret)
        && (iins->uInstr.mnemonic != UD_Ileave) && (iins->uInstr.mnemonic != UD_Icall)) {
        if (isNonEscapePrefixCode(iins->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitSource1));
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitSource2));
            
        } else {
            use |= LynxRegisterToBitVector(tableEntry.x_implicitSource1);
            use |= LynxRegisterToBitVector(tableEntry.x_implicitSource2);
        }
    }
    
    //handle special cases (e.g. string operations)
    use |= GetImplicitNonMemRegUsed(iins);
    
    /* Special Cases:
     *
     * CMPXCHG
     * The reg-used depends on the ZF flag of next instruction in dynamic trace
     * (OP2 must be a register)
     */
    if (iins->uInstr.mnemonic == UD_Icmpxchg) {
        unsigned int eax1, eax2;
        eax1 = iins->eax;
        Instruction *temp_ins = GetInstruction(iList, (iins->order) + 1);
        eax2 = temp_ins->eax;
        RInstrDestroy(iList, temp_ins);
        if (eax1 == eax2) { //ZF is set in next flag, i.e. OP1 == AL/AX/EAX/RAX
            ; //nothing to dom since OP1, OP2 and EAX all should be in reg use-set
        } else { //ZF is cleared in next flag, i.e. OP1 != AL/AX/EAX/RAX
            // then we have to remove OP2 from reg use-set
            use &= ~(UDRegisterToBitVector(iins->uInstr.operand[1].base));
        }
    }// end CMPXCHG

    
    iins->nonMemRegUsed = use;
    if(iins->uInstr.mnemonic == UD_Ipushad){
        iins->nonMemRegUsed = iins->memRegUsed;
        iins->memRegUsed = 0;
    }
    iins->regFlagsKnown |= NonMemRegUsedKnown;
    
    /////////////////////// FlagsDef /////////////////////
    //X86op tableEntry = GetTableEntry(instr);
    PSW_BITVECTOR flags = tableEntry.x_psw;
    
    /* TODO: this does not work I guess for floating point
     * instructions since there is another table just for
     * floating points but above method gives us an entry
     * from grps
     */
    if(iins->uInstr.mnemonic == UD_Ifnstcw || iins->uInstr.mnemonic == UD_Ifnstsw || iins->uInstr.mnemonic == UD_Ixchg){
    	flags = 0x0;
    }
    
    //tableEntry sets flags defined to all for call instructions, but
    //only applies to calls that task switch.
    if (!ALLOW_TASKSWITCH_CALLS && (iins->uInstr.mnemonic == UD_Icall)) {
        flags = 0x0;
    }
    iins->regFlagsKnown |= FlagsDefKnown;
    iins->flagsDef = flags;
    
    /////////////////////// FlagsUsed /////////////////////
    
    //int flags;
    //X86op tableEntry = GetTableEntry(instr);

    flags = tableEntry.x_pswUsed;
    if(iins->uInstr.mnemonic == UD_Ifnstcw || iins->uInstr.mnemonic == UD_Ifnstsw){
    	flags = 0x0;
    }
    iins->flagsUsed = flags;
    iins->regFlagsKnown |= FlagsUsedKnown;
    
    return;
}

REGISTER_BITVECTOR RegDef(InstrList *iList, Instruction *instr) {
    if (instr->regFlagsKnown & RegDefKnown) {
        return (instr->regDef);
    }

    REGISTER_BITVECTOR def = 0x0;
    X86op tableEntry = GetTableEntry(instr);
    ud_operand_t op0 = instr->uInstr.operand[0];
    ud_operand_t op1 = instr->uInstr.operand[1];

    /*
     printf("%6d: [0x%08x] <0x%08x> %3d:: <0x%08x> %3d:: <0x%08x> %3d:: %s\n",
     instr->order, instr->addr,
     (unsigned int)tableEntry.x_op1, tableEntry.x_bytemode1,
     (unsigned int)tableEntry.x_op2, tableEntry.x_bytemode2,
     (unsigned int)tableEntry.x_op3, tableEntry.x_bytemode3,
     instr->uInstr.name);
     */

    //if x_op1 is not null, it represents a destination
    if (tableEntry.x_op1) {
        if (tableEntry.x_bytemode1 >= 98) {
            if (isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66) {
                def |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode1));
            } else
                def |= LynxRegisterToBitVector(tableEntry.x_bytemode1);
        } else if (op0.type == UD_OP_REG) {
            def |= UDRegisterToBitVector(op0.base);
        }
    }

    //need to handle xchng instr as special case, it has 2 dest's
    //xadd also has the second operand as destination. added by Babak.
    if (instr->uInstr.mnemonic == UD_Ixchg || instr->uInstr.mnemonic == UD_Ixadd) {
        if (op1.type == UD_OP_REG) {
            def |= UDRegisterToBitVector(op1.base);
        }
    }

    // else if "internal" call, then return value calculation should be
    // visible, no need to guess or set it

    //op2 will never have dest registers, so skip it

    //need to include implicit destination information, if any
    if (isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
        def |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitDest1));
        def |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitDest2));
    } else {
        def |= LynxRegisterToBitVector(tableEntry.x_implicitDest1);
        def |= LynxRegisterToBitVector(tableEntry.x_implicitDest2);
    }


    /*
     * mov ax,[ss:eax]
     * above code does not handle this!
     * so add it manually
     */
    if(instr->uInstr.mnemonic == UD_Imov && instr->uInstr.operand[0].type == UD_OP_REG && def == 0){
    	def = UDRegisterToBitVector(instr->uInstr.operand[0].base);
    }

    //handle special cases (e.g. string operations)
    def |= GetImplicitRegDef(instr);

    //if(isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66){	//chnage all 32 bit regs to 16 bits
    //fprintf(stderr, "[RegDef]: %d: %s regsDef: %d\n", instr->order, instr->uInstr.name, def);

    //}

    /*
     * popad doesn't pop ESP
     */
    if (instr->uInstr.mnemonic == UD_Ipopa || instr->uInstr.mnemonic == UD_Ipopad) {
        def |= LYNX_ALL;
    }
    /*
     * 1. CMPXCHG
     * The reg-def depends on the ZF flag of next instruction in dynamic trace
     * (OP2 must be a register)
     */
    if (instr->uInstr.mnemonic == UD_Icmpxchg) {
        unsigned int eax1, eax2;
        eax1 = instr->eax;
        Instruction *temp_ins = GetInstruction(iList, (instr->order) + 1);
        eax2 = temp_ins->eax;
        if (eax1 == eax2) {
            if (instr->uInstr.operand[0].type == UD_OP_REG) {
                //if OP1 is EAX
                if (UDRegisterToBitVector(instr->uInstr.operand[0].base) == LYNX_EAX) {
                    ; //do nothing
                }
                // if OP1 is not EAX, remove EAX from def
                else {
                    def &= ~LYNX_EAX;
                }
            }
            //instr->uInstr.operand[0].type == UD_OP_MEM
            else {
                def &= ~LYNX_EAX;
            }
        }
        //ZF is cleared in next flag, i.e. OP1 != AL/AX/EAX/RAX
        else {
            if (instr->uInstr.operand[0].type == UD_OP_REG)
                def &= ~(UDRegisterToBitVector(instr->uInstr.operand[0].base));
        }
    }
    if(instr->uInstr.mnemonic == UD_Icall){
    	def |= LYNX_ESP;
    }
    return (def);
}

REGISTER_BITVECTOR MemRegUsed(Instruction *instr) {
    if (instr->regFlagsKnown & MemRegUsedKnown) {
        return (instr->memRegUsed);
    }

    int i;
    REGISTER_BITVECTOR use = 0x0;
    X86op tableEntry = GetTableEntry(instr);

    for (i = 0; i < 3; i++) {
        if ((instr->uInstr.mnemonic != UD_Ilea) && (instr->uInstr.operand[i].type == UD_OP_MEM)) {
            //handle lea instructions as non-memory
            use |= UDRegisterToBitVector(instr->uInstr.operand[i].base);
            use |= UDRegisterToBitVector(instr->uInstr.operand[i].index);
        }
    }
    //handle implicit memory reg sources TODO: others?
    if (IsPush(instr) || IsPop(instr) || (instr->uInstr.mnemonic == UD_Iret)
            || (instr->uInstr.mnemonic == UD_Ileave) || (instr->uInstr.mnemonic == UD_Icall)) {
        use |= LynxRegisterToBitVector(tableEntry.x_implicitSource1);
        use |= LynxRegisterToBitVector(tableEntry.x_implicitSource2);
    }

    //handle special cases (e.g. string operations)
    use |= GetImplicitMemRegUsed(instr);

    return (use);
}

/*
 * calculate any registers used by instruction that do NOT
 * contribute to calculation of a memory address
 */
REGISTER_BITVECTOR NonMemRegUsed(InstrList *iList, Instruction *instr) {
    if (instr->regFlagsKnown & NonMemRegUsedKnown) {
        return (instr->nonMemRegUsed);
    }

    int i;
    REGISTER_BITVECTOR use = 0x0;
    X86op tableEntry = GetTableEntry(instr);
    ud_operand_t op0 = instr->uInstr.operand[0];

    //first (0th) operand may be src, dest, or both. Handle as special case
    if (FirstOpIsSrc(instr)) {
        if (tableEntry.x_bytemode1 > 98) {
            if (isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
                use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode1));
            } else {
                use |= LynxRegisterToBitVector(tableEntry.x_bytemode1);
            }
        }
        if (op0.type == UD_OP_REG) {
            use |= UDRegisterToBitVector(op0.base);
        }
    }

    //use bytemode info
    if (tableEntry.x_bytemode2 > 98) {
        if (isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode2));
        } else {
            use |= LynxRegisterToBitVector(tableEntry.x_bytemode2);
        }
    }
    if (tableEntry.x_bytemode3 > 98) {
        if (isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_bytemode3));
        } else {
            use |= LynxRegisterToBitVector(tableEntry.x_bytemode3);
        }
    }

    for (i = 1; i < 3; i++) {
        if (instr->uInstr.operand[i].type == UD_OP_REG) {
            use |= UDRegisterToBitVector(instr->uInstr.operand[i].base);
        } else if ((instr->uInstr.mnemonic == UD_Ilea) && (instr->uInstr.operand[i].type
                == UD_OP_MEM)) {
            //handle lea as non-memory use of registers
            use |= UDRegisterToBitVector(instr->uInstr.operand[i].base);
            use |= UDRegisterToBitVector(instr->uInstr.operand[i].index);
        }
    }

    //handle implicit sources TODO: Other exceptions besides push/pop?
    if (!IsPush(instr) && !IsPop(instr) && (instr->uInstr.mnemonic != UD_Iret)
            && (instr->uInstr.mnemonic != UD_Ileave) && (instr->uInstr.mnemonic != UD_Icall)) {
        if (isNonEscapePrefixCode(instr->uInstr.bytes[0]) == 0x66) { //chnage all 32 bit regs to 16 bits
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitSource1));
            use |= LynxRegisterToBitVector(Thirty2To16BitRegs(tableEntry.x_implicitSource2));

        } else {
            use |= LynxRegisterToBitVector(tableEntry.x_implicitSource1);
            use |= LynxRegisterToBitVector(tableEntry.x_implicitSource2);
        }
    }

    //handle special cases (e.g. string operations)
    use |= GetImplicitNonMemRegUsed(instr);

    /* Special Cases:
     *
     * CMPXCHG
     * The reg-used depends on the ZF flag of next instruction in dynamic trace
     * (OP2 must be a register)
     */
    if (instr->uInstr.mnemonic == UD_Icmpxchg) {
        unsigned int eax1, eax2;
        eax1 = instr->eax;
        Instruction *temp_ins = GetInstruction(iList, (instr->order) + 1);
        eax2 = temp_ins->eax;
        RInstrDestroy(iList, temp_ins);
        if (eax1 == eax2) { //ZF is set in next flag, i.e. OP1 == AL/AX/EAX/RAX
            ; //nothing to dom since OP1, OP2 and EAX all should be in reg use-set
        } else { //ZF is cleared in next flag, i.e. OP1 != AL/AX/EAX/RAX
            // then we have to remove OP2 from reg use-set
            use &= ~(UDRegisterToBitVector(instr->uInstr.operand[1].base));
        }
    }// end CMPXCHG

    //TODO: should we include ENTER and LEAVE?

    return (use);
}

REGISTER_BITVECTOR RegUsed(InstrList *iList, Instruction *instr) {
    REGISTER_BITVECTOR memRegUsed = 0x0, nonMemRegUsed = 0x0;

    if (instr->regFlagsKnown & MemRegUsedKnown) {
        memRegUsed = instr->memRegUsed;
    } else {
        memRegUsed = MemRegUsed(instr);
    }

    if (instr->regFlagsKnown & NonMemRegUsedKnown) {
        nonMemRegUsed = instr->nonMemRegUsed;
    } else {
        nonMemRegUsed = NonMemRegUsed(iList, instr);
    }

    return (memRegUsed | nonMemRegUsed);
}

/*
 * RegUsedInclSP() returns the registers used by an instruction, including
 * esp (for push/pop instrs).
 */
REGISTER_BITVECTOR RegUsedInclSP(InstrList *iList, Instruction *iins) {
    REGISTER_BITVECTOR used;

    used = RegUsed(iList, iins);
    if (IsPush(iins) || IsPop(iins)) {
        used |= LYNX_ESP;
        used |= LYNX_ESP;
    }

    return used;
}

/*
 * RegDefInclSP() returns the registers defined by an instruction, including
 * esp (for push/pop instrs).
 */
REGISTER_BITVECTOR RegDefInclSP(InstrList *iList, Instruction *iins) {
    REGISTER_BITVECTOR def, used;

    def = RegDef(iList, iins);
    used = RegUsed(iList, iins);
    if (IsPush(iins) || IsPop(iins)) {
        def |= (LYNX_ESP | used);
        def |= (LYNX_ESP | used);
    }

    return def;
}

PSW_BITVECTOR FlagsDef(Instruction *instr) {
    if (instr->regFlagsKnown & FlagsDefKnown) {
        return (instr->flagsDef);
    }
    X86op tableEntry = GetTableEntry(instr);
    PSW_BITVECTOR flags = tableEntry.x_psw;

    /* TODO: this does not work I guess for floating point
     * instructions since there is another table just for
     * floating points but above method gives us an entry
     * from grps
     */
    if(instr->uInstr.mnemonic == UD_Ifnstcw || instr->uInstr.mnemonic == UD_Ifnstsw || instr->uInstr.mnemonic == UD_Ixchg){
    	flags = 0x0;
    }

    //tableEntry sets flags defined to all for call instructions, but
    //only applies to calls that task switch.
    if (!ALLOW_TASKSWITCH_CALLS && (instr->uInstr.mnemonic == UD_Icall)) {
        flags = 0x0;
    }
    instr->regFlagsKnown |= FlagsDefKnown;
    instr->flagsDef = flags;
    return (flags);
}

PSW_BITVECTOR FlagsUsed(Instruction *instr) {
    if (instr->regFlagsKnown & FlagsUsedKnown) {
        return (instr->flagsUsed);
    }
    int flags;
    X86op tableEntry = GetTableEntry(instr);
    instr->regFlagsKnown |= FlagsUsedKnown;
    flags = instr->flagsUsed = tableEntry.x_pswUsed;
    if(instr->uInstr.mnemonic == UD_Ifnstcw || instr->uInstr.mnemonic == UD_Ifnstsw){
    	flags = 0x0;
    }
    return flags;
}

WriteSet *MemDef(InstrList *iList, Instruction *instr) {
    //if already calculated
    if (instr->regFlagsKnown & MemDefKnown) {
        WriteSet *ws = WriteSetCreate();
        if (instr->memDefMin != 0x0) {
            //if known not to be null
            Range *r = RangeCreate(instr->memDefMin, instr->memDefMin + instr->memDefSize-1);
            AddRangeToWriteSet(ws, r);
        }

        return (ws);
    }

    //else calculate it now
    WriteSet *ret = GetWriteTargets(iList, instr);

    /* Special Cases
     *
     * CMPXCHG
     * The reg-used depends on the ZF flag of next instruction in dynamic trace
     * (OP2 must be a register)
     */
    if (instr->uInstr.mnemonic == UD_Icmpxchg) {
        unsigned int eax1, eax2;
        eax1 = instr->eax;
        Instruction *temp_ins = GetInstruction(iList, (instr->order) + 1);
        eax2 = temp_ins->eax;
        if (eax1 == eax2) { //ZF is set in next flag, i.e. OP1 == AL/AX/EAX/RAX
            ;
        } else { //ZF is cleared in next flag, i.e. OP1 != AL/AX/EAX/RAX
            if (instr->uInstr.operand[0].type == UD_OP_MEM) {
                WriteSetDestroy(ret);
                ret = WriteSetCreate();
            }
        }
    }// end CMPXCHG

    return (ret);
}

WriteSet *MemUsed(InstrList *iList, Instruction *instr) {
    //if already calculated
    if (instr->regFlagsKnown & MemUsedKnown) {
        WriteSet *ws = WriteSetCreate();
        if (instr->memUsedMin != 0x0) {
            //if known not to be null
            Range *r = RangeCreate(instr->memUsedMin, instr->memUsedMin + instr->memUsedSize-1);
            AddRangeToWriteSet(ws, r);
        }

        return (ws);
    }

    //else calculate it now
    int i, start = 1;
    Range *target = NULL;
    WriteSet *ws = WriteSetCreate();

    //first (0th) operand may be src, dest, or both. Handle as special case
    if (FirstOpIsSrc(instr)) {
        start = 0;
    }

    //for the last two operands
    for (i = start; i < 3; i++) {
        if (instr->uInstr.operand[i].type == UD_OP_MEM) {
            target = GetTarget(instr, instr->uInstr.operand[i]);
            AddRangeToWriteSet(ws, target);
        }
    }

    // get implicit targets
    target = GetImplicitMemSource(iList, instr);
    AddRangeToWriteSet(ws, target);

    return (ws);
}

/*
 * This function should be called once when a new instrFile is being
 * created.  It makes a single pass to pre-calculate mem, reg, and flag uses
 * and defs so that these values need not be calculated over and over
 * for later use.
 *
 * Note that we do not include this calculation when reading in the std
 * trace file, because some mem def and use calculation depend on the
 * instr *after* the current instr (which would not have been read yet).
 */
void CalcKnownDefsAndUses(InstrList *iList) {
    int i = 0;
    Instruction *prev = NULL;
    Instruction *next;

    /*iList->numAddrRanges = 0;
    iList->addrRangeHead = malloc(sizeof(AddrRange));
    iList->addrRangeHead->start = -1;
    iList->addrRangeHead->end = -1;
    iList->addrRangeHead->next = NULL;*/

    for (i = 0; i < numInstrs; i ++) {
        next = GetInstruction(iList, i);

        next->ControlDependentOn = -1;      // not dependent yet!
        InstrRegsFalgsUseDef(iList, next);

        WriteSet *ws = MemDef(iList, next);
        if (WriteSetIsEmpty(ws)) {
            next->regFlagsKnown |= MemDefKnown;
            next->memDefMin = 0;
            next->memDefSize = 0;
        } else {

            int numRanges = al_size(ws->ranges);
            if (numRanges == 1) {
                next->regFlagsKnown |= MemDefKnown;
                Range *r = al_get(ws->ranges, 0);
                next->memDefMin = r->min;
                next->memDefSize = r->max - r->min + 1;
            }
            if(next->memDefMin != 0){
            	INSTR_SET_FLAG(next, INSTR_WRITES_MEM);
            }
        }
        WriteSetDestroy(ws);

        ws = MemUsed(iList, next);
        if (WriteSetIsEmpty(ws)) {
            next->regFlagsKnown |= MemUsedKnown;
            next->memUsedMin = 0;
            next->memUsedSize = 0;
        } else {

            int numRanges = al_size(ws->ranges);
            if (numRanges == 1) {
                next->regFlagsKnown |= MemUsedKnown;
                Range *r = al_get(ws->ranges, 0);
                next->memUsedMin = r->min;
                next->memUsedSize = r->max - r->min + 1;
            }
            if(next->memUsedMin){
            	INSTR_SET_FLAG(next, INSTR_READS_MEM);
            }
        }
        WriteSetDestroy(ws);

        if(prev && IsCondJump(prev)){
            ADDRESS target = prev->uInstr.operand[0].lval.sdword+prev->uInstr.length+prev->addr;
            if(target == next->addr){
                INSTR_SET_FLAG(prev, INSTR_CONDJUMP_TAKEN);
            }
        }


        SaveInstrChange(iList, next);
        prev = next;
    }
}

/*void *CalcKnownDefsAndUsesPar(void *args) {
    int i = 0;
    //Instruction *prev = NULL;

    ThreadArgs arg = *(ThreadArgs *) args;
    int Start = arg.first;
    int End = arg.last;
    InstrList *iList = arg.i_list;

    for (i = Start; i < End; i++) {
        Instruction *next = GetInstruction(iList, i);

        next->ControlDependentOn = -1;      // not dependent yet!

        next->regDef = RegDef(iList, next);
        next->regFlagsKnown |= RegDefKnown;

        next->memRegUsed = MemRegUsed(next);
        next->regFlagsKnown |= MemRegUsedKnown;
        next->nonMemRegUsed = NonMemRegUsed(iList, next);
        if(next->uInstr.mnemonic == UD_Ipushad){
            next->nonMemRegUsed = next->memRegUsed;
        }
        next->regFlagsKnown |= NonMemRegUsedKnown;

        next->flagsDef = FlagsDef(next);
        next->regFlagsKnown |= FlagsDefKnown;

        next->flagsUsed = FlagsUsed(next);
        next->regFlagsKnown |= FlagsUsedKnown;

        WriteSet *ws = MemDef(iList, next);
        if (WriteSetIsEmpty(ws)) {
            next->regFlagsKnown |= MemDefKnown;
            next->memDefMin = 0;
        } else {
            INSTR_SET_FLAG(next, INSTR_WRITES_MEM);
            int numRanges = al_size(ws->ranges);
            if (numRanges == 1) {
                next->regFlagsKnown |= MemDefKnown;
                Range *r = al_get(ws->ranges, 0);
                next->memDefMin = r->min;
            }
        }
        WriteSetDestroy(ws);

        ws = MemUsed(iList, next);
        if (WriteSetIsEmpty(ws)) {
            next->regFlagsKnown |= MemUsedKnown;
            next->memUsedMin = 0;
        } else {
            INSTR_SET_FLAG(next, INSTR_READS_MEM);
            int numRanges = al_size(ws->ranges);
            if (numRanges == 1) {
                next->regFlagsKnown |= MemUsedKnown;
                Range *r = al_get(ws->ranges, 0);
                next->memUsedMin = r->min;
            }
        }
        WriteSetDestroy(ws);

        SaveInstrChange(iList, next);
        RInstrDestroy(iList, next);

        //if(prev && INSTR_HAS_FLAG(next, INSTR_IN_BASE_MODULE) && !INSTR_HAS_FLAG(prev, INSTR_IN_BASE_MODULE)){
        //  INSTR_SET_FLAG(prev, INSTR_USED_AS_CALL);
        //  SaveInstrChange(iList, prev);
        //}

        //free(prev);
        //prev = next;

    }
    return NULL;
}*/


/**********************************************************************
 *                                                                    *
 *             CREATING UDIS86 INSTRUCTION REPRESENTATIONS            *
 *                                                                    *
 **********************************************************************/

/*
 * UDCreatePushReg() -- create an instruction to push a register.
 */
ud_t *UDCreatePushReg(enum ud_type reg) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0x50 + (reg - UD_R_EAX);
        udins = alloc(sizeof(ud_t));
        InitUdis(udins);
        ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
        ud_disassemble(udins);
        return udins;
    }

    if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0x50 + (reg - UD_R_AX);
        udins = alloc(sizeof(ud_t));
        InitUdis(udins);
        ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
        ud_disassemble(udins);
        return udins;
    }

    fprintf(stderr, "[UDCreatePushReg]: unrecognized register value 0x%x\n", reg);
    return NULL;
}

void ExcludeNonRealMain(InstrList *iList){
	Instruction *iins, *lastGood = NULL;
	SetCurrentInstr(iList, iList->startMarker);

	while((iins = FetchNextInstr(iList))){
		if(iins->uInstr.mnemonic == UD_Iadd && iins->uInstr.operand[1].type == UD_OP_IMM && iins->uInstr.operand[1].lval.sdword == 0xabc0000){
			lastGood = GetInstruction(iList, iins->prev);
		}
		if(iins->uInstr.mnemonic == UD_Iadd && iins->uInstr.operand[1].type == UD_OP_IMM && iins->uInstr.operand[1].lval.sdword == 0xabc9999 && lastGood){
			iins = FetchNextInstr(iList);
			lastGood->next = iins->order;
			iins->prev = lastGood->order;
			SaveInstrChange(iList, iins);
			SaveInstrChange(iList, lastGood);
			lastGood = NULL;
		}
	}

}

ud_t *UDCreateCallToAddr(int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i, j;
    i = 0;

    buf[i++] = 0xe8;
    for (j = 0; j < 4; j++) {
        buf[i++] = (val & 0xff);
        val = (val >> 8);
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreatePopReg() -- create an instruction to pop a register.
 */
ud_t *UDCreatePopReg(enum ud_type reg) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0x58 + (reg - UD_R_EAX);
        udins = alloc(sizeof(ud_t));
        InitUdis(udins);
        ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
        ud_disassemble(udins);
        return udins;
    }

    if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0x58 + (reg - UD_R_AX);
        udins = alloc(sizeof(ud_t));
        InitUdis(udins);
        ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
        ud_disassemble(udins);
        return udins;
    }

    fprintf(stderr, "[UDCreatePopReg]: unrecognized register value 0x%x\n", reg);
    return NULL;
}

/*
 * UDCreateRet() -- create a return instruction with an
 * with an arbitrary value to pop from stack.
 */

ud_t *UDCreateRet(int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    buf[0] = 0xc2;
    buf[1] = (val & 0xff);
    buf[2] = ((val >> 8) & 0xff);

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    if (val < 0)
        fprintf(stderr, "[UDCreateRet] Error: %d is less than 0\n", val);
    return udins;
}

/*
 * ChangeXchgMemRegToMov() -- change a xchg instruction with first memory
 * operand and the second register operand to a move with same operands
 * exchangeOps: when 1, second operand in xchg becomes first in mov and so on
 * in the same order otherwise.
 */
ud_t *ChangeXchgMemRegToMov(unsigned char *bytes, int exchangeOps) {
    int reg;
    ud_t *udinsold, *udins;
    unsigned char *buff = malloc(MAX_INSTRLEN);
    memcpy(buff, bytes, MAX_INSTRLEN);
    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);
    reg = udinsold->operand[1].base;
    if (buff[0] == 0x66) {
        if ((reg >= UD_R_EAX && reg <= UD_R_EDI) || (reg >= UD_R_AX && reg <= UD_R_DI)) {
            if (exchangeOps)
                buff[1] = 0x8B;
            else
                buff[1] = 0x89;
        } else {
            if (exchangeOps)
                buff[1] = 0x8A;
            else
                buff[1] = 0x88;
        }
    } else {
        if ((reg >= UD_R_EAX && reg <= UD_R_EDI) || (reg >= UD_R_AX && reg <= UD_R_DI)) {
            if (exchangeOps)
                buff[0] = 0x8B;
            else
                buff[0] = 0x89;
        } else {
            if (exchangeOps)
                buff[0] = 0x8A;
            else
                buff[0] = 0x88;
        }
    }
    ud_set_input_buffer(udins, buff, MAX_INSTRLEN);
    ud_disassemble(udins);

    /* check: operands should be same!
     if(!((!(exchangeOps) && udinsold->operand[0].type == udins->operand[0].type && udinsold->operand[0].base == udins->operand[0].base &&
     udinsold->operand[0].index == udins->operand[0].index && udinsold->operand[0].offset == udins->operand[0].offset &&
     udinsold->operand[0].scale == udins->operand[0].scale)) || !(((exchangeOps) && udinsold->operand[1].type == udins->operand[0].type &&
     udinsold->operand[1].base == udins->operand[0].base && udinsold->operand[1].index == udins->operand[0].index &&
     udinsold->operand[1].offset == udins->operand[0].offset && udinsold->operand[1].scale == udins->operand[0].scale)))
     fprintf(stderr, "[ChangeXchgMemRegToMov] Warning: %s changed to %s incorrectly!\n", ud_insn_asm(udinsold), ud_insn_asm(udins));
     */
    free(udinsold);
    return udins;
}

ud_t *UDCreateImulImmWithReg(enum ud_type reg, unsigned int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateImulImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xb8 + (reg - UD_R_EAX);
        buf[1] = (val >> 0) & 0xff;
        buf[2] = (val >> 8) & 0xff;
        buf[3] = (val >> 16) & 0xff;
        buf[4] = (val >> 24) & 0xff;
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xb8 + (reg - UD_R_AX);
        buf[2] = (val >> 0) & 0xff;
        buf[3] = (val >> 8) & 0xff;
    } else { // 8-bit
        buf[0] = 0xb0 + (reg - UD_R_AL);
        buf[1] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

ud_t *UDCreateMovRegToReg(enum ud_type destination, enum ud_type source) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if(UDRegSize(source) != UDRegSize(destination)){
        //fprintf(stderr, "[UDCreateMovRegToReg] Error, operands are not the same size!\n");
        return NULL;
    }
    if ((destination >= UD_R_EAX && destination <= UD_R_EDI)) {
        buf[0] = 0x89;
        buf[1] = (0xc0 | (destination - UD_R_EAX) | ((source - UD_R_EAX) << 3));       // mod/reg/rm
    } else if (destination >= UD_R_AX && destination <= UD_R_DI) { // 16-bit
    	buf[0] = 0x66;
        buf[1] = 0x89;
        buf[2] = (0xc0 | (destination - UD_R_AX) | ((source - UD_R_AX) << 3));       // mod/reg/rm
    } else {
        buf[0] = 0x88;
        buf[1] = (0xc0 | (destination - UD_R_AL) | ((source - UD_R_AL) << 3));       // mod/reg/rm
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}
/*
 * UDCreateMovImmToReg() -- create an instruction to move an
 * immediate value into a register.
 */
ud_t *UDCreateMovImmToReg(enum ud_type reg, unsigned int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
    	fprintf(stderr, "[UDCreateMovImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xb8 + (reg - UD_R_EAX);
        buf[1] = (val >> 0) & 0xff;
        buf[2] = (val >> 8) & 0xff;
        buf[3] = (val >> 16) & 0xff;
        buf[4] = (val >> 24) & 0xff;
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xb8 + (reg - UD_R_AX);
        buf[2] = (val >> 0) & 0xff;
        buf[3] = (val >> 8) & 0xff;
    } else { // 8-bit
        buf[0] = 0xb0 + (reg - UD_R_AL);
        buf[1] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

ud_t *UDCreateShlImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreatesShlImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xc1;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x4) << 3);
        buf[2] = (val >> 0) & 0xff; // imm values are just 1 byte size in all instruction formats
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x4) << 3);
        buf[3] = (val >> 0) & 0xff;
    } else { // 8-bit
        buf[0] = 0xc0;
        buf[1] = 0xc0 + ((reg - UD_R_AL + 0x4) << 3);
        buf[2] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}

ud_t *UDCreateSalImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateShrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xc1;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x4) << 3);
        buf[2] = (val >> 0) & 0xff; // imm values are just 1 byte size in all instruction formats
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x4) << 3);
        buf[3] = (val >> 0) & 0xff;
    } else { // 8-bit
        buf[0] = 0xc0;
        buf[1] = 0xc0 + ((reg - UD_R_AL + 0x4) << 3);
        buf[2] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}

ud_t *UDCreateShrImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateShrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xc1;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x5) << 3);
        buf[2] = (val >> 0) & 0xff; // imm values are just 1 byte size in all instruction formats
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x5) << 3);
        buf[3] = (val >> 0) & 0xff;
    } else { // 8-bit
        buf[0] = 0xc0;
        buf[1] = 0xc0 + ((reg - UD_R_AL + 0x5) << 3);
        buf[2] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}

ud_t *UDCreateSarImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateShrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xc1;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x7) << 3);
        buf[2] = (val >> 0) & 0xff; // imm values are just 1 byte size in all instruction formats
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x7) << 3);
        buf[3] = (val >> 0) & 0xff;
    } else { // 8-bit
        buf[0] = 0xc0;
        buf[1] = 0xc0 + ((reg - UD_R_AL + 0x7) << 3);
        buf[2] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}
ud_t *UDCreateRorImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateShrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xc1;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x1) << 3);
        buf[2] = (val >> 0) & 0xff; // imm values are just 1 byte size in all instruction formats
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x1) << 3);
        buf[3] = (val >> 0) & 0xff;
    } else { // 8-bit
        buf[0] = 0xc0;
        buf[1] = 0xc0 + ((reg - UD_R_AL + 0x1) << 3);
        buf[2] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}
ud_t *UDCreateRolImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateShrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xc1;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x0) << 3);
        buf[2] = (val >> 0) & 0xff; // imm values are just 1 byte size in all instruction formats
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x0) << 3);
        buf[3] = (val >> 0) & 0xff;
    } else { // 8-bit
        buf[0] = 0xc0;
        buf[1] = 0xc0 + ((reg - UD_R_AL + 0x0) << 3);
        buf[2] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}
ud_t *UDCreateImulImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI)) {
        fprintf(stderr, "[UDCreateShrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0x69;
        buf[1] = 0xc0 + ((reg - UD_R_EAX + 0x0) << 3) + ((reg - UD_R_EAX + 0x0)) ;
        for(i = 0; i < 4; i++){
            buf[2+i] = (val >> 0) & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xc1;
        buf[2] = 0xc0 + ((reg - UD_R_AX + 0x0) << 3) + ((reg - UD_R_AX + 0x0)) ;
        for(i = 0; i < 4; i++){
            buf[3+i] = (val >> 0) & 0xff;
            val = val >> 8;
        }
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}

/*
 * UDCreateCmpMemWithImm() -- create an instruction to compare a
 * memory location with an immediate value. TODO: review correctness of newly created instructions!
 */
ud_t *UDCreateCmpMemWithImm(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);
    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateCmpImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    int opcode2 = (MODR | RM | (7 << 3));
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }
    if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }
    free(udinsold);
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    // check if destination is correct
    free(udinsold);
    return udins;
}

/*
 * UDCreateXorMemWithImm() -- create an instruction to xor a
 * memory location with an immediate value.
 */
ud_t *UDCreateXorMemWithImm(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);
    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateXorImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    int opcode2 = (MODR | RM | (6 << 3));
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }
    if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }
    free(udinsold);
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    // check if destination is correct
    free(udinsold);
    return udins;
}

ud_t *UDCreateNotOnReg(enum ud_type reg) {
    unsigned char buf[MAX_INSTRLEN];

    ud_t *udins;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateNotOnReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[0] = 0xf7;
        buf[1] = 0xd0 + ((reg - UD_R_EAX + 0x0));
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[0] = 0x66;
        buf[1] = 0xf7;
        buf[2] = 0xd0 + ((reg - UD_R_AX));

    } else { // 8-bit
        buf[0] = 0xf6;
        buf[1] = 0xd0 + ((reg - UD_R_AL));
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}

ud_t *UDCreateNotMem(int offset, int size) {
    unsigned char buf[MAX_INSTRLEN];

    ud_t *udins;
    int i = 0;
    int j = 0;

    if (size == 4) { // 32-bit
        buf[j++] = 0xf7;
        buf[j++] = 0x15;
    } else if (size == 2) { // 16-bit
    	buf[j++] = 0x66;
        buf[j++] = 0xf7;
        buf[j++] = 0x15;
    } else { // 8-bit
        buf[j++] = 0xf6;
        buf[j++] = 0x15;
    }
    for(i =0; i < 4; i++){
    	buf[j++] = (offset & 0xff);
    	offset = offset >> 8;
    }
    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;

}

/*
 * UDCreateAndMemWithImm() -- create an instruction to and a
 * memory location with an immediate value.
 */
ud_t *UDCreateAndMemWithImm(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);
    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateAndImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    int opcode2 = (MODR | RM | (4 << 3));
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }
    if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }
    free(udinsold);
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    free(udinsold);
    return udins;
}

/*
 * UDCreateOrMemWithImm() -- create an instruction to or a
 * memory location with an immediate value.
 */
ud_t *UDCreateOrMemWithImm(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);
    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateOrImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    int opcode2 = (MODR | RM | (1 << 3));
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }
    if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }
    free(udinsold);
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    free(udinsold);
    return udins;
}

/*
 * UDCreateSubImmFromMem() -- create an instruction to subtract an
 * immediate value from a memory address. TODO: review correctness of newly created instructions!
 */
ud_t *UDCreateSubImmFromMem(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);

    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateSubImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    int opcode2 = (MODR | (0x5 << 3) | RM);
    if (RM == 0x04) {
        udinsold->inp_curr++;
    }

    if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }

    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    free(udinsold);
    return udins;
}
/*ud_t *UDCreatePushReg(enum ud_type dest){
 unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
 ud_t *udins;
 if ( !(dest >= UD_R_EAX && dest <= UD_R_EDI)
 && !(dest >= UD_R_AX && dest <= UD_R_DI)) {
 fprintf(stderr, "[UDCreatePushReg]: unrecognized register value 0x%x\n", dest);
 return NULL;
 }
 if ((dest >= UD_R_EAX && dest <= UD_R_EDI))
 buf[0] = 0x50 + (dest - UD_R_EAX);
 else if((dest >= UD_R_AX && dest <= UD_R_DI)){
 buf[0] = 0x66;
 buf[1] = 0x50 + (dest - UD_R_AX);
 }
 InitUdis(udins);
 ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
 ud_disassemble(udins);
 return udins;
 }*/

ud_t *UDCreatePopOffset(unsigned char *bytes, int offset) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (bytes[0] != 0x66) {
        buf[i++] = 0x8f;
        buf[i++] = 0x05;
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    } else {
        buf[i++] = 0x66;
        buf[i++] = 0x8f;
        buf[i++] = 0x35;
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    }

    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;
}

ud_t *UDCreatePushOffset(unsigned char *bytes, int offset) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (bytes[0] != 0x66) {
        buf[i++] = 0xff;
        buf[i++] = 0x35;
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    } else {
        buf[i++] = 0x66;
        buf[i++] = 0xff;
        buf[i++] = 0x35;
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    }

    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;
}


/* UDCreateMovImmToOffset
 * Creates an instruction in form of:
 * mov [offs.], const.
 *
 */

ud_t *UDCreateMovImmToOffset(int offset, int value, int size) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if(size == 4){
        buf[i++] = 0xc7;       // mov opcode
        buf[i++] = 0x05;       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 4; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else if (size == 2){
        buf[i++] = 0x66;        // 16 bit prefix
        buf[i++] = 0xc7;       // mov opcode
        buf[i++] = 0x05;       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 2; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else {
        buf[i++] = 0xc6;       // mov opcode
        buf[i++] = 0x05;       // modr/m
        for(j = 0; j < 4; j++){                 // I know it is not a loop!
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 1; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;

}

ud_t *UDCreateAddImmToOffset(int offset, int value, int size) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if(size == 4){
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0x05;       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 4; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else if (size == 2){
        buf[i++] = 0x66;        // 16 bit prefix
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0x05;       // modr/m
        for(j = 0; j < 2; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 2; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else {
        buf[i++] = 0x80;       // mov opcode
        buf[i++] = 0x05;       // modr/m
        for(j = 0; j < 1; j++){                 // I know it is not a loop!
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 1; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;

}

ud_t *UDCreateSubImmFromOffset(int offset, int value, int size) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if(size == 4){
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0x2d;       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 4; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else if (size == 2){
        buf[i++] = 0x66;        // 16 bit prefix
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0x2d;       // modr/m
        for(j = 0; j < 2; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 2; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else {
        buf[i++] = 0x80;       // mov opcode
        buf[i++] = 0x2d;       // modr/m
        for(j = 0; j < 1; j++){                 // I know it is not a loop!
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 1; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;

}
/*
 * Create cmp instruction with one offset operand
 * and a register operand.
 */
ud_t *UDCreateCmpOffsetWithReg(int offset, enum ud_type dest, int type) {

    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateMovImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    if ((dest >= UD_R_EAX && dest <= UD_R_EDI)){		// 32
        buf[i++] = type;       // mov opcode
        buf[i++] = 0x05 + ((dest - UD_R_EAX) << 3);       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }

    } else if ((dest >= UD_R_AX && dest <= UD_R_DI)){
    	buf[i++] = 0x66;
        buf[i++] = type;       // mov opcode
        buf[i++] = 0x05 + ((dest - UD_R_AX) << 3);       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    } else {
    	buf[i++] = type;       // mov opcode
    	buf[i++] = 0x05 + ((dest - UD_R_AL) << 3);       // modr/m
    	for(j = 0; j < 4; j++){
    		buf[i++] = (offset & 0xff);
    		offset = (offset >> 8);
    	}

    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;
}

ud_t *UDCreateCmpRegWithImm(enum ud_type src, int value) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(src >= UD_R_EAX && src <= UD_R_EDI) && !(src >= UD_R_AX && src <= UD_R_DI) && !(src
            >= UD_R_AL && src <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateMovImmToMem]: unrecognized register value 0x%x\n", src);
        return NULL;
    }

    if ((src >= UD_R_EAX && src <= UD_R_EDI)){		// 32
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0xf8 + (src - UD_R_EAX);       // modr/m

        for(j = 0; j < 4; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }

    } else if ((src >= UD_R_AX && src <= UD_R_DI)){
    	buf[i++] = 0x81;
        buf[i++] = 0x39;       // mov opcode
        buf[i++] = 0xf8 + (src - UD_R_AX);       // modr/m
        for(j = 0; j < 2; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else {
    	buf[i++] = 0x80;       // mov opcode
    	buf[i++] = 0xf8 + (src - UD_R_AL);       // modr/m

        for(j = 0; j < 1; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;
}

/*
 * Create cmp instruction with one offset operand
 * and a register operand.
 */
ud_t *UDCreateCmpOffsetWithImm(int offset, int value, int size) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;


    if (size == 4){		// 32
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0x3d;       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 4; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }

    } else if (size == 2){
    	buf[i++] = 0x66;
        buf[i++] = 0x81;       // mov opcode
        buf[i++] = 0x3d;       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
        for (j = 0; j < 2; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }
    } else {
    	buf[i++] = 0x80;       // mov opcode
    	buf[i++] = 0x3d;       // modr/m
    	for(j = 0; j < 4; j++){
    		buf[i++] = (offset & 0xff);
    		offset = (offset >> 8);
    	}
        for (j = 0; j < 1; j++){
            buf[i++] = (value & 0xff);
            value = (value >> 8);
        }

    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;
}

ud_t *UDCreateAddRegToOffset(int offset, enum ud_type src, int size) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if(size == 4){
    	if (!(src >= UD_R_EAX && src <= UD_R_EDI)){
			fprintf(stderr, "[UDCreateAddRegToOffset]: 0x%x register is not 32!\n", src);
			return NULL;
		}
        buf[i++] = 0x01;       // mov opcode
        buf[i++] = 0x05 + ((src - UD_R_EAX) << 3);       // modr/m
        for(j = 0; j < 4; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }

    } else if (size == 2){
    	if (!(src >= UD_R_AX && src <= UD_R_DI)){
			fprintf(stderr, "[UDCreateAddRegToOffset]: 0x%x register is not 16!\n", src);
			return NULL;
		}
        buf[i++] = 0x66;        // 16 bit prefix
        buf[i++] = 0x01;       // add opcode
        buf[i++] = 0x05+ ((src - UD_R_AX) << 3);       // modr/m
        for(j = 0; j < 2; j++){
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }

    } else {
    	if (!(src >= UD_R_AL && src <= UD_R_BH)){
			fprintf(stderr, "[UDCreateAddRegToOffset]: 0x%x register is not 32!\n", src);
			return NULL;
		}
        buf[i++] = 0x00;       // add opcode
        buf[i++] = 0x05+ ((src - UD_R_AL) << 3);       // modr/m
        for(j = 0; j < 1; j++){                 // I know it is not a loop!
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    }

    // create instruction;
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;

}

/* UDCreateMovOffsetToReg
 * Creates a mov instruction to copy a register to 
 * a memory offset or vice versa.
 * Direction: 1 is for reg to memory offset
 * 0 is for offset to reg
 */



ud_t *UDCreateMovOffsetToReg(enum ud_type dest, int offset, int direction) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;
    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateMovOffsetToReg]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    if ((dest >= UD_R_EAX && dest <= UD_R_EDI)) {
        if (!direction)
            buf[i++] = 0x8b;
        else
            buf[i++] = 0x89;
        buf[i++] = 0x05 | ((dest - UD_R_EAX) << 3);
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    } else if ((dest >= UD_R_AX && dest <= UD_R_DI)) {
        buf[i++] = 0x66;
        if (!direction)
            buf[i++] = 0x8b;
        else
            buf[i++] = 0x89;
        buf[i++] = 0x05 | ((dest - UD_R_AX) << 3);
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    } else {
        if (!direction)
            buf[i++] = 0x8a;
        else
            buf[i++] = 0x88;
        buf[i++] = 0x05 | ((dest - UD_R_AL) << 3);
        for (j = 0; j < 4; j++) {
            buf[i++] = (offset & 0xff);
            offset = (offset >> 8);
        }
    }
    udins = malloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    return udins;
}

/*
 * UDCreateMovImmToMem() -- create an instruction to move an
 * immediate value to a memory address.
 */
ud_t *UDCreateMovImmToMem(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);
    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateMovImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }
    if (bytes[0] == 0xa3 || bytes[0] == 0xa2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = 0x5;
            buf[i++] = bytes[1]; //displacement
            buf[i++] = bytes[2]; //displacement
            buf[i++] = bytes[3]; //displacement
            buf[i++] = bytes[4]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = 0x5;
            buf[i++] = bytes[1]; //displacement
            buf[i++] = bytes[2]; //displacement
            buf[i++] = bytes[3]; //displacement
            buf[i++] = bytes[4]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            buf[i++] = 0x5;
            buf[i++] = bytes[1]; //displacement
            buf[i++] = bytes[2]; //displacement
            buf[i++] = bytes[3]; //displacement
            buf[i++] = bytes[4]; //displacement
            buf[i++] = val & 0xff;
        }

    } else if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }
    free(udinsold);
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    free(udinsold);
    return udins;
}

ud_t *UDCreateMovImmToMemFromNonMov(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];// = {0x66, 0x89, 0x9d, 0xec, 0xfe, 0xff, 0xff};
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);
    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateMovImmToMemFromNonMov]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }
    if (bytes[0] == 0xa3 || bytes[0] == 0xa2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = 0x5;
            buf[i++] = bytes[1]; //displacement
            buf[i++] = bytes[2]; //displacement
            buf[i++] = bytes[3]; //displacement
            buf[i++] = bytes[4]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = 0x5;
            buf[i++] = bytes[1]; //displacement
            buf[i++] = bytes[2]; //displacement
            buf[i++] = bytes[3]; //displacement
            buf[i++] = bytes[4]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            buf[i++] = 0x5;
            buf[i++] = bytes[1]; //displacement
            buf[i++] = bytes[2]; //displacement
            buf[i++] = bytes[3]; //displacement
            buf[i++] = bytes[4]; //displacement
            buf[i++] = val & 0xff;
        }

    } else if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0xc7;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0xc6;
            buf[i++] = (MODR | RM);
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }
    free(udinsold);
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);

    free(udinsold);
    return udins;
}
/*
 * UDCreateAddImmToMem() -- create an instruction to Add an
 * immediate value to a memory address.
 */
ud_t *UDCreateAddImmToMem(unsigned char *bytes, enum ud_type dest, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udinsold, *udins;
    int SIB = 0, MODR = 0, RM = 0, i;

    udins = malloc(sizeof(ud_t));
    udinsold = malloc(sizeof(ud_t));
    InitUdis(udins);
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    clear_insn(udinsold);
    get_prefixes(udinsold);

    for (i = 0; i < udinsold->inp_curr; i++)
        buf[i] = bytes[i];

    if (!(dest >= UD_R_EAX && dest <= UD_R_EDI) && !(dest >= UD_R_AX && dest <= UD_R_DI) && !(dest
            >= UD_R_AL && dest <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateSubImmToMem]: unrecognized register value 0x%x\n", dest);
        return NULL;
    }

    MODR |= ((bytes[udinsold->inp_curr + 1] & 0xc0));
    RM |= (bytes[udinsold->inp_curr + 1] & 0x07);
    SIB |= bytes[udinsold->inp_curr + 2];
    int opcode2 = (MODR | RM);
    if (RM == 0x4) {
        udinsold->inp_curr++;
    }

    if ((MODR >> 6) == 1) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //disp
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = val & 0xff;
        }
    } else if ((MODR >> 6) == 0x2) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
            buf[i++] = bytes[udinsold->inp_curr + 3];
            buf[i++] = bytes[udinsold->inp_curr + 4];
            buf[i++] = bytes[udinsold->inp_curr + 5];
            buf[i++] = (val >> 0) & 0xff;
        }

    } else if (MODR == 0) {
        if (dest >= UD_R_EAX && dest <= UD_R_EDI) { // 32-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
            buf[i++] = (val >> 16) & 0xff;
            buf[i++] = (val >> 24) & 0xff;
        } else if (dest >= UD_R_AX && dest <= UD_R_DI) { // 16-bit
            buf[i++] = 0x81;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
            buf[i++] = (val >> 8) & 0xff;
        } else { // 8-bit
            buf[i++] = 0x80;
            buf[i++] = opcode2;
            if (RM == 0x4)
                buf[i++] = SIB;
            if (RM == 0x5) {
                buf[i++] = bytes[udinsold->inp_curr + 2]; //displacement
                buf[i++] = bytes[udinsold->inp_curr + 3];
                buf[i++] = bytes[udinsold->inp_curr + 4];
                buf[i++] = bytes[udinsold->inp_curr + 5];
            }
            buf[i++] = (val >> 0) & 0xff;
        }
    }

    udinsold = malloc(sizeof(ud_t));
    InitUdis(udinsold);
    ud_set_input_buffer(udinsold, bytes, MAX_INSTRLEN);
    ud_disassemble(udinsold);

    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    free(udinsold);
    return udins;
}

ud_t *UDCreateImulImm8ToReg(enum ud_type reg, int val){
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0;

    if((val & ~0xff) != 0){
        fprintf(stderr, "[UDCreateImulImmToReg] Error, immidiate value is greater than one byte!\n ");
        return NULL;
    }

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI)) {
        fprintf(stderr, "[UDCreateImulImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit

        buf[i++] = 0x6B;
        buf[i++] = 0xc0 + (reg - UD_R_EAX);

        buf[i++] = val & 0xff;
    }
    else{
        buf[i++] = 0x6B;
        buf[i++] = 0xc0 + (reg - UD_R_AX);

        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * create jz instruction
 * Don't know how the address is
 * being handled!
 */

ud_t *UDCreateJz(int val){
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0;

    buf[i++] = 0x74;
    buf[i++] = 0xcb;

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * create sub dst, src
 */

ud_t *UDCreateSubRegFromReg(enum ud_type src, enum ud_type dst) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0;

    if (!(src >= UD_R_EAX && src <= UD_R_EDI) && !(src >= UD_R_AX && src <= UD_R_DI) && !(src
            >= UD_R_AL && src <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateAddImmToReg]: unrecognized register value 0x%x\n", src);
        return NULL;
    }

    if (src >= UD_R_EAX && src <= UD_R_EDI) { // 32-bit
        buf[i++] = 0x29;
        buf[i++] = 0xc0 + (dst - UD_R_EAX) + ((src - UD_R_EAX) << 3);
    } else if (src >= UD_R_AX && src <= UD_R_DI) { // 32-bit
    	buf[i++] = 0x66;
        buf[i++] = 0x29;
        buf[i++] = 0xc0 + (dst - UD_R_AX) + ((src - UD_R_AX) << 3);
    } else {
        buf[i++] = 0x28;
        buf[i++] = 0xc0 + (dst - UD_R_AL) + ((src - UD_R_AL) << 3);
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * create add dst, src
 */
ud_t *UDCreateAddRegToReg(enum ud_type src, enum ud_type dst) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0;

    if (!(src >= UD_R_EAX && src <= UD_R_EDI) && !(src >= UD_R_AX && src <= UD_R_DI) && !(src
            >= UD_R_AL && src <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateAddImmToReg]: unrecognized register value 0x%x\n", src);
        return NULL;
    }

    if (src >= UD_R_EAX && src <= UD_R_EDI) { // 32-bit
        buf[i++] = 0x01;
        buf[i++] = 0xc0 + (dst - UD_R_EAX) + ((src - UD_R_EAX) << 3);
    } else if (src >= UD_R_AX && src <= UD_R_DI) { // 32-bit
    	buf[i++] = 0x66;
        buf[i++] = 0x01;
        buf[i++] = 0xc0 + (dst - UD_R_AX) + ((src - UD_R_AX) << 3);
    } else {
        buf[i++] = 0x00;
        buf[i++] = 0xc0 + (dst - UD_R_AL) + ((src - UD_R_AL) << 3);
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateAddImmToReg() -- create an instruction to add
 * an immediate value to a register.
 */
ud_t *UDCreateAddImmToReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateAddImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0x05;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xc0 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0x05;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xc0 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0x04;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xc0 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateSubImmFromReg() -- create an instruction to subtract
 * an immediate value from a register.
 */
ud_t *UDCreateSubImmFromReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateSubImmFromReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0x2d;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xe8 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0x2d;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xe8 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0x2c;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xe8 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateXorImmWithReg() -- create an instruction to xor
 * an immediate value with a register.
 */
ud_t *UDCreateXorImmWithReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateAddImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0x35;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xf0 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0x35;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xf0 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0x34;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xf0 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateAndImmWithReg() -- create an instruction to and
 * an immediate value with a register.
 */
ud_t *UDCreateAndImmWithReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateAndImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0x25;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xe0 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0x25;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xe0 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0x24;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xe0 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateCmpImmWithReg() -- create an instruction to compare
 * an immediate value with a register.
 */
ud_t *UDCreateCmpImmWithReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateCmpImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0x3d;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xf8 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0x3d;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xf8 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0x3c;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xf8 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateTestImmWithReg() -- create an instruction to test
 * an immediate value with a register.
 */
ud_t *UDCreateTestImmWithReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateTestImmWithReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0xa9;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xc0 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0xa9;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xc0 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0xa8;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xc0 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateOrImmWithReg() -- create an instruction to or
 * an immediate value with a register.
 */
ud_t *UDCreateOrImmWithReg(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreateOrImmToReg]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        if (reg == UD_R_EAX) {
            buf[i++] = 0x0d;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xc8 + (reg - UD_R_EAX);
        }
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        if (reg == UD_R_AX) {
            buf[i++] = 0x0d;
        } else {
            buf[i++] = 0x81;
            buf[i++] = 0xc8 + (reg - UD_R_AX);
        }
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        if (reg == UD_R_AL) {
            buf[i++] = 0x0c;
        } else {
            buf[i++] = 0x80;
            buf[i++] = 0xc8 + (reg - UD_R_AL);
        }
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * push val
 */
ud_t *UDCreatePushImm(enum ud_type reg, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    if (!(reg >= UD_R_EAX && reg <= UD_R_EDI) && !(reg >= UD_R_AX && reg <= UD_R_DI) && !(reg
            >= UD_R_AL && reg <= UD_R_BH)) {
        fprintf(stderr, "[UDCreatePushImm]: unrecognized register value 0x%x\n", reg);
        return NULL;
    }

    if (reg >= UD_R_EAX && reg <= UD_R_EDI) { // 32-bit
        buf[i++] = 0x68;
        for (j = 0; j < 4; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) { // 16-bit
        buf[i++] = 0x66;
        buf[i++] = 0x68;
        for (j = 0; j < 2; j++) {
            buf[i++] = val & 0xff;
            val = val >> 8;
        }
    } else { // 8-bit
        buf[i++] = 0x6A;
        buf[i++] = val & 0xff;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * jmp addr
 */
ud_t *UDCreateJump(int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    //if (op->size == 32) {     // 32-bit
    buf[i++] = 0xe9;
    for (j = 0; j < 4; j++) {
        buf[i++] = val & 0xff;
        val = val >> 8;
    }
    //}
    /*
     else if (op->size == 16) {  // 16-bit
     buf[i++] = 0xFF;
     for (j = 0; j < 2; j++) {
     buf[i++] = val & 0xff;
     val = val >> 8;
     }
     }
     else {                                     // 8-bit
     buf[i++] = 0xEB;
     buf[i++] = val & 0xff;
     }
     */
    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

ud_t *UDCreateJumpToAddr(int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0, j;

    buf[i++] = 0xEA;
    for (j = 0; j < 4; j++) {
        buf[i++] = val & 0xff;
        val = val >> 8;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateIndirectStoreConst() -- create an indirect store of a
 * 32-bit constant.
 */
ud_t *UDCreateIndirectStoreConst(enum ud_type base, enum ud_type index, int disp, int val) {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i, j;

    if (index != UD_NONE) {
        fprintf(stderr, "Indirect store involving index reg not yet implemented\n");
        return NULL;
    }

    if ((disp & 0xff) != disp) {
        fprintf(stderr, "Indirect store with disp > 8 bits not yet implemented\n");
        return NULL;
    }

    if (base == UD_R_ESP) {
        if ((index != UD_NONE) || (((uint) disp) > 0xff)) {
            fprintf(
                    stderr,
                    "index: %d, disp %d: Indirect store through ESP only implemented for index == NONE and disp 8bit\n",
                    index, disp);
        }

        buf[0] = 0xc7;
        buf[1] = 0x44;
        buf[2] = 0x24;
        buf[3] = disp & 0xff;

        for (i = 4, j = 0; j < 4; i++, j++) {
            buf[i] = val & 0xff;
            val = val >> 8;
        }
    } else {
        buf[0] = 0xc7;
        buf[1] = 0x40 + (base - UD_R_EAX);
        buf[2] = disp & 0xff;

        for (i = 3, j = 0; j < 4; i++, j++) {
            buf[i] = val & 0xff;
            val = val >> 8;
        }
    }
    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDCreateNop() -- create a no operation, or "nop", instruction
 */
ud_t *UDCreateNop() {
    unsigned char buf[MAX_INSTRLEN];
    ud_t *udins;
    int i = 0;

    buf[0] = 0x90; // the nop byte code

    //clear the rest of the buffer, just to be safe
    for (i = 1; i < MAX_INSTRLEN; i++) {
        buf[i] = 0x0;
    }

    udins = alloc(sizeof(ud_t));
    InitUdis(udins);
    ud_set_input_buffer(udins, buf, MAX_INSTRLEN);
    ud_disassemble(udins);
    return udins;
}

/*
 * UDOpIsRegister(op) checks whether its operand is a general-purpose
 * register.  If it is, it returns the register; otherwise it returns
 * the value UD_NONE.
 */
enum ud_type UDOpIsRegister(ud_operand_t *op) {
    if (op->type == UD_OP_REG) {
        return op->base;
    }

    return UD_NONE;
}

/*
 * UDOpIsIndirectThruReg(op) checks whether its operand is an
 * indirect memory access through a register.  If it is, it returns
 * the register; otherwise it returns UD_NONE.
 */
enum ud_type UDOpIsIndirectThruReg(ud_operand_t *op) {
    if (op->type == UD_OP_MEM && op->base != UD_NONE && op->index == UD_NONE && op->scale == 1
            && op->offset == 0) {
        return op->base;
    }

    return UD_NONE;
}

int RegMask(REGISTER_BITVECTOR reg) {
    if (reg >= UD_R_EAX && reg <= UD_R_EDI) {
        return 0xffffffff;
    } else if (reg >= UD_R_AX && reg <= UD_R_DI) {
        return 0xffff;
    } else if (reg >= UD_R_AL && reg <= UD_R_BH) {
        switch (reg) {
        case UD_R_AL:
        case UD_R_CL:
        case UD_R_DL:
        case UD_R_BL:
            return 0xff;
        case UD_R_AH:
        case UD_R_CH:
        case UD_R_DH:
        case UD_R_BH:
            return 0xff00;
        default:
            return 0;
        }
    }
    fprintf(stderr, "[RegMask] Warning: Bad register %d\n", (int) reg);
    return 0x0;
}

WriteSet *MemoryMask(int offset, REGISTER_BITVECTOR reg) {
    WriteSet *tmp = WriteSetCreate();
    Range *r = NULL;
    if (reg >= LYNX_EAX && reg <= LYNX_EDI) {
        r = RangeCreate(offset, offset + 3);
    } else if (reg >= LYNX_AX && reg <= LYNX_DI) {
        r = RangeCreate(offset, offset + 1);
    } else {
        switch (reg) {
        case LYNX_AL:
        case LYNX_BL:
        case LYNX_CL:
        case LYNX_DL:
            r = RangeCreate(offset, offset);
            break;
        case LYNX_AH:
        case LYNX_CH:
        case LYNX_DH:
        case LYNX_BH:
            r = RangeCreate(offset + 1, offset + 1);
            break;
        default:
            break;
        }
    }
    if (!r) {
        fprintf(stderr, "[MemoryMask] Warning: Not a valid register: %d\n", (int) reg);
    }
    AddRangeToWriteSet(tmp, r);
    return tmp;
}

void CarefullCopyUdins2InsStr(InstrList *iList, Instruction *iins, ins_structure *uIns, ud_t *udIns,
        int doNotChangeFlags) {

    if (udIns == NULL) {
        udIns = malloc(sizeof(ud_t));
        return;
    }
    PSW_BITVECTOR tmpFlagsDef = 0, tmpFlagsUsed = 0;

    CopyUdins2InsStr(uIns, udIns);

    if (doNotChangeFlags) {
        //iins->regFlagsKnown |= FlagsUsedKnown; //flags used already stored in flagsUsed
        //iins->regFlagsKnown |= FlagsDefKnown; //flags def already stored
        tmpFlagsDef = iins->flagsDef;
        tmpFlagsUsed = iins->flagsUsed;

    }
    InstrRegsFalgsUseDef(iList, iins);
    if (doNotChangeFlags) {
        iins->flagsDef = tmpFlagsDef;
        iins->flagsUsed = tmpFlagsUsed;
    }
    /*iins->regFlagsKnown &= ~MemRegUsedKnown; //mem regs used already stored
    iins->regFlagsKnown &= ~NonMemRegUsedKnown; //non mem regs used already stored
    iins->regFlagsKnown &= ~RegDefKnown;*/ //regs def already stored
    iins->regFlagsKnown &= ~MemUsedKnown; //mem used stored in abbreviated form
    iins->regFlagsKnown &= ~MemDefKnown;



    iins->memUsedMin = 0;
    iins->memUsedSize = 0;
    iins->memDefMin = 0;
    iins->memDefSize = 0;

    WriteSet *ws = MemDef(iList, iins);
    iins->regFlagsKnown |= MemDefKnown;
    if (WriteSetIsEmpty(ws)) {
        INSTR_REMOVE_FLAG(iins, INSTR_WRITES_MEM);
    } else {
        INSTR_SET_FLAG(iins, INSTR_WRITES_MEM);
        int numRanges = al_size(ws->ranges);
        if (numRanges == 1) {
            iins->regFlagsKnown |= MemDefKnown;
            Range *r = al_get(ws->ranges, 0);
            iins->memDefMin = r->min;
            iins->memDefSize = r->max - r->min + 1;
        }
    }
    WriteSetDestroy(ws);

    ws = MemUsed(iList, iins);
    iins->regFlagsKnown |= MemUsedKnown;
    if (WriteSetIsEmpty(ws)) {
        INSTR_REMOVE_FLAG(iins, INSTR_READS_MEM);
    } else {
        INSTR_SET_FLAG(iins, INSTR_READS_MEM);
        int numRanges = al_size(ws->ranges);
        if (numRanges == 1) {
            iins->regFlagsKnown |= MemUsedKnown;
            Range *r = al_get(ws->ranges, 0);
            iins->memUsedMin = r->min;
            iins->memUsedSize = r->max - r->min + 1;
        }
    }
    WriteSetDestroy(ws);
}



/*
 * OpcodeIsArith(mnemonic) -- returns true if the mnemonic given is one
 * of a set of recognized arithmetic operations.
 */
bool OpcodeIsArith(enum ud_mnemonic_code opcode) {
    switch (opcode) {
    case UD_Iadc:
    case UD_Iadd:
    case UD_Ixadd:
    case UD_Iand:
    case UD_Idec:
    case UD_Iinc:
    case UD_Inot:
    case UD_Ior:
    case UD_Ishl:
    case UD_Ishr:
    case UD_Isub:
    case UD_Isbb:
    case UD_Ixor:
    case UD_Ineg:
    case UD_Irol:
    case UD_Iror:
    case UD_Iimul:
    case UD_Imul:
    case UD_Isar:
    case UD_Isal:
    case UD_Iidiv:
    case UD_Idiv:
    case UD_Ircr:
    case UD_Ircl:
    case UD_Ibswap:
    case UD_Ibtc:
    case UD_Ibtr:
    case UD_Ibts:
    case UD_Ibt:
    case UD_Icwde:
    case UD_Icbw:
    case UD_Ishrd:
    case UD_Ishld:
    case UD_Ibsf:
    case UD_Ibsr:

        return true;
    default:
        return false;
    }
    /* NOTREACHED */
}

/*
 * OpcodeIsArithWithArithmetic(mnemonic) -- returns true if the mnemonic given is one
 * of a set of recognized arithmetic operations and if the second operand
 *  is 0, the result would be the same.
 */bool OpcodeIsArithWithNoEffect(enum ud_mnemonic_code opcode) {
    switch (opcode) {
    case UD_Iadd:
    case UD_Ior:
    case UD_Ishl:
    case UD_Ishr:
    case UD_Isub:
    case UD_Isbb:
    case UD_Irol:
    case UD_Iror:
    case UD_Isar:
    case UD_Isal:
    case UD_Ircr:
    case UD_Ircl:
        return true;
    default:
        return false;
    }
    /* NOTREACHED */
}

#define ROT_LEFT 0x0
#define ROT_RIGHT !ROT_LEFT

int rotate(unsigned int invalue, int places, int direction, int size) {
    int outvalue;

    /*  Rotating left or right?   */
    if (direction == ROT_LEFT) {
        /*  First a normal shift  */
        outvalue = invalue << (places % (8 * size));
        outvalue |= invalue >> ((8 * size) - places % (8 * size));
    } else {
        /*  First a normal shift  */
        outvalue = invalue >> (places % (8 * size));
        /*  Then place the part that's shifted off the screen at the end  */
        outvalue |= invalue << ((8 * size) - places % (8 * size));
    }
    return outvalue;
}

/*
 * implements bswap instruction
 * bits 0 through 7 are swapped with bits 24 through 31,
 * and bits 8 through 15 are swapped with bits 16 through 23
 * Flags are not
 */
int bSwap(int val) {

    int byte0 = (val & 0xff);
    int byte1 = (val & 0xff00);
    int byte2 = (val & 0xff0000);
    int byte3 = (val & 0xff000000);
    return ((byte0) << 24 | (byte1 << 8) | (byte2 >> 8) | (byte3 >> 24));
}

/*
 * ArithOp(op, val1, val2) -- return the result of performing the arithmetic
 * operation op on the values val1 and val2 (for binary op) or on val1
 * (for unary op).
 */
unsigned int ArithOp(enum ud_mnemonic_code op, unsigned int val1, int val2, int size,
        REGISTER_BITVECTOR *flags, Instruction *iins) {
    int tmp;
    int res = 0, i=0;
    switch (op) {
    case UD_Iadd:
    case UD_Ixadd:
    case UD_Iadc: // TODO: we are treating adc like add! Should be fixed by considering the carry flag.
        return val1 + val2;

    case UD_Iand:
        return val1 & val2;

    case UD_Idec:
        return val1 - 1;

    case UD_Iinc:
        return val1 + 1;

    case UD_Inot:
        return ~val1;

    case UD_Ior:
        return val1 | val2;

    case UD_Ishl:
    case UD_Isal:
    case UD_Ishld:              // TODO
        return val1 << val2;

    case UD_Ishr:
    case UD_Isar:
    case UD_Ishrd:              // TODO: this is not really shr!
        return val1 >> val2;

    case UD_Isub:
    case UD_Isbb:
        return val1 - val2;

    case UD_Ixor:
        return val1 ^ val2;

    case UD_Ineg:
        return ~val1 + 1;

    case UD_Iimul:
    case UD_Imul:
        return val1 * val2;

    case UD_Iidiv:
    case UD_Idiv:
        if (val2)
            return val1 / val2;
        else
            return 0;

    case UD_Iror:
        return rotate(val1, val2, ROT_RIGHT, size);
    case UD_Irol:
        return rotate(val1, val2, ROT_LEFT, size);

    case UD_Ibswap:
        return bSwap(val1);

    case UD_Ircr:
        return rotate(val1, val2, ROT_RIGHT, size);
    case UD_Ircl:
        //fprintf(stderr, "[ArithOp] Warning: operation cannot be competed inside the function!");
        return rotate(val1, val2, ROT_LEFT, size);         // TODO: implement rcr and rcl arithmetics1

    case UD_Ibt:
        *flags |= PSW_CF;
        break;

    case UD_Ibts:
        *flags |= PSW_CF;
        tmp = (1 << val2);
        return (val1 | tmp);
    case UD_Ibtr:
        *flags |= PSW_CF;
        tmp = ~(1 << val2);
        return (val1 & tmp);
    case UD_Ibtc:
        *flags |= PSW_CF;
        tmp = (1 << val2);
        if (val1 & tmp) {
            return (val1 & ~(1 << val2));
        } else {
            return (val1 | (1 << val2));
        }
    case UD_Icbw:
    case UD_Icwde:
        return SignExtend(val1);

    case UD_Iaaa:
        return val1;

    case UD_Ibsf:
        while(!(val2 & (1 << i++)));
        res = i-1;
        break;

    case UD_Ibsr:
        i = size*8;
        while(!(val2 & (1 << i--)));
        res = i-1;
        break;

    default:
        fprintf(stderr, "ERROR [ArithOp]: unrecognized operation %d\n", op);
        return -1;
    }
    return res;
}

/*REGISTER_BITVECTOR MaskDefinedRegOld(REGISTER_BITVECTOR reg) {
    int result = 0, i, pos;
    result |= reg;
    for (i = 0; i < 14; i++) {
        if (((reg >> i) & 1)) { // 32 bit registers
            result |= (result | (1 << (i + 14)));
        }
    }
    reg = result;
    for (pos = 0; i < 18; i++, pos++) { // ax through bx registers
        if (((reg >> i) & 1)) {
            result |= (result | (3 << (i + pos + 8)));
        }
    }
    return result;
}*/

REGISTER_BITVECTOR MaskDefinedReg(REGISTER_BITVECTOR reg) {
    REGISTER_BITVECTOR result = 0;
    result |= reg;
    if((reg & LYNX_EAX)){
        result |= (LYNX_AX | LYNX_AL | LYNX_AH);
    } else if((reg & LYNX_AX)){
        result |= (LYNX_AL | LYNX_AH);
    }

    if((reg & LYNX_ECX)){
        result |= (LYNX_CX | LYNX_CL | LYNX_CH);
    } else if((reg & LYNX_CX)){
        result |= (LYNX_CL | LYNX_CH);
    }
    if((reg & LYNX_EDX)){
        result |= (LYNX_DX | LYNX_DL | LYNX_DH);
    } else if((reg & LYNX_DX)){
        result |= (LYNX_DL | LYNX_DH);
    }
    if((reg & LYNX_EBX)){
        result |= (LYNX_BX | LYNX_BL | LYNX_BH);
    } else if((reg & LYNX_BX)){
        result |= (LYNX_BL | LYNX_BH);
    }

    if ((reg & LYNX_ESP))
        result |= LYNX_SP;
    if ((reg & LYNX_EBP))
        result |= LYNX_BP;
    if ((reg & LYNX_EDI))
        result |= LYNX_DI;
    if ((reg & LYNX_ESI))
        result |= LYNX_SI;

    /*int tmp = MaskDefinedRegOld(reg);
    if(tmp != result){
        fprintf(stderr, "Erorr: Different results for reg: %d\n", (int)reg);
    }*/

    return result;
}

REGISTER_BITVECTOR UpMaskDefinedReg(REGISTER_BITVECTOR reg) {
    REGISTER_BITVECTOR res = reg;
    if ((reg & LYNX_AL) || (reg & LYNX_AH) || (reg & LYNX_AX))
        res |= LYNX_EAX | LYNX_AX;
    if ((reg & LYNX_CL) || (reg & LYNX_CH) || (reg & LYNX_CX))
        res |= LYNX_ECX | LYNX_CX;
    if ((reg & LYNX_DL) || (reg & LYNX_DH) || (reg & LYNX_DX))
        res |= LYNX_EDX | LYNX_DX;
    if ((reg & LYNX_BL) || (reg & LYNX_BH) || (reg & LYNX_BX))
        res |= LYNX_EBX | LYNX_BX;
    if ((reg & LYNX_SP))
        res |= LYNX_ESP;
    if ((reg & LYNX_BP))
        res |= LYNX_EBP;
    if ((reg & LYNX_DI))
        res |= LYNX_EDI;
    if ((reg & LYNX_SI))
        res |= LYNX_ESI;
    return res;

}
REGISTER_BITVECTOR UpToAllRegs(REGISTER_BITVECTOR reg) {

    REGISTER_BITVECTOR res = reg;
    if ((reg & LYNX_AL) || (reg & LYNX_AH) || (reg & LYNX_AX))
        res |= LYNX_EAX | LYNX_AX | LYNX_AH | LYNX_AL;
    if ((reg & LYNX_CL) || (reg & LYNX_CH) || (reg & LYNX_CX))
        res |= LYNX_ECX | LYNX_CX | LYNX_CH | LYNX_CL;
    if ((reg & LYNX_DL) || (reg & LYNX_DH) || (reg & LYNX_DX))
        res |= LYNX_EDX | LYNX_DX | LYNX_DH | LYNX_DL;
    if ((reg & LYNX_BL) || (reg & LYNX_BH) || (reg & LYNX_BX))
        res |= LYNX_EBX | LYNX_BX | LYNX_BH | LYNX_BL;
    if ((reg & LYNX_SP))
        res |= LYNX_ESP | LYNX_SP;
    if ((reg & LYNX_BP))
        res |= LYNX_EBP | LYNX_BP;
    if ((reg & LYNX_DI))
        res |= LYNX_EDI | LYNX_DI;
    if ((reg & LYNX_SI))
        res |= LYNX_ESI | LYNX_SI;
    return res;
}


/*
 * return [lg(v)]
 */
int floor_log2(REGISTER_BITVECTOR v) {
    int i;
    for (i = 0; i < 30; i++) {
        if (((v >> i) & 1))
            return i;
    }
    return -1;
}

bool OperandsAreEqual(ud_operand_t *op0, ud_operand_t *op1) {
    if (op0->type != op1->type)
        return false;
    if (op0->type == UD_OP_MEM) {
        if (!(op0->base == op1->base && op0->index == op1->index && op0->offset == op1->offset
                && op0->scale == op1->scale)) {
            return false;
        }
    } else if (op0->type == UD_OP_REG) {
        if (op0->base != op1->base) {
            return false;
        }
    } else if (op0->type == UD_OP_IMM || op0->type == UD_OP_CONST)
        if (op0->lval.sdword != op1->lval.sdword)
            return false;
    return true;
}


Instruction *ThreeXorsToXchg(InstrList *iList, Instruction *iins) {

    ud_operand_t *op0, *op1;

    Instruction *second;
    int regs;

    op0 = &(iins->uInstr.operand[0]);
    op1 = &(iins->uInstr.operand[1]);
    regs = (((op0->type == UD_OP_REG ? op0->base : 0) | (op1->type == UD_OP_REG ? op1->base : 0)));
    while ((second = FetchNextInstr(iList)) != NULL) {
        if (second->uInstr.mnemonic == UD_Ixor
                && OperandsAreEqual(op0, &(second->uInstr.operand[1]))
                && OperandsAreEqual(op1, &(second->uInstr.operand[0]))) {
            Instruction *third;
            while ((third = FetchNextInstr(iList)) != NULL) {
                if (third->uInstr.mnemonic == UD_Ixor
                        && OperandsAreEqual(op0, &(third->uInstr.operand[0]))
                        && OperandsAreEqual(op1, &(third->uInstr.operand[1]))) {
                    if(iins->uInstr.bytes[0] != 0x66){
                    	if(iins->uInstr.bytes[0] == 0x31){
                    		iins->uInstr.bytes[0] = 0x87;
                    	} else {
                    		iins->uInstr.bytes[0] = 0x86;
                    	}
                    } else {
                        iins->uInstr.bytes[1] = 0x87;
                    }
                    ud_t *udins = malloc(sizeof(ud_t));
                    InitUdis(udins);
                    ud_set_input_buffer(udins, iins->uInstr.bytes, MAX_INSTRLEN);
                    ud_disassemble(udins);
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 0);
                    //INSTR_SET_FLAG(second, INSTR_IS_DELETED);
                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    SaveInstrChange(iList, iins);

                    INSTR_SET_FLAG(second, INSTR_IS_DELETED);
                    INSTR_SET_FLAG(third, INSTR_IS_DELETED);

                    SaveInstrChange(iList, third);
                    SaveInstrChange(iList, second);

                    free(udins);
                    return iins;

                } else {
                    if (third->uInstr.operand[0].type == UD_OP_REG
                            && (regs & third->uInstr.operand[0].base)) {
                        RInstrDestroy(iList, third);
                        return NULL;
                    }
                    if (third->uInstr.operand[1].type == UD_OP_REG
                            && (regs & third->uInstr.operand[1].base)) {
                        RInstrDestroy(iList, third);
                        return NULL;
                    }
                    if (INSTR_HAS_FLAG(third, INSTR_READS_MEM)) {
                        if ((iins->memUsedMin >= third->memUsedMin
                                && iins->memUsedMin <= third->memUsedMin + third->memUsedSize - 1)
                                || (iins->memUsedMin + iins->memUsedSize - 1 <= third->memUsedMin + third->memUsedSize - 1
                                        && iins->memUsedMin + iins->memUsedSize - 1 >= third->memUsedMin)) {
                            RInstrDestroy(iList, third);
                            return NULL;
                        }
                    } else if (INSTR_HAS_FLAG(third, INSTR_WRITES_MEM)) {
                        if ((iins->memUsedMin >= third->memDefMin
                                && iins->memUsedMin <= third->memDefMin + third->memDefSize - 1)
                                || (iins->memDefMin + iins->memDefSize - 1 <= third->memDefMin + third->memDefSize - 1
                                        && iins->memDefMin + iins->memDefSize - 1 >= third->memDefMin)) {
                            RInstrDestroy(iList, third);
                            return NULL;
                        }
                    }
                }
            }

        } else {
            if (second->uInstr.operand[0].type == UD_OP_REG
                    && (regs & second->uInstr.operand[0].base)) {
                RInstrDestroy(iList, second);
                return NULL;
            }
            if (second->uInstr.operand[1].type == UD_OP_REG
                    && (regs & second->uInstr.operand[1].base)) {
                RInstrDestroy(iList, second);
                return NULL;
            }
            if (INSTR_HAS_FLAG(second, INSTR_READS_MEM)) {
                if ((iins->memUsedMin >= second->memUsedMin
                        && iins->memUsedMin <= second->memUsedMin + second->memUsedSize - 1)
                        || (iins->memUsedMin + iins->memUsedSize - 1 <= second->memUsedMin + second->memUsedSize - 1
                                && iins->memUsedMin + iins->memUsedSize - 1 >= second->memUsedMin)) {
                    RInstrDestroy(iList, second);
                    return NULL;
                }
            } else if (INSTR_HAS_FLAG(second, INSTR_WRITES_MEM)) {
                if ((iins->memUsedMin >= second->memDefMin && iins->memUsedMin <= second->memDefMin + second->memDefSize - 1)
                        || (iins->memUsedMin + iins->memUsedSize - 1 <= second->memDefMin + second->memDefSize - 1
                                && iins->memUsedMin + iins->memUsedSize - 1 >= second->memDefMin)) {
                    RInstrDestroy(iList, second);
                    return NULL;
                }
            }

        }
        RInstrDestroy(iList, second);
    }
    return NULL;
}

Instruction *TransformXchgNotXchgToNot(InstrList *iList, Instruction *iins) {
    Instruction *second;
    int regs;

    ud_operand_t *op0, *op1;

    op0 = &(iins->uInstr.operand[0]);
    op1 = &(iins->uInstr.operand[1]);
    regs = (((op0->type == UD_OP_REG ? op0->base : 0) | (op1->type == UD_OP_REG ? op1->base : 0)));
    while ((second = FetchNextInstr(iList)) != NULL) {
        if (second->uInstr.mnemonic == UD_Inot
                && OperandsAreEqual(op1, &(second->uInstr.operand[0]))) {
            Instruction *third;
            while ((third = FetchNextInstr(iList)) != NULL) {
                if (third->uInstr.mnemonic == UD_Ixchg
                        && OperandsAreEqual(op0, &(third->uInstr.operand[0]))
                        && OperandsAreEqual(op1, &(third->uInstr.operand[1]))) {

                    ud_t *udins;
                    if (op0->type == UD_OP_MEM) {
                        udins = UDCreateNotMem(iins->memDefMin,
                                iins->memDefSize);
                    } else {
                        udins = UDCreateNotOnReg(op0->base);
                    }
                    CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins, 0);

                    INSTR_SET_FLAG(iins, INSTR_TMP3);
                    SaveInstrChange(iList, iins);
                    INSTR_SET_FLAG(second, INSTR_IS_DELETED);
                    SaveInstrChange(iList, second);
                    INSTR_SET_FLAG(third, INSTR_IS_DELETED);
                    SaveInstrChange(iList, third);
                    free(udins);
                    RInstrDestroy(iList, third);

                    //N += 2;
                    //fprintf(stderr, "[xor to xchg]: %d: %s\n",
                    //      iins->order, ud_insn_asm(LoadUDisInstrObj(iins)));

                    return iins;
                } else {
                    if (third->uInstr.operand[0].type == UD_OP_REG
                            && (regs & third->uInstr.operand[0].base)) {
                        RInstrDestroy(iList, third);
                        return NULL;
                    }
                    if (third->uInstr.operand[1].type == UD_OP_REG
                            && (regs & third->uInstr.operand[1].base)) {
                        RInstrDestroy(iList, third);
                        return NULL;
                    }
                    if (INSTR_HAS_FLAG(third, INSTR_READS_MEM)) {
                        if ((iins->memUsedMin >= third->memUsedMin
                                && iins->memUsedMin <= third->memUsedMin + third->memUsedSize - 1)
                                || (iins->memUsedMin + iins->memUsedSize - 1 <= third->memUsedMin + third->memUsedSize - 1
                                        && iins->memUsedMin + iins->memUsedSize - 1 >= third->memUsedMin)) {
                            RInstrDestroy(iList, third);
                            return NULL;
                        }
                    } else if (INSTR_HAS_FLAG(third, INSTR_WRITES_MEM)) {
                        if ((iins->memUsedMin >= third->memDefMin
                                && iins->memUsedMin <= third->memDefMin + third->memDefSize - 1)
                                || (iins->memUsedMin + iins->memUsedSize - 1  <= third->memDefMin + third->memDefSize - 1
                                        && iins->memUsedMin + iins->memUsedSize - 1  >= third->memDefMin)) {
                            RInstrDestroy(iList, third);
                            return NULL;
                        }
                    }
                }
            }

        } else {
            if (second->uInstr.operand[0].type == UD_OP_REG
                    && (regs & second->uInstr.operand[0].base)) {
                RInstrDestroy(iList, second);
                return NULL;
            }
            if (second->uInstr.operand[1].type == UD_OP_REG
                    && (regs & second->uInstr.operand[1].base)) {
                RInstrDestroy(iList, second);
                return NULL;
            }
            if (INSTR_HAS_FLAG(second, INSTR_READS_MEM)) {
                if ((iins->memUsedMin >= second->memUsedMin
                        && iins->memUsedMin <= second->memUsedMin + second->memUsedSize - 1)
                        || (iins->memUsedMin + iins->memUsedSize - 1 <= second->memUsedMin + second->memUsedSize - 1
                                && iins->memUsedMin + iins->memUsedSize - 1 >= second->memUsedMin)) {
                    RInstrDestroy(iList, second);
                    return NULL;
                }
            } else if (INSTR_HAS_FLAG(second, INSTR_WRITES_MEM)) {
                if ((iins->memUsedMin >= second->memDefMin && iins->memUsedMin <= second->memDefMin + second->memDefSize - 1)
                        || (iins->memUsedMin + iins->memUsedSize - 1 <= second->memDefMin + second->memDefSize - 1
                                && iins->memUsedMin + iins->memUsedSize - 1 >= second->memDefMin)) {
                    RInstrDestroy(iList, second);
                    return NULL;
                }
            }

        }
        RInstrDestroy(iList, second);
    }
    return NULL;
}


int SignExtend(int value) {
    if (value & (1 << 15))
        return (value | 0xffff0000);
    else if (value & (1 << 7))
        return (value | 0xffffff00);
    return value;
}

int range_compare(void *fir, void *sec) {
    Range *r1 = (Range *) fir;
    Range *r2 = (Range *) sec;
    if (r1->min < r2->min)
        return -1;
    return 0;

}

/*
 * map memory used in the program to a
 * global (duplicate) memory for using
 * in liveness and simplification process.
 */
void ConstantMemoryMap(InstrList* iList) {
    BytetoByteMap *PageDir = NULL;

    Instruction *iins=NULL;
    WriteSet *memUsed, *memDef;
    WriteSet *globalMem;
    enum ud_mnemonic_code opcode;
    ud_operand_t *op0, *op1;
    bool startMarker = false, endMarker = false;

    //uint64_t  eightPages = PAGE_SIZE*8;

    globalMem = WriteSetCreate();
    globalMem->ranges->al_compare = &range_compare;

    iList->currentInstr = iList->firstInstr;
    while ((iins = FetchNextInstr(iList)) != NULL) {

        opcode = iins->uInstr.mnemonic;
        op0 = &(iins->uInstr.operand[0]);
        op1 = &(iins->uInstr.operand[1]);
        if(opcode == UD_Imovsd){
            Instruction *tmp = GetInstruction(iList, iins->next);
            while(tmp && tmp->addr == iins->addr){
                INSTR_SET_FLAG(tmp, INSTR_IS_DELETED);
                SaveInstrChange(iList, tmp);
                tmp = GetInstruction(iList, tmp->next);
            }
        }

#if MAIN_MARKER
        if (startMarker == false && opcode == UD_Ior && op0->type == UD_OP_REG && op1->type == UD_OP_REG && op0->base == UD_R_EAX && op1->base == UD_R_EAX) {
            Instruction *second = GetInstruction(iList, iins->next);
            if(second->uInstr.mnemonic == UD_Ior && second->uInstr.operand[0].type == UD_OP_REG && second->uInstr.operand[0].base == UD_R_EAX && second->uInstr.operand[1].type == UD_OP_REG && second->uInstr.operand[1].base == UD_R_EAX) {
                Instruction *third = GetInstruction(iList, second->next);
                if(third->uInstr.mnemonic == UD_Ior && third->uInstr.operand[0].type == UD_OP_REG && third->uInstr.operand[0].base == UD_R_EAX && third->uInstr.operand[1].type == UD_OP_REG && third->uInstr.operand[1].base == UD_R_EAX) {
                    startMarker = true;
                    INSTR_SET_FLAG(third, INSTR_UNSIMPLIFIABLE);
                    SaveInstrChange(iList, third);
                    iList->startMarker = iins->order;
                }
            }
        }

        if (endMarker == false && opcode == UD_Ior && op0->type == UD_OP_REG && op1->type == UD_OP_REG && op0->base == UD_R_EBX && op1->base == UD_R_EBX) {
            Instruction *second = GetInstruction(iList, iins->next);
            if(second->uInstr.mnemonic == UD_Ior && second->uInstr.operand[0].type == UD_OP_REG && second->uInstr.operand[0].base == UD_R_EBX && second->uInstr.operand[1].type == UD_OP_REG && second->uInstr.operand[1].base == UD_R_EBX) {
                Instruction *third = GetInstruction(iList, second->next);
                if(third->uInstr.mnemonic == UD_Ior && third->uInstr.operand[0].type == UD_OP_REG && third->uInstr.operand[0].base == UD_R_EBX && third->uInstr.operand[1].type == UD_OP_REG && third->uInstr.operand[1].base == UD_R_EBX) {
                    endMarker = true;
                    iList->endMarker = iins->order;
                    INSTR_SET_FLAG(iins, INSTR_UNSIMPLIFIABLE);
                    SaveInstrChange(iList, iins);
                }
            }
        }
        if(iList->endMarker == 0) {
            iList->endMarker = iList->numInstrs;
        }
#endif

        if (opcode == UD_Ixor && InstrXorAsMov(iins)) {
            ud_t *udins_new = UDCreateMovImmToReg(op0->base, 0);

            CarefullCopyUdins2InsStr(iList, iins, &(iins->uInstr), udins_new, 1);
            SaveInstrChange(iList, iins);
        }

        if (!INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM) && !INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
            continue;
        }
        memUsed = MemUsed(iList, iins);
        memDef = MemDef(iList, iins);

        if (INSTR_HAS_FLAG(iins, INSTR_READS_MEM)) {
            if (!WriteSetSubsetOfWriteSet(memUsed, globalMem)) {
                globalMem = MergeWriteSets(globalMem, memUsed);
            }

        }
        if (INSTR_HAS_FLAG(iins, INSTR_WRITES_MEM)) {
            if (!WriteSetSubsetOfWriteSet(memDef, globalMem)) {
                globalMem = MergeWriteSets(globalMem, memDef);
            }

        }
#if CHECKSUM_DETECT
        if(INSTR_HAS_FLAG(iins, INSTR_IN_BASE_MODULE)){
            Range *exec = RangeCreate(iins->addr, iins->addr + iins->uInstr.length);
            AddRangeToWriteSet(globalMem, exec);
        }

#endif

        WriteSetDestroy(memDef);
        WriteSetDestroy(memUsed);
        RInstrDestroy(iList, iins);
    }

    al_sort(globalMem->ranges);
    /*Merging those ranges which are within a page*/
    int r = 0;
    for (; r < al_size(globalMem->ranges) - 1; r++) {
        Range *r1 = al_get(globalMem->ranges, r);
        Range *r2 = al_get(globalMem->ranges, r + 1);
        uint diff=0;
        if (r1->min < r2->min) {
            diff = r2->min - r1->max;
        } else {
            fprintf(stderr, "Warning: arraylist is not sorted\n");
        }
        if (diff <= PAGE_SIZE) {
            //fprintf(stderr, " 0x%08x -- 0x%08x combined with 0x%08x -- 0x%08x resuting: ", r1->min, r1->max, r2->min, r2->max);
            r1->max = r2->max;
            //fprintf(stderr, " 0x%08x -- 0x%08x\n", r1->min, r1->max);
            al_removeAt(globalMem->ranges, r + 1);
            r--;
            RangeDestroy(r2);
        }
        if (diff <= 0)
            fprintf(stderr,
                    "Error in calculating ranges: Range1: 0x%08x -- 0x%08x, Range2: 0x%08x -- 0x%08x\n",
                    r1->min, r1->max, r2->min, r2->max);
    }

    /*breaking ranges longer that one page*/
    int totalPageDirEntries = 0;
    for (r = 0; r < al_size(globalMem->ranges); r++) {
        Range *r1 = al_get(globalMem->ranges, r);
        int pages = (r1->max - r1->min) / PAGE_SIZE;
        if (pages >= 1) {
            int newPages = pages;
            int nextMin = r1->min + PAGE_SIZE;
            r1->max = nextMin - 1;
            for (; newPages >= 0; newPages--) {
                Range *newRane = RangeCreate(nextMin, nextMin + PAGE_SIZE - 1);
                al_add(globalMem->ranges, newRane);
                nextMin = nextMin + PAGE_SIZE;
            }
        }
    }
    totalPageDirEntries = al_size(globalMem->ranges);
    iList->constantNumPageEntries = totalPageDirEntries / 8 + 1;
    PageDir = malloc((iList->constantNumPageEntries) * sizeof(BytetoByteMap));
    bzero(PageDir, (iList->constantNumPageEntries) * sizeof(BytetoByteMap));
    Range *r1;
    for (r = 0; r < iList->constantNumPageEntries; r++) {
        int j;
        PageDir[r].Constatns = malloc(PAGE_SIZE);
        for (j = 0; j < 8 && r * 8 + j < totalPageDirEntries; j++) {
            r1 = al_get(globalMem->ranges, r * 8 + j);
            PageDir[r].startAddrs[j] = (int) r1->min;
            /*
             * *sizeof(uint64_t) is for dependency analysis which we need 64 bit (instruction's order number)
             * for each entry (does not matter how big is the instruction's read or write memory)
             */
            PageDir[r].pages[j] = malloc(PAGE_SIZE * sizeof(uint64_t));
            memset(PageDir[r].pages[j], 0, PAGE_SIZE * sizeof(uint64_t));
        }
    }

    iList->constantPageDir = PageDir;
}
